//package ij.plugin;
import ij.*;
import ij.gui.*;
import ij.util.Tools;
import ij.plugin.*;
import java.io.*;
import java.net.*;
import java.util.*;
import java.lang.reflect.*;


/** This plugin implements the Help/Update AstroImageJ command. */
public class AstroImageJ_Updater implements PlugIn {

    public static final String URL = "http://www.astro.louisville.edu/software/astroimagej/updates";
    
	public void run(String arg) {
		
//        if (arg.equals("menus"))
//			{updateMenus(); return;}
		if (IJ.getApplet()!=null) return;
		//File file = new File(Prefs.getHomeDir() + File.separator + "ij.jar");
		//if (isMac() && !file.exists())
		//	file = new File(Prefs.getHomeDir() + File.separator + "ImageJ.app/Contents/Resources/Java/ij.jar");
		URL url = getClass().getResource("/ij/IJ.class");
		String ij_jar = url == null ? null : url.toString().replaceAll("%20", " ");
		if (ij_jar==null || !ij_jar.startsWith("jar:file:")) {
			error("Could not determine location of ij.jar");
			return;
		}
		int exclamation = ij_jar.indexOf('!');
		ij_jar = ij_jar.substring(9, exclamation);
		if (IJ.debugMode) IJ.log("Updater: "+ij_jar);
		File file1 = new File(ij_jar);
		if (!file1.exists()) {
			error("File not found: "+file1.getPath());
			return;
		}
		if (!file1.canWrite()) {
			String msg = "No write access: "+file1.getPath();
			error(msg);
			return;
		}
        
		url = getClass().getResource("/AstroImageJ_Updater.class");
		String Astronomy_jar = url == null ? null : url.toString().replaceAll("%20", " ");
		if (Astronomy_jar==null || !Astronomy_jar.startsWith("jar:file:")) {
			error("Could not determine location of Astronomy_.jar");
			return;
		}
		exclamation = Astronomy_jar.indexOf('!');
		Astronomy_jar = Astronomy_jar.substring(9, exclamation);
		if (IJ.debugMode) IJ.log("Updater: "+Astronomy_jar);
		File file2 = new File(Astronomy_jar);
		if (!file2.exists()) {
			error("File not found: "+file2.getPath());
			return;
		}
		if (!file2.canWrite()) {
			String msg = "No write access: "+file2.getPath();
			error(msg);
			return;
		}        
        
//		url = getClass().getResource(file.getParent().replace('\\', '/') +"/macros/StartupMacros.txt");
//		String Startup_Macros = url == null ? null : url.toString().replaceAll("%20", " ");
//        IJ.showMessage(Startup_Macros);
//		if (Startup_Macros==null || !Startup_Macros.startsWith("jar:file:")) {
//			error("Could not determine location of StartupMacros.txt");
//			return;
//		}
//		exclamation = Startup_Macros.indexOf('!');
//		Startup_Macros = Startup_Macros.substring(9, exclamation);
//		if (IJ.debugMode) IJ.log("Updater: "+Startup_Macros);
		File file3 = new File(file2.getParent() + "/../macros/StartupMacros.txt");
		if (!file3.exists()) {
			error("File not found: "+file3.getPath());
			return;
		}
		if (!file3.canWrite()) {
			String msg = "No write access: "+file3.getPath();
			error(msg);
			return;
		}         
        
        
		String[] list = openUrlAsList(URL+"/versions.txt");
		int count = list.length + 1;
		String[] versions = new String[count];
		String[] urls = new String[count];
        String[] AstronomyUrls = new String[count];
        String[] StartupMacrosUrls = new String[count];
//		String uv = getUpgradeVersion();
//		if (uv==null) return;
		versions[0] = "daily build";//"v"+uv;
		urls[0] = URL+"/ij.jar";//"/upgrade/ij.jar";
        AstronomyUrls[0] = URL+"/Astronomy_.jar";
        StartupMacrosUrls[0] = URL+"/StartupMacros.txt";
		if (versions[0]==null) return;
		for (int i=1; i<count; i++) {
			String version = list[i-1];
			versions[i] = version;//.substring(0,version.length()-1); // remove letter
			urls[i] = URL+"/ij"+version+".jar";//.substring(1,2)+version.substring(3,6)+".jar";
            AstronomyUrls[i] = URL+"/Astronomy_"+version+".jar";
            StartupMacrosUrls[i] = URL+"/StartupMacros"+version+".txt";
		}
//		versions[count-1] = "daily build";
//		urls[count-1] = URL+"/ij.jar";
		int choice = showDialog(versions);
		if (choice==-1) return;
//		if (!versions[choice].startsWith("daily") && versions[choice].compareTo("v1.39")<0
//		&& Menus.getCommands().get("AstroImageJ Updater")==null) {
//			String msg = "This command is not available in versions of ImageJ prior\n"+
//			"to 1.39 so you will need to install the plugin version at\n"+
//			"<"+URL+"/plugins/imagej-updater.html>.";
//			if (!IJ.showMessageWithCancel("Update AstroImageJ", msg))
//				return;
//		}
		byte[] jar = getFile(urls[choice], "ij"+(choice==0?"":versions[choice])+".jar");
		if (jar==null) {
			error("Unable to download ij.jar from "+urls[choice]);
			return;
		}
		byte[] Astro_jar = getFile(AstronomyUrls[choice], "Astronomy_"+(choice==0?"":versions[choice])+".jar");
		if (Astro_jar==null) {
			error("Unable to download Astronomy_.jar from "+AstronomyUrls[choice]);
			return;
		}
        byte[] StartupMacro = getFile(StartupMacrosUrls[choice], "StartupMacros"+(choice==0?"":versions[choice])+".txt");
		if (StartupMacro==null) {
			error("Unable to download StartupMacros.txt from "+StartupMacrosUrls[choice]);
			return;
		}
		//file.renameTo(new File(file.getParent()+File.separator+"ij.bak"));
//		if (version().compareTo("1.37v")>=0)
		Prefs.savePreferences();
		// if (!renameJar(file)) return; // doesn't work on Vista
		saveFile(file1, jar);
        saveFile(file2, Astro_jar);
        saveFile(file3, StartupMacro);
        
//		if (choice<count-1) // force macro Function Finder to download fresh list
//			new File(IJ.getDirectory("macros")+"functions.html").delete();
		System.exit(0);
	}

	int showDialog(String[] versions) {
		GenericDialog gd = new GenericDialog("AstroImageJ Updater");
		gd.addChoice("Upgrade To:", versions, versions[(versions.length > 0 ? 1 : 0)]);
		String msg = 
			"You are currently running AstroImageJ "+IJ.getAstroVersion()+".\n"+
			" \n"+
			"Click \"OK\", to download and install the selected upgrade.\n"+
            "After a successful download, AstroImageJ will exit.\n"+
			"Restart AstroImageJ to run the upgraded version.\n";

		gd.addMessage(msg);
        gd.addHelp(URL+"/release_notes.html");
        gd.setHelpLabel("Release Notes");
		gd.showDialog();
		if (gd.wasCanceled())
			return -1;
		else
			return gd.getNextChoiceIndex();
	}

//	String getUpgradeVersion() {
//		String url = IJ.URL+"/notes.html";
//		String notes = openUrlAsString(url, 20);
//		if (notes==null) {
//			error("Unable to connect to "+IJ.URL+". You\n"
//				+"may need to use the Edit>Options>Proxy Settings\n"
//				+"command to configure ImageJ to use a proxy server.");
//			return null;
//		}
//		int index = notes.indexOf("Version ");
//		if (index==-1) {
//			error("Release notes are not in the expected format");
//			return null;
//		}
//		String version = notes.substring(index+8, index+13);
//		return version;
//	}

	String openUrlAsString(String address, int maxLines) {
		StringBuffer sb;
		try {
			URL url = new URL(address);
			InputStream in = url.openStream();
			BufferedReader br = new BufferedReader(new InputStreamReader(in));
			sb = new StringBuffer();
			int count = 0;
			String line;
			while ((line=br.readLine())!=null && count++<maxLines)
				sb.append (line + "\n");
			in.close ();
		} catch (IOException e) {sb = null;}
			return sb!=null?new String(sb):null;
	}

	byte[] getFile(String address, String name) {
		byte[] data;
		try {
			URL url = new URL(address);
			IJ.showStatus("Connecting to "+URL);
			URLConnection uc = url.openConnection();
			int len = uc.getContentLength();
			if (len<=0)
				return null;
			IJ.showStatus("Downloading "+name+" ("+IJ.d2s((double)len/1048576,1)+"MB)");
			InputStream in = uc.getInputStream();
			data = new byte[len];
			int n = 0;
			while (n < len) {
				int count = in.read(data, n, len - n);
				if (count<0)
					throw new EOFException();
	   			 n += count;
				IJ.showProgress(n, len);
			}
			in.close();
		} catch (IOException e) {
			return null;
		}
		return data;
	}

	/*Changes the name of ij.jar to ij-old.jar
	boolean renameJar(File f) {
		File backup = new File(Prefs.getHomeDir() + File.separator + "ij-old.jar");
		if (backup.exists()) {
			if (!backup.delete()) {
				error("Unable to delete backup: "+backup.getPath());
				return false;
			}
		}
		if (!f.renameTo(backup)) {
			error("Unable to rename to ij-old.jar: "+f.getPath());
			return false;
		}
		return true;
	}
	*/

	void saveFile(File f, byte[] data) {
		try {
			FileOutputStream out = new FileOutputStream(f);
			out.write(data, 0, data.length);
			out.close();
		} catch (IOException e) {
		}
	}

	String[] openUrlAsList(String address) {
		IJ.showStatus("Connecting to "+URL);
		Vector v = new Vector();
		try {
			URL url = new URL(address);
			InputStream in = url.openStream();
			BufferedReader br = new BufferedReader(new InputStreamReader(in));
			String line;
			while (true) {
				line = br.readLine();
				if (line==null) break;
				if (!line.equals("")) v.addElement(line);
			}
			br.close();
		} catch(Exception e) { }
		String[] lines = new String[v.size()];
		v.copyInto((String[])lines);
		IJ.showStatus("");
		return lines;
	}

	// Use reflection to get version since early versions
	// of ImageJ do not have the IJ.getVersion() method.
//	String version() {
//		String version = "";
//		try {
//			Class ijClass = ImageJ.class;
//			Field field = ijClass.getField("VERSION");
//			version = (String)field.get(ijClass);
//		}catch (Exception ex) {}
//		return version;
//	}

	boolean isMac() {
		String osname = System.getProperty("os.name");
		return osname.startsWith("Mac");
	}
	
	void error(String msg) {
		IJ.error("AstroImageJ Updater", msg);
	}
	
//	void updateMenus() {
//		if (IJ.debugMode) {
//			long start = System.currentTimeMillis();
//			Menus.updateImageJMenus();
//			IJ.log("Refresh Menus: "+(System.currentTimeMillis()-start)+" ms");
//		} else
//			Menus.updateImageJMenus();
//	}

}
