// FITS_Header_Editor.java

import ij.*;
import ij.process.*;
import ij.gui.*;
import ij.plugin.filter.*;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.table.*;
import java.util.*;
import java.util.Calendar;
import java.text.NumberFormat;

public class FITS_Header_Editor_Standalone implements PlugInFilter, ListSelectionListener, ActionListener, ItemListener
	{
	ImagePlus imp;
	String[][] header = null;
	String[][] rows = null;

	JTable table = null;
	JFrame frame = null;
    JPanel panel;
    JScrollPane scrollPane;
	JCheckBox keywordLockCB;

	public static String SAVE      = "SAVE";
    public static String SAVEFILE  = "SAVE FILE";
	public static String CANCEL    = "CANCEL";
	public static String DELETE    = "DELETE";
	public static String INSERT    = "INSERT";
	public static String VERSION   = "2012-05-01";

	boolean keywordLock = true;
	boolean changed = false;

    int frameX = -999999;
    int frameY = -999999;
    int frameWidth = 600;
    int frameHeight = 600;

    FitsJ fitsJ = new FitsJ();
    
	public int setup(String arg, ImagePlus imp)
		{
		this.imp = imp;
		return DOES_ALL;
		}

	public void run(ImageProcessor ip)
		{
        getPrefs();

		String[] hdr = fitsJ.getHeader(imp);
		int l = hdr.length;
		header = new String[l][5];
		for (int i=0; i < l; i++)
			{
			String card = hdr[i];
			String ctype = fitsJ.getCardType(card);

			header[i][0] = ""+(i+1);
			header[i][4] = ctype;

			if (ctype == "C" || ctype == "H")	// COMMENT OR HISTORY
				{
				header[i][1] = card.substring(0,7);
				header[i][2] = card.substring(8,card.length()).trim();
				header[i][3] = null;
				}
			else if (ctype == "S")
				{
				header[i][1] = fitsJ.getCardKey (card);
				header[i][2] = fitsJ.getCardValue (card);
				header[i][3] = fitsJ.getCardComment (card);
				}
			else if (ctype == "E")
				{
				header[i][1] = "END";
				header[i][2] = null;
				header[i][3] = null;
				}
			else
                {
				header[i][1] = fitsJ.getCardKey (card);
				header[i][2] = fitsJ.getCardValue (card);
				header[i][3] = fitsJ.getCardComment (card);
				}
			}

		// CREATE GUI

		frame = new JFrame ("FITS Header Editor");
        frame.setDefaultCloseOperation(javax.swing.WindowConstants.DISPOSE_ON_CLOSE);
        frame.addWindowListener(new WindowAdapter(){
            @Override
            public void windowClosing(WindowEvent e){
            savePrefs();}});

//		frame.add (new JLabel("Image: "+imp.getShortTitle()));

        buildTable();

		}


    public void buildTable()
        {
        boolean newTable = true;
        int col1size = 100;
        int col2size = 100;
        int col3size = 100;
        int col4size = 100;
        int col5size = 100;
        if (table != null)
            {
            newTable = false;
            }
        if (!newTable)
            {
            col1size = table.getColumnModel().getColumn(0).getWidth();
            col2size = table.getColumnModel().getColumn(1).getWidth();
            col3size = table.getColumnModel().getColumn(2).getWidth();
            col4size = table.getColumnModel().getColumn(3).getWidth();
            col5size = table.getColumnModel().getColumn(4).getWidth();
            frameWidth = frame.getWidth();
            frameHeight = frame.getHeight();
            frameX = frame.getX();
            frameY = frame.getY();
            frame.remove(panel);
            }
        panel = new JPanel(new BorderLayout());
		table = new JTable (new FITSTableModel());
		// table.setFillsViewportHeight(true);		Java 1.6
		table.setShowGrid(true);

        table.setColumnSelectionAllowed(false);


		ListSelectionModel model = table.getSelectionModel();
		model.addListSelectionListener(this);
		FontMetrics metrics = table.getFontMetrics(table.getFont());

        int w = metrics.stringWidth("M");
        if (newTable)
            {
            col1size = 5*w;
            col2size = 9*w;
            col3size = 50*w;
            col4size = 50*w;
            col5size = 5*w;
            }

		TableColumn col = table.getColumnModel().getColumn(0);
		col.setPreferredWidth(col1size);
		col = table.getColumnModel().getColumn(1);
		col.setPreferredWidth (col2size);
        col = table.getColumnModel().getColumn(2);
		col.setPreferredWidth (col3size);
        col = table.getColumnModel().getColumn(3);
		col.setPreferredWidth (col4size);
		col = table.getColumnModel().getColumn(4);
		col.setPreferredWidth (col5size);
		DefaultTableCellRenderer renderer = new DefaultTableCellRenderer();
		renderer.setToolTipText("<html>B = boolean<br>C = comment<br>E = end<br>H = history<br>I = integer<br>R = real<br>S = string<br>? = unknown</html>");
		col.setCellRenderer(renderer);
        table.setPreferredScrollableViewportSize(new Dimension(col1size+col2size+col3size+col4size+col5size,table.getRowHeight()*header.length>600 ? 600 : table.getRowHeight()*header.length));
        table.setAutoResizeMode(JTable.AUTO_RESIZE_OFF);
        table.getTableHeader().setReorderingAllowed(false);
		scrollPane = new JScrollPane (table);
		panel.add (scrollPane, BorderLayout.CENTER);

		// ADD BUTTONS

		JPanel gui = new JPanel();


		keywordLockCB = new JCheckBox("Lock keyword values", keywordLock);
        keywordLockCB.setToolTipText("Lock or unlock editing of the keyword field");
        keywordLockCB.addItemListener(this);
        gui.add (keywordLockCB);

		JButton deleterow = new JButton (DELETE);
        deleterow.setToolTipText("Delete selected row(s)");
		deleterow.addActionListener (this);
		gui.add (deleterow);

		JButton insertrow = new JButton (INSERT);
        insertrow.setToolTipText("Insert a row below the (first) selected row");
		insertrow.addActionListener (this);
		gui.add (insertrow);

		JButton save = new JButton (SAVE);
        save.setToolTipText("Save header changes to AstroImageJ memory and exit the editor");
		save.addActionListener (this);
		gui.add (save);

		JButton saveFile = new JButton (SAVEFILE);
        saveFile.setToolTipText("Save header changes and file to hard drive and exit the editor");
		saveFile.addActionListener (this);
		gui.add (saveFile);

		JButton cancel = new JButton (CANCEL);
        cancel.setToolTipText("Cancel changes and exit the editor");
		cancel.addActionListener (this);
		gui.add (cancel);

		// JButton cutcol = new JButton (CUT);
		// cutcol.addActionListener (this);
		// gui.add (cutcol);

		// JButton pastecol = new JButton (PASTE);
		// pastecol.addActionListener (this);
		// gui.add (pastecol);



        gui.setMinimumSize(new Dimension(10,400));
		panel.add (gui, BorderLayout.SOUTH);
		table.doLayout();
		frame.add (panel);

		frame.pack ();
        if (!newTable)
            {
            frame.setSize(frameWidth, frameHeight);
            }
        Dimension mainScreenSize = Toolkit.getDefaultToolkit().getScreenSize();
        if (!Prefs.isLocationOnScreen(new Point(frameX, frameY)))
            {
            frameX = mainScreenSize.width/2 - frame.getWidth()/2;
            frameY = mainScreenSize.height/2 - frame.getHeight()/2;
            }
        frame.setLocation(frameX, frameY);

		frame.setVisible (true);
    }


    public void itemStateChanged (ItemEvent e)
		{
		Object source = e.getItemSelectable();
		if (e.getStateChange() == ItemEvent.SELECTED)
			{
			if (source == keywordLockCB)
				keywordLock = true;
            }
        else
            {
			if (source == keywordLockCB)
				keywordLock = false;
            }
        }

	// ActionListener METHODS

	public void actionPerformed (ActionEvent e)
		{
		String cmd = e.getActionCommand();
		if (cmd.equals(CANCEL))
			{
			frame.setVisible (false);
            savePrefs();
            frame.dispose();
			frame = null;
			return;
			}
		else if (cmd.equals(SAVE))
			{
            if (table.getCellEditor() != null) //finalize the last edit if enter has not been pressed
                {
                table.getCellEditor().stopCellEditing();
                }

            String[] hdr = extractHeader();
            if (hdr == null)
                {
                IJ.showMessage ("Header edit error, changes not saved!");
                return;
                }

            fitsJ.putHeader(imp,hdr);

			frame.setVisible (false);
            savePrefs();
            frame.dispose();
			frame = null;
			return;
			}
		else if (cmd.equals(SAVEFILE))
			{
            if (table.getCellEditor() != null) //finalize the last edit if enter has not been pressed
                {
                table.getCellEditor().stopCellEditing();
                }

            String[] hdr = extractHeader();
            if (hdr == null)
                {
                IJ.showMessage ("Header edit error, changes not saved!");
                return;
                }

            fitsJ.putHeader(imp,hdr);
            IJ.saveAs(""+imp.getTitle().substring(imp.getTitle().lastIndexOf('.')),""+imp.getTitle());
			frame.setVisible (false);
            savePrefs();
            frame.dispose();
			frame = null;
			return;
			}
		else if (cmd.equals(DELETE))
			{
			int[] selectedrow = table.getSelectedRows();
            int numrows = selectedrow.length;
            if (numrows < 1)
                return;
            if (selectedrow[numrows-1]==header.length - 1)
                numrows--;
            if (numrows < 1)
                return;
            String[][] newHeader = new String[header.length-numrows][5];
            int numfound = 0;
            for (int i=0; i<header.length; i++)
                {
                if (numfound<numrows && i==selectedrow[numfound])
                    numfound++;
                else
                    {
                    newHeader[i-numfound][0] = ""+(i-numfound+1);
                    newHeader[i-numfound][1] = header[i][1];
                    newHeader[i-numfound][2] = header[i][2];
                    newHeader[i-numfound][3] = header[i][3];
                    newHeader[i-numfound][4] = header[i][4];
                    }
                }

            header = newHeader;
            buildTable();
            if (numrows==1) table.changeSelection(selectedrow[0],1,false,false);
			}
		else if (cmd.equals(INSERT))
			{
			int row = table.getSelectedRow();
			if (row < 0 || row == table.getRowCount()-1) row = table.getRowCount()-2;
            String[][] newHeader = new String[header.length+1][5];
            for (int j=0; j<5; j++)
                {
                for (int i=0; i <= row; i++) {newHeader[i][j] = header[i][j];}
                if (j==0) newHeader[row+1][j]=""+(row+2);
                else if (j==1 || j==2 || j==3) newHeader[row+2][j]="";
                else newHeader[row+1][j]=" ";
                for (int i=row+2; i < newHeader.length; i++)
                    {
                    if (j==0)
                        newHeader[i][j] = ""+(i+1);
                    else
                        newHeader[i][j] = header[i-1][j];
                    }
                }
            header = newHeader;
            buildTable();
            table.changeSelection(row+1,1,false,false);
			}
		}

	protected String[] extractHeader()
		{
		int n=0;
		int l = header.length;
		String[] hdr = new String[l];
		String key,val,comment;

		// COPY BASIC HEADER, CHECKING FOR 80 CHARACTER LIMIT AND FOR EMPTY FIELDS

		for (int i=0; i < l; i++)
			{
			if (header[i][1] != null && (header[i][2] != null || header[i][4].trim().equals("H") ||
                header[i][4].trim().equals("C") || header[i][4].trim().equals("E")))
				{
				key = header[i][1].trim();
                if (key.length() > 8)
                    {
                    IJ.showMessage("Row "+(i+1)+" 'Keyword' entry is too long: length must be <= 8 characters!");
                    table.changeSelection(i,0,false,false);
                    return null;
                    }
				while (key.length() < 8)
                    {
                    key += " ";
                    }
				val = header[i][2] != null ? header[i][2].trim() : "";

                // AKEYWORD=_VALUE_/_COMMENT
                // 1234567890     123		= 10 CHARS FOR KEY, > 8 FOR COMMENT
                if (val.length() > 70)
                    {
                    IJ.showMessage("Row "+(i+1)+" 'Value' entry is too long: length must be <= 70 characters!");
                    table.changeSelection(i,1,false,false);
                    return null;
                    }
                comment = "";
                if (val.length() < 67 && header[i][3] != null)
                    {
                    comment = " / "+header[i][3].trim();
                    if (comment.length() > (67-val.length()))
                        comment = comment.substring(0,67-val.length());
                    while ((val.length()+comment.length()) < 70) {val += " ";}
                    }
                else	{
                    while ((val.length()) < 70) {val += " ";}
                    }
                hdr[n++] = key+(header[i][4].equals("H")||header[i][4].equals("C")||header[i][4].equals("E") ? "" : "= ")+val+comment;
				}
			}

		// ADD A HISTORY ENTRY
//        hdr = FitsJ.addHistory("Header modified by AstroIJ FITS editor, Version "+VERSION,hdr);
		return hdr;

		}

	// ListListener METHODS

	public void valueChanged (ListSelectionEvent e)
		{
		int i1 = e.getFirstIndex();
		int i2 = e.getLastIndex();
		}

	double parseDouble(String s)
        {
        double defaultValue = Double.NaN;
		if (s==null) return defaultValue;
		try {
			defaultValue = Double.parseDouble(s);
		    }
        catch (NumberFormatException e) {defaultValue = Double.NaN;}
		return defaultValue;
        }

	Integer parseInteger(String s)
        {
        Integer defaultValue = Integer.MIN_VALUE;
		if (s==null) return defaultValue;
		try {
			defaultValue = Integer.parseInt(s);
		    }
        catch (NumberFormatException e) {defaultValue = Integer.MIN_VALUE;}
		return defaultValue;
        }

    void getPrefs()
        {
        frameX = (int) Prefs.get ("fitsedit.frameX",frameX);
        frameY = (int) Prefs.get ("fitsedit.frameY",frameY);
        keywordLock = Prefs.get ("fitsedit.editKeys",keywordLock);
        }

    void savePrefs() {
        Prefs.set ("fitsedit.frameX",frame.getLocation().x);
        Prefs.set ("fitsedit.frameY",frame.getLocation().y);
        Prefs.set ("fitsedit.editKeys",keywordLock);
        }

	/**
	 * A private TableModel for FITS data
	 */
	class FITSTableModel extends AbstractTableModel
		{
		private String[] columnNames = { "#","Keyword","Value","Comment","Type" };
        String keyword;
        String value;
        String comment;
        String type;
		public int getColumnCount() { return 5; }
		public int getRowCount() { return header.length; }
		public String getColumnName(int col) { return columnNames[col]; }
		public Object getValueAt(int row, int col) { return header[row][col]; }

		/*
		 * The 1st and last columns aren't editable, and the 2nd also if ....
		 */
		public boolean isCellEditable(int row, int col)
			{
			if (col == 0 || (col == 4))
				return false;
			else if (col == 3 && (header[row][4].equals("C") || header[row][4].equals("H")))  //comment column not editiable in comment or history card
				return false;
			else if (col == 1 && keywordLock && !header[row][4].equals(" "))
				return false;
            else if (row == header.length - 1 && header[row][4].equals("E"))  //don't allow user to edit last row if it contains END card
                return false;
			return true;
			}

		public void setValueAt(Object newvalue, int row, int col)
			{
			if (newvalue instanceof String)
				{
                keyword = col == 1 ? ((String)newvalue).toUpperCase().trim() : (header[row][1]==null ? "" : header[row][1].trim());
				value = col == 2 ? ((String)newvalue).trim() : (header[row][2]==null ? "" : header[row][2].trim());
                comment = col == 3 ? ((String)newvalue).trim() : header[row][3];
                type = header[row][4].trim();
                if (col == 1)
                    {
                    if (keyword.equalsIgnoreCase("END"))
                        {
                        keyword =  header[row][1].trim();
                        IJ.showMessage("Row "+(row+1)+" keyword value 'END' is not allowed!");
                        }
                    else if (!keyword.matches("[-A-Z0-9_]{"+keyword.length()+"}"))
                        {
                        keyword =  header[row][1].trim();
                        IJ.showMessage("Row "+(row+1)+" keyword '"+keyword+"' is not allowed! Valid characters are A-Z_0-9-");
                        }
                    else
                        keyword = keyword.substring(0, keyword.length() < 8 ? keyword.length() : 8).toUpperCase();
                    }
                getType();
				header[row][1] = keyword;
                header[row][2] = value;
                header[row][3] = comment;
                header[row][4] = type;
				changed = true;
				}
			fireTableRowsUpdated(row, row);
			}



        void getType()
            {
            if (keyword.equals(""))
                {
                type = " ";
                }
            else if (keyword.equals("COMMENT"))
                {
                type = "C";
                }
            else if (keyword.equals("HISTORY"))
                {
                type = "H";
                }
            else if ((value.startsWith("'") && value.endsWith("'")) || (value.startsWith("\"") && value.endsWith("\"")))
                {
                value = "'"+value.substring(1, value.length()-1)+"'";
                type = "S";
                }
            else if (value.equalsIgnoreCase("T") || value.equalsIgnoreCase("F"))
                {
                value = value.toUpperCase();
                type = "B";
                }
            else if (parseInteger(value)!=Integer.MIN_VALUE)
                {
                type = "I";
                }
            else if (!Double.isNaN(parseDouble(value.toUpperCase().replace("D", "E"))))
                {
                value = value.toUpperCase();
                type = "R";
                }
            else
                {
                if (value.startsWith("'") || value.startsWith("\"")) value = value.substring(1, value.length());
                if (value.endsWith("'") || value.endsWith("\"")) value = value.substring(0, value.length()-1);
                value = "'"+value+"'";
                type = "S";
                }
            }
//        public void insertRow()
//            {
//			int row = table.getSelectedRow();
//			if (row < 0) row = table.getRowCount()-1;
//
//
//            }


		}



/**
 * A collection of helpful static methods to read, manipulate, write, and query FITS-based images within ImageJ.
 * The header is manipulated as an array of Strings which can be read and written via the "Info" property
 * (a Properties object wouldn't keep the cards in the right order and can't be used with Stacks).
 */
public class FitsJ
	{
	 public int 	NO_CARD = -1;		// TYPES OF FITS CARDS
	 public int 	STRING_CARD = 0;
	 public int 	INTEGER_CARD = 1;
	 public int	DOUBLE_CARD = 2;
	 public int	BOOLEAN_CARD = 3;
	 public int 	COMMENT_CARD = 4;
	 public int 	HISTORY_CARD = 5;

	public  int 	KEY_PART = 0;		// PARTS OF A CARD PARSE BY THE cardParts() METHOD
	public  int 	STRING_PART= 1;
	public  int 	DOUBLE_PART = 2;
	public  int 	INTEGER_PART = 3;
	public  int 	BOOLEAN_PART = 4;
	public  int 	COMMENT_PART = 5;
	public  int 	TYPE_PART = 6;

	/**
	 * Extracts the original FITS header from the Properties object of the
	 * ImagePlus image (or from the slice label in the case of an ImageStack)
	 * and returns it as a String object representing all card.
	 *
	 * @param img		The ImagePlus image which has the FITS header in it's "Info" property.
	 */
	public  String getHeaderString (ImagePlus img)
		{
		String content = null;

		int depth = img.getStackSize();
		if (depth == 1)
			{
			Properties props = img.getProperties();
			if (props == null)
				return null;
			content = (String)props.getProperty ("Info");
			}
		else if (depth > 1)
			{
			int slice = img.getCurrentSlice();
			ImageStack stack = img.getStack();
			content = stack.getSliceLabel(slice);
			}
		return content;
		}

	/**
	 * Extracts the original FITS header from the Properties object of the
	 * ImagePlus image (or from the slice label in the case of an ImageStack)
	 * and returns it as an array of String objects representing each card.
	 *
	 * @param img		The ImagePlus image which has the FITS header in it's "Info" property.
	 */
	public  String[] getHeader (ImagePlus img)
		{
		String content = getHeaderString(img);
		if (content == null) return null;

		// PARSE INTO LINES

		String[] lines = content.split("\n");

		// FIND "SIMPLE" AND "END" KEYWORDS

		int istart = 0;
		for (; istart < lines.length; istart++)
			{
			if (lines[istart].trim().startsWith("SIMPLE") ) break;
			}
		if (istart == lines.length) return null;

		int iend = istart+1;
		for (; iend < lines.length; iend++)
			{
			String s = lines[iend].trim();
			if ( s.equals("END") || s.startsWith ("END ") ) break;
			}
		if (iend >= lines.length) return null;

		int l = iend-istart+1;
		String header = "";
		for (int i=0; i < l; i++)
			header += lines[istart+i]+"\n";
		return header.split("\n");
		}

	/**
	 * Checks to make sure the FITS header agrees with the ImageJ image parameters.
	 *
	 * @param img			The ImagePlus image.
	 * @param hdr			The FITS header stored in a String array.
	 */
	public  boolean isConsistent (ImagePlus img, String[] hdr)
		{
		int nx = img.getWidth();
		int ny = img.getHeight();
		int nz = img.getStackSize();
		int n = 2;
		if (nz > 1) n++;
		int naxis=0;
		int naxis1=0;
		int naxis2=0;
		int naxis3=0;
		try	{
			naxis  = findIntValue ("NAXIS",hdr);
			naxis1 = findIntValue ("NAXIS1",hdr);
			naxis2 = findIntValue ("NAXIS2",hdr);
			if (nz > 1) naxis3 = findIntValue ("NAXIS3",hdr);
			}
		catch (NumberFormatException e)
			{
			return false;
			}
		if (n != naxis) return false;
		if (nx != naxis1) return false;
		if (ny != naxis2) return false;
		if (nz > 1 && nz != naxis3) return false;
		return true;
		}

	/**
	 * Insures that the FITS header agrees with the ImageJ image parameters.
	 *
	 * @param img			The ImagePlus image.
	 * @param hdr			The FITS header stored in a String array.
	 */
	public  void makeConsistent (ImagePlus img, String[] hdr)
		{
		int nx = img.getWidth();
		int ny = img.getHeight();
		int nz = img.getStackSize();
		int n = 2;
		if (nz > 1) n++;
		int naxis=0;
		int naxis1=0;
		int naxis2=0;
		int naxis3=0;
		try	{
			naxis  = findIntValue ("NAXIS",hdr);
			naxis1 = findIntValue ("NAXIS1",hdr);
			naxis2 = findIntValue ("NAXIS2",hdr);
			if (nz > 1) naxis3 = findIntValue ("NAXIS3",hdr);
			}
		catch (NumberFormatException e)
			{
			IJ.beep(); IJ.beep(); IJ.beep();
			IJ.log("Error making FITS image header consistent with ImageJ image properties!");
			return;
			}
		setCard ("NAXIS",n,null,hdr);
		setCard ("NAXIS1",nx,null,hdr);
		setCard ("NAXIS2",ny,null,hdr);
		if (nz > 1)
			setCard ("NAXIS3",nz,null,hdr);
		}

	/**
	 * Sets a FITS card with an integer keyword
	 *
	 * @param key			The name of the FITS keyword (e.g. "NAXIS3").
	 * @param property		The integer value corresponding to this keyword.
	 * @param comment		The FITS comment string for this keyword.
	 * @param cards			The FITS header as an array of Strings.
	 */
	public  String[] setCard (String key, int property, String comment, String[] cards)
		{
		return set (key,""+property,comment, cards);
		}

	/**
	 * Sets a FITS card with a double keyword.
	 *
	 * @param key			The name of the FITS keyword (e.g. "JD").
	 * @param property		The double value corresponding to this keyword.
	 * @param comment		The FITS comment string for this keyword.
	 * @param cards			The FITS header as an array of Strings.
	 */
	public  String[] setCard (String key, double property, String comment, String[] cards)
		{
		return set (key,""+property,comment, cards);
		}

	/**
	 * Sets a FITS card with a boolean keyword.
	 *
	 * @param key			The name of the FITS keyword (e.g. "EXTENDED").
	 * @param property		The boolean value corresponding to this keyword.
	 * @param comment		The FITS comment string for this keyword.
	 * @param cards			The FITS header as an array of Strings.
	 */
	public  String[] setCard (String key, boolean property, String comment, String[] cards)
		{
		if (property)
			return set (key,"T       ",comment, cards);
		else
			return set (key,"F       ",comment, cards);
		}

	/**
	 * Sets a FITS card with a string keyword.
	 *
	 * @param key			The name of the FITS keyword (e.g. "DATE-OBS").
	 * @param property		The string value corresponding to this keyword.
	 * @param comment		The FITS comment string for this keyword.
	 * @param cards			The FITS header as an array of Strings.
	 */
	public  String[] setCard (String key, String property, String comment, String[] cards)
		{
		return set (key,"'"+property+"'",comment, cards);
		}

	/**
	 * Creates a FITS card in the form of a String.
	 *
	 * @param key			The name of the FITS keyword.
	 * @param property		The value corresponding to this keyword.
	 * @param comment		The FITS comment string for this keyword.
	 */
	protected  String createCard (String key, String val, String comment)
		{
		String k = key.trim();

		String sp = "";
		for (int l=k.length(); l < 8; l++)
			sp += " ";
		String card = k+sp+"= "+val;
		if (comment != null)
			card += " / "+comment.trim();
		return card;
		}

	/**
	 * Sets a FITS card in the form of a String to an ImagePlus image.
	 *
	 * @param key			The name of the FITS keyword.
	 * @param property		The value corresponding to this keyword.
	 * @param comment		The FITS comment string for this keyword.
	 * @param cards			The FITS header as an Array of strings.
	 */
	protected  String[] set (String key, String val, String comment, String[] cards)
		{
		if (cards == null) return null;
		String k = key.trim();
		String card = null;

		// GET OLD VALUE AND COMMENT FROM CARD CONTAINING KEYWORD

		int icard = findCardWithKey (k,cards);
		if (icard >= 0)
			card = cards[icard];
		String old = getCardValue(card);
		String comm = getCardComment (card);

		// IF THERE'S A NEW OR OLD COMMENT, USE IT

		String sp = "";
		for (int l=k.length(); l < 8; l++)
			sp += " ";
		String v = k+sp+"= "+val;
		if (comment != null)
			v += " / "+comment.trim();
		else if (comm != null)
			v += " / "+comm;

		// SAVE NEW HEADER

		if (icard >= 0)
			cards[icard] = v;
		else
			cards = addCard (v,cards);
		return cards;
		}

	/**
	 * Saves a FITS header in an array of Strings back into an ImagePlus's "Info" property.
	 */
	public  void putHeader (ImagePlus img, String[] cards)
		{
		String s = unsplit(cards,"\n");

		int depth = img.getStackSize();
		if (depth == 1)
			img.setProperty ("Info", s);
		else if (depth > 1)
			{
			int slice = img.getCurrentSlice();
			ImageStack stack = img.getStack();
			String label = stack.getShortSliceLabel(slice);
			stack.setSliceLabel(label+"\n"+s, slice);
			}
		}

	/**
	 * Saves a FITS header in an array of Strings back into an ImageStack's current slice.
	 */
	public  void putHeader (ImageStack stack, String[] cards, int slice)
		{
		String s = unsplit(cards,"\n");
		String label = stack.getShortSliceLabel(slice);
		stack.setSliceLabel(label+"\n"+s, slice);
		}

	/**
	 * Adds a FITS card to the FITS header stored in a String array.  The resulting modified header is returned.
	 *
	 * @param card		A FITS card image to be added to the FITS header String array.
	 * @param cards		A String array holding the contents of a FITS header.
	 */
	public  String[] addCard (String card, String[] cards)
		{
		int l = cards.length;
		String[] hdr = new String[l+1];
		for (int i=0; i < l; i++)
			hdr[i] = new String (cards[i]);
		hdr[l-1] = new String (card);
		hdr[l] = "END";
		return hdr;
		}

	/**
	 * Inserts a blank card at a particular position.
	 *
	 * @param cards			The array of strings containing the FITS header.
	 * @param pos			The card position where a new blank card should be inserted.
	 */
	public  String[] insertBlankCard (String[] cards, int pos)
		{
		int l = cards.length;
		if (pos == l-1 && cards[l-1].equals("END"))
			return addCard ("",cards);

		String[] hdr = new String[l+1];
		for (int i=0; i < pos; i++)
			hdr[i] = new String (cards[i]);
		hdr[pos] = "";
		for (int i=pos+1; i <= l; i++)
			hdr[i] = new String (cards[i-1]);
		return hdr;
		}

	/**
	 * Removes all FITS cards with the given key from the FITS header stored in a String array.
	 *
	 * @param key		A FITS card keyword for those cards to be removed from the FITS header String array.
	 * @param cards		A String array holding the contents of a FITS header.
	 */
	public  String[] removeCards (String key, String[] cards)
		{
		int l = cards.length;
		String[] hdr = new String[l+1];
		for (int i=0; i < l; i++)
			{
			if (! getCardKey (cards[i]).equals(key))
				hdr[i] = new String (cards[i]);
			}
		return hdr;
		}

	/**
	 * Finds the location of a FITS card in a String array having the FITS keyword "key".
	 *
	 * @param header	A String array holding the contents of a FITS header.
	 * @param key		A String containing the FITS keyword to be searched for.
	 */
	public  int findCardWithKey (String key, String[] header)
		{
		if (key == null) return -1;

		int n=header.length;
		String k = key.trim();
		for (int i=0; i < n; i++)
			{
			String[] l = header[i].trim().split("=");
			if (l.length > 0 && l[0].trim().equals(k)) return i;
			}

		return -1;
		}

	/**
	 * Adds a comment to the image's FITS header.  The resulting modified header is returned.
	 *
	 * @param img			The image whose comments are to be extended.
	 * @param comment		The FITS comment string.
	 */
	public  String[] addComment (String comment, String[] cards)
		{
		if (cards == null) return null;

		String[] parts = segmentString(comment,68);
		if (parts.length > 0)
			cards = addCard (pad("COMMENT "+parts[0],80),cards);
		for (int i=1; i < parts.length; i++)
			cards = addCard (pad("COMMENT ... "+parts[i],80),cards);
		return cards;
		}

	/**
	 * Adds a comment to the image's FITS header after a particular position.  The resulting modified header is returned.
	 *
	 * @param img			The image whose comments are to be extended.
	 * @param comment		The FITS comment string.
	 * @param pos			The card number of the previous card.
	 */
	public  String[] addCommentAfter (String comment, String[] cards, int pos)
		{
		if (cards == null || pos < 0 || pos >= cards.length) return null;

		String[] parts = segmentString(comment,68);	// len("COMMENT ... ")+68=80
		String[] hdr = insertBlankCard (cards, pos+1);
		hdr[pos+1] = pad("COMMENT "+parts[0],80);
		for (int i=1; i < parts.length; i++)
			{
			hdr = insertBlankCard (hdr, pos+1+i);
			hdr[pos+1+i] = pad("COMMENT ... "+parts[i],80);
			}
		return hdr;
		}

	/**
	 * Adds a FITS history card to the image's FITS header.  The resulting modified header is returned.
	 *
	 * @param img			The ImagePlus whose history entries are to be extended.
	 * @param comment		The FITS history string.
	 */
	public  String[] addHistory (String history, String[] cards)
		{
		if (cards == null) return null;

		String[] parts = segmentString(history,68);	// len("HISTORY ... ")+68 = 80
		if (parts.length > 0)
			{
			// IJ.log(parts[0]);
			cards = addCard (pad("HISTORY "+parts[0],80),cards);
			}
		for (int i=1; i < parts.length; i++)
			{
			// IJ.log(parts[i]);
			cards = addCard (pad("HISTORY ... "+parts[i],80),cards);
			}
		return cards;
		}

	/**
	 * Tells the type of FITS value in a card:
	 *	"B"=BOOLEAN,
	 *	"C"=COMMENT,
	 *	"E"=END,
	 *	"H"=HISTORY,
	 *	"I"=INTEGER NUMBER,
	 *	"R"=REAL NUMBER,
	 *	"S"=STRING,
	 *	" "=BLANK
	 *	"?"=OTHER
	 *
	 * @param card		The FITS card image from which the comment should be extracted.
	 */
	public  String getCardType (String card)
		{
		if (card == null) return null;
		int equals = -1;
		if (card.substring(0,card.length() < 8 ? card.length() : 8).trim().equals(""))
			return " ";
		if (card.startsWith("COMMENT"))
			return "C";
		if (card.startsWith("HISTORY"))
			return "H";
		if (card.trim().equals("END"))
			return "E";
		equals = card.indexOf("=");
		if (equals < 0)
			return "?";
		String s = card.substring(equals).trim();
		String val = getCardValue(card).trim();
		if (val.startsWith("'") && val.endsWith("'"))
			return "S";
		if (val.startsWith("\"") && val.endsWith("\""))
			return "S";
		if (val.trim().equals("T") || val.trim().equals("F"))
			return "B";
		if (val.indexOf(".") < 0)
			return "I";
		if (val.indexOf(".") >= 0)
			return "R";
		return "X";
		}

	/**
	 * Extracts the FITS keyword from a card.
	 *
	 * @param card		The FITS card image from which the comment should be extracted.
	 */
	public  String getCardKey (String card)
		{
		if (card == null) return null;
		int equals = -1;
		if (card.startsWith("COMMENT") || card.startsWith("HISTORY"))
			return null;
		equals = card.indexOf("=");
		if (equals < 0)
			return null;		// NO VALUE (e.g. COMMENT?)
		return card.substring(0,equals).trim();
		}

	/**
	 * Extracts the FITS value for a given keyword from a FITS image with a given name.  This special routine is meant to be used
	 * within ImageJ macros via the
	 *
	 *	exptime = call ("FitsJ.getCardFromImage","EXPTIME","myimage.fits");
	 *
	 * @param key		The FITS keyword.
	 * @param filename	The name of the file used by ImageJ (CAUTION: may be bereft of a file suffix or the name of a stack's directory!).
	 */
	public  String getCardValueFromImage (String key, String image)
		{
		ImagePlus img = null;
		if (image == null || image.equals(""))
			img = WindowManager.getCurrentImage();
		else
			img = WindowManager.getImage(image);
		if (img == null)
			{
			IJ.log("FitsJ.getCardValueFromFile: no image "+image);
			return null;
			}

		String[] cards = getHeader(img);
		if (cards == null)
			{
			IJ.log("FitsJ.getCardValueFromFile: no FITS header for "+image);
			return null;
			}

		int icard = findCardWithKey (key, cards);
		if (icard < 0) return null;

		String ctype = getCardType(cards[icard]);
		if (ctype.equals("S"))
			return getCardStringValue(cards[icard]);
		else
			return getCardValue(cards[icard]);
		}

	/**
	 * Sets the FITS value of a given keyword to an image with a given name.  This special routine is meant to be used
	 * within ImageJ macros via the
	 *
	 *	call ("FitsJ.setCardOfImage","EXPTIME","1.2345","comment","double","myimage.fits");
	 *
	 * @param key		The FITS keyword.
	 * @param filename	The name of the file used by ImageJ (CAUTION: may be bereft of a file suffix or the name of a stack's directory!).
	 */
	public  void setCardOfImage (String key, String val, String comment, String typ, String image)
		{
		ImagePlus img = null;
		if (image == null || image.equals(""))
			img = WindowManager.getCurrentImage();
		else
			img = WindowManager.getImage(image);
		if (img == null)
			{
			IJ.log("FitsJ.setCardOfImage: no image "+image);
			return;
			}

		String[] cards = getHeader(img);
		if (cards == null)
			{
			IJ.log("FitsJ.setCardOfImage: no FITS header for "+image);
			return;
			}
		if (typ.equals("string"))
			setCard(key,val,comment,cards);
		else if (typ.equals("boolean"))
			{
			if (val.startsWith("T") || val.startsWith("t"))
				setCard(key,true,comment,cards);
			else if (val.startsWith("F") || val.startsWith("f"))
				setCard(key,false,comment,cards);
			}
		else if (typ.equals("integer"))
			{
			try	{
				setCard(key,Integer.parseInt(val),comment,cards);
				}
			catch (NumberFormatException e)
				{
				}
			}
		else if (typ.equals("double") || typ.equals("float") || typ.startsWith("real"))
			{
			try	{
				setCard(key,Double.parseDouble(val),comment,cards);
				}
			catch (NumberFormatException e)
				{
				}
			}
		putHeader(img,cards);
		}

	/**
	 * Replaces keyword of a FITS card.
	 *
	 * @param newkey	The new FITS keyword.
	 * @param card		The FITS card image from which the key should be exchanged.
	 */
	public  String replaceKey (String key, String card)
		{
		if (key == null || card == null) return null;
		String val = getCardValue(card);
		String comment = getCardComment(card);
		return createCard(key,val,comment);
		}

	/**
	 * Extracts the FITS value from a card
	 *
	 * @param card		The FITS card image from which the comment should be extracted.
	 */
	public  String getCardValue (String card)
		{
		if (card == null) return null;

		int q1 = -1;
		int q2 = -1;
		int slash = -1;
		int equals = -1;

		// LOOK FOR CARDS WITH NO KEYWORD VALUE PAIR

		if (card.startsWith("COMMENT") || card.startsWith("HISTORY"))
			return card.substring(8);

		// FIND EQUALS SIGN

		equals = card.indexOf("=");
		if (equals < 0) return null;		// NO VALUE (e.g. COMMENT?)

		// LOOK FOR QUOTED VALUE

		q1 = card.indexOf("'");
		if (q1 >= 0)
			{
			q2 = card.indexOf("'",q1+1);
			if (q2 < 0) q1 = -1;
			}

		// OR FOR A DOUBLE QUOTED VALUE

		if (q1 < 0)
			{
			q1 = card.indexOf("\"");
			if (q1 > 0)
				{
				q2 = card.indexOf("\"",q1+1);
				if (q2 < 0) q1 = -1;
				}
			}

		// LOOK FOR COMMENT

		slash = card.indexOf("/");
// IJ.log(""+card);
// IJ.log("\tq1="+q1+", q2="+q2+", slash="+slash);

		if (slash < 0)			// NO COMMENT PRESENT, RETURN EVERYTHING RIGHT OF '='
			{
			if (q2 > 0)		// AS VALUE IN QUOTES
				return card.substring (q1,q2+1);
			else			// AS UNQUOTED VALUE
				return card.substring (equals+1).trim();
			}
		else if (q1 > 0 && slash < q2)		// MATCHING QUOTES IN VALUE, RETURN STRING WITH QUOTES
			return card.substring (q1,q2+1);

		// NO MATCHING QUOTES PRESENT, RETURN STRING BETWEEN '=' AND SLASH AS COMMENT
		return card.substring (equals+1,slash).trim();
		}

	/**
	 * Extracts the FITS comment from a card, including something like    "DATE = '12/34/56' / A date."
	 *
	 * @param card		The FITS card image from which the comment should be extracted.
	 */
	public  String getCardComment (String card)
		{
		if (card == null) return null;

		if (card.startsWith("COMMENT") || card.startsWith("HISTORY"))
			return null;

		int q1 = -1;
		int q2 = -1;
		int slash = -1;
		int equals = -1;

		// FIND EQUALS SIGN

		equals = card.indexOf("=");
		if (equals < 0) return null;		// NO VALUE (e.g. COMMENT?)

		// LOOK FOR SIMPLE QUOTE IN VALUE

		q1 = card.indexOf("'");
		if (q1 >= 0)
			q2 = card.indexOf("'",q1+1);

		// OR FOR A DOUBLE QUOTE IN VALUE

		if (q1 < 0)
			{
			q1 = card.indexOf("\"");
			if (q1 >= 0)
				q2 = card.indexOf("\"",q1+1);
			}

		// LOOK FOR COMMENT

		slash = card.indexOf("/");

		if (slash < 0)			// NO COMMENT PRESENT
			return null;
		else if (q2 < 0)		// NO MATCHING QUOTES PRESENT, RETURN STRING RIGHT OF SLASH AS COMMENT
			return card.substring (slash+1);
		else if (slash > q2)		// MATCHING QUOTES IN VALUE, RETURN STRING RIGHT OF SLASH AS COMMENT
			return card.substring (slash+1);

		slash = card.indexOf("/",q2+1);
		return card.substring (slash+1);
		}

	/**
	 * Extracts the double value from a FITS card.
	 *
	 * @param card		The FITS card image from which the value should be extracted.
	 * @changes		Accepts double values within strings (2009-01-10).
	 */
	public  double getCardDoubleValue (String card) throws NumberFormatException
		{
		double d = Double.NaN;
		String s = getCardValue (card).trim();
		if (s == null)
			return Double.NaN;
		if (s.startsWith("\'") || s.startsWith("\""))
			d = Double.parseDouble(s.substring(1,s.length()-1));
		else
			d = Double.parseDouble(s);
		return d;
		}

	/**
	 * Extracts a string value from a FITS card.
	 *
	 * @param card		The FITS card image from which the value should be extracted.
	 */
	public  String getCardStringValue (String card)
		{
		String s = getCardValue (card).trim();
		int l = s.length();
		if (s.startsWith("\"") && s.endsWith("\""))
			return s.substring(1,l-1);
		if (s.startsWith("'") && s.endsWith("'"))
			return s.substring(1,l-1);
		return s;
		}

	/**
	 * Extracts an int value from a FITS card.
	 *
	 * @param card		The FITS card image from which the value should be extracted.
	 */
	public  int getCardIntValue (String card) throws NumberFormatException
		{
		int i = 0;
		String s = getCardValue (card);
		i = Integer.parseInt(s);
		return i;
		}

	/**
	 * Extracts a boolean value from a FITS card.
	 *
	 * @param card		The FITS card image from which the value should be extracted.
	 */
	public  boolean getCardBooleanValue (String card)
		{
		String s = getCardValue (card);
		if (s.equals("T"))
			return true;
		else
			return false;
		}

	/**
	 * Finds and extracts a double value from a FITS header stored in a String array.
	 *
	 * @param key		The FITS keyword that should be found and parsed.
	 * @param cards		The FITS header.
	 */
	public  double findDoubleValue (String key, String[] cards) throws NumberFormatException
		{
		double d = Double.NaN;
		int icard = findCardWithKey (key, cards);
		if (icard < 0) return d;
		return getCardDoubleValue (cards[icard]);
		}

	/**
	 * Finds and extracts an integer value from a FITS header stored in a String array.
	 *
	 * @param key		The FITS keyword that should be found and parsed.
	 * @param cards		The FITS header.
	 */
	public  int findIntValue (String key, String[] cards) throws NumberFormatException
		{
		int icard = findCardWithKey (key, cards);
		if (icard < 0) throw new NumberFormatException();
		return getCardIntValue (cards[icard]);
		}

	/**
	 * Expands a Properties (key,val) pair into proper FITS format.
	 *
	 * @param key		The property key, which may contain a prefix showing that it is a FITS entry.
	 * @param val		The string value.
	 */
	protected  String expandKeyValue (String key, String val)
		{
		if (key == null || val == null) return null;

		int l = val.length();
		if (l > 70) l=70;
		String v = val.substring(0,l);

		l = key.length();
		String k = key;
		if (key.startsWith ("COMMENT"))
			return new String ("COMMENT "+v);
		else if (key.startsWith ("HISTORY"))
			return new String ("HISTORY "+v);
		while (l++ < 8) k += " ";
		return new String (k+"= "+v.trim());
		}

	/**
	 * Extracts the different parts of a FITS header card.
	 *
	 * @param card		A FITS card image.
	 */
	public  Object[] cardParts (String card)
		{
		String key = null;
		String val = null;
		String comment = null;
		double d = 0.0;
		int i = 0;
		boolean b = false;
		int typ = NO_CARD;

		// System.err.println("card="+card);

		String s = new String(card);

		// COMMENT

		if (card.startsWith ("COMMENT"))
			{
			key ="COMMENT";
			val = card.substring (7);
			comment = null;
			typ = COMMENT_CARD;
			}

		// HISTORY

		else if (card.startsWith ("HISTORY"))
			{
			key = "HISTORY";
			val = card.substring (7);
			comment = null;
			typ= HISTORY_CARD;
			}

		else
			{
			int eq = s.indexOf ("=");
			// System.err.println("eq="+eq);
			if (eq < 0) return null;
			key = s.substring (0,eq);
			// System.err.println("key="+key);
			if (key == null) return null;
			val = s.substring (eq+1);
			// System.err.println("val="+val);

			// COMMENT

			comment = getCardComment (s.substring(eq+1));
			// System.err.println ("comment="+comment);
			if (comment != null && !comment.equals(""))
				{
				int slash = s.indexOf (comment);
				// System.err.println ("slash=["+s.substring(slash-1,slash+1)+"]");
				val = s.substring (eq+1,slash-1).trim();
				// System.err.println ("val=["+val+"]");
				}

			// STRING

			if (val.startsWith("\'") || val.startsWith("\""))
				{
				s = val;
				val = s.substring(1,s.length()-1);
				// System.err.println ("val=["+val+"]");
				typ = STRING_CARD;
				}

			// BOOLEAN

			else if (val.equals("T") || val.equals("F"))
				{
				b = val.equals("T");
				typ = BOOLEAN_CARD;
				}

			// INTEGER OR DOUBLE

			else
				{
				try	{
					i = Integer.parseInt(val);
					typ = INTEGER_CARD;
					}
				catch (NumberFormatException e)
					{
					try	{
						d = Double.parseDouble(val);
						typ = DOUBLE_CARD;
						}
					catch (NumberFormatException nfe)
						{
						typ = NO_CARD;
						}
					}
				}
			}

		Object[] arr = new Object[]
				{key, val, new Double(d), new Integer(i), new Boolean(b), comment, new Integer(typ)};
		return arr;
		}

	/**
	 * Copies a FITS header from one image to another.
	 */
	public  void copyHeader (ImagePlus imFrom, ImagePlus imTo)
		{
		String[] hdrFrom = getHeader (imFrom);
		if (hdrFrom != null)
			{
			hdrFrom = addHistory ("Complete FITS header copied from image "+imFrom.getShortTitle()+", slice "+imFrom.getCurrentSlice(), hdrFrom);
			putHeader (imTo, hdrFrom);
			}
		}

	/**
	 * Copies some cards from one image's FITS header to another.
	 */
	public  void copyHeader (ImagePlus imFrom, ImagePlus imTo, String[] keys)
		{
		// GET HEADERS

		String[] hdrFrom = getHeader (imFrom);
		String[] hdrTo = getHeader (imTo);

		// TRANSFER CARDS WITH THE SELECTED KEYWORDS

		hdrTo = addHistory ("FITS header cards copied from image "+imFrom.getShortTitle()+", slice "+imFrom.getCurrentSlice(), hdrTo);
		for (int i=0; i < keys.length; i++)
			{
			int icard = findCardWithKey (keys[i], hdrFrom);
			if (icard >= 0)
				hdrTo = addCard (hdrFrom[icard], hdrTo);
			}

		// SAVE RESULTING HEADER

		putHeader (imTo,hdrTo);
		}

	/**
	 * Pads a string to a given total length.
	 *
	 * @param s		The input string.
	 * @param length	The length to which the string should be padded.
	 */
	public  String pad (String s, int length)
		{
		int l = s.length();
		if (l >= length)
			return s.substring(0,length);
		else	{
			String blanks="";
			while (l++ < length) blanks += " ";
			return s+blanks;
			}
		}

	/**
	 * Separates a string into segements of a given length.
	 *
	 * @param s		The input string.
	 * @param length	The desired maximum length
	 */
	public  String[] segmentString (String s, int length)
		{
		if (s == null) return null;
		int l = s.length();
		int n = l/length;
		if (n == 0)
			n = 1;
		else if (n*length < l)
			n++;
		String[] arr = new String[n];
		int m1=0;
		for (int i=0; i < n; i++)
			{
			int m2 = m1+length;
			if (m2 >= l) m2=l;
			arr[i] = s.substring(m1,m2);
			m1 += length;
			}
		return arr;
		}

	/**
	 * Unsplits a string.
	 *
	 * @param arr		A String array
	 * @param sep		The desired separator.
	 */
	public  String unsplit (String[] arr, String sep)
		{
		String s = arr[0];
		for (int i=1; i < arr.length; i++)
			s += sep+arr[i];
		return s;
		}


	/********************************** DATE, TIME, JD ROUTINES ***********************************************/


	/**
	 * Extracts a DateTime string either from an explict DateTime entry or builds one from
	 * separate date and time entries.
	 */
	public  String getDateTime (String[] cards)
		{
		String dt = getExplicitDateTime (cards);
		if (dt != null) return dt;

		String date = getDate (cards);
		if (date == null) return null;
		String time = getTime (cards);
		if (time == null) return null;
		dt = date+"T"+time;
		return dt;
		}

	/**
	 * Extracts explicit DateTime string from the FITS "DATE-OBS" entry.
	 */
	public  String getExplicitDateTime (String[] cards)
		{
		String datum = getDateObs(cards);
		if (datum == null) return null;

		// MAKE SURE IT'S REALLY AN ISO DATETIME WITH yyyy-{m}m-{d}dT{hh:mm:ss}

		int i = datum.indexOf("T");
		int j = datum.indexOf("-");

		if (i > 7 && j == 4)
			return datum;
		return null;
		}

	/**
	 * Extracts calendar date from the FITS header stored in a String array.
	 */
	public  String getDateObs (String[] cards)
		{
		String dateobs = null;

		// TRY "DATE-OBS"

		int icard = findCardWithKey ("DATE-OBS", cards);
		if (icard > 0)
			dateobs = getCardStringValue (cards[icard]);

		// TRY "DATEOBS"

		if (dateobs == null)
			{
			icard = findCardWithKey ("DATEOBS", cards);
			if (icard > 0)
				dateobs = getCardStringValue (cards[icard]);
			}

		// TRY "DATE_OBS"

		if (dateobs == null)
			{
			icard = findCardWithKey ("DATE_OBS", cards);
			if (icard > 0)
				dateobs = getCardStringValue (cards[icard]);
			}

		return dateobs;
		}

	/**
	 * Extracts calendar date from the FITS header stored in a String array.
	 */
	public  String getDate (String[] cards)
		{
		String datum = getDateObs (cards);
		if (datum == null) return null;

		// RE-ARRANGE INTO ISO FORMAT

		String dt="";

		// CHECK FOR dd/mm/yy

		if (datum.length() == 8 && datum.charAt(2) == '/' && datum.charAt(5) == '/')
			dt = new String("19"+datum.substring(6,8)+"-"+datum.substring(3,5)+"-"+datum.substring(1,3));

		// CHECK FOR dd/mm/yyyy

		else if (datum.length() == 10 && datum.charAt(2) == '/' && datum.charAt(5) == '/')
			dt = new String(datum.substring(6,10)+"-"+datum.substring(3,5)+"-"+datum.substring(1,3));

		// CHECK FOR yyyy-mm-dd

		else if (datum.length() == 10 && datum.charAt(4) == '-' && datum.charAt(7) == '-')
			dt = new String(datum.substring(0,4)+"-"+datum.substring(5,7)+"-"+datum.substring(8,10));

		// CHECK FOR yy-mm-dd

		else if (datum.length() == 8 && datum.charAt(2) == '-' && datum.charAt(5) == '-')
			dt = new String("19"+datum.substring(0,2)+"-"+datum.substring(3,5)+"-"+datum.substring(6,8));

		// OR GIVE UP

		else
			{
			IJ.log ("Unable to parse date "+datum);
			return null;
			}
		return dt;
		}

	/**
	 * Extracts UT Time from a FITS header in the form of a String array.
	 */
	public  String getTimeObs (String[] cards)
		{
		String timeobs = null;

		// TRY "TIME-OBS"

		int icard = findCardWithKey ("TIME-OBS", cards);
		if (icard > 0)
			timeobs = getCardStringValue (cards[icard]);

		// TRY "TIMEOBS"

		if (timeobs == null)
			{
			icard = findCardWithKey ("TIMEOBS", cards);
			if (icard > 0)
				timeobs = getCardStringValue (cards[icard]);
			}

		// TRY "TIME_OBS"

		if (timeobs == null)
			{
			icard = findCardWithKey ("TIME_OBS", cards);
			if (icard > 0)
				timeobs = getCardStringValue (cards[icard]);
			}

		// OR EXTRACT FROM "TM-START"

		if (timeobs == null)
			{
			icard = findCardWithKey ("TM-START", cards);
			if (icard > 0)
				timeobs = getCardStringValue (cards[icard]);
			}

		// OR EXTRACT FROM "TM_START"

		if (timeobs == null)
			{
			icard = findCardWithKey ("TM_START", cards);
			if (icard > 0)
				timeobs = getCardStringValue (cards[icard]);
			}

		// OR EXTRACT FROM "UT"

		if (timeobs == null)
			{
			icard = findCardWithKey ("UT", cards);
			if (icard > 0)
				timeobs = getCardStringValue (cards[icard]);
			}

		// OR EXTRACT FROM "UTC"

		if (timeobs == null)
			{
			icard = findCardWithKey ("UTC", cards);
			if (icard > 0)
				timeobs = getCardStringValue (cards[icard]);
			}

		// OR EXTRACT FROM "UTSTART"

		if (timeobs == null)
			{
			icard = findCardWithKey ("UTSTART", cards);
			if (icard > 0)
				timeobs = getCardStringValue (cards[icard]);
			}

		// OR EXTRACT FROM "UT-START"

		if (timeobs == null)
			{
			icard = findCardWithKey ("UT-START", cards);
			if (icard > 0)
				timeobs = getCardStringValue (cards[icard]);
			}

		// OR EXTRACT FROM "UT_START"

		if (timeobs == null)
			{
			icard = findCardWithKey ("UT_START", cards);
			if (icard > 0)
				timeobs = getCardStringValue (cards[icard]);
			}

		return timeobs;
		}

	/**
	 * Extracts UT Time in format hh:mm:ss from a FITS header in a String array.
	 */
	public  String getTime (String[] cards)
		{
		String datum = getTimeObs (cards);
		String dt="";

		// CHECK FOR hh:mm:ss.sss

		if (datum.indexOf(":") > 0)
			dt = datum;

		// OR CHECK FOR FLOATING POINT NUMBER

		else	{
			try	{
				double fp = Double.parseDouble(datum);
				int hh = (int)fp;
				int mm = (int)((fp-(double)hh)*60.0);
				double ss = (fp-(double)hh-(double)mm/60.0)/60.0;

				String sh=null;
				String sm=null;

				if (hh < 10)
					sh = "0"+hh;
				else
					sh = ""+hh;
				if (mm < 10)
					sm = "0"+mm;
				else
					sm = ""+mm;
				dt = sh+":"+sm+":"+ss;
				}
			catch (NumberFormatException e)
				{
				IJ.error("Unable to parse time "+datum);
				return null;
				}
			}
		return dt;
		}

	/**
	 * Returns time of day in seconds.
	 */
	public  double getDecimalTime (String[] cards)
		{
		double t = Double.NaN;
		String time = getTime(cards);
		if (time == null) return Double.NaN;

		try	{
			int i = time.indexOf(":");
			double hh = Double.parseDouble (time.substring (0,i));
			double mm = Double.parseDouble (time.substring (i+1,i+3));
			double ss = Double.parseDouble (time.substring (i+4));
			t = 3600.0*hh+60.0*mm+ss;
			}
		catch (NumberFormatException e)
			{
			IJ.error ("Unable to parse time "+time);
			}
		return  t;
		}

	/**
	 * Extracts exposure time from the FITS header in a  String array.
	 *
	 * Version 2009-01-10: accepts double values in strings via new getCardDoubleValue().
	 */
	public  double getExposureTime (String[] cards)
		{
		double tstart=0.0;
		double tend=0.0;
		int icard = 0;
		int icard2 = 0;

		try	{
			// CHECK FOR STANDARD KEYWORD "EXPTIME" (SECS)

			icard = findCardWithKey ("EXPTIME",cards);
			if (icard >= 0)
				{
				tstart = getCardDoubleValue (cards[icard]);
				if (! Double.isNaN(tstart))
					return tstart;
				}

			// CHECK FOR KEYWORD "EXPOSURE" (e.g. Mount Stromlo)

			icard = findCardWithKey ("EXPOSURE",cards);
			if (icard >= 0)
				{
				tstart = getCardDoubleValue (cards[icard]);
				if (! Double.isNaN(tstart))
					return tstart;
				}

			// OR CHECK FOR 'TM-START' AND 'TM-END' (SECS)

			icard = findCardWithKey ("TM-START",cards);
			icard2 = findCardWithKey ("TM-END",cards);

			// OR CHECK FOR 'TM_START' AND 'TM_END' (SECS)

			if (icard < 0 || icard2 < 0)
				{
				icard  = findCardWithKey ("TM_START",cards);
				icard2 = findCardWithKey ("TM_END",cards);
				}

			// OR CHECK FOR 'UT-START' AND 'UT-END' (SECS)

			if (icard < 0 || icard2 < 0)
				{
				icard  = findCardWithKey ("UT-START",cards);
				icard2 = findCardWithKey ("UT-END",cards);
				}

			// OR CHECK FOR 'UT_START' AND 'UT_END' (SECS)

			if (icard < 0 || icard2 < 0)
				{
				icard  = findCardWithKey ("UT_START",cards);
				icard2 = findCardWithKey ("UT_END",cards);
				}

			// OR GIVE UP

			if (icard < 0 || icard2 < 0)
				return Double.NaN;

			tstart = getCardDoubleValue (cards[icard]);
			tend   = getCardDoubleValue (cards[icard2]);
			if (Double.isNaN(tstart) || Double.isNaN(tend))
				return Double.NaN;
			}
		catch (NumberFormatException e)
			{
			IJ.error ("Unable to extract exposure time from FITS header: "+e.getMessage());
			return Double.NaN;
			}

		// WATCH OUT FOR CHANGE OF DAYS

		if (tend < tstart) tend += 3600*24.0;

		// RETURN DIFFERENCE BETWEEN START AND END TIMES

		return (tend-tstart);
		}

	/**
	 * Returns mid-exposure dateTime using a Properties list.
	 */
	public  String getMeanDateTime (String[] cards)
		{
		String dt = getDateTime (cards);
		double t = getExposureTime (cards);
		if (dt == null || Double.isNaN(t)) return null;

		t *= 0.5;

		Duration dur;
		Date date;
        DateParser dateParser = new DateParser();

		try	{
			dur = new Duration("P"+t+"S");
			date = dateParser.parse (dt);
			}
		catch (InvalidDateException e)
			{
			IJ.error(e.getMessage());
			return null;
			}
		GregorianCalendar cal = new GregorianCalendar();
		cal.setTime (date);
		try	{
			Calendar result = dur.addTo(cal);
			return dateParser.getIsoDate(result.getTime());
			}
		catch (InvalidDateException e)
			{
			IJ.error ("Unable to add half of exposure time = "+t+" to datetime "+dt);
			return null;
			}
		}


/**************************************** JD METHODS **************************************************/


	/**
	 * Returns JD from a FITS header stored in a String array.
	 */
	public  double getJD (String[] cards)
		{
		boolean modified = false;
		double julian = Double.NaN;

		// TRY TO GET JD FROM FITS HEADER

		julian = findDoubleValue ("JD-OBS", cards);
		if (Double.isNaN(julian))
			julian = findDoubleValue ("JD", cards);
		if (Double.isNaN(julian))
			{
			julian = findDoubleValue ("MJD-OBS", cards);
			if (! Double.isNaN(julian))
				modified = true;
			}
		if (Double.isNaN(julian))
			{
			julian = findDoubleValue ("MJD", cards);
			if (! Double.isNaN(julian))
				modified = true;
			}

		// OTHERWISE DERIVE FROM DATETIME

		if (Double.isNaN(julian))
			{
			String dt = getDateTime (cards);
			if (dt == null) return Double.NaN;
            JulianDate julianDate = new JulianDate();
			julian = julianDate.JD (dt);
			}
		if (Double.isNaN(julian))
			return Double.NaN;

		if (modified) julian += 2400000.0;
		return julian;
		}

	/**
	 * Returns mid-exposure Julian Date from a FITS header stored in a String array.
	 */
	public  double getMeanJD (String[] cards)
		{
		double jd = getJD (cards);
		double texp = getExposureTime (cards);
		if (Double.isNaN(jd) || Double.isNaN(texp))
			return Double.NaN;
		else
			return jd+0.5*(texp/3600.0)/24.0;
		}

	/**
	 * Returns MJD from a FITS heaader stored in a String array.
	 */
	public  double getMJD (String[] cards)
		{
		double jd = getJD (cards);
		if (!Double.isNaN(jd)) jd -= 2400000.0;
		return jd;
		}

	/**
	 * Returns mid-exposure MJD from a FITS header stored in a String array.
	 */
	public  double getMeanMJD (String[] cards)
		{
		double jd = getMeanJD(cards);
		if (!Double.isNaN(jd))
			jd -= 2400000.0;
		return jd;
		}
	}

public class Duration
	{
	private int plusMinus;
	private double years;
	private double months;
	private double days;
	private double hours;
	private double minutes;
	private double seconds;

	/**
	 * Instatiates a Duration from an ISO 8601 format duration string
	 */
	public Duration (String isodur) throws InvalidDateException
		{
		boolean isTime = false;
		String value = null;
		String delim = null;

		plusMinus=1;
		years   = -1.0;
		months  = -1.0;
		days    = -1.0;
		hours   = -1.0;
		minutes = -1.0;
		seconds = -1.0;

		// DURATION FORMAT IS: (-)PnYnMnDTnHnMnS
		StringTokenizer st = new StringTokenizer(isodur, "-PYMDTHS", true);

		try	{

			// OPTIONAL SIGN

			value = st.nextToken();

			if (value.equals("-"))
				{
				plusMinus= -1;
				value=st.nextToken();
				}

			// DURATION MUST START WITH A "P"

			if (!value.equals("P"))
				throw new InvalidDateException(isodur+
					" : "+value+
					" : no P deliminator for duration");

			// GET NEXT FIELD

			while (st.hasMoreTokens())
				{

				// VALUE

	    			value = new String(st.nextToken());
				if (value.equals("T"))
					{
					if (!st.hasMoreTokens())
						throw new InvalidDateException(isodur+
							" : "+value+
							": no values after duration T delimitor");
					value = st.nextToken();
					isTime = true;
					}

				// DELIMINATOR

				if (!st.hasMoreTokens())
					throw new InvalidDateException(isodur+
						" : "+value+
						"No deliminator for duration");
				delim = new String(st.nextToken());

				// YEAR

				if (delim.equals("Y"))
					{
					years = Double.parseDouble(value);
					}

				// MONTH

				else if (delim.equals("M") && isTime == false)
					{
					months = Double.parseDouble(value);
					if (months != (double)((int)months))
						throw new InvalidDateException(
							"Cannot process decimal months!");
					}

				// DAYS

				else if (delim.equals("D"))
					{
					days = Double.parseDouble(value);
					}

				// HOURS

				else if (delim.equals("H"))
					{
					hours = Double.parseDouble(value);
					isTime = true;
					}

				// MINUTES

				else if (delim.equals("M") && isTime == true)
					{
					minutes = Double.parseDouble(value);
					}

				// SECONDS

				else if (delim.equals("S"))
					{
					seconds = Double.parseDouble(value);
					}
				else	{
					throw new InvalidDateException(isodur+
						": what duration delimiter is "+delim+"?");
					}
				}
			}
		catch (NumberFormatException ex)
			{
			throw new InvalidDateException("["+ex.getMessage()+
					   "] is not valid");
			}
		}

	/**
	* Add Duration to a Calendar
	* @return modified Calendar
	*/
	public Calendar addTo(Calendar cal) throws InvalidDateException
		{
		int iyears,imonths,idays,ihours,imins,isecs,millis;

		if (years > 0)
			{
			iyears = (int)years;
			imonths = 0;			// AVOID USING MONTHS
			idays  = (int)(365.254*(years
					-iyears));
			ihours = (int)(24.0*365.254*(years
					-iyears
					-idays/365.254));
			imins  = (int)(60.0*24.0*365.254*(years
					-iyears
					-idays/365.254
					-ihours/24.0/365.254));
			isecs  = (int)(60.0*60.0*24.0*365.254*(years
					-iyears
					-idays/365.254
					-ihours/24.0/365.254
					-imins/60.0/24.0/365.254));
			millis  = (int)(1000.0*60.0*60.0*24.0*365.254*(years
					-iyears
					-idays/365.254
					-ihours/24.0/365.254
					-imins/60.0/24.0/365.254
					-isecs/60.0/60.0/24.0/365.254));
			cal.add(Calendar.YEAR,        plusMinus*iyears);
			cal.add(Calendar.DAY_OF_YEAR, plusMinus*idays);
			cal.add(Calendar.HOUR,        plusMinus*ihours);
			cal.add(Calendar.MINUTE,      plusMinus*imins);
			cal.add(Calendar.SECOND,      plusMinus*isecs);
			cal.add(Calendar.MILLISECOND, plusMinus*millis);
			}

		if (months > 0)
			{
	 		imonths = (int)months;
			if (months != (double)imonths)
				throw new InvalidDateException(
					"Cannot add decimal months!");
			cal.add(Calendar.MONTH, plusMinus*imonths);
			}

		if (days > 0)
			{
			idays  = (int)days;
			ihours = (int)(24.0*(days
					-idays));
			imins  = (int)(60.0*24.0*(days
					-idays
					-ihours/24.0));
			isecs  = (int)(60.0*60.0*24.0*(days
					-idays
					-ihours/24.0
					-imins/24.0/60.0));
			millis  = (int)(1000.0*60.0*60.0*24.0*(days
					-idays
					-ihours/24.0
					-imins/60.0/24.0
					-isecs/60.0/60.0/24.0));
			cal.add(Calendar.DAY_OF_YEAR, plusMinus*idays);
			cal.add(Calendar.HOUR,        plusMinus*ihours);
			cal.add(Calendar.MINUTE,      plusMinus*imins);
			cal.add(Calendar.SECOND,      plusMinus*isecs);
			cal.add(Calendar.MILLISECOND, plusMinus*millis);
			}

		if (hours > 0)
			{
			ihours  = (int)hours;
			imins  = (int)(60.0*(hours
					-ihours));
			isecs  = (int)(60.0*60.0*(hours
					-ihours
					-imins/60.0));
			millis  = (int)(1000.0*60.0*60.0*(hours
					-ihours
					-imins/60.0
					-isecs/60.0/60.0));
			cal.add(Calendar.HOUR,        plusMinus*ihours);
			cal.add(Calendar.MINUTE,      plusMinus*imins);
			cal.add(Calendar.SECOND,      plusMinus*isecs);
			cal.add(Calendar.MILLISECOND, plusMinus*millis);
			}

		if (minutes > 0)
			{
			imins  = (int)minutes;
			isecs  = (int)(60.0*(minutes
					-imins));
			millis  = (int)(1000.0*60.0*(minutes
					-imins
					-isecs/60.0));
			cal.add(Calendar.MINUTE,      plusMinus*imins);
			cal.add(Calendar.SECOND,      plusMinus*isecs);
			cal.add(Calendar.MILLISECOND, plusMinus*millis);
			}

		if (seconds > 0)
			{
			isecs  = (int)seconds;
			millis  = (int)(1000.0*(seconds
					-isecs));
			cal.add(Calendar.SECOND,      plusMinus*isecs);
			cal.add(Calendar.MILLISECOND, plusMinus*millis);
			}
		return cal;
		}

	/**
	* Generate a string representation of an ISO 8601 duration
	* @return a string representing the duration in the ISO 8601 format
	*/
	public String toString()
		{
		StringBuffer buffer = new StringBuffer();

		// OPTIONAL SIGN

		if (plusMinus == -1)
			buffer.append("-");

		// REQUIRED "P"

		buffer.append("P");

		if (years > 0)
			{
			if (years == (double)((int)years))
				buffer.append((int)years);
			else
				buffer.append(years);
			buffer.append("Y");
			}
		if (months > 0)
			{
			if (months == (double)((int)months))
				buffer.append((int)months);
			else
				buffer.append(months);
			buffer.append("M");
			}
		if (days > 0)
			{
			if (days == (double)((int)days))
				buffer.append((int)days);
			else
				buffer.append(days);
			buffer.append("D");
			}

		// DATE-TIME SEPARATOR (IF NEEDED)

		if ((years > 0 || months > 0 || days > 0) &&
		    (hours > 0 || minutes > 0 || seconds > 0))
			buffer.append("T");


		if (hours > 0)
			{
			if (hours == (double)((int)hours))
				buffer.append((int)hours);
			else
				buffer.append(hours);
			buffer.append("H");
			}
		if (minutes > 0)
			{
			if (minutes == (double)((int)minutes))
				buffer.append((int)minutes);
			else
				buffer.append(minutes);
			buffer.append("M");
			}
		if (seconds > 0)
			{
			buffer.append(seconds);
			buffer.append("S");
			}

		return buffer.toString();
		}

	public void main(String args[])
		{
		String isodur;
		Duration dur;

		isodur = new String("-P1997Y07M16DT19H20M30.45S");
		System.out.println("Testing negative duration:"+isodur);
		try	{
			dur = new Duration(isodur);
			System.out.println(isodur+"="+dur.toString());
			}
		catch (InvalidDateException e)
			{
			System.out.println(e.getMessage());
			}

		isodur = new String("P1Y2.34M5D");
		System.out.println("\nTesting decimal month error:"+isodur);
		try	{
			dur = new Duration(isodur);
			System.out.println(isodur+"="+dur.toString());
			}
		catch (InvalidDateException e)
			{
			System.out.println(e.getMessage());
			}

		isodur = new String("P4H5M6.789S");
		System.out.println("\nTesting date versus time:"+isodur);
		try	{
			dur = new Duration(isodur);
			System.out.println(isodur+"="+dur.toString());
			}
		catch (InvalidDateException e)
			{
			System.out.println(e.getMessage());
			}

		isodur = new String("P1M2DT4H5M0.789S");
		System.out.println("\nTesting addition to calendar date:"+isodur);
		try	{
			Calendar cal = new GregorianCalendar(TimeZone.getTimeZone("UTC"));
			dur = new Duration(isodur);
			System.out.println(
				"["+cal.getTime().toString()+
				"] + "+dur.toString());
			dur.addTo(cal);
			System.out.println(
				" = ["+cal.getTime().toString()+"]"
				);
			}
		catch (InvalidDateException e)
			{
			System.out.println(e.getMessage());
			}

		isodur = new String("-P1M2DT4H5M0.789S");
		System.out.println("\nTesting subtraction from calendar date:"+isodur);
		try	{
			Calendar cal = new GregorianCalendar(TimeZone.getTimeZone("UTC"));
			dur = new Duration(isodur);
			System.out.println(
				"["+cal.getTime().toString()+
				"] + "+dur.toString());
			dur.addTo(cal);
			System.out.println(
				" = ["+cal.getTime().toString()+"]"
				);
			}
		catch (InvalidDateException e)
			{
			System.out.println(e.getMessage());
			}

		}

	}


public class DateParser {

    private boolean check(StringTokenizer st, String token)
	throws InvalidDateException
    {
	try {
	    if (st.nextToken().equals(token)) {
		return true;
	    } else {
		throw new InvalidDateException("Missing ["+token+"]");
	    }
	} catch (NoSuchElementException ex) {
	    return false;
	}
    }

    private Calendar getCalendar(String isodate)
    	throws InvalidDateException
    {
	// YYYY-MM-DDThh:mm:ss.sTZD
	StringTokenizer st = new StringTokenizer(isodate, "-T:.+Z", true);

	Calendar calendar = new GregorianCalendar(TimeZone.getTimeZone("UTC"));
	calendar.clear();
	try {
	    // Year
	    if (st.hasMoreTokens()) {
		int year = Integer.parseInt(st.nextToken());
		calendar.set(Calendar.YEAR, year);
	    } else {
		return calendar;
	    }
	    // Month
	    if (check(st, "-") && (st.hasMoreTokens())) {
		int month = Integer.parseInt(st.nextToken()) -1;
		calendar.set(Calendar.MONTH, month);
	    } else {
		return calendar;
	    }
	    // Day
	    if (check(st, "-") && (st.hasMoreTokens())) {
		int day = Integer.parseInt(st.nextToken());
		calendar.set(Calendar.DAY_OF_MONTH, day);
	    } else {
		return calendar;
	    }
	    // Hour
	    if (check(st, "T") && (st.hasMoreTokens())) {
		int hour = Integer.parseInt(st.nextToken());
		calendar.set(Calendar.HOUR_OF_DAY, hour);
	    } else {
		calendar.set(Calendar.HOUR_OF_DAY, 0);
		calendar.set(Calendar.MINUTE, 0);
		calendar.set(Calendar.SECOND, 0);
		calendar.set(Calendar.MILLISECOND, 0);
		return calendar;
	    }
	    // Minutes
	    if (check(st, ":") && (st.hasMoreTokens())) {
		int minutes = Integer.parseInt(st.nextToken());
		calendar.set(Calendar.MINUTE, minutes);
	    } else {
		calendar.set(Calendar.MINUTE, 0);
		calendar.set(Calendar.SECOND, 0);
		calendar.set(Calendar.MILLISECOND, 0);
		return calendar;
	    }

	    //
	    // Not mandatory now
	    //

	    // Secondes
	    if (! st.hasMoreTokens()) {
		return calendar;
	    }
	    String tok = st.nextToken();
	    if (tok.equals(":")) { // secondes
		if (st.hasMoreTokens()) {
		    int secondes = Integer.parseInt(st.nextToken());
		    calendar.set(Calendar.SECOND, secondes);
		    if (! st.hasMoreTokens()) {
			return calendar;
		    }
		    // frac sec
		    tok = st.nextToken();
		    if (tok.equals(".")) {
			// bug fixed, thx to Martin Bottcher
			String nt = st.nextToken();
			while(nt.length() < 3) {
			    nt += "0";
			}
			nt = nt.substring( 0, 3 ); //Cut trailing chars..
			int millisec = Integer.parseInt(nt);
			//int millisec = Integer.parseInt(st.nextToken()) * 10;
			calendar.set(Calendar.MILLISECOND, millisec);
			if (! st.hasMoreTokens()) {
			    return calendar;
			}
			tok = st.nextToken();
		    } else {
			calendar.set(Calendar.MILLISECOND, 0);
		    }
		} else {
		    throw new InvalidDateException("No secondes specified");
		}
	    } else {
		calendar.set(Calendar.SECOND, 0);
		calendar.set(Calendar.MILLISECOND, 0);
	    }
	    // Timezone
	    if (! tok.equals("Z")) { // UTC
		if (! (tok.equals("+") || tok.equals("-"))) {
		    throw new InvalidDateException("only Z, + or - allowed");
		}
		boolean plus = tok.equals("+");
		if (! st.hasMoreTokens()) {
		    throw new InvalidDateException("Missing hour field");
		}
		int tzhour = Integer.parseInt(st.nextToken());
		int tzmin  = 0;
		if (check(st, ":") && (st.hasMoreTokens())) {
		    tzmin = Integer.parseInt(st.nextToken());
		} else {
		    throw new InvalidDateException("Missing minute field");
		}
		if (plus) {
		    calendar.add(Calendar.HOUR, tzhour);
		    calendar.add(Calendar.MINUTE, tzmin);
		} else {
		    calendar.add(Calendar.HOUR, -tzhour);
		    calendar.add(Calendar.MINUTE, -tzmin);
		}
	    }
	} catch (NumberFormatException ex) {
	    throw new InvalidDateException("["+ex.getMessage()+
					   "] is not an integer");
	}
	return calendar;
    }

    /**
     * Parse the given string in ISO 8601 format and build a Date object.
     * @param isodate the date in ISO 8601 format
     * @return a Date instance
     * @exception InvalidDateException if the date is not valid
     */
    public Date parse(String isodate)
    	throws InvalidDateException
    {
	Calendar calendar = getCalendar(isodate);
	return calendar.getTime();
    }

    private String twoDigit(int i) {
	if (i >=0 && i < 10) {
	    return "0"+String.valueOf(i);
	}
	return String.valueOf(i);
    }

    /**
     * Generate a ISO 8601 date
     * @param date a Date instance
     * @return a string representing the date in the ISO 8601 format
     */
    public String getIsoDate(Date date) {
	Calendar calendar = new GregorianCalendar(TimeZone.getTimeZone("UTC"));
	calendar.setTime(date);
	StringBuffer buffer = new StringBuffer();
	buffer.append(calendar.get(Calendar.YEAR));
	buffer.append("-");
	buffer.append(twoDigit(calendar.get(Calendar.MONTH) + 1));
	buffer.append("-");
	buffer.append(twoDigit(calendar.get(Calendar.DAY_OF_MONTH)));
	buffer.append("T");
	buffer.append(twoDigit(calendar.get(Calendar.HOUR_OF_DAY)));
	buffer.append(":");
	buffer.append(twoDigit(calendar.get(Calendar.MINUTE)));
	buffer.append(":");
	buffer.append(twoDigit(calendar.get(Calendar.SECOND)));
	buffer.append(".");
	buffer.append(twoDigit(calendar.get(Calendar.MILLISECOND) / 10));
	buffer.append("Z");
	return buffer.toString();
    }

    public void test(String isodate) {
	System.out.println("----------------------------------");
	try {
	    Date date = parse(isodate);
	    System.out.println(">> "+isodate);
	    System.out.println(">> "+date.toString()+" ["+date.getTime()+"]");
	    System.out.println(">> "+getIsoDate(date));
	} catch (InvalidDateException ex) {
	    System.err.println(isodate+" is invalid");
	    System.err.println(ex.getMessage());
	}
	System.out.println("----------------------------------");
    }

    public void test(Date date) {
	String isodate = null;
	System.out.println("----------------------------------");
	try {
	    System.out.println(">> "+date.toString()+" ["+date.getTime()+"]");
	    isodate = getIsoDate(date);
	    System.out.println(">> "+isodate);
	    date = parse(isodate);
	    System.out.println(">> "+date.toString()+" ["+date.getTime()+"]");
	} catch (InvalidDateException ex) {
	    System.err.println(isodate+" is invalid");
	    System.err.println(ex.getMessage());
	}
	System.out.println("----------------------------------");
    }

    public void main(String args[]) {
	test("1997-07-16T19:20:30.45-02:00");
	test("1997-07-16T19:20:30+01:00");
	test("1997-07-16T19:20:30+01:00");
	test("1997-07-16T19:20");
	test("1997-07-16");
	test("1997-07");
	test("1997");
	test(new Date());
    }

}

public class InvalidDateException extends Exception
	{
	public InvalidDateException(String msg)
		{
		super(msg);
		}
	}

public class JulianDate
	{
	// Gregorian Calendar adopted Oct. 15, 1582 (2299161)

	/**
	 * Gets JD from a dateTime string.
	 */
	public double JD (String dt)
		{
		if (dt == null) return Double.NaN;
		int[] YMDhmsm = new int[7];
		try	{
			int i1=dt.indexOf("-");
			if (i1 < 0) return Double.NaN;
			YMDhmsm[0] = Integer.parseInt(dt.substring(0,i1));
			i1++;
			int i2=dt.indexOf("-",i1);
			if (i2 < 0) return Double.NaN;
			YMDhmsm[1] = Integer.parseInt(dt.substring(i1,i2));
			i1=i2+1;
			i2=dt.indexOf("T",i1);
			if (i2 < 0) return JD(YMDhmsm);
			YMDhmsm[2] = Integer.parseInt(dt.substring(i1,i2));
			i1=i2+1;
			i2=dt.indexOf(":",i1);
			if (i2 < 0) return Double.NaN;
			YMDhmsm[3] = Integer.parseInt(dt.substring(i1,i2));
			i1=i2+1;
			i2=dt.indexOf(":",i1);
			if (i2 < 0) return Double.NaN;
			YMDhmsm[4] = Integer.parseInt(dt.substring(i1,i2));
			double secs = Double.parseDouble(dt.substring(i2+1));
			YMDhmsm[5] = (int)(secs+0.5);
			YMDhmsm[6] = (int)(1000*(secs-(double)YMDhmsm[5]));
			return JD(YMDhmsm);
			}
		catch (NumberFormatException e)
			{
			return Double.NaN;
			}
		}

	/**
	 * Gets JD from a Calendar object.
	 */
	public double JD (Calendar cal)
		{
		int[] YMDhmsm = new int[]{
			cal.get(Calendar.YEAR),
			cal.get(Calendar.MONTH)+1,
			cal.get(Calendar.DAY_OF_MONTH),
			cal.get(Calendar.HOUR_OF_DAY),
			cal.get(Calendar.MINUTE),
			cal.get(Calendar.SECOND),
			cal.get(Calendar.MILLISECOND)
			};
		return JD(YMDhmsm);
		}

	/**
	 * Gets JD from integer year,month,day,hour,minute,second,millisecond array.
	 */
	public double JD (int[] YMDhmsm)
		{
		if (YMDhmsm == null) return Double.NaN;

		int ijd = julday(YMDhmsm);	// STANDARD NR INTEGER JD AT NOON OF GIVEN DAY

		// CONSTRUCT THE REST

		int hour=0;
		int minute=0;
		int second=0;
		int millisecond=0;
		if (YMDhmsm.length > 3)
			hour=YMDhmsm[3];
		if (YMDhmsm.length > 4)
			minute=YMDhmsm[4];
		if (YMDhmsm.length > 5)
			second=YMDhmsm[5];
		if (YMDhmsm.length > 6)
			millisecond=YMDhmsm[6];

		double julian = (double)ijd+(hour+(double)minute/60.0+(second+0.001*(double)millisecond)/3600.0)/24.0;
		julian -= 0.5;

		return julian;
		}

	/**
	 * Converts a Julian day to a Calendar object.
	 */
	public Calendar calendar (double injulian)
		{
		int[] YMDhmsm = dateArray(injulian);
		Calendar cal = Calendar.getInstance();
		cal.set (Calendar.YEAR,         YMDhmsm[0]);
		cal.set (Calendar.MONTH,        YMDhmsm[1]-1);
		cal.set (Calendar.DAY_OF_MONTH, YMDhmsm[0]);
		cal.set (Calendar.HOUR,         YMDhmsm[0]);
		cal.set (Calendar.MINUTE,       YMDhmsm[0]);
		cal.set (Calendar.SECOND,       YMDhmsm[0]);
		cal.set (Calendar.MILLISECOND,  YMDhmsm[0]);
		return cal;
		}

	/**
	 * Converts a Julian day to integer year,month,day,hour,minute,second,millisecond
	 */
	public int[] dateArray (double jd)
		{
		int ijd = (int)(jd+0.5);
		double diff = jd-(double)ijd;	// RESIDUAL DAY

		int[] darr = caldat(ijd);	// RESULT CONTAINS YEAR,MONTH,DAY AT NOON OF GIVEN DAY

		// NOW DO THE FRACTIONAL PART

		diff += 0.5;
		int hour = (int)(diff*24.0);
		diff = 24.0*diff-(double)hour;		// HOURS
		int minute = (int)(diff*60.0);
		diff = 60.0*diff-(double)minute;	// MINUTES
		int second = (int)(diff*60.0+0.5);
		diff = 60.0*diff-(double)second;	// SECONDS
		int millisecond = (int)(diff*1000.0);

		darr[3] = hour;
		darr[4] = minute;
		darr[5] = second;
		darr[6] = millisecond;
		return darr;
		}

	/**
	 * Converts a JD to a dateTime string.
	 */
	public String dateTime (double julian)
		{
		int[] YMDhmsm = dateArray(julian);
		return dateTime (YMDhmsm);
		}

	/**
	 * Converts a dateTime array to a dateTime string.
	 */
	public String dateTime (int[] YMDhmsm)
		{
		double secs = YMDhmsm[5]+0.001*(double)YMDhmsm[6];

		NumberFormat fi = NumberFormat.getIntegerInstance ();
		fi.setMinimumIntegerDigits (2);
		NumberFormat fd = NumberFormat.getInstance();
		fd.setMinimumIntegerDigits (2);

		// DecimalFormat f = new DecimalFormat("##");

		String sjd = new String(""
			+YMDhmsm[0]+"-"+fi.format(YMDhmsm[1])+"-"+fi.format(YMDhmsm[2])
			+"T"
			+fi.format(YMDhmsm[3])+":"+fi.format(YMDhmsm[4])+":"+fd.format(secs));
		return sjd;
		}

	/**
	 * Test of JulianDay methods.
	 */
	public void main(String args[])
		{
		// TEST ARRAY METHOD

		System.out.println("Test #1 : Julian date for May 23, 1968, 11:59:59 : "
					+ JD( new int[] {1968,5,23,11,59,59,0} )
					+ " = slightly under 2440000");
		int results[] = dateArray(JD(new int[] {1968,5,23,11,59,59,0} ));
		double secs = results[5]+0.001*(double)results[6];
		System.out.println ("\t... back to calendar : "+dateTime(results));

		System.out.println("Test #2 : Julian date for May 23, 1968, 12:00:01 : "
					+ JD( new int[] {1968,5,23,12,0,1,0} )
					+ " = slightly over 2440000");
		results = dateArray(JD(new int[] {1968,5,23,12,00,01,0} ));
		secs = results[5]+0.001*(double)results[6];
		System.out.println ("\t... back to calendar : "+dateTime(results));

		System.out.println("Test #3 : Julian date for May 23, 1968, 23:59:59.99 : "
					+ JD( new int[] {1968,5,23,23,59,59,99} )
					+ " = slightly under 2440000.5");
		results = dateArray(JD(new int[] {1968,5,23,23,59,59,990} ));
		secs = results[5]+0.001*(double)results[6];
		System.out.println ("\t... back to calendar : "+dateTime(results));

		// TEST Calendar METHODS

		Calendar today = Calendar.getInstance();
		double todayJulian = JD(new int[]{
			today.get(Calendar.YEAR),
			today.get(Calendar.MONTH)+1,
			today.get(Calendar.DAY_OF_MONTH),
			today.get(Calendar.HOUR),
			today.get(Calendar.MINUTE),
			today.get(Calendar.SECOND),
			today.get(Calendar.MILLISECOND)
			});
		System.out.println("Test #4 : Julian date for today : " + todayJulian);
		System.out.println ("\t... back to dateTime : "+dateTime(todayJulian));

		// TEST DATETIME METHODS

		double date1 = JD("2005-01-01T02:34:56.789");
		double date2 = date1+Math.PI;
		System.out.println("Test #5 : Between"+dateTime(date1)+" and "+dateTime(date2)+" : "
				+ (date2 - date1) + " days");
		}

	/**
	 * Returns the whole number JD at noon of the day represented by year,month,day.
	 * Numerical Recipes in C, 2nd ed., Cambridge University Press 1992
	 */
	public int julday (int[] YMDhmsm)
		{
		int year=YMDhmsm[0];
		int month=YMDhmsm[1]; // jan=1, feb=2,... IS ONE MORE THAN THE java.util.Calendar STANDARD!!
		int day=YMDhmsm[2];

		int ja,jul,jy,jm;
		int IGREG = 15+31*(10+12*1582);

		jy = year;
		if (jy < 0) ++jy;
		if (month > 2)
			{
			jm = month+1;
			}
		else	{
			jy--;
			jm = month+13;
			}
		jul = (int)(java.lang.Math.floor(365.25*jy)+java.lang.Math.floor(30.6001*jm)+day+1720995);
		if ((day+31*(month+12*year)) >= IGREG)
			{
			ja = (int)(0.01*jy);
			jul += 2-ja+(int)(0.25*ja);
			}
		return jul;
		}

	/**
	 * Returns the year,month,day where the whole number JD took place at noon.
	 * Numerical Recipes in C, 2nd ed., Cambridge University Press 1992
	 */
	public int[] caldat (int julian)
		{
		int ja,jalpha,jb,jc,jd,je;
		int IGREG = 2299161;

		int[] YMDhmsm = new int[7];	// ONLY NON-NR PART!

		if (julian >= IGREG)
			{
			jalpha = (int)(((julian-1867216)-0.25)/36524.25);
			ja = julian+1+jalpha-(int)(0.25*jalpha);
			}
		else if (julian < 0)
			{
			ja = julian+36525*(1-julian/36525);
			}
		else	{
			ja = julian;
			}
		jb = ja + 1524;
		jc = (int)(6680.0+((double)(jb-2439870)-122.1)/365.25);
		jd = (int)(365*jc+(0.25*jc));
		je = (int)((jb-jd)/30.6001);
		int day = jb-jd-(int)(30.6001*je);
		int mm = je-1;
		if (mm > 12) mm -= 12;
		int iyyy = jc-4715;
		if (mm > 2) iyyy--;
		if (iyyy <= 0) iyyy--;
		if (julian < 0) iyyy -= 100*(1-julian/36525);

		YMDhmsm[0] = iyyy;
		YMDhmsm[1] = mm;
		YMDhmsm[2] = day;
		YMDhmsm[3] = 0;		// HOURS
		YMDhmsm[4] = 0;		// MINUTES
		YMDhmsm[5] = 0;		// SECONDS
		YMDhmsm[6] = 0;		// MILLISECONDS
		return YMDhmsm;
		}

	}


	}
