// Plot_Table_Columns.java

import ij.*;
import ij.gui.*;
import ij.text.*;
import ij.util.*;
import ij.measure.ResultsTable;
import ij.plugin.*;
import ij.plugin.PlugIn;
import ij.Prefs;
import ij.process.*;
import ij.io.*;
import ij.measure.CurveFitter;
import flanagan.analysis.*; 

import javax.swing.*;
import javax.swing.border.TitledBorder;
import javax.swing.event.*;
import java.util.Hashtable;

import java.awt.*;
import java.awt.geom.*;
import java.util.*;
import java.io.*;
import java.awt.event.*;

import java.lang.Math.*;
import java.text.*;

// astroj ------
import astroj.*;
import java.awt.image.BufferedImage;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import javax.imageio.IIOImage;
import javax.imageio.ImageIO;
import javax.imageio.ImageWriteParam;
import javax.imageio.stream.FileImageOutputStream;
// ------ astroj


/**
 * This plugin plots any number of columns from a Results or MeasurementTable.  This plugin
 * is based on the Plot_Columns plugin by F.V. Hessman.
 *
 * @author F.V. Hessman, Georg-August-Universitaet Goettingen
 * @date 2009-FEB-09
 * @version 1.0
 *
 * @author K.A. Collins, University of Louisville, Kentucky USA
 * @date 2010-JUN-09
 * @version 2.1
 * @changes New GUI interface, support for multiple column plotting, error plotting
 * support, interactive shifting and scaling, titles, subtitles, and legends, custom,
 * semi-custom, and auto scaling
 *
 *
 * 
 */

/**
 * This plugin plots any two columns from a Results or MeasurementTable.
 * @author F.V. Hessman, Georg-August-Universitaet Goettingen
 * @date 2009-FEB-09
 * @version 1.0
 *
 * To create version outside of the Astronomy plugins, comment in the "// plain ..." lines and comment out the "// astroj ..." lines
 */
public class MultiPlot_ implements PlugIn
        {
        static String title;
        static double titlePosX;
        static double titlePosY;
        static String subtitle;
        static double subtitlePosX;
        static double subtitlePosY;
        static double legendPosX;
        static double legendPosY;
        static boolean button2Drag;
        static String xlabeldefault;
        static int xcolumndefault;
        static int plotSizeX;
        static int plotSizeY;
        static double pltMinX;
        static double pltMinY;
        static double pltMaxX;
        static double pltMaxY;
        static double plotMinX;
        static double plotMinY;
        static double plotMaxX;
        static double plotMaxY;
        static int xExponent;
        static int yExponent;
        static int mmagrefs;
        static int firstCurve;
        static String xLegend;
        static String yLegend;
        static boolean legendLeft;
        static boolean legendRight;
        static boolean useTitle;
        static boolean useSubtitle;
        static boolean useXCustomName;
        static boolean useYCustomName;
        static boolean useXColumnName;
        static boolean useYColumnName;
        static boolean xTics;
        static boolean yTics;
        static boolean xGrid;
        static boolean yGrid;
        static boolean xNumbers;
        static boolean yNumbers;
        static boolean plotAutoMode;
        static boolean nextPanel;
        static boolean canceled;
        static boolean waitSecs;
        static boolean openDataSetWindow, openRefStarWindow;
        static boolean rememberWindowLocations;
        static boolean autoScaleX;
        static boolean useFirstX;
        static boolean autoScaleY;
        static boolean showToolTips;
        static boolean showXAxisNormal;
        static boolean showXAxisAsPhase;
        static boolean showXAxisAsHoursSinceTc;
        static boolean showXAxisAsDaysSinceTc;
        static double T0;
        static double period;
        static boolean showXScaleInfo;
        static boolean showYScaleInfo, showYNormInfo;
        static boolean showYShiftInfo;
        static boolean showLScaleInfo;
        static boolean showLRelScaleInfo;
        static boolean showLShiftInfo;
        static boolean showLRelShiftInfo;
        static boolean showYBinInfo;
        static boolean showLBinInfo;
        static boolean showYmmagInfo;
        static boolean showLmmagInfo, showLdetrendInfo, showLnormInfo;
        static boolean showYSymbolInfo;
        static boolean showLSymbolInfo;
        static boolean useTwoLineLegend;
        static boolean useWideDataPanel;
        static boolean subframeWasShowing;
        static boolean newPlotWindow;
        static boolean useDefaultSettings;
        static boolean tableHasText;
        static boolean setSubpanelVisible, refStarPanelWasShowing, addAstroDataFrameWasShowing;
        static boolean usePixelScale;
        static boolean showVMarker1, showVMarker2;
        static boolean showDMarkers, useDMarker1, useDMarker4;
        static boolean divideNotSubtract;
        static boolean leftDragReleased;
        static boolean prioritizeColumns;
        static boolean invertYAxis;
        static int invertYAxisSign;
        static boolean negateMag;
        static boolean selectAnotherTableCanceled;
        static boolean panelShiftDown;
        static boolean panelControlDown;
        static boolean panelAltDown;
        static boolean keepFileNamesOnAppend;
        static boolean updatePlotRunning;
        static boolean saveNewXColumn;
        static boolean saveNewYColumn;
        static boolean saveNewYErrColumn;
        static boolean updatePlotEnabled;
        static boolean disableUpdatePlotBox;
        static boolean astroConverterUpdating;

        static double pixelScale;
        static double zoom;
        static double xMin;
        static double xMax;
        static double xWidth;
        static double yMin;
        static double yMax;
        static double xOffset;
        static double yOffset;
        static double xautoscalemin;
        static double xautoscalemax;
        static double yautoscalemin;
        static double yautoscalemax;
        static double yRange;
        static double dashLength;
        static double numDashes;
        static double mouseX;
        static double mouseY;
        static double totalPanOffsetX;
        static double totalPanOffsetY;
        static double newPanOffsetX;
        static double newPanOffsetY;

        static int cur;
        static int plotOptions;
        static int mainFrameLocationX;
        static int mainFrameLocationY;
        static int subFrameLocationX, refStarFrameLocationX, addAstroDataFrameLocationX;
        static int subFrameLocationY, refStarFrameLocationY, addAstroDataFrameLocationY;
        static int plotFrameLocationX;
        static int plotFrameLocationY;

        static int excludedHeadSamples;
        static int excludedTailSamples;
        static int holdExcludedHeadSamples;
        static int holdExcludedTailSamples;
        static int excludedHeadSamplesStep;
        static int excludedTailSamplesStep;
        static int refStarHorzWidth;
        static int lastRefStar;
        static boolean cycleEnabledStarsLess1PressedConsecutive;

        static double xWidthStep;
        static double xMaxStep;
        static double xMinStep;
        static double yMaxStep;
        static double yMinStep;
        static double yMid;
        static double firstXmin;
        static int plotSizeXStep;
        static int plotSizeYStep;
        static int nearestLine;
        static double nearestX;
        static int boldedDatum;
        static boolean useBoldedDatum;
        static boolean useUpdateStack;
        static boolean multiUpdate;
        static int selectedRow;

        static String vMarker1TopText;
        static String vMarker1BotText;
        static String vMarker2TopText;
        static String vMarker2BotText;
        static boolean saveImage;
        static boolean savePlot;
        static boolean saveConfig;
        static boolean saveTable;
        static boolean saveApertures;
        static boolean saveLog;
        static boolean saveAllPNG;          
        static String imageSuffix;
        static String aperSuffix;
        static String logSuffix;
        static String plotSuffix;
        static String configSuffix;
        static String dataSuffix;
        static String appendDestinationSuffix;
        static String appendSourceSuffix;
        static String combinedTableName;
        static String requestedTableName;
        static String templateDir;
        static String JDColumn;

        static int n;
        static int[] nn;
        static int[] nnr;
        static double[] coeffs;        
        static String[] xlabel2;
        static int selectedRowStart, selectedRowEnd;
        static java.util.List<String> deletedRowList;
        static java.io.File[] dragAndDropFiles;
        static double[][] x;
        static double[][] y;
        static double[][] yerr;
        static double[][] yop;
        static double[][] yoperr;
        static double[][][] detrend;
        static double[][] detrendVars;

        static double[][] xc1;
        static double[][] xc2;
        static double[][] yc1;
        static double[][] yc2;
        static double xMultiplierFactor;
        static double yMultiplierFactor;
        static int excluded;

        static String xlab;
        static String ylab;
        static String priorityColumns;
        static String priorityDetrendColumns;
        static double dx;
        static double dy;
        static double xPlotMin;
        static double xPlotMax;
        static double yPlotMin;
        static double yPlotMax;
        
                
        static boolean addAirmass, addAltitude, addAzimuth, addBJD, addBJDCorr, addDecNow, addRaNow;
        static boolean addDec2000, addRA2000, addGJD, addHJD, addHJDCorr, addHourAngle, addZenithDistance;
        
        static String airmassName, altitudeName, azimuthName, bjdName, bjdCorrName, decNowName, raNowName;
        static String dec2000Name, ra2000Name, gjdName, hjdName, hjdCorrName, hourAngleName, zenithDistanceName;
        
        
        static JCheckBox airmassCB, altitudeCB, azimuthCB, bjdCB, bjdCorrCB, decNowCB, raNowCB;
        static JCheckBox dec2000CB, ra2000CB, gjdCB, hjdCB, hjdCorrCB, hourAngleCB, zenithDistanceCB;
                
        static JTextField airmassField, altitudeField, azimuthField, bjdField, bjdCorrField, decNowField, raNowField;        
        static JTextField dec2000Field, ra2000Field, gjdField, hjdField, hjdCorrField, hourAngleField, zenithDistanceField;

        static int startDragX, startDragY, endDragX, endDragY;
        static int startDragSubImageX, startDragSubImageY;
        static int startDragScreenX, startDragScreenY;
        static int newPositionX, newPositionY;

        static String temptextfield;
        static Integer tempIntField;

        static Plot plot;
        static ImagePlus iplus;
        static int sliceCol;
        static JFrame mainFrame;
        static JFrame subFrame, refStarFrame, addAstroDataFrame;
        static JFrame getMaxCurvesFrame;
//        static Frame plotFrame;
        static Panel plotpanel;
        static JPanel mainpanel;
        static JScrollPane mainscrollpane;
        static JPanel mainsubpanel, refStarMainPanel;
        static JScrollPane subscrollpane, refStarScrollPane;

        static JMenuBar mainmenubar;
        static JMenu filemenu;
        static JMenu preferencesmenu;
        static JMenu xaxismenu;
        static JMenu yaxismenu;
        static JMenu tablemenu;
        static JCheckBoxMenuItem opendatasetCB, openrefstarCB;
        static JCheckBoxMenuItem usewidedataCB, divideNotSubtractCB;
        static JCheckBoxMenuItem rememberwindowlocationsCB;
        static JCheckBoxMenuItem showtooltipsCB;
        static JCheckBoxMenuItem usedefaultsettingsCB;
        static JCheckBoxMenuItem usepixelscaleCB;
        static JCheckBoxMenuItem useBoldedDatumCB, useUpdateStackCB;
        static JCheckBoxMenuItem xTicsCB, yTicsCB, xGridCB, yGridCB, xNumbersCB, yNumbersCB;
        static JRadioButtonMenuItem showXAxisNormalCB, showXAxisAsPhaseCB,showXAxisAsHoursSinceTcCB, showXAxisAsDaysSinceTcCB;
        static ButtonGroup xPhaseGroup;
        static ButtonGroup[] morelegendRadioGroup, detrendVarRadioGroup;

        static JMenuItem setephemerismenuitem, setrefstarhorzsizemenuitem;
        static JMenuItem changepixelscalemenuitem, setverticalmarkertextmenuitem;
        static JMenuItem savedatamenuitem, saveimagemenuitem, savedataconfigmenuitem, saveplotconfigmenuitem,
                         opendatamenuitem, openplotconfigmenuitem, opendataconfigmenuitem, saveplottemplatemenuitem, openplottemplatemenuitem,
                         opentablemenuitem, transposetablemenuitem, addnewcolsfromplotmenuitem, addnewcolsfromastroCCmenuitem;
        static JCheckBoxMenuItem usePriorityColumnsCB;
        static JMenuItem changePriorityColumnsmenuitem;
        static JMenuItem changemaxdatalengthmenuitem;

        static JTextField titleField, subtitleField, legendField, xlegendField, ylegendField;
        static JLabel statusLabel, xcolumnlabel, ycolumnslabel, ycolumnlabel, operatorlabel;
        static JLabel xcolumnlabelsub, ycolumnlabelsub, operatorlabelsub;
        static JLabel opcolumnlabel, markerlabel, markercolorlabel;
        static JLabel opcolumnlabelsub, markerlabelsub, markercolorlabelsub;
        static JSlider titlePosYSlider;
        static JSlider titlePosXSlider;
        static JSlider subtitlePosYSlider;
        static JSlider subtitlePosXSlider;
        static JSlider legendPosYSlider;
        static JSlider legendPosXSlider;
        static JComboBox xdatacolumndefault;

        static JRadioButton autoxButton, firstxButton, customxButton;
        static JRadioButton autoyButton, firstyButton, customyButton;
        
        static JRadioButton useGJDButton, useHJDButton, useBJDButton;
        static boolean useGJD = true, useHJD = false, useBJD = false;
        static ButtonGroup JDRadioGroup;

        static SpinnerModel mmagrefsmodel;
        static JSpinner mmagrefsspinner;
        static SpinnerModel maxcurvesmodel, maxdetrendvarsmodel;
        static JSpinner maxcurvesspinner, maxdetrendvarsspinner;

        static JSpinner xwidthstepspinner,xmaxstepspinner, xminstepspinner,ymaxstepspinner,yminstepspinner;
        static JSpinner plotwidthstepspinner, plotheightstepspinner;

        static JPopupMenu xwidthsteppopup, xmaxsteppopup, xminsteppopup,ymaxsteppopup,yminsteppopup;
        static JPopupMenu plotwidthsteppopup, plotheightsteppopup, trimdataheadsteppopup, trimdatatailsteppopup;
        static JPopupMenu xaxispopup, yaxispopup, legendpopup;
        static JPopupMenu vmarkersteppopup;
        static JPopupMenu dmarkersteppopup;

        static SpinnerModel xwidthstepmodel, xmaxstepmodel, xminstepmodel,ymaxstepmodel,yminstepmodel;
        static SpinnerModel plotwidthstepmodel, plotheightstepmodel;
        static SpinnerModel xwidthspinnermodel, xminspinnermodel, xmaxspinnermodel;
        static SpinnerModel yminspinnermodel, ymaxspinnermodel, plotheightspinnermodel, plotwidthspinnermodel;
        static SpinnerModel xmultipliermodel, ymultipliermodel;
        static SpinnerModel trimdataheadmodel, trimdatatailmodel, trimdataheadstepmodel, trimdatatailstepmodel;
        static SpinnerModel vmarkerstepmodel, vmarker1spinnermodel, vmarker2spinnermodel;
        static SpinnerModel dmarkerstepmodel, dmarker1spinnermodel, dmarker2spinnermodel, dmarker3spinnermodel, dmarker4spinnermodel;

        static JSpinner xwidthspinner, xminspinner, xmaxspinner, xmultiplierspinner, ymultiplierspinner;
        static JSpinner yminspinner, ymaxspinner, plotheightspinner, plotwidthspinner;
        static JSpinner trimdataheadspinner, trimdatatailspinner, trimdataheadstepspinner, trimdatatailstepspinner;
        static JSpinner vmarkerstepspinner, vmarker1spinner, vmarker2spinner;
        static JSpinner dmarkerstepspinner, dmarker1spinner, dmarker2spinner, dmarker3spinner, dmarker4spinner;        


//        static JCheckBox usextickscheckbox, useytickscheckbox, usexgridcheckbox;
//        static JCheckBox useygridcheckbox, usexnumberscheckbox, useynumberscheckbox;
        static JCheckBox showVMarker1CB, showVMarker2CB;
        static JCheckBox showDMarkersCB, useDMarker1CB, useDMarker4CB;

        static JButton moreybutton, closebutton, grabautoxbutton, grabautoybutton, updateplotbutton, refStarButton, OKbutton;

        static String tableName;
        static String[] columns, unfilteredColumns, oldUnfilteredColumns;
        static String[] columnswd, columnsDetrend;
        
        static MeasurementTable table;
        static TextPanel tpanel;
        static JLabel dummylabel1, binsizespinnerlabel, mmagrefsspinnerlabel, dummylabel4;

        static ImageWindow plotWindow;
        static ImagePlus plotImage;
//        static ImageProcessor ip;
        static ImageCanvas plotImageCanvas;
        static OverlayCanvas plotOverlayCanvas;
        static Panel plotbottompanel;
        static Label plotcoordlabel;
        static Vector<Roi> list;
        static GeneralPath path;
        static Roi roi;
        static TitledBorder dataSectionBorder = BorderFactory.createTitledBorder("Data ()");;        


        static int maxCurves, maxDetrendVars;
        static int[] detrendVarDisplayed;
        static int maxColumnLength;
        static double vMarkerStep=0.01;
        static double vMarker1Value=0.5, vMarker2Value=0.7;
        static double dMarkerStep=0.01;
        static double dMarker1Value=0.3, dMarker2Value=0.5, dMarker3Value=0.7, dMarker4Value=0.9;        
        static String[] xlabel;
        static String[] ylabel;
        static String[][] detrendlabel;
        static String[][] detrendlabelhold;
        static int[] xcolumn;
        static int[] ycolumn;
        static int[] errcolumn;
        static int[] operrcolumn;
        static int[][] detrendcolumn;
        static int[] xc1column, yc1column,xc2column,yc2column;
        static String[] oplabel;
        static int[] opcolumn;
//        static String[] errlabel;
//        static String[] operrlabel;
        static boolean[] showErrors;  //user has requested to plot auto-error
        static boolean[] showErr; //retains showErrors from previous plot update
        static boolean[] showOpErrors;
        static boolean[] lines;
        static int[] marker;
        static Color[] color;
        static int[] markerIndex;
        static int[] colorIndex;
        static int[] operatorIndex;
        static String[] cblabels;
        static boolean[] moreOptions;
        static double[] autoScaleFactor;
        static double[] autoScaleStep;
        static double[] autoShiftFactor;
        static double[] autoShiftStep;
        static double[] customScaleFactor;
        static double[] customScaleStep;
        static double[] customShiftFactor;
        static double[] customShiftStep;
        static double[][] detrendFactor;
        static double[][] detrendFactorStep;        
        static double[] baseline;
        static double[] subtotalScaleFactor;
        static double[] subtotalShiftFactor;
        static double[] totalScaleFactor;
        static double[] totalShiftFactor;
        static double[] yWidth;
        static double[] xMinimum;
        static double[] xMaximum;
        static double[] yMinimum;
        static double[] yMn;
        static double[] yMaximum;
        static double[] yMx;
        static double[] yMidpoint;
        static double[] nullX = null;
        static double[] nullY = null;

        static boolean[] mmag;
        static boolean[] fromMag;
        static int[] normIndex;
        static int[] detrendFitIndex;
        static int[][] detrendIndex;
        static boolean[] plotY;
        static boolean[] useColumnName;
        static boolean[] useLegend;
        static boolean[] force;
        static boolean[] ASInclude;

        static int[] binSize;

        static String[] legend;
        static String[] xyc1label;
        static String[] xyc2label;

        static JCheckBox[] usecurvebox;
        static JCheckBox[] errorcolumnbox;
        static JCheckBox[] operrorcolumnbox;
        static JCheckBox[] usebinbox;
        static JCheckBox[] usemmagbox, fromMagBox;
        static JComboBox[] normtypecombobox;
        static JComboBox[] detrendtypecombobox;
        static JComboBox   jdcolumnbox;
        static JCheckBox[] uselinesbox;
        static JCheckBox[] autoscaleincludebox;
        static JCheckBox[] forcebox;
        
        static JPanel[] refStarPanel;
        static JLabel[] refStarLabel;
        static JCheckBox[] refStarCB;
        static boolean[] isRefStar;
        static int numAps;

        static JComboBox[] xdatacolumn;
        static JComboBox[] ydatacolumn;
        static JComboBox[] operatorselection;
        static JComboBox[] operatorcolumn;
        static JComboBox[] detrendbox;
        static JComboBox[] markersymbolselection;
        static JComboBox[] markercolorselection;

        static SpinnerModel[] binsizespinnermodel;
        static JSpinner[] binsizespinner;
        static JPanel[] morelegendradiopanelgroup;
        static JPanel[] autoscalepanelgroup;
        static JPanel[] customscalepanelgroup;
        static JPanel[] detrendpanelgroup, normpanelgroup;
        static JPanel[] savecolumnpanelgroup;
        static JRadioButton[] legendnoneButton;
        static JRadioButton[] legendcustomNameButton;
        static JRadioButton[] legendcolumnNameButton;
        static JRadioButton[][] detrendVarButton;
        static JRadioButton[] autoscaleButton;
        static JRadioButton[] customscaleButton;
        static ButtonGroup[] autoscaleRadioGroup;
        static JTextField[] morelegendField;
        static SpinnerModel[] autoshiftmodel;
        static SpinnerModel[] autoshiftstepmodel;
        static SpinnerModel[] autoscalemodel;
        static SpinnerModel[] autoscalestepmodel;
        static SpinnerModel[] customshiftmodel;
        static SpinnerModel[] customshiftstepmodel;
        static SpinnerModel[] customscalemodel;
        static SpinnerModel[] customscalestepmodel;
        static SpinnerModel[] detrendfactormodel;
        static SpinnerModel[] detrendfactorstepmodel;
        static JSpinner[] autoshiftspinner;
        static JSpinner[] autoshiftstepspinner;
        static JSpinner[] autoscalespinner;
        static JSpinner[] autoscalestepspinner;
        static JSpinner[] customshiftspinner;
        static JSpinner[] customshiftstepspinner;
        static JSpinner[] customscalespinner;
        static JSpinner[] customscalestepspinner;
        static JSpinner[] detrendfactorspinner;
        static JSpinner[] detrendfactorstepspinner;        
        static JPopupMenu[] autoscalesteppopup;
        static JPopupMenu[] autoshiftsteppopup;
        static JPopupMenu[] customscalesteppopup;
        static JPopupMenu[] customshiftsteppopup;
        static JPopupMenu[] detrendfactorsteppopup;
        static JPanel[] autoscalesteppanel;
        static JPanel[] customscalesteppanel;
        static JPanel[] autoshiftsteppanel;
        static JPanel[] customshiftsteppanel;
        static JPanel[] detrendfactorsteppanel;
        static JLabel[] autoscalesteplabel;
        static JLabel[] autoshiftsteplabel;
        static JLabel[] customscalesteplabel;
        static JLabel[] customshiftsteplabel;
        static JLabel[] detrendfactorsteplabel;
        static JLabel[] curvelabel;
        static JLabel[] othercurvelabel;
        static JPanel[] grabautopanel;
        static JButton[] grabautobutton, savecolumnbutton;
        static ImageIcon forceIcon;
        static ImageIcon configureIcon;
        static ImageIcon helpIcon;
        static ImageIcon plotIcon;
        static ImageIcon editTextIcon, insertColumnIcon, copyVMarkersIcon;
        static JButton legendconfigureButton;
        static JButton vmarker1edittextButton, vmarker2edittextButton;
        static JButton xlegendconfigureButton;
        static JButton ylegendconfigureButton;
        static JButton legendslabelconfigureButton;
        static JPanel legendgroup;
        static JPanel mainpanelgroupa;
        
        static Color defaultBackground;
        static Color defaultOKForeground;
        

        static JSpinner.NumberEditor[] customscaleeditor;
        static DecimalFormat[] customscaleformat;
        static JSpinner.NumberEditor[] customshifteditor;
        static DecimalFormat[] customshiftformat;
        static JSpinner.NumberEditor[] detrendfactoreditor;
        static DecimalFormat[] detrendfactorformat;        

        static DecimalFormat fourPlaces;

        static String[] markers;
        static String[] colors;
        static ImageIcon[] normiconlist;
        static ImageIcon[] detrendiconlist;

        static String[] operators;

        static String[] opSymbol;

        static String[] spinnerscalelist;

        static String[] integerspinnerscalelist;
        
        static TimerTask delayedUpdateTask;
        static java.util.Timer delayedUpdateTimer;   
        
        static AstroConverter acc;

       
    
	public void run(String inTableName)
		{
        //SET DEFAULT SYSTEM LOOK AND FEEL
        if (IJ.isWindows())
                {
//                        try {UIManager.setLookAndFeel("javax.swing.plaf.metal.MetalLookAndFeel");}
                try {UIManager.setLookAndFeel("com.sun.java.swing.plaf.windows.WindowsLookAndFeel");}
//                        try {UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());}
                catch (Exception e) { }
                }
        else if (IJ.isLinux())
                {
                try {UIManager.setLookAndFeel("javax.swing.plaf.metal.MetalLookAndFeel");}
                catch (Exception e) { }
                }
//                    try {UIManager.setLookAndFeel("com.sun.java.swing.plaf.windows.WindowsLookAndFeel");}
//                    try {UIManager.setLookAndFeel("com.sun.java.swing.plaf.nimbus.NimbusLookAndFeel");}
//                    try {UIManager.setLookAndFeel("com.sun.java.swing.plaf.motif.MotifLookAndFeel");}
        else if (IJ.isMacOSX())
                {

                try {UIManager.setLookAndFeel("javax.swing.plaf.metal.MetalLookAndFeel");}
//                        try {UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());}
                catch (Exception e) { }
                }

        plotAutoMode = Prefs.get("plot.automode",plotAutoMode);
        if (plotAutoMode)
            {
            updatePlot();
            plotAutoMode = false;
            Prefs.set("plot.automode",false);
            return;
            }
        
        if (mainFrame == null || !mainFrame.isShowing())
            {
            if (mainFrame != null)
                {
                mainFrame.dispose();
                subFrame.dispose();
                refStarFrame.dispose();
                mainFrame=null;
                subFrame=null;
                refStarFrame=null;
                }
            initializeVariables();
            setupArrays();
            useDefaultSettings = Prefs.get("plot2.useDefaultSettings", useDefaultSettings);
            if (useDefaultSettings)
                    {
                    useDefaultSettings = false;
                    Prefs.set("plot2.useDefaultSettings", useDefaultSettings);
                    }
            else
                    getPreferences();
            }

		// FIND ALL TABLES
        if (inTableName != null && !inTableName.equals("")) requestedTableName = inTableName;
        findTables(false);
        }


    static void findTables(boolean forceUpdate)
        {
        String[] tables = MeasurementTable.getMeasurementTableNames();
        if (tables == null || tables.length == 0)
            {
            makeDummyTable();
            }
        else
            {
            //filter out B&C, Log, Errors windows if they exist
            int j = 0;
            for (int i=0; i < tables.length; i++)
                if (tables[i].equals("B&C") || tables[i].startsWith("Log") || tables[i].equals("Errors"))
                    j++;
            String[] filteredTables = new String[tables.length-j];
            j = 0;
            for (int i=0; i < tables.length; i++)
                if (!tables[i].equals("B&C") && !tables[i].startsWith("Log") && !tables[i].equals("Errors"))
                    {
                    filteredTables[j] = tables[i];
                    j++;
                    }
            if (requestedTableName != null)
                {
                for (int i=0; i < filteredTables.length; i++)
                    {
                    if (requestedTableName.equals(filteredTables[i]))
                        {
                        tableName = requestedTableName;
                        requestedTableName = null;
                        setTable(MeasurementTable.getTable(tableName), forceUpdate);
                        break;
                        }
                    }
                if (requestedTableName != null)
                    {
                    requestedTableName = null;
                    makeDummyTable();
                    }
                }
            else if (filteredTables.length == 0)
                {
                makeDummyTable();
                }
            else if (filteredTables.length == 1)
                {
                tableName = filteredTables[0];
                setTable(MeasurementTable.getTable(tableName), forceUpdate);
                }
            else                            // IF MORE THAN ONE, ASK WHICH TABLE SHOULD BE USED
                {
                GenericDialog gd = new GenericDialog("Plot Table Columns");
                gd.addMessage ("Select table to be plotted.");
                gd.addChoice ("Table to be plotted",filteredTables,"Measurements");
                gd.showDialog();
                if (gd.wasCanceled())
                    {
                    makeDummyTable();
                    }
                else
                    {
                    tableName = gd.getNextChoice();
                    setTable(MeasurementTable.getTable(tableName), forceUpdate);
                    }
                }
            }
        }
    static public void setTable(MeasurementTable inTable, boolean forceUpdate)
        {
        table = inTable;
        if (table == null)
            {
            makeDummyTable();
            }
        else
            {
            tableName = table.shortTitle();
            tpanel = MeasurementTable.getTextPanel(MeasurementTable.longerName(tableName));
            dataSectionBorder.setTitle("Data ("+MeasurementTable.shorterName(tableName) +")");                 
            unfilteredColumns = table.getColumnHeadings().split("\t");
            if (unfilteredColumns == null || unfilteredColumns.length == 0)
                {
                IJ.showMessage ("No data columns to plot.");
                makeDummyTable();
                }
            else
                {
                updateColumnLists();
                Prefs.set("plot2.tableName",tableName);
                finishSetup(forceUpdate);
                }
            }
        }
    
    static public MeasurementTable getTable()
        {
        return table;
        }

    static void finishSetup(boolean forceUpdate)
        {
        if (mainFrame == null || !mainFrame.isShowing())
            showMainJPanel();
//        if (!Arrays.equals(unfilteredColumns, oldUnfilteredColumns))
//            {
//            IJ.log("columns are different");
//            for (int i=0; i<(oldUnfilteredColumns==null?0:oldUnfilteredColumns.length); i++)
//                IJ.log("oldUnfilteredColumn["+i+"]="+oldUnfilteredColumns[i]);
//            for (int i=0; i<(unfilteredColumns==null?0:unfilteredColumns.length); i++)
//                IJ.log("unfilteredColumn["+i+"]="+unfilteredColumns[i]);              
//            }
        else if (forceUpdate || !Arrays.equals(unfilteredColumns, oldUnfilteredColumns))
            updatePanels();
        oldUnfilteredColumns = unfilteredColumns.clone();

        updatePlot();
		}

//------------------------------------------------------------------------------

static public void makeDummyTable()
        {
        tableName = "No Table Selected";
        table = null; //MeasurementTable.getTable(tableName);
        dataSectionBorder.setTitle("Data ("+MeasurementTable.shorterName(tableName) +")");
        unfilteredColumns = new String[1];
        unfilteredColumns[0] = "";
        columns = unfilteredColumns.clone();
        columnsDetrend = unfilteredColumns.clone();
        columnswd = unfilteredColumns.clone();
        finishSetup(true);
//        Prefs.set("plot2.tableName",tableName);
        }

static public void updateColumnLists()
        {
        int j = 0;
        for (int i=0; i < unfilteredColumns.length; i++)
                if (unfilteredColumns[i].equals("Label") || unfilteredColumns[i].equals("image"))
                       j++;
        columns = new String[unfilteredColumns.length - j];
        j = 0;
        for (int i=0; i < unfilteredColumns.length; i++)
                if (!unfilteredColumns[i].equals("Label") && !unfilteredColumns[i].equals("image") )
                        {
                        columns[j] = unfilteredColumns[i];
                        j++;
                        }
        columnsDetrend = columns.clone();
        if (prioritizeColumns) sortColumns();
        columnswd = new String[columns.length+1];
        columnswd[0] = columns[0];
        columnswd[1] = "default";
        for (int i=2; i< columnswd.length; i++)
            columnswd[i]=columns[i-1];
        }

    static public void sortColumns()
        {
        String[] pcols = priorityColumns.split(",");
        if (pcols.length < 1 || columns.length < 2) return;
        String[] cols = columns.clone();
        int j = 1;
        String pcol;
        for (int k=0; k<pcols.length; k++)
            {
            pcol = pcols[k].toLowerCase().trim();
            if (pcol.equals("rel_flux") || pcol.equals("rel_flux_") || pcol.equals("rel_flux_t") || pcol.equals("rel_flux_c"))
                {     //treat rel_flux, rel_flux_T, rel_flux_C as a special cases since rel_flux matches rel_flux_err, etc.
                for (int i=1; i<cols.length; i++)
                    {
                    if (cols[i].toLowerCase().startsWith("rel_flux_t") || cols[i].toLowerCase().startsWith("rel_flux_c"))
                        {
                        columns[j]=cols[i];
                        cols[i]="";
                        j++;
                        }

                    }
                }
            else
                {
                for (int i=1; i<cols.length; i++)
                    {                
                    if (cols[i].toLowerCase().startsWith(pcol))
                        {
                        columns[j]=cols[i];
                        cols[i]="";
                        j++;
                        }
                    }
                }
            
            }
        for (int i=1; i<cols.length; i++) //get all non-priority columns
            {
            if (!cols[i].equals(""))
                {
                columns[j]=cols[i];
                j++;
                }

            }

        pcols = priorityDetrendColumns.split(",");
        if (pcols.length < 1 || columnsDetrend.length < 2) return;
        cols = columnsDetrend.clone();
        j = 1;
        for (int k=0; k<pcols.length; k++)
            {
            pcol = pcols[k].toLowerCase().trim();
            if (pcol.equals("rel_flux") || pcol.equals("rel_flux_") || pcol.equals("rel_flux_t") || pcol.equals("rel_flux_c"))
                {
                for (int i=1; i<cols.length; i++)
                    {
                    if (cols[i].toLowerCase().startsWith("rel_flux_t") || cols[i].toLowerCase().startsWith("rel_flux_c"))
                        {
                        columnsDetrend[j]=cols[i];
                        cols[i]="";
                        j++;
                        }

                    }
                }
            else
                {
                for (int i=1; i<cols.length; i++)
                    {                
                    if (cols[i].toLowerCase().startsWith(pcol))
                        {
                        columnsDetrend[j]=cols[i];
                        cols[i]="";
                        j++;
                        }
                    }
                }
            
            }
        for (int i=1; i<cols.length; i++)
            {
            if (!cols[i].equals(""))
                {
                columnsDetrend[j]=cols[i];
                j++;
                }

            }

        }


    static public void updatePanels()
        {
        if(mainFrame.isShowing())
            {
            if(subFrame.isShowing())
                {
                subFrameLocationX = subFrame.getLocation().x;
                subFrameLocationY = subFrame.getLocation().y;
                Prefs.set("plot2.subFrameLocationX",subFrameLocationX);
                Prefs.set("plot2.subFrameLocationY",subFrameLocationY);
                }
            subFrame.dispose();
            if(refStarFrame != null)
                {
                if (refStarFrame.isShowing()) 
                    refStarPanelWasShowing = true;
                else
                    refStarPanelWasShowing = false;
                closeRefStarFrame();
                }
            else
                {
                refStarPanelWasShowing = false;
                }
            if(addAstroDataFrame != null)
                {
                if (addAstroDataFrame.isShowing()) 
                    addAstroDataFrameWasShowing = true;
                else
                    addAstroDataFrameWasShowing = false;
                closeAddAstroDataFrame();
                }
            else
                {
                addAstroDataFrameWasShowing = false;
                }
            if (acc != null)
                {
                acc.saveAndClose();
                }
            mainFrameLocationX = mainFrame.getLocation().x;
            mainFrameLocationY = mainFrame.getLocation().y;
            Prefs.set("plot2.mainFrameLocationX",mainFrameLocationX);
            Prefs.set("plo2.mainFrameLocationY",mainFrameLocationY);
            mainFrame.dispose();
            }
        showMainJPanel();
        oldUnfilteredColumns = unfilteredColumns.clone();
        }

    static public void setPlotAutoMode(boolean mode)
        {
        plotAutoMode = mode;
        }
    
    static public void clearPlot()
        {
        checkAndLockTable();
        table = new MeasurementTable(tableName);
        table.setLock(false);
//        table.show();        
        if (plot != null)
            {
            plotImage = WindowManager.getImage("Plot of "+tableName);
            if (plotImage != null)
                {
                plot = new Plot ("Plot of "+tableName,xlab,ylab,nullX,nullY,plotOptions);
                plot.setSize(plotSizeX, plotSizeY);
                plot.setLimits (plotMinX, plotMaxX, plotMinY, plotMaxY);
                plot.setJustification(plot.CENTER);
                if (useTitle)
                        {
                        plot.changeFont(new java.awt.Font("Dialog",java.awt.Font.PLAIN,18));
                        plot.addTitle(titlePosX, titlePosY/plotSizeY>1?1:titlePosY/plotSizeY,title);
                        }
                if (useSubtitle)
                        {
                        plot.changeFont(new java.awt.Font("Dialog",java.awt.Font.PLAIN,14));
                        plot.addTitle(subtitlePosX, subtitlePosY/plotSizeY>1?1:subtitlePosY/plotSizeY,subtitle);
                        }
                plot.draw();
                ImageProcessor ip = plot.getProcessor();
                plotImage.setProcessor("Plot of "+tableName,ip);
                }
            }
        else
            {
            IJ.showStatus("No plot to clear");
            }
//        IJ.log("deleting rows");
//        for (int i=table.getCounter(); i>=0; i--)
//            {
//            IJ.log("deleting row "+i);
//            table.deleteRow(i);
//            }
//        table.show();
        }
    
	/**
	 * Starts timer which handles simultaneous updatePlot requests.
	 */
	static protected void startDelayedUpdateTimer ()
        {
		try	{
            if (delayedUpdateTimer != null) delayedUpdateTimer.cancel();
            if (delayedUpdateTask != null) delayedUpdateTask.cancel();
            
			delayedUpdateTask = new TimerTask()
                {
                public void run()
                    {
                    updatePlot();
//                    IJ.log("delayed updatePlot() completed");
                    }
                };
            delayedUpdateTimer = new java.util.Timer();
            delayedUpdateTimer.schedule(delayedUpdateTask,200);
            }
        catch(Exception e)
            {
            IJ.showMessage ("Error starting delayed plot update timer : "+e.getMessage());
            }
        }    


//------------------------UPDATEPLOT()---------------------------------------

        static public void updatePlot()
                {
                if ((table != null && table.isLocked()) || !updatePlotEnabled || updatePlotRunning)
                    {
//                    IJ.log("starting delayed updatePlot() timer");
                    startDelayedUpdateTimer();
                    return;
                    }
                updatePlotRunning = true;
                checkAndLockTable();
//                table = MeasurementTable.getTable (tableName);
//                tpanel = MeasurementTable.getTextPanel(MeasurementTable.longerName(tableName));
                
                if (tableName.equals("No Table Selected"))
                    {
                    updatePlotRunning = false;
                    if (table != null) table.setLock(false);
                    return;
                    }

                n = table.getCounter();
                if (n < 1)
                        {
//                        clearPlot();
                        updatePlotRunning = false;
                        if (table != null) table.setLock(false);
                        return;
                        }
                if (n > maxColumnLength)
                    {
                    maxColumnLength = n > 2*maxColumnLength ? n : 2*maxColumnLength;
                    setupDataBuffers();
                    }
                unfilteredColumns = table.getColumnHeadings().split("\t");
                if (!Arrays.equals(unfilteredColumns, oldUnfilteredColumns))
                    {
                    updatePlotRunning = false;
                    if (table != null) table.setLock(false);                  
                    updateColumnLists();
                    updatePanels();
                    }

                savePreferences();

                for (int curve=0; curve<maxCurves; curve++)
                        {
                        if (binSize[curve] < 1)
                            {
                            binSize[curve] = 1;
                            binsizespinner[curve].setValue(1);
                            }
                        }

                holdExcludedHeadSamples = excludedHeadSamples;
                holdExcludedTailSamples = excludedTailSamples;
                if (excludedHeadSamples + excludedTailSamples >= n)  //Handle case for more samples excluded than in dataset
                    {
                    excludedHeadSamples = excludedHeadSamples < n ? excludedHeadSamples : n - 1;
                    excludedTailSamples = n - excludedHeadSamples - 1;
                    }
                excluded = excludedHeadSamples + excludedTailSamples;

                int magSign = negateMag?1:-1;
                for (int i=0;i<maxCurves;i++)
                        {
                        marker[i] = markerOf(markerIndex[i]);
                        color[i] = colorOf(colorIndex[i]);
                        }

                // INITIALIZE BASELINE FOR RELATIVE MMAG CALCULATIONS
                for (int i=0; i<maxCurves;i++)
                        {
                        baseline[i]=0;
                        }
                // GET DATA

                if (n > x[0].length)
                    {
                    nn = new int[maxCurves];
                    xlabel2 = new String[maxCurves];
                    x = new double[maxCurves][n];
                    y = new double[maxCurves][n];
                    yerr = new double[maxCurves][n];
                    yop = new double [maxCurves][n];
                    yoperr = new double [maxCurves][n];
                    detrend = new double [maxCurves][maxDetrendVars][n];
                    xc1 = new double [maxCurves][n];
                    xc2 = new double [maxCurves][n];
                    yc1 = new double [maxCurves][n];
                    yc2 = new double [maxCurves][n];
                    }

                xMultiplierFactor = Math.pow(10, (double)xExponent);
                yMultiplierFactor = Math.pow(10, (double)yExponent);
                
                int[] detrendVarsUsed = new int[maxCurves];
                
                for (int curve=0; curve<maxCurves; curve++)
                        {
                        nn[curve] = (n-excluded)/binSize[curve];
                        nnr[curve] = (n-excluded)%binSize[curve];
                        if (nnr[curve]>0) nn[curve]++;
                        detrendVarsUsed[curve] = 0;
                        }
                
                for (int curve=0; curve<maxCurves; curve++)
                        {
                        if (xlabel[curve].trim().length() == 0 ||
                            (xlabel[curve].equalsIgnoreCase("default") && xlabeldefault.trim().length() == 0))
                            {
                            for (int j=0; j < nn[curve]; j++)
                                    x[curve][j] = (double)(j+1);
                            xlabel2[curve]="Sample Number";
                            }
                        else
                            {
                            if (xlabel[curve].equalsIgnoreCase("default"))
                                {
                                xcolumn[curve] = table.getColumnIndex (xlabeldefault);
                                if (xcolumn[curve] == ResultsTable.COLUMN_NOT_FOUND)
                                    {
                                    xcolumn[curve] = 0;
                                    plotY[curve] = false;
                                    usecurvebox[curve].setSelected(false);
                                    savecolumnbutton[curve].setEnabled(plotY[curve]);
//                                    if (xlabeldefault.trim().length() != 0)
//                                        IJ.showMessage ("Cannot access default X-axis table column "+xlabeldefault+" !");
//                                    xlabeldefault = columns[0];
//                                    return;
                                    }
                                xlabel2[curve]=xlabeldefault.trim();
                                }
                            else 
                                {
                                xcolumn[curve] = table.getColumnIndex (xlabel[curve]);
                                if (xcolumn[curve] == ResultsTable.COLUMN_NOT_FOUND)
                                        {
                                        xcolumn[curve] = 0;
                                        plotY[curve] = false;
                                        usecurvebox[curve].setSelected(false);
                                        savecolumnbutton[curve].setEnabled(plotY[curve]);
//                                        if (xlabel[curve].trim().length() != 0)
//                                            IJ.showMessage ("Cannot access X-axis table column "+xlabel[curve]+" for dataset "+(curve+1)+"!");
//                                        xlabel[curve] = "default";
//                                        xdatacolumn[curve].setSelectedItem(xlabel[curve]);
//                                        return;
                                        }
                                xlabel2[curve]=xlabel[curve].trim();
                                }
                            
                            if (plotY[curve])
                                {
                                int bucketSize = binSize[curve];

                                for (int j=0; j < nn[curve]; j++)
                                    {
                                    double xin = 0;
                                    int numNaN = 0;
                                    if (nnr[curve] > 0 && j == nn[curve] - 1)
                                        {
                                        bucketSize = nnr[curve];
                                        }
                                    else
                                        {
                                        bucketSize = binSize[curve];
                                        }
                                    x[curve][j] = 0;
                                    for (int k=excludedHeadSamples; k < (bucketSize+excludedHeadSamples); k++)
                                        {
                                        xin = table.getValueAsDouble (xcolumn[curve],j*binSize[curve]+k);
                                        if (Double.isNaN(xin))
                                            {
                                            numNaN += 1;
                                            if (numNaN == bucketSize)
                                                {
                                                bucketSize = 1;
                                                x[curve][j] = Double.NaN;
                                                numNaN = 0;
                                                break;
                                                }                                               
                                            }
                                        else
                                            {
                                            x[curve][j] += xin;
                                            }
                                        }
                                    bucketSize -= numNaN;
                                    x[curve][j] = x[curve][j]*xMultiplierFactor/(double)bucketSize;
                                    }
                                if (xlabel2[curve].startsWith("J.D.-2400000"))
                                    {
                                    for (int j=0; j < nn[curve]; j++)
                                        {
                                        x[curve][j] += 2400000;
                                        }
                                    }
                                if (!showXAxisNormal)
                                    {
                                    if (showXAxisAsPhase)
                                        {
                                        for (int j=0; j < nn[curve]; j++)
                                            {
                                            x[curve][j] = ((x[curve][j] - T0)%period)/period;
                                            if (x[curve][j] > 0.5) x[curve][j] -= 1.0;
                                            }
                                        }   
                                    else if (showXAxisAsDaysSinceTc)
                                        {
                                        double halfPeriod = period/2.0;
                                        for (int j=0; j < nn[curve]; j++)
                                            {
                                            x[curve][j] = ((x[curve][j] - T0)%period);
                                            if (x[curve][j] > halfPeriod) x[curve][j] -= period;
                                            }                                      
                                        } 
                                    else if (showXAxisAsHoursSinceTc)
                                        {
                                        double halfPeriod = period/2.0;
                                        for (int j=0; j < nn[curve]; j++)
                                            {
                                            x[curve][j] = ((x[curve][j] - T0)%period);
                                            if (x[curve][j] > halfPeriod) x[curve][j] -= period;
                                            x[curve][j] *= 24;
                                            }                                      
                                        }
                                    }
                                }
                            }
                        }
                for (int curve=0;curve<maxCurves;curve++)
                        {
                        if (plotY[curve])
                                {
                                if (ylabel[curve].trim().length() == 0)
                                        {
                                        for (int j=0; j < nn[curve]; j++)
                                                y[curve][j] = (double)(j+1);
                                        }
                                else	{
                                        ycolumn[curve] = table.getColumnIndex (ylabel[curve]);
                                        if (ycolumn[curve] == ResultsTable.COLUMN_NOT_FOUND)
                                                {
//                                                if (ylabel[curve].trim().length() != 0)
//                                                    IJ.log ("Cannot access Y"+curve+" table column "+ylabel[curve]+" !");
                                                ycolumn[curve] = 0;
                                                plotY[curve] = false;
                                                usecurvebox[curve].setSelected(false);
                                                savecolumnbutton[curve].setEnabled(plotY[curve]);
                                                }
                                        else if(operatorIndex[curve] == 5)  //calculate distance
                                                {
                                                xc1column[curve] = table.getColumnIndex (getPositionColumn(ylabel[curve],"X"));
                                                yc1column[curve] = table.getColumnIndex (getPositionColumn(ylabel[curve],"Y"));
                                                xyc1label[curve] = getSourceID(ylabel[curve]);
                                                if (xc1column[curve] == ResultsTable.COLUMN_NOT_FOUND )
                                                        {
                                                        if (getPositionColumn(ylabel[curve],"X").trim().length() != 0)
                                                            IJ.showMessage ("Cannot access "+getPositionColumn(ylabel[curve],"X")+" source position data column for distance calculation !");
                                                        operatorIndex[curve] = 0;
                                                        operatorselection[curve].setSelectedIndex(operatorIndex[curve]);
                                                        plotY[curve] = false;
                                                        usecurvebox[curve].setSelected(false);
                                                        savecolumnbutton[curve].setEnabled(plotY[curve]);
                                                        }
                                                if (yc1column[curve] == ResultsTable.COLUMN_NOT_FOUND )
                                                        {
                                                        if (getPositionColumn(ylabel[curve],"Y").trim().length() != 0)
                                                            IJ.showMessage ("Cannot access "+getPositionColumn(ylabel[curve],"Y")+" source position data column for distance calculation !");
                                                        operatorIndex[curve] = 0;
                                                        operatorselection[curve].setSelectedIndex(operatorIndex[curve]);
                                                        plotY[curve] = false;
                                                        usecurvebox[curve].setSelected(false);  
                                                        savecolumnbutton[curve].setEnabled(plotY[curve]);
                                                        }
                                                }
                                        if (detrendFitIndex[curve] != 0)
                                            {
                                            for (int v = 0; v < maxDetrendVars; v++)
                                                {
                                                if (detrendIndex[curve][v] != 0)
                                                    { 
                                                    detrendcolumn[curve][v] = table.getColumnIndex(detrendlabel[curve][v]);

                                                    if (detrendcolumn[curve][v] == ResultsTable.COLUMN_NOT_FOUND)
                                                        {
                                                        detrendIndex[curve][v] = 0;
                                                        if (detrendVarDisplayed[curve] == v) detrendbox[curve].setSelectedIndex(0);
                                                        }
                                                    detrendVarsUsed[curve]++;
                                                    }
                                                }
                                            }
                                        if (showErrors[curve] == true || operatorIndex[curve] == 6)
                                                {
                                                errcolumn[curve] = ResultsTable.COLUMN_NOT_FOUND;
                                                if (operatorIndex[curve] == 6)   //custom error
                                                    {
                                                    errcolumn[curve] = table.getColumnIndex (oplabel[curve]);
                                                    showErrors[curve] = true;
                                                    }
                                                else if (ylabel[curve].startsWith("rel_flux_T") || ylabel[curve].startsWith("rel_flux_C"))
                                                    {
                                                    errcolumn[curve] = table.getColumnIndex ("rel_flux_err_"+ylabel[curve].substring(9));
                                                    }
                                                else if (ylabel[curve].startsWith("Source-Sky_"))
                                                    {
                                                    errcolumn[curve] = table.getColumnIndex ("Source_Error_"+ylabel[curve].substring(11));
                                                    }
                                                else if (ylabel[curve].startsWith("tot_C_cnts"))
                                                    {
                                                    errcolumn[curve] = table.getColumnIndex ("tot_C_err"+ylabel[curve].substring(10));
                                                    }                                                
                                               
                                                if (errcolumn[curve] == ResultsTable.COLUMN_NOT_FOUND)
                                                    {
                                                    showErrors[curve] = false;
                                                    }
                                                }

                                        
//                                        if (!errlabel[curve].equals(""))
//                                                {
//                                                errcolumn[curve] = table.getColumnIndex (errlabel[curve]);	
//                                                if (errcolumn[curve] == ResultsTable.COLUMN_NOT_FOUND)
//                                                        {
////                                                        if (errlabel[curve].trim().length() != 0)
////                                                            IJ.showMessage ("Cannot access Y"+curve+" error column "+errlabel[curve]+" !");
//                                                        showErrors[curve] = false;
////                                                        errorcolumnbox[curve].setSelected(false);
//                                                        }
//                                                else
//                                                        showErrors[curve] = true;
//                                                }
//                                        else
//                                                showErrors[curve] = false;                                        
                                        

                                        if (operatorIndex[curve] != 0)
                                                {
                                                opcolumn[curve] = table.getColumnIndex (oplabel[curve]);	
                                                if (opcolumn[curve] == ResultsTable.COLUMN_NOT_FOUND)
                                                        {
//                                                        if (oplabel[curve].trim().length() != 0)
//                                                            IJ.showMessage ("Cannot access Y"+curve+" operator data column "+oplabel[curve]+" !");
                                                        operatorIndex[curve] = 0;
                                                        operatorselection[curve].setSelectedIndex(operatorIndex[curve]);
                                                        }
                                                else if(operatorIndex[curve] == 5) //calculate distance
                                                        {
                                                        xc2column[curve] = table.getColumnIndex (getPositionColumn(oplabel[curve],"X"));
                                                        yc2column[curve] = table.getColumnIndex (getPositionColumn(oplabel[curve],"Y"));
                                                        xyc2label[curve] = getSourceID(oplabel[curve]);
                                                        if (xc2column[curve] == ResultsTable.COLUMN_NOT_FOUND )
                                                                {
                                                                if (getPositionColumn(oplabel[curve],"X").trim().length() != 0)
                                                                    IJ.showMessage ("Cannot access "+getPositionColumn(oplabel[curve],"X")+" source position data column for distance calculation !");
                                                                operatorIndex[curve] = 0;
                                                                operatorselection[curve].setSelectedIndex(operatorIndex[curve]);
                                                                }
                                                        if (yc2column[curve] == ResultsTable.COLUMN_NOT_FOUND )
                                                                {
                                                                if (getPositionColumn(oplabel[curve],"Y").trim().length() != 0)
                                                                    IJ.showMessage ("Cannot access "+getPositionColumn(oplabel[curve],"Y")+" source position data column for distance calculation !");
                                                                operatorIndex[curve] = 0;
                                                                operatorselection[curve].setSelectedIndex(operatorIndex[curve]);
                                                                }
                                                        }
                                                }
//                                        if (!operrlabel[curve].equals(""))
//                                                {
//                                                operrcolumn[curve] = table.getColumnIndex (operrlabel[curve]);	
//                                                if (operrcolumn[curve] == ResultsTable.COLUMN_NOT_FOUND)
//                                                        {
////                                                        if (operrlabel[curve].trim().length() != 0)
////                                                            IJ.showMessage ("Cannot access Y"+curve+" operator error column "+operrlabel[curve]+" !");
//                                                        showOpErrors[curve] = false;
////                                                        operrorcolumnbox[curve].setSelected(false);
//                                                        }
//                                                else
//                                                        showOpErrors[curve] = true;
//                                                }
//                                        else
//                                                showOpErrors[curve] = false;
                                        
                                        if (operatorIndex[curve] > 0 && operatorIndex[curve] < 5 && showErrors[curve] == true)
                                            {
                                            operrcolumn[curve] = ResultsTable.COLUMN_NOT_FOUND;
                                            if (oplabel[curve].startsWith("rel_flux_T") || oplabel[curve].startsWith("rel_flux_C"))
                                                {
                                                operrcolumn[curve] = table.getColumnIndex ("rel_flux_err_"+oplabel[curve].substring(9));
                                                }
                                            else if (oplabel[curve].startsWith("Source-Sky_"))
                                                {
                                                operrcolumn[curve] = table.getColumnIndex ("Source_Error_"+oplabel[curve].substring(11));
                                                }
                                            else if (oplabel[curve].startsWith("tot_C_cnts"))
                                                {
                                                operrcolumn[curve] = table.getColumnIndex ("tot_C_err"+oplabel[curve].substring(10));
                                                }
                                            if (operrcolumn[curve] == ResultsTable.COLUMN_NOT_FOUND)
                                                {
                                                showOpErrors[curve] = false;
                                                showErrors[curve] = false;
                                                }
                                            else
                                                {
                                                showOpErrors[curve] = true;
                                                }
                                            } 
                                        else
                                            showOpErrors[curve] = false;                                        

                                        int bucketSize = binSize[curve];
                                        for (int j=0; j < nn[curve]; j++)
                                                {
                                                double yin = 0;
                                                double errin = 0;
                                                double opin = 0;
                                                double operrin = 0;
                                                int numNaN = 0;
                                                y[curve][j] = 0;
                                                yerr[curve][j] = 0;
                                                yop[curve][j] = 0;
                                                for (int v = 0; v < maxDetrendVars; v++)
                                                    {
                                                    detrend[curve][v][j] = 0;
                                                    }
                                                xc1[curve][j] = 0;
                                                xc2[curve][j] = 0;
                                                yc1[curve][j] = 0;
                                                yc2[curve][j] = 0;
                                                // BIN DATA IF APPLICABLE
                                                if (nnr[curve] > 0 && j == nn[curve] - 1)
                                                    bucketSize = nnr[curve];
                                                else
                                                    bucketSize = binSize[curve];                                               
                                                for (int k=excludedHeadSamples; k < (bucketSize+excludedHeadSamples); k++)
                                                        {
                                                        yin = table.getValueAsDouble (ycolumn[curve],j*binSize[curve]+k);
                                                        if (Double.isNaN(yin))
                                                            {
                                                            numNaN += 1;
                                                            if (numNaN == bucketSize)
                                                                {
                                                                bucketSize = 1;
                                                                y[curve][j] = Double.NaN;
                                                                if (showErrors[curve])
                                                                        yerr[curve][j]  = Double.NaN;
                                                                if (operatorIndex[curve] != 0)
                                                                        yop[curve][j] = Double.NaN;
                                                                if (detrendFitIndex[curve] != 0)
                                                                    {
                                                                    for (int v = 0; v < maxDetrendVars; v++)
                                                                        {
                                                                        if (detrendIndex[curve][v] != 0)
                                                                            {
                                                                            detrend[curve][v][j] = Double.NaN;
                                                                            }
                                                                        }              
                                                                    }

                                                                if (operatorIndex[curve] == 5) //calculate distance
                                                                    {
                                                                    xc1[curve][j] = Double.NaN;
                                                                    xc2[curve][j] = Double.NaN;
                                                                    yc1[curve][j] = Double.NaN;
                                                                    yc2[curve][j] = Double.NaN;
                                                                    }
                                                                if (showOpErrors[curve])
                                                                        yoperr[curve][j] = Double.NaN;
                                                                numNaN = 0;
                                                                break;
                                                                }                                                                
                                                            }
                                                        else
                                                            {
                                                            if (fromMag[curve])
                                                                {
                                                                yin = Math.pow(10, -yin/2.5);
                                                                } 
                                                            y[curve][j] += yin;
                                                         
                                                            if (showErrors[curve])
                                                                {
                                                                errin = table.getValueAsDouble (errcolumn[curve],j*binSize[curve]+k);
                                                                if (fromMag[curve])
                                                                    {
                                                                    errin = yin * (-Math.pow(10, -errin/2.5) + 1);
                                                                    } 
                                                                yerr[curve][j] += errin * errin;
                                                                }
                                                            if (operatorIndex[curve] != 0)
                                                                {
                                                                opin = table.getValueAsDouble (opcolumn[curve],j*binSize[curve]+k);
                                                                if (fromMag[curve])
                                                                    {
                                                                    opin = Math.pow(10,-opin/2.5);
                                                                    }
                                                                yop[curve][j] += opin;
                                                                }
                                                            if (detrendFitIndex[curve] != 0)
                                                                {
                                                                for (int v = 0; v < maxDetrendVars; v++)
                                                                    {
                                                                    
                                                                    if (detrendIndex[curve][v] != 0)
                                                                        {
                                                                        detrend[curve][v][j] += table.getValueAsDouble (detrendcolumn[curve][v],j*binSize[curve]+k);
                                                                        }
                                                                    }              
                                                                }                                                            
                                                            if (operatorIndex[curve] == 5) //calculate distance
                                                                {
                                                                xc1[curve][j] += table.getValueAsDouble (xc1column[curve],j*binSize[curve]+k);
                                                                xc2[curve][j] += table.getValueAsDouble (xc2column[curve],j*binSize[curve]+k);
                                                                yc1[curve][j] += table.getValueAsDouble (yc1column[curve],j*binSize[curve]+k);
                                                                yc2[curve][j] += table.getValueAsDouble (yc2column[curve],j*binSize[curve]+k);
                                                                }
                                                            if (showOpErrors[curve])
                                                                {
                                                                operrin = table.getValueAsDouble (operrcolumn[curve],j*binSize[curve]+k);
                                                                if (fromMag[curve])
                                                                    {
                                                                    operrin = opin * (-Math.pow(10, -operrin/2.5) + 1);
                                                                    }
                                                                yoperr[curve][j] += operrin * operrin;
                                                                }
                                                            }
                                                        }
                                                bucketSize -= numNaN;
                                                y[curve][j] = y[curve][j]*yMultiplierFactor/bucketSize;

                                                if (showErrors[curve])
                                                        yerr[curve][j] = yMultiplierFactor*Math.sqrt(yerr[curve][j])/bucketSize;
                                                else
                                                        yerr[curve][j] = 0.0;
                                                if (detrendFitIndex[curve] != 0)
                                                    {
                                                    for (int v = 0; v < maxDetrendVars; v++)
                                                        {
                                                        if (detrendIndex[curve][v] != 0)
                                                            {
                                                            detrend[curve][v][j] /= bucketSize;
                                                            }
                                                        }              
                                                    }                                                 
                                                if (operatorIndex[curve] != 0)
                                                        {
                                                        yop[curve][j] = yop[curve][j]*yMultiplierFactor/bucketSize;
                                                        if (operatorIndex[curve] == 5)
                                                            {
                                                            xc1[curve][j] = xc1[curve][j]*yMultiplierFactor/bucketSize;
                                                            xc2[curve][j] = xc2[curve][j]*yMultiplierFactor/bucketSize;
                                                            yc1[curve][j] = yc1[curve][j]*yMultiplierFactor/bucketSize;
                                                            yc2[curve][j] = yc2[curve][j]*yMultiplierFactor/bucketSize;
                                                            }
                                                        }
                                                if (showOpErrors[curve])
                                                        yoperr[curve][j] = yMultiplierFactor*Math.sqrt(yoperr[curve][j])/bucketSize;
                                                else
                                                        yoperr[curve][j] = 0.0;

                                                //APPLY OPERATOR/OPERROR FUNCTIONS TO YDATA AND YERROR

                                                if (operatorIndex[curve] == 0)  //no operator
                                                        {

                                                        }
                                                else if (operatorIndex[curve] == 1)  //divide by
                                                        {
                                                        if (yop[curve][j]==0)
                                                            {
                                                            yerr[curve][j] = 1.0e+100;
                                                            y[curve][j] = 1.0e+100;
                                                            }
                                                        else
                                                            {
                                                            if (showErrors[curve] || showOpErrors[curve])
                                                                {
                                                                yerr[curve][j] = yMultiplierFactor*Math.sqrt(
                                                                ((yerr[curve][j]*yerr[curve][j])/(yop[curve][j]*yop[curve][j])) +
                                                                ((y[curve][j]*y[curve][j]*yoperr[curve][j]*yoperr[curve][j])/
                                                                 (yop[curve][j]*yop[curve][j]*yop[curve][j]*yop[curve][j])));
                                                                }
                                                            y[curve][j] = y[curve][j]*yMultiplierFactor/yop[curve][j];
                                                            }
                                                        }
                                                else if (operatorIndex[curve] == 2)  //multiply by
                                                        {
                                                        if (showErrors[curve] || showOpErrors[curve])
                                                                yerr[curve][j] = Math.sqrt(yop[curve][j]*yop[curve][j]*yerr[curve][j]*yerr[curve][j]+
                                                                    y[curve][j]*y[curve][j]*yoperr[curve][j]*yoperr[curve][j])/yMultiplierFactor;
                                                        y[curve][j] = y[curve][j]*yop[curve][j]/yMultiplierFactor;
                                                        }
                                                else if (operatorIndex[curve] == 3)  //subtract
                                                        {
                                                        if (showErrors[curve] || showOpErrors[curve])
                                                                yerr[curve][j] = Math.sqrt(yerr[curve][j]*yerr[curve][j]+yoperr[curve][j]*yoperr[curve][j]);
                                                        y[curve][j] = y[curve][j]-yop[curve][j];
                                                        }
                                                else if (operatorIndex[curve] == 4)  //add
                                                        {
                                                         if (showErrors[curve] || showOpErrors[curve])
                                                                yerr[curve][j] = Math.sqrt(yerr[curve][j]*yerr[curve][j]+yoperr[curve][j]*yoperr[curve][j]);
                                                        y[curve][j] = y[curve][j]+yop[curve][j];
                                                        }
                                                else if (operatorIndex[curve] == 5)  //distance from x1,y1 to x2,y2
                                                        {
                                                        y[curve][j] = (usePixelScale?pixelScale:1.0)*Math.sqrt(((xc1[curve][j]-xc2[curve][j])*(xc1[curve][j]-xc2[curve][j]))+
                                                                                ((yc1[curve][j]-yc2[curve][j])*(yc1[curve][j]-yc2[curve][j])));
                                                        }           
                                                }
                                        if (ylabel[curve].trim().startsWith("J.D.-2400000"))
                                            {
                                            for (int j=0; j < nn[curve]; j++)
                                                {
                                                y[curve][j] += 2400000;
                                                }
                                            } 
                                        }
                                }
                        }
                
                // PERFORM X-AUTOSCALING TO ONE OR MORE CURVES
                dx = 0.0;

                xPlotMin = xMin;
                xPlotMax = xMax;
                
                xautoscalemin = Double.POSITIVE_INFINITY;
                xautoscalemax = Double.NEGATIVE_INFINITY;

                firstCurve = -1;
                for (int curve = 0; curve<maxCurves; curve++)
                    {
                    if (plotY[curve])
                        {
                        xMinimum[curve] = minOf(x[curve],nn[curve]); //FIND MIN AND MAX X OF EACH SELECTED DATASET
                        xMaximum[curve] = maxOf(x[curve],nn[curve]);
                        if ((firstCurve == -1))
                                firstCurve = curve; //FIND THE FIRST CURVE TO DISPLAY - IT IS USED FOR THE Y-AXIS LABEL
                        if (ASInclude[curve])// && !force[curve])
                            {
                            if (xMinimum[curve] < xautoscalemin) xautoscalemin = xMinimum[curve];
                            if (xMaximum[curve] > xautoscalemax) xautoscalemax = xMaximum[curve];
                            }
                        }
                    }

                if (firstCurve == -1)   //IF NO CURVES SELECTED FOR DISPLAY, USE THE FIRST CURVE
                        firstCurve = 0;
                
                if (showVMarker1 || showVMarker2)
                    {
                    double v1 = vMarker1Value;
                    double v2 = vMarker2Value; 
                    if ((xlabel2[firstCurve].contains("J.D.") || xlabel2[firstCurve].contains("JD")) && showXAxisNormal)
                        {
                        v1 += (int)xautoscalemin;
                        v2 += (int)xautoscalemin;
                        }
                    if (showVMarker1)
                        {
                        if (v1 < xautoscalemin && (!showVMarker2 || (showVMarker2 && v1 <= v2))) xautoscalemin = v1; 
                        if (v1 > xautoscalemax && (!showVMarker2 || (showVMarker2 && v1 >= v2))) xautoscalemax = v1; 
                        }
                    if (showVMarker2)
                        {
                        if (v2 < xautoscalemin && (!showVMarker1 || (showVMarker1 && v2 < v1))) xautoscalemin = v2; 
                        if (v2 > xautoscalemax && (!showVMarker1 || (showVMarker1 && v2 > v1))) xautoscalemax = v2; 
                        }           
                    }

                if (Double.isInfinite(xautoscalemin)) xautoscalemin = Double.isInfinite(xautoscalemax) ? 0.0 : xautoscalemax - 1.0;
                if (Double.isInfinite(xautoscalemax)) xautoscalemax = xautoscalemin + 1.0;


                if (autoScaleX)
                    {
                    xPlotMin = xautoscalemin;
                    xPlotMax = (xautoscalemin == xautoscalemax) ? xautoscalemin+0.01 : xautoscalemax;
                    }
                else if (useFirstX)
                    {
                    firstXmin = xautoscalemin;                    
                    xPlotMin = xautoscalemin;
                    xPlotMax = (xWidth <= 0) ? xautoscalemin + 0.2 : xautoscalemin+xWidth;
                    }
                dx = (xPlotMax-xPlotMin)/99.0;

                for (int curve = 0; curve<maxCurves; curve++)
                    {
                    if (plotY[curve])
                        {
                        boolean atLeastOne = false;
                        boolean detrendYNotConstant = false;
                        boolean[] detrendYDNotConstant = new boolean[maxDetrendVars];
                        for (int v = 0; v <maxDetrendVars; v++)
                            {
                            if (detrendFitIndex[curve] == 1) detrendYDNotConstant[v] = true;
                            else detrendYDNotConstant[v] = false;
                            }                        
                        if (detrendFitIndex[curve] != 0)
                            {
                            for (int v = 0; v <maxDetrendVars; v++)
                                {
                                if (detrendIndex[curve][v] != 0) atLeastOne = true;
                                }
                            }
                        if (atLeastOne)
                            {
                            double[] detrendAverage = new double[maxDetrendVars];
                            int[] detrendPower = new int[maxDetrendVars];
                            for(int v = 0; v < maxDetrendVars; v++)
                                {
                                detrendAverage[v] = 0.0;
                                detrendPower[v] = 1;
                                for (int j=0; j < nn[curve]; j++)
                                    { 
                                    detrendAverage[v] += detrend[curve][v][j]/(double)nn[curve];
                                    }
                                for (int j=0; j < nn[curve]; j++)
                                    { 
                                    detrend[curve][v][j] -= detrendAverage[v]; 
                                    }
                                if (v > 0)
                                    {
                                    for (int u = 0; u < v; u++)
                                        {
                                        if (detrendIndex[curve][u] == detrendIndex[curve][v])
                                            detrendPower[v]++;
                                        }
                                    }
                                if (detrendPower[v] > 1)
                                    {
                                    detrendAverage[v] = 0.0;
                                    for (int j=0; j < nn[curve]; j++)
                                        { 
                                        detrendAverage[v] += detrend[curve][v][j];
                                        }
                                    detrendAverage[v] /= nn[curve];
                                    for (int j=0; j < nn[curve]; j++)
                                        { 
                                        detrend[curve][v][j] -= detrendAverage[v]; 
                                        }                                    
                                    }
                                } 
                            for(int v = 0; v < maxDetrendVars; v++)
                                {
                                if (detrendPower[v] == 2)
                                    {
                                    for (int j=0; j < nn[curve]; j++)
                                        { 
                                        detrend[curve][v][j] *= detrend[curve][v][j]; 
                                        }          
                                    }
                                else if (detrendPower[v] > 2)
                                    {
                                    for (int j=0; j < nn[curve]; j++)
                                        { 
                                        detrend[curve][v][j] = Math.pow(detrend[curve][v][j], detrendPower[v]); 
                                        }          
                                    }
                                }
                            
                            double yAverage = 0.0;
                            int avgCount = 0;
                            int detrendCount = 0; 
                            if (detrendFitIndex[curve] != 1)
                                {
                                double detrendMin = useDMarker1 ? dMarker1Value : Double.NEGATIVE_INFINITY;
                                double detrendMax = useDMarker4 ? dMarker4Value : Double.POSITIVE_INFINITY;
                                double detrendLeft = dMarker2Value;
                                double detrendRight = dMarker3Value;
                                if ((xlabel2[firstCurve].contains("J.D.") || xlabel2[firstCurve].contains("JD")) && showXAxisNormal)                                
                                    {
                                    detrendMin += (int)xPlotMin;
                                    detrendMax += (int)xPlotMin;    
                                    detrendLeft += (int)xPlotMin;
                                    detrendRight += (int)xPlotMin;
                                    }                             
                                switch(detrendFitIndex[curve])
                                    {
                                    case 2: // left of D2
                                        detrendMax = detrendLeft;
                                        break;
                                    case 3: // right of D3
                                        detrendMin = detrendRight;
                                        break;
                                    case 4: // outside D2 and D3
                                        break;
                                    case 5: // inside D2 and D3
                                        detrendMin = detrendLeft;
                                        detrendMax = detrendRight;
                                        break;
                                    case 6: // left of D3
                                        detrendMax = detrendRight;
                                        break;
                                    case 7: // right of D2
                                        detrendMin = detrendLeft;
                                        break;
                                    case 8: // use all data
                                        break;                            
                                    default: // use all data
                                        detrendFitIndex[curve] = 0;
//                                        detrendMax = detrendMin;
                                        break;
                                    }
                                double[] detrendY = new double[nn[curve]];
                                double[][] detrendYD = new double[maxDetrendVars][nn[curve]];
                                boolean noNaNs = true;
                                if (detrendFitIndex[curve] == 4)
                                    {
                                    for (int j=0; j < nn[curve]; j++)
                                        {
                                        noNaNs = true;
                                        if (!Double.isNaN(y[curve][j]))
                                            {
                                            for(int v = 0; v < maxDetrendVars; v++)
                                                {                                                
                                                if (detrendIndex[curve][v] != 0 && Double.isNaN(detrend[curve][v][j]))
                                                    {
                                                    noNaNs = false;
                                                    }
                                                }
                                            }
                                        else
                                            {
                                            noNaNs = false;
                                            }
                                        if (noNaNs)
                                            {
                                            avgCount++;                                             
                                            yAverage += y[curve][j];
//                                            for(int v = 0; v < maxDetrendVars; v++)
//                                                { 
//                                                detrendAverage[v] += detrend[curve][v][j];
//                                                }
                                            if ((x[curve][j] > detrendMin && x[curve][j] < detrendLeft) ||
                                                (x[curve][j] > detrendRight && x[curve][j] < detrendMax))
                                                {
                                                detrendY[detrendCount] = y[curve][j];
                                                if (detrendY[0] != detrendY[detrendCount]) detrendYNotConstant = true;
                                                for (int v = 0; v < maxDetrendVars; v++)
                                                    {
                                                    detrendYD[v][detrendCount] = detrend[curve][v][j];
                                                    if (detrendYD[v][0] != detrendYD[v][detrendCount]) detrendYDNotConstant[v] = true;
                                                    }
                                                detrendCount++;
                                                }
                                            }
                                        }

                                    }
                                else
                                    {
                                    for (int j=0; j < nn[curve]; j++)
                                        {
                                        noNaNs = true;
                                        if (!Double.isNaN(y[curve][j]))
                                            {
                                            for(int v = 0; v < maxDetrendVars; v++)
                                                {                                                
                                                if (detrendIndex[curve][v] != 0 && Double.isNaN(detrend[curve][v][j]))
                                                    {
                                                    noNaNs = false;
                                                    }
                                                }
                                            }
                                        else
                                            {
                                            noNaNs = false;
                                            }
                                        if (noNaNs)
                                            {
                                            avgCount++;                                             
                                            yAverage += y[curve][j];
//                                            for(int v = 0; v < maxDetrendVars; v++)
//                                                {  
//                                                detrendAverage[v] += detrend[curve][v][j];
//                                                } 
                                            if (x[curve][j] > detrendMin && x[curve][j] < detrendMax)
                                                {
                                                detrendY[detrendCount] = y[curve][j];
                                                if (detrendY[0] != detrendY[detrendCount]) detrendYNotConstant = true;
                                                for (int v = 0; v < maxDetrendVars; v++)
                                                    {
                                                    detrendYD[v][detrendCount] = detrend[curve][v][j];
                                                    if (detrendYD[v][0] != detrendYD[v][detrendCount]) detrendYDNotConstant[v] = true;
                                                    }
                                                detrendCount++;
                                                } 
                                            }
                                        }
                                    }
                                if (detrendCount>0)
                                    {
                                    yAverage /= avgCount; //avgCount is always >= detrendCount, so > 0 here
//                                    for(int v = 0; v < maxDetrendVars; v++)
//                                        {                                                
//                                        detrendAverage[v] /= avgCount; //avgCount is always >= detrendCount
//                                        }                                     
                                    for (int j=0; j < detrendCount; j++)
                                        {  
                                        detrendY[j] -= yAverage;
//                                        for(int v = 0; v < maxDetrendVars; v++)
//                                            {                                                
//                                            detrendYD[v][j] -= detrendAverage[v]; 
//                                            }                                         
                                        }
                                    for (int v=0; v<maxDetrendVars; v++)
                                        {
                                        if (detrendIndex[curve][v] != 0 && !detrendYDNotConstant[v]) detrendVarsUsed[curve]--;
                                        }
                                    detrendVars = new double[detrendVarsUsed[curve]][detrendCount];
                                    int varCount = 0;
                                    for (int v=0; v<maxDetrendVars; v++)
                                        {
                                        if (detrendIndex[curve][v] != 0 && detrendYDNotConstant[v])
                                            {
                                            detrendVars[varCount] = Arrays.copyOf(detrendYD[v],detrendCount);
                                            varCount++;
                                            }
                                        }
                                    if (detrendVarsUsed[curve] > 0 && detrendYNotConstant && detrendCount > detrendVarsUsed[curve]+2) //need enough data to provide degrees of freedom for successful fit
                                        {
//                                        IJ.log("Setting up Regression, Curve = "+curve+" ,VarsUsed = "+detrendVarsUsed[curve]+" ,detrendCount = "+detrendCount);
                                        Regression regression = new Regression(detrendVars, Arrays.copyOf(detrendY,detrendCount));
                                        

//                                        IJ.log("Starting Regression");
                                        regression.linear();
                                        if (regression.getDegFree()>0)
                                            {                                            
    //                                        IJ.log("Getting coeffs");
                                            coeffs = regression.getCoeff();
    //                                        for (int i=0; i<coeffs.length; i++)
    //                                            {
    //                                            IJ.log("coeff["+i+"]="+coeffs[i]);
    //                                            }
                                            varCount = 0;
                                            for (int v = 0; v < maxDetrendVars; v++)
                                                {
                                                if (detrendIndex[curve][v] != 0 && detrendYDNotConstant[v])
                                                    {
                                                    detrendFactor[curve][v] = coeffs[varCount+1];
                                                    if (detrendVarDisplayed[curve] == v)
                                                        {
                                                        detrendfactorspinner[curve].setValue(coeffs[varCount+1]);
                                                        }
                                                    varCount++;
                                                    }
                                                }
                                            }
                                        else
                                            {
                                            IJ.log("Detrend degrees of freedom < 1");
                                            }
                                        }
                                    }
                                }
                            else
                                {
                                for (int j=0; j < nn[curve]; j++)
                                    {
                                    boolean noNaNs = true;
                                    if (!Double.isNaN(y[curve][j]))
                                        {
                                        for(int v = 0; v < maxDetrendVars; v++)
                                            {                                                
                                            if (detrendIndex[curve][v] != 0 && Double.isNaN(detrend[curve][v][j]))
                                                {
                                                noNaNs = false;
                                                }
                                            }
                                        }
                                    else
                                        {
                                        noNaNs = false;
                                        }
                                    if (noNaNs)
                                        {
                                        yAverage += y[curve][j];
//                                        for(int v = 0; v < maxDetrendVars; v++)
//                                            {                                                
//                                            detrendAverage[v] += detrend[curve][v][j];
//                                            }                                         
                                        avgCount++;
                                        }
                                    }
                                if (avgCount>0)
                                    {
                                    yAverage /= avgCount;
//                                    for(int v = 0; v < maxDetrendVars; v++)
//                                        {                                                
//                                        detrendAverage[v] /= avgCount;
//                                        }                                     
                                    }  
                                }

                            if (divideNotSubtract)
                                {
                                double trend;
                                if (yAverage != 0.0)
                                    {
                                    for (int j=0; j < nn[curve]; j++)
                                        {
                                        trend = yAverage;
                                        for(int v = 0; v < maxDetrendVars; v++)
                                            {
                                            if (detrendIndex[curve][v] != 0 && detrendYDNotConstant[v])
                                                {
                                                trend += detrendFactor[curve][v]*(detrend[curve][v][j]);//-detrendAverage[v]);
                                                }
                                            }
                                        trend /= yAverage;
                                        if (trend != 0.0)
                                            {
                                            y[curve][j] /= trend;
                                            if (showErrors[curve] || showOpErrors[curve])
                                                {
                                                yerr[curve][j] /= trend;
                                                }           
                                            }
                                        }
                                    }
                                }
                            else
                                {
                                double trend;
                                for (int j=0; j < nn[curve]; j++)
                                    {
                                    trend = 0.0;
                                    for(int v = 0; v < maxDetrendVars; v++)
                                        {
                                        if (detrendIndex[curve][v] != 0 && detrendYDNotConstant[v])
                                            {
                                            trend += detrendFactor[curve][v]*(detrend[curve][v][j]);//-detrendAverage[v]);
                                            }
                                        }                                    
                                    if (showErrors[curve] || showOpErrors[curve])
                                        {
                                        yerr[curve][j] /= (y[curve][j]/(y[curve][j]-trend));
                                        }                                     
                                    y[curve][j] -= trend;
                                    }
                                }                            
                            }
                        
                        if (normIndex[curve] != 0)
                            {
                            double normMin = useDMarker1 ? dMarker1Value : Double.NEGATIVE_INFINITY;
                            double normMax = useDMarker4 ? dMarker4Value : Double.POSITIVE_INFINITY;
                            double normLeft = dMarker2Value;
                            double normRight = dMarker3Value;

                            if ((xlabel2[firstCurve].contains("J.D.") || xlabel2[firstCurve].contains("JD")) && showXAxisNormal)                                
                                {
                                normMin += (int)xPlotMin;
                                normMax += (int)xPlotMin;    
                                normLeft += (int)xPlotMin;
                                normRight += (int)xPlotMin;
                                }                             
                            double normAverage = 0.0;
                            int normCount = 0; 
                            
                            switch(normIndex[curve])
                                {
                                case 1: // left of D2
                                    normMax = normLeft;
                                    break;
                                case 2: // right of D3
                                    normMin = normRight;
                                    break;
                                case 3: // outside D2 and D3
                                    break;
                                case 4: // inside D2 and D3
                                    normMin = normLeft;
                                    normMax = normRight;
                                    break;
                                case 5: // left of D3
                                    normMax = normRight;
                                    break;
                                case 6: // right of D2
                                    normMin = normLeft;
                                    break;
                                case 7: // use all data
                                    break;                            
                                default:
                                    normIndex[curve] = 0;
                                    normMax = normMin;
                                    break;
                                }

                            if (normIndex[curve] == 3)
                                {
                                for (int j=0; j < nn[curve]; j++)
                                    {                                
                                    if (!Double.isNaN(y[curve][j]) && !Double.isNaN(x[curve][j]) &&
                                       ((x[curve][j] > normMin && x[curve][j] < normLeft) || (x[curve][j] > normRight && x[curve][j] < normMax)))
                                        {
                                        normAverage += y[curve][j];
                                        normCount++;                                        
                                        }
                                    }
                                }
                            else if (normIndex[curve] != 0)
                                {
                                for (int j=0; j < nn[curve]; j++)
                                    {
                                    if (!Double.isNaN(y[curve][j]) && !Double.isNaN(x[curve][j]) && x[curve][j] > normMin && x[curve][j] < normMax)
                                        {
                                        normAverage += y[curve][j];
                                        normCount++;
                                        } 
                                    }
                                }

                            if (normAverage == 0.0 || normCount == 0)
                                {
                                normAverage = 1.0;
                                }
                            else
                                {
                                normAverage /= normCount;
                                }

                            for (int j=0; j < nn[curve]; j++)
                                {
                                if (showErrors[curve] || showOpErrors[curve])
                                    {
                                    yerr[curve][j] /= normAverage;
                                    }
                                y[curve][j] /= normAverage;
                                }
                            }

                        //PERFORM MAGNITUDE CONVERSION IF APPLICABLE
                        if (mmag[curve])
                            {
                            int cnt = 0;
                            int numgood = 0;
                            while (numgood<mmagrefs && cnt<nn[curve])  //Prepare mmag calulation baseline reference level
                                {
                                if (!Double.isNaN(y[curve][cnt]) && !Double.isNaN(x[curve][cnt]))
                                    {
                                    baseline[curve] += y[curve][cnt];
                                    numgood++;
                                    }
                                cnt++;
                                }
                            baseline[curve] /= (numgood == 0 ? 1 : numgood);
                            baseline[curve] = (baseline[curve]==0 ? 1 : baseline[curve]);
                            for (int j=0; j < nn[curve]; j++)
                                {
                                if (showErrors[curve] || showOpErrors[curve])
                                    {
                                    yerr[curve][j] = yMultiplierFactor*(2.5*Math.log10(1.0+yerr[curve][j]/y[curve][j]));
                                    }
                                y[curve][j] = magSign*yMultiplierFactor*2.5*Math.log10(y[curve][j]/baseline[curve]);
                                }
                            }
                        }

                    }


                // PERFORM Y-AUTOSCALING TO ONE OR MORE CURVES

                dy = 0.0;

                yPlotMin = yMin;
                yPlotMax = yMax;
                
                yautoscalemin = Double.POSITIVE_INFINITY;
                yautoscalemax = Double.NEGATIVE_INFINITY;                

                for (int curve = 0; curve<maxCurves; curve++)
                    {
                    if (plotY[curve])
                        {
                        yMn[curve] = showErrors[curve]?minOf(y[curve],yerr[curve],nn[curve]):minOf(y[curve],nn[curve]);      //FIND MIN AND MAX Y OF EACH SELECTED DATASET
                        yMx[curve] = showErrors[curve]?maxOf(y[curve],yerr[curve],nn[curve]):maxOf(y[curve],nn[curve]);
                        if (customScaleFactor[curve] >= 0)
                            {
                            yMinimum[curve] = force[curve] ? yMn[curve] : yMn[curve]*customScaleFactor[curve] + customShiftFactor[curve]; //FIND MIN AND MAX Y OF EACH SELECTED DATASET
                            yMaximum[curve] = force[curve] ? yMx[curve] : yMx[curve]*customScaleFactor[curve] + customShiftFactor[curve];
                            }
                        else
                            {
                            yMinimum[curve] = force[curve] ? yMn[curve] : yMx[curve]*customScaleFactor[curve] + customShiftFactor[curve]; //FIND MIN AND MAX Y OF EACH SELECTED DATASET
                            yMaximum[curve] = force[curve] ? yMx[curve] : yMn[curve]*customScaleFactor[curve] + customShiftFactor[curve];
                            }
                        if (ASInclude[curve] && !force[curve])
                            {
                            if (yMinimum[curve] < yautoscalemin) yautoscalemin = yMinimum[curve];
                            if (yMaximum[curve] > yautoscalemax) yautoscalemax = yMaximum[curve];
                            }
                        }
                    }

                if (Double.isInfinite(yautoscalemin)) yautoscalemin = Double.isInfinite(yautoscalemax) ? 0.0 : yautoscalemax - 1.0;
                if (Double.isInfinite(yautoscalemax)) yautoscalemax = yautoscalemin + 1.0;                        

                if (autoScaleY)
                    {
                    yPlotMin = yautoscalemin;
                    yPlotMax = (yautoscalemin == yautoscalemax) ? yautoscalemin+0.01 : yautoscalemax;
                    }
                dy = (yPlotMax-yPlotMin)/99.0;

                // IF MJD, SHOW DECIMAL DATE DUE TO float API OF PlotWindow

                xOffset = 0.0;
                yOffset = 0.0;
                int jd=0;
                xlab ="";
                ylab = "";
                boolean xGoodLabel = false;
                boolean yGoodLabel = false;

                // MAKE X-AXIS LABEL

                for (int curve=0; curve<maxCurves; curve++)
                    {
                    if (plotY[curve])
                        {
                        if (showXAxisAsPhase)
                                {
                                xlab = "Orbital Phase (periods) (T0 = "+T0+", P = "+period+")"; 
                                xGoodLabel = true;
                                }
                        else if (showXAxisAsDaysSinceTc)
                                {
                                xlab = "Days Since Tc (T0 = "+T0+", P = "+period+")"; 
                                xGoodLabel = true;
                                }                        
                        else if (showXAxisAsHoursSinceTc)
                                {
                                xlab = "Hours Since Tc (T0 = "+T0+", P = "+period+")"; 
                                xGoodLabel = true;
                                }                        
                        else if (xlabel2[curve].contains("J.D.") || xlabel2[curve].contains("JD"))                            
                                {
                                if (curve == firstCurve)
                                    {
                                    if (xExponent != 0) xmultiplierspinner.setValue(0);
                                    jd = (int)xPlotMin;
                                    xOffset = (double)jd;
                                    xPlotMin -= xOffset;
                                    xPlotMax -= xOffset;
                                    if (useXColumnName)
                                            {
                                            if (xlabel2[curve].trim().startsWith("J.D.-2400000"))
                                                xlab = "Geocentric Julian Date - "+jd;
                                            else if (xlabel2[curve].trim().startsWith("BJD_MOBS"))
                                                xlab = "Barycentric Julian Date - "+jd+" (mid-exposure)";
                                            else if (xlabel2[curve].trim().startsWith("BJD_SOBS"))
                                                xlab = "Barycentric Julian Date - "+jd+" (exposure start)";
                                            else if (xlabel2[curve].trim().startsWith("BJD"))
                                                xlab = "Barycentric Julian Date - "+jd;
                                            else if (xlabel2[curve].trim().startsWith("HJD_MOBS"))
                                                xlab = "Heliocentric Julian Date - "+jd+" (mid-exposure)";
                                            else if (xlabel2[curve].trim().startsWith("HJD_SOBS"))
                                                xlab = "Heliocentric Julian Date - "+jd+" (exposure start)";
                                            else if (xlabel2[curve].trim().startsWith("HJD"))
                                                xlab = "Heliocentric Julian Date - "+jd;
                                            else if (xlabel2[curve].trim().startsWith("JD_SOBS"))
                                                xlab = "Geocentric Julian Date - "+jd+" (exposure start)";
                                            else if (xlabel2[curve].trim().startsWith("JD_MOBS"))
                                                xlab = "Geocentric Julian Date - "+jd+" (mid-exposure)";
                                            else if (xlabel2[curve].trim().contains("JD"))
                                                xlab = "Julian Date - "+jd;
                                            else
                                                xlab = "Label Error";
                                            xGoodLabel = true;
                                            }
                                    else if (useXCustomName)
                                            {
                                            xlab = xLegend;
                                            xGoodLabel = true;
                                            }
                                    }
                                for (int j=0; j < nn[curve]; j++)
                                    x[curve][j] -= xOffset;
                                }
                        else if (useXColumnName && curve == firstCurve)
                                {
                                xlab = xlabel2[curve];
                                xGoodLabel = true;
                                }
                        else if (useXCustomName && curve == firstCurve)
                                {
                                xlab = xLegend;
                                xGoodLabel = true;
                                }
                        }
                     }

                if (xGoodLabel && (xExponent != 0) && showXScaleInfo)
                        xlab += " x 1E"+xExponent;

                // MAKE Y-AXIS LABEL

                for (int curve=0; curve<maxCurves; curve++)
                    {
                    if (plotY[curve])
                        {
                        if (ylabel[curve].contains("J.D.") || ylabel[curve].contains("JD"))                            
                                {
                                if (curve == firstCurve)
                                    {
                                    if (yExponent != 0) ymultiplierspinner.setValue(0);
                                    jd = (int)yPlotMin;
                                    yOffset = (double)jd;
                                    yPlotMin -= yOffset;
                                    yPlotMax -= yOffset;
                                    if (useYColumnName && operatorIndex[curve] != 5)
                                            {
                                            if (ylabel[curve].trim().startsWith("J.D.-2400000"))
                                                ylab = "Geocentric Julian Date - "+jd;
                                            else if (ylabel[curve].trim().startsWith("BJD_MOBS"))
                                                ylab = "Barycentric Julian Date - "+jd+" (mid-exposure)";
                                            else if (ylabel[curve].trim().startsWith("BJD_SOBS"))
                                                ylab = "Barycentric Julian Date - "+jd+" (exposure start)";
                                            else if (ylabel[curve].trim().startsWith("BJD"))
                                                ylab = "Barycentric Julian Date - "+jd;
                                            else if (ylabel[curve].trim().startsWith("HJD_MOBS"))
                                                ylab = "Heliocentric Julian Date - "+jd+" (mid-exposure)";
                                            else if (ylabel[curve].trim().startsWith("HJD_SOBS"))
                                                ylab = "Heliocentric Julian Date - "+jd+" (exposure start)";
                                            else if (ylabel[curve].trim().startsWith("HJD"))
                                                ylab = "Heliocentric Julian Date - "+jd;         
                                            else if (ylabel[curve].trim().startsWith("JD_MOBS"))
                                                ylab = "Geocentric Julian Date - "+jd+" (mid-exposure)";
                                            else if (ylabel[curve].trim().startsWith("JD_SOBS"))
                                                ylab = "Geocentric Julian Date - "+jd+" (exposure start)";
                                            else if (ylabel[curve].trim().contains("JD"))
                                                ylab = "Julian Date - "+jd;
                                            else
                                                ylab = "Label Error";
                                            yGoodLabel = true;
                                            }
                                    else if (useYCustomName)
                                            {
                                            ylab = yLegend;
                                            yGoodLabel = true;
                                            }
                                    else if(useYColumnName && operatorIndex[curve] == 5)
                                            {
                                            ylab = xyc1label[curve]+" - "+xyc2label[curve]+" centroid distance";
                                            ylab += usePixelScale ? " (arcsecs)" : " (pixels)";
                                            yGoodLabel = true;
                                            }
                                    }
                                for (int j=0; j < nn[curve]; j++)
                                    y[curve][j] -= yOffset;
                                }
                        else if (useYColumnName && curve == firstCurve)
                                {
                                if (operatorIndex[curve] != 5)
                                    {
                                    ylab=ylabel[firstCurve];
                                    if (operatorIndex[firstCurve] != 0)
                                            ylab += opSymbol[operatorIndex[firstCurve]]+oplabel[firstCurve];   
                                    }
                                else
                                    {
                                    ylab = xyc1label[firstCurve]+" - "+xyc2label[firstCurve]+" centroid distance";
                                    ylab += usePixelScale ? " (arcsecs)" : " (pixels)";
                                    }
                                yGoodLabel = true;
                                }
                        else if (useYCustomName && curve == firstCurve)
                                {
                                ylab = yLegend;
                                yGoodLabel = true;
                                }
                        }
                    }
                //FINISH Y-LABEL AFTER FINDING SCALING FACTORS BELOW

              
                //SCALE AND SHIFT SECONDARY CURVES TO FIT ON PLOT

                yRange = yPlotMax - yPlotMin;
                yMid = yPlotMin + yRange/2.;
                for (int curve=0; curve<maxCurves; curve++)
                    {
                    if (plotY[curve])
                        {
                        yMidpoint[curve] = (yMaximum[curve] + yMinimum[curve])/2.;
                        yWidth[curve] = yMaximum[curve] - yMinimum[curve];
                        if (yWidth[curve] == 0) yWidth[curve] = 1.0e-10;
                        if (force[curve])
                                {
                                subtotalScaleFactor[curve] = (invertYAxis ? -1 : 1)*autoScaleFactor[curve]*yRange/yWidth[curve];
                                subtotalShiftFactor[curve] = (yMid + yRange*(invertYAxis ? -1 : 1)*autoShiftFactor[curve] - subtotalScaleFactor[curve]*yMidpoint[curve]);
                                totalScaleFactor[curve] = 1.0/(subtotalScaleFactor[curve]*yMultiplierFactor);
                                totalShiftFactor[curve] = subtotalShiftFactor[curve];
                                }
                        else
                                {
                                subtotalScaleFactor[curve] = customScaleFactor[curve];
                                subtotalShiftFactor[curve] = customShiftFactor[curve];
                                totalScaleFactor[curve] = 1.0/(subtotalScaleFactor[curve]*yMultiplierFactor);
                                totalShiftFactor[curve] = subtotalShiftFactor[curve];
                                }
                        for (int j=0; j<nn[curve]; j++)
                            {
                            if (showErrors[curve] || showOpErrors[curve])
                                {
                                yerr[curve][j] = subtotalScaleFactor[curve]*yerr[curve][j];
                                }
                            y[curve][j] = subtotalScaleFactor[curve]*y[curve][j] + subtotalShiftFactor[curve];
                            }
                        }
                    }

                if (yGoodLabel)
                        {
                        if (showYScaleInfo || showYShiftInfo)
                                ylab += scaleShiftText(showYScaleInfo, showYShiftInfo, mmag[firstCurve], showYmmagInfo, totalScaleFactor[firstCurve], totalShiftFactor[firstCurve]);
                        if (showYNormInfo && normIndex[firstCurve] != 0) ylab += " (normalized)";
                        if (mmag[firstCurve] && showYmmagInfo)
                                if (totalScaleFactor[firstCurve] == 0.001)
                                        ylab += " (mmag)";
                                else
                                        ylab += " (mag)";
                        
                        if ((binSize[firstCurve]!=1) && showYBinInfo)
                                ylab += " (bin size = "+binSize[firstCurve]+")";
                        if (showYSymbolInfo)
                                ylab += " ("+colors[colorIndex[firstCurve]]+" "+markers[markerIndex[firstCurve]]+")";
                        }



 //----------------Set up plot options------------------------------------
                plotOptions = 0;
                if (xTics) plotOptions += ij.gui.Plot.X_TICKS;
                if (yTics) plotOptions += ij.gui.Plot.Y_TICKS;
                if (xGrid) plotOptions += ij.gui.Plot.X_GRID;
                if (yGrid) plotOptions += ij.gui.Plot.Y_GRID;
                if (xNumbers) plotOptions += ij.gui.Plot.X_NUMBERS;
                if (yNumbers) plotOptions += ij.gui.Plot.Y_NUMBERS;

                plot = new Plot ("Plot of "+tableName,xlab,ylab,nullX,nullY,plotOptions);
                plot.setSize(plotSizeX, plotSizeY);
                pltMinX = xPlotMin-5.*dx;
                pltMaxX = xPlotMax+5.*dx;
                pltMinY = yPlotMin-5.*dy;
                pltMaxY = yPlotMax+5.*dy;

                if (invertYAxis)
                    {
                    double yMaxTemp = pltMaxY;
                    pltMaxY = pltMinY;
                    pltMinY = yMaxTemp;
                    }

                if (plotImageCanvas != null) //zoom != 0.0 &&
                    {
                    plotMinX = totalPanOffsetX+newPanOffsetX+pltMinX+(pltMaxX-pltMinX)*(mouseX/(plotSizeX-(Plot.LEFT_MARGIN+Plot.RIGHT_MARGIN+1)))*zoom;
                    plotMaxX = totalPanOffsetX+newPanOffsetX+pltMaxX-(pltMaxX-pltMinX)*((plotSizeX-(Plot.LEFT_MARGIN+Plot.RIGHT_MARGIN+1)-mouseX)/(plotSizeX-(Plot.LEFT_MARGIN+Plot.RIGHT_MARGIN+1)))*zoom;
                    plotMinY = totalPanOffsetY+newPanOffsetY+pltMinY+(pltMaxY-pltMinY)*((plotSizeY-(Plot.TOP_MARGIN+Plot.BOTTOM_MARGIN)-mouseY)/(plotSizeY-(Plot.TOP_MARGIN+Plot.BOTTOM_MARGIN)))*zoom;
                    plotMaxY = totalPanOffsetY+newPanOffsetY+pltMaxY-(pltMaxY-pltMinY)*(mouseY/(plotSizeY-(Plot.TOP_MARGIN+Plot.BOTTOM_MARGIN)))*zoom;
                    }
                else
                    {
                    plotMinX = pltMinX;
                    plotMaxX = pltMaxX;
                    plotMinY = pltMinY;
                    plotMaxY = pltMaxY;
                    zoom = 0.0;
                    totalPanOffsetX=0.0;
                    totalPanOffsetY=0.0;
                    newPanOffsetX=0.0;
                    newPanOffsetY=0.0;
                    leftDragReleased=false;
                    }
                
                plot.setLimits (plotMinX, plotMaxX, plotMinY, plotMaxY);
                double legPosY = legendPosY;
                for (int curve=0; curve<maxCurves; curve++)
                    {
                    if (plotY[curve])
                        {
                        plot.setColor (color[curve]);
                        if (marker[curve] == ij.gui.Plot.DOT)
                            plot.setLineWidth(4);
                        else
                            plot.setLineWidth(1);

                        plot.addPoints (Arrays.copyOf(x[curve],nn[curve]),Arrays.copyOf(y[curve],nn[curve]),marker[curve]);

                        plot.setLineWidth(1);

                        if (showErrors[curve] || showOpErrors[curve])     //code to replace plot.addErrorBars
                            {               //since plot.addErrorBars only plots with lines enabled
                            for (int j=0; j < nn[curve]; j++)
                                {
                                plot.drawLine(x[curve][j], y[curve][j]-yerr[curve][j], x[curve][j], y[curve][j]+yerr[curve][j]);
                                plot.drawLine(x[curve][j]-(3.0*(plotMaxX-plotMinX)/plotSizeX), y[curve][j]+yerr[curve][j], x[curve][j]+(3.0*(plotMaxX-plotMinX)/plotSizeX), y[curve][j]+yerr[curve][j]);
                                plot.drawLine(x[curve][j]-(3.0*(plotMaxX-plotMinX)/plotSizeX), y[curve][j]-yerr[curve][j], x[curve][j]+(3.0*(plotMaxX-plotMinX)/plotSizeX), y[curve][j]-yerr[curve][j]);
                                }
                            }

                        if (lines[curve] && !(marker[curve] == ij.gui.Plot.LINE))
                            {
                             for (int j=0; j < nn[curve]-1; j++)
                                {
                                plot.drawLine(x[curve][j],y[curve][j],x[curve][j+1],y[curve][j+1]);
                                }
                            }

                        plot.setLineWidth(1);
                        if (legendLeft)
                            plot.setJustification(ij.process.ImageProcessor.LEFT_JUSTIFY);
                        else if (legendRight)
                            plot.setJustification(ij.process.ImageProcessor.RIGHT_JUSTIFY);
                        else
                            plot.setJustification(ij.process.ImageProcessor.CENTER_JUSTIFY);

                        // MAKE FULL LEGEND STRING FOR CURVE
                        String llab = "";

                        if (useColumnName[curve])
                            {
                            if (operatorIndex[curve] == 5)
                                {
                                llab = xyc1label[curve]+" - "+xyc2label[curve]+" centroid distance";
                                if (showLdetrendInfo && detrendFitIndex[curve] != 0)
                                    {
                                    boolean atLeastOne = false;
                                    for (int v = 0; v < maxDetrendVars; v++)
                                        {
                                        if (detrendIndex[curve][v] != 0)
                                            {
                                            if (atLeastOne)
                                                llab += "+"+detrendlabel[curve][v];
                                            else
                                                llab += " ("+detrendlabel[curve][v];
                                            atLeastOne = true;
                                            }
                                        if (v == maxDetrendVars - 1 && atLeastOne)
                                            llab += " detrended)";
                                        }
                                    }
                                if (showLnormInfo && normIndex[curve] != 0)
                                    {
                                    llab += " (normalized)";
                                    }
                                else
                                    {
                                    llab += usePixelScale ? " (arcsecs)" : " (pixels)";
                                    }
                                }
                            else
                                {
                                llab=ylabel[curve];
                                if (operatorIndex[curve] != 0)
                                    llab += opSymbol[operatorIndex[curve]]+oplabel[curve];
                                if (showLdetrendInfo && detrendFitIndex[curve] != 0)
                                    {
                                    boolean atLeastOne = false;
                                    for (int v = 0; v < maxDetrendVars; v++)
                                        {
                                        if (detrendIndex[curve][v] != 0)
                                            {
                                            if (atLeastOne)
                                                llab += "+"+detrendlabel[curve][v];
                                            else
                                                llab += " ("+detrendlabel[curve][v];
                                            atLeastOne = true;
                                            }
                                        if (v == maxDetrendVars - 1 && atLeastOne)
                                            llab += " detrended)";
                                        }
                                    }                               
                                if (showLnormInfo && normIndex[curve] != 0) llab += " (normalized)";
                                }
                            }
                        else if (useLegend[curve])
                            llab = legend[curve];
                        if (useColumnName[curve] || useLegend[curve])
                            {
                            if (mmag[curve] && showLmmagInfo)
                                if (mmag[firstCurve] && totalScaleFactor[firstCurve] == 0.001)
                                    llab += " (mmag)";
                                else if (totalScaleFactor[curve] == 0.001)
                                    llab += " (mmag)";
                                else
                                    llab += " (mag)";
                            if (!force[curve]&&(showLScaleInfo || showLShiftInfo))
                                llab += scaleShiftText(showLScaleInfo, showLShiftInfo, mmag[curve], showLmmagInfo, totalScaleFactor[curve], totalShiftFactor[curve]);
                            else if (force[curve]&&(showLRelScaleInfo || showLRelShiftInfo))
                                llab += scaleShiftText(showLRelScaleInfo, showLRelShiftInfo, mmag[curve], showLmmagInfo, totalScaleFactor[curve], totalShiftFactor[curve]);
                            if (useTwoLineLegend)
                                {
                                drawLegendSymbol(curve, legPosY, llab);
                                plot.addLabel(legendPosX, legPosY, llab);
                                legPosY += 18./plotSizeY;
                                llab = " ";
                                }

                            if ((binSize[curve] != 1) && showLBinInfo)
                                llab += " (bin size = "+binSize[curve]+")";
                            if (showLSymbolInfo)
                                llab += " ("+markers[markerIndex[curve]]+")";
                            if (!llab.equals(" "))
                                {
                                if (!useTwoLineLegend) drawLegendSymbol(curve, legPosY, llab);
                                plot.addLabel(legendPosX, legPosY, llab);
                                legPosY += 18./plotSizeY;
                                }
                            }
                        }
                    }

                if(useBoldedDatum && boldedDatum != -1)
                    {
                    double[] xx = new double[1];
                    double[] yy = new double[1];
                    for (int curve=0; curve<maxCurves; curve++)
                        {
                        if (plotY[curve])
                            {
                            plot.setColor (color[curve]);
                            if (marker[curve] == ij.gui.Plot.DOT)
                                plot.setLineWidth(8);
                            else
                                plot.setLineWidth(2);
                            if (curve == firstCurve)
                                {
                                xx[0] = x[curve][boldedDatum];
                                yy[0] = y[curve][boldedDatum];                                
                                }
                            else
                                {
                                xx[0] = x[curve][boldedDatum*binSize[firstCurve]/(binSize[curve])];
                                yy[0] = y[curve][boldedDatum*binSize[firstCurve]/(binSize[curve])];
                                }
                            plot.addPoints(xx,yy,marker[curve]);
                            }
                        }
                    }

                plot.setLineWidth(1);
                plot.setJustification(plot.CENTER);
                
                if (showDMarkers)
                    {
                    plot.setColor (Color.GRAY);
                    double[] samples1 = new double[10];
                    double[] samples2 = new double[10];
                    double[] samples3 = new double[10];
                    double[] samples4 = new double[10];
                    for (int i = 0; i < 10; i++)
                        {
                        samples1[i] = Double.NaN;
                        samples2[i] = Double.NaN;
                        samples3[i] = Double.NaN;
                        samples4[i] = Double.NaN;
                        }                    
                    
                    double preDmark1Ref = invertYAxis?yPlotMin:yPlotMax;
                    double preDmark2Ref = preDmark1Ref;
                    double postDMarker3Ref = preDmark1Ref;
                    double postDMarker4Ref = preDmark1Ref;
                    int nBefore2 = 0;
                    int nAfter2 = 0;
                    int nBefore1 = 0;
                    int nAfter1 = 0;
                    int nBefore3 = 0;
                    int nAfter3 = 0;
                    int nBefore4 = 0;
                    int nAfter4 = 0;
                    for (int i=0; i<nn[firstCurve]; i++)
                        {
                        if (!Double.isNaN(x[firstCurve][i]) && !Double.isNaN(y[firstCurve][i]))
                            {
                            if (x[firstCurve][i] < dMarker2Value)
                                {
                                samples2[4]=samples2[3]; samples2[3]=samples2[2]; samples2[2]=samples2[1]; samples2[1]=samples2[0]; samples2[0]=y[firstCurve][i];
                                nBefore2++;
                                }
                            else if (nAfter2 < 5)
                                {
                                samples2[5+nAfter2]=y[firstCurve][i]; 
                                nAfter2++;
                                }
                                
                            if (useDMarker1 && x[firstCurve][i] < dMarker1Value)
                                {
                                samples1[4]=samples1[3]; samples1[3]=samples1[2]; samples1[2]=samples1[1]; samples1[1]=samples1[0]; samples1[0]=y[firstCurve][i];
                                nBefore1++;
                                }   
                            else if (useDMarker1 && nAfter1 < 5)
                                {
                                samples1[5+nAfter1]=y[firstCurve][i]; 
                                nAfter1++;
                                }
                            
                            if (x[firstCurve][i] < dMarker3Value)
                                {
                                samples3[4]=samples3[3]; samples3[3]=samples3[2]; samples3[2]=samples3[1]; samples3[1]=samples3[0]; samples3[0]=y[firstCurve][i];
                                nBefore3++;
                                }
                            else if (nAfter3 < 5)
                                {
                                samples3[5+nAfter3]=y[firstCurve][i]; 
                                nAfter3++;
                                }
                                
                            if (useDMarker4 && x[firstCurve][i] < dMarker4Value)
                                {
                                samples4[4]=samples4[3]; samples4[3]=samples4[2]; samples4[2]=samples4[1]; samples4[1]=samples4[0]; samples4[0]=y[firstCurve][i];
                                nBefore4++;
                                }   
                            else if (useDMarker4 && nAfter4 < 5)
                                {
                                samples4[5+nAfter4]=y[firstCurve][i]; 
                                nAfter4++;
                                }                           
                            }
                        }
                    if (nBefore2+nAfter2 > 0)
                        {
                        preDmark2Ref = invertYAxis?minOf(samples2,10):maxOf(samples2,10);
                        preDmark2Ref = invertYAxis?Math.max(preDmark2Ref, plotMaxY):Math.min(preDmark2Ref, plotMaxY);
                        }
                    if (nBefore3+nAfter3 > 0)
                        {
                        postDMarker3Ref = invertYAxis?minOf(samples3,10):maxOf(samples3,10);
                        postDMarker3Ref = invertYAxis?Math.max(postDMarker3Ref, plotMaxY):Math.min(postDMarker3Ref, plotMaxY);
                        }                    
                    dashLength = 5*(plotMaxY - plotMinY)/plotSizeY;
                    
                    numDashes = -10 + (preDmark2Ref - plotMinY)/dashLength;        //plot dMarker2
                    for (int dashCount = 0; dashCount < numDashes; dashCount +=2)    
                        {
                        plot.drawLine(dMarker2Value, preDmark2Ref-dashLength*dashCount, dMarker2Value, preDmark2Ref-dashLength*(dashCount+1));
                        }
                    plot.setJustification(plot.CENTER);                    
                    plot.addLabel((dMarker2Value-plotMinX)/(plotMaxX-plotMinX), 1 - 16.0/plotSizeY, "Left");

                    numDashes = -10 + (postDMarker3Ref - plotMinY)/dashLength;     //plot dMarker3
                    for (int dashCount = 0; dashCount < numDashes; dashCount +=2)     
                        {
                        plot.drawLine(dMarker3Value, postDMarker3Ref-dashLength*dashCount, dMarker3Value, postDMarker3Ref-dashLength*(dashCount+1));
                        }
                    plot.addLabel((dMarker3Value-plotMinX)/(plotMaxX-plotMinX), 1 - 16.0/plotSizeY, "Right");
                    
                    if (useDMarker1)   //plot dMarker1
                        {
                        if (nBefore1+nAfter1 > 0)
                            {
                            preDmark1Ref = invertYAxis?minOf(samples1,10):maxOf(samples1,10);
                            preDmark1Ref = invertYAxis?Math.max(preDmark1Ref, plotMaxY):Math.min(preDmark1Ref, plotMaxY);
                            }                        
                        numDashes = -10 + (preDmark1Ref - plotMinY)/dashLength;        //plot dMarker1
                        for (int dashCount = 0; dashCount < numDashes; dashCount +=2)    
                            {
                            plot.drawLine(dMarker1Value, preDmark1Ref-dashLength*dashCount, dMarker1Value, preDmark1Ref-dashLength*(dashCount+1));
                            }     
                        plot.addLabel((dMarker1Value-plotMinX)/(plotMaxX-plotMinX), 1 - 25.0/plotSizeY, "Left");
                        plot.addLabel((dMarker1Value-plotMinX)/(plotMaxX-plotMinX), 1 - 7.0/plotSizeY, "Trim");                         
                        }
                    if (useDMarker4)   //plot dMarker4
                        {
                        if (nBefore4+nAfter4 > 0)
                            {
                            postDMarker4Ref = invertYAxis?minOf(samples4,10):maxOf(samples4,10);
                            postDMarker4Ref = invertYAxis?Math.max(postDMarker4Ref, plotMaxY):Math.min(postDMarker4Ref, plotMaxY);
                            }                         
                        numDashes = -10 + (postDMarker4Ref - plotMinY)/dashLength;     //plot dMarker4
                        for (int dashCount = 0; dashCount < numDashes; dashCount +=2)     
                            {
                            plot.drawLine(dMarker4Value, postDMarker4Ref-dashLength*dashCount, dMarker4Value, postDMarker4Ref-dashLength*(dashCount+1));
                            }       
                        plot.addLabel((dMarker4Value-plotMinX)/(plotMaxX-plotMinX), 1 - 25.0/plotSizeY, "Right");
                        plot.addLabel((dMarker4Value-plotMinX)/(plotMaxX-plotMinX), 1 - 7.0/plotSizeY, "Trim");                        
                        }                  
                    }                
                
                if (showVMarker1) drawVMarker(vMarker1Value, vMarker1TopText, vMarker1BotText);

                if (showVMarker2) drawVMarker(vMarker2Value, vMarker2TopText, vMarker2BotText);

                plot.setColor (java.awt.Color.black);
                plot.setJustification(plot.CENTER);
                if (useTitle)
                        {
                        plot.changeFont(new java.awt.Font("Dialog",java.awt.Font.PLAIN,18));
                        plot.addTitle(titlePosX, titlePosY/plotSizeY>1?1:titlePosY/plotSizeY,title);
                        }
                if (useSubtitle)
                        {
                        plot.changeFont(new java.awt.Font("Dialog",java.awt.Font.PLAIN,14));
                        plot.addTitle(subtitlePosX, subtitlePosY/plotSizeY>1?1:subtitlePosY/plotSizeY, subtitle);
                        }

                plotImage = WindowManager.getImage("Plot of "+tableName);
                
                if (plotImage == null)
                    {
                    plotFrameLocationX=(int)Prefs.get("plot2.plotFrameLocationX", plotFrameLocationX);
                    plotFrameLocationY=(int)Prefs.get("plot2.plotFrameLocationY", plotFrameLocationY);
                    if (!Prefs.isLocationOnScreen(new Point(plotFrameLocationX,plotFrameLocationY)))
                        {
                        plotFrameLocationX = 10;
                        plotFrameLocationY = 10;
                        Prefs.set("plot2.plotFrameLocationX", plotFrameLocationX);
                        Prefs.set("plot2.plotFrameLocationY", plotFrameLocationY);
                        }
                    plotWindow = plot.show();
                    plotWindow.setIconImage(plotIcon.getImage());
                    plotImage = plotWindow.getImagePlus();//WindowManager.getImage("Plot of "+tableName);
                    
//                    plotWindow = plotImage.getWindow();
//                    plotWindow = new PlotWindow(plot);
//                    plotImage = plotWindow.getImagePlus();
//                    plotFrame = WindowManager.getFrame("Plot of "+tableName);
//                    if ((plotImage != null) && rememberWindowLocations)
//                        {
//                        plotFrameLocationX=(int)Prefs.get("plot2.plotFrameLocationX", plotFrameLocationX);
//                        plotFrameLocationY=(int)Prefs.get("plot2.plotFrameLocationY", plotFrameLocationY);
//                        if (!Prefs.isLocationOnScreen(new Point(plotFrameLocationX,plotFrameLocationY)))
//                            {
//                            plotFrameLocationX = 10;
//                            plotFrameLocationY = 10;
//                            }
//                        plotWindow.setLocation(plotFrameLocationX, plotFrameLocationY);
//                        }

//                    plotImage = WindowManager.getImage("Plot of "+tableName);
//                    plotFrame = WindowManager.getFrame("Plot of "+tableName);
//                    ImageProcessor ip = plot.getProcessor();
//                    plotImage.setProcessor("Plot of "+tableName,ip);
                    
                    plotImageCanvas = plotImage.getCanvas();
                    plotOverlayCanvas = new OverlayCanvas(plotImage);
                    list.clear();

                    MouseMotionListener[] mml = plotImageCanvas.getMouseMotionListeners();
                    if (mml.length>0)
                            for (int i=0; i<mml.length; i++)
                                    plotImageCanvas.removeMouseMotionListener(mml[i]);

                    MouseWheelListener[] mwl = plotImageCanvas.getMouseWheelListeners();
                    if (mwl.length>0)
                            for (int i=0; i<mwl.length; i++)
                                    plotImageCanvas.removeMouseWheelListener(mwl[i]);

                    MouseWheelListener[] mwl3 = plotWindow.getMouseWheelListeners();
                    if (mwl3.length>0)
                            for (int i=0; i<mwl3.length; i++)
                                    plotWindow.removeMouseWheelListener(mwl3[i]);

                    MouseListener[] ml = plotImageCanvas.getMouseListeners();
                    if (ml.length>0)
                            for (int i=0; i<ml.length; i++)
                                    plotImageCanvas.removeMouseListener(ml[i]);

                    plotImageCanvas.addMouseWheelListener(plotMouseWheelListener);
                    plotImageCanvas.addMouseMotionListener(plotMouseMotionListener);
                    plotImageCanvas.addMouseListener(plotMouseListener);
                    plotImageCanvas.setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));
                    zoom = 0.0;
                    }
                else
                    {
                    plotWindow = plotImage.getWindow();
//                    plotFrame = (Frame)plotWindow;
                    ImageProcessor ip = plot.getProcessor();
                    plotImage.setProcessor("Plot of "+tableName,ip);
                    plotImageCanvas = plotImage.getCanvas();
                    plotImageCanvas.setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));
                    }

                plotbottompanel = (Panel) plotWindow.getComponent(1);
                plotbottompanel.getComponentCount();
                plotbottompanel.setSize(600,30);
                plotcoordlabel = (Label) plotbottompanel.getComponent(plotbottompanel.getComponentCount()-1);
                plotcoordlabel.setSize(400,20);
                plotcoordlabel.setBackground(Color.white);

                //Replot to clean up blank areas

//                plotImage = WindowManager.getImage("Plot of "+tableName);
//                ImageProcessor ip = plot.getProcessor();
//                plotImage.setProcessor("Plot of "+tableName,ip);
                
                excludedHeadSamples = holdExcludedHeadSamples;
                excludedTailSamples = holdExcludedTailSamples;
                for (int i=0; i<maxCurves; i++)
                    {
                    showErr[i] = showErrors[i];
                    showErrors[i] = errorcolumnbox[i].isSelected();
                    }
                table.setLock(false);    
                updatePlotRunning = false;
                }
        
    static void drawVMarker (double vMarkerValue, String vMarkerTopText, String vMarkerBotText)
                    {
                    plot.setColor (Color.red);
                    double[] samples = new double[10];
                    for (int i = 0; i < 10; i++)
                        {
                        samples[i] = Double.NaN;
                        }
                    double preVmarkRef= invertYAxis?yPlotMin:yPlotMax;
                    int nBefore = 0;
                    int nAfter = 0;
                    for (int i=0; i<nn[firstCurve]; i++)
                        {
                        if (!Double.isNaN(x[firstCurve][i]) && !Double.isNaN(y[firstCurve][i]))
                            {
                            if (x[firstCurve][i] < vMarkerValue)
                                {
                                samples[4]=samples[3]; samples[3]=samples[2]; samples[2]=samples[1]; samples[1]=samples[0]; samples[0]=y[firstCurve][i];
                                nBefore++;
                                }
                            else if (nAfter < 5)
                                {
                                samples[5+nAfter]=y[firstCurve][i]; 
                                nAfter++;
                                }
                            else
                                {
                                break;
                                }
                            }
                        }
                    if (nBefore+nAfter > 0)
                        {
                        preVmarkRef = invertYAxis?minOf(samples,10):maxOf(samples,10);
                        preVmarkRef = invertYAxis?Math.max(preVmarkRef, plotMaxY):Math.min(preVmarkRef, plotMaxY);
                        }
                    dashLength = 5*(plotMaxY - plotMinY)/plotSizeY;
                    numDashes = -10 + (preVmarkRef - plotMinY)/dashLength;
                    for (int dashCount = 0; dashCount < numDashes; dashCount +=2)    //plot vMarker1
                        {
                        plot.drawLine(vMarkerValue, preVmarkRef-dashLength*dashCount, vMarkerValue, preVmarkRef-dashLength*(dashCount+1));
                        }
                    plot.setJustification(plot.CENTER);
                    plot.addLabel((vMarkerValue-plotMinX)/(plotMaxX-plotMinX), 1 - 25.0/plotSizeY, vMarkerTopText);
                    plot.addLabel((vMarkerValue-plotMinX)/(plotMaxX-plotMinX), 1 - 7.0/plotSizeY, vMarkerBotText);
                    }        
        
    static void drawLegendSymbol(int curve, double legPosY, String llab)
        {
        double xShift = 6.0;
        double yShift = 6.0;
        Font font = new Font("Dialog", Font.PLAIN, 12);
		int h = 0;
		int w = 0; 
        if (mainpanel != null)
            {
            Graphics g = mainpanel.getGraphics();
            FontMetrics metrics = g.getFontMetrics (font);
            h = metrics.getHeight();
            w = metrics.stringWidth(llab);
            }
    
        if (legendRight)
            xShift += w;
        else if (!legendLeft)
            xShift += w/2;      
        plot.setColor (color[curve]);
        if (marker[curve] == ij.gui.Plot.DOT)
            {
            xShift += 2;
            yShift += 2;
            plot.setLineWidth(4);
            }
        else if (marker[curve] == ij.gui.Plot.LINE)
            {
            xShift += 7;
            yShift += 2;            
            plot.setLineWidth(1);
            }        
        else
            {
            xShift += 3;
            yShift += 3;            
            plot.setLineWidth(1);
            }
        double x = plotMinX + (plotMaxX-plotMinX)*(legendPosX-xShift/(plotSizeX-Plot.LEFT_MARGIN-Plot.RIGHT_MARGIN));
        double y = plotMinY + (plotMaxY-plotMinY)*(1-legPosY+yShift/(plotSizeY-Plot.TOP_MARGIN-Plot.BOTTOM_MARGIN));
        double[] xx = {x};
        double[] yy = {y};  
        if (marker[curve] != ij.gui.Plot.LINE)
            {
            plot.addPoints (xx,yy,marker[curve]);
            }
        else
            {
            plot.drawLine(x, y, x + (plotMaxX-plotMinX)*6.0/(plotSizeX-Plot.LEFT_MARGIN-Plot.RIGHT_MARGIN), y);
            }
        plot.setLineWidth(1);  
        }
        
        

    static String getPositionColumn(String label, String XorY)
        {
        String nums = "";
        for (int i = 0; i<label.length(); i++)
            {
            if (label.charAt(i) >= '0' && label.charAt(i) <= '9')
                {
                nums = nums.concat(label.substring(i, i+1));
                }
            }
        int ap =  parseInteger(nums, 1);
        if (ap < 1 || ap > isRefStar.length) ap = 1;
        if (table.getColumnIndex(XorY+"(IJ)_"+(isRefStar[ap -1]?"C":"T")+ap) == ResultsTable.COLUMN_NOT_FOUND)
            {
            if (table.getColumnIndex(XorY+"(FITS)_"+(isRefStar[ap -1]?"C":"T")+ap) == ResultsTable.COLUMN_NOT_FOUND)
                {
                return "no_position_data_found";
                }
            else
                {
                return XorY+"(FITS)_"+(isRefStar[ap -1]?"C":"T")+ap;
                }
            }
        return XorY+"(IJ)_"+(isRefStar[ap -1]?"C":"T")+ap;
        }

    static String getSourceID(String label)
        {
        String nums = "";
        for (int i = 0; i<label.length(); i++)
            {
            if (label.charAt(i) >= '0' && label.charAt(i) <= '9')
                {
                nums = nums.concat(label.substring(i, i+1));
                }
            }
        int ap =  parseInteger(nums, 1);
        if (ap < 1 || ap > isRefStar.length) ap = 1;
        return (isRefStar[ap -1]?"C":"T")+ap;
        }
    
	static int parseInteger(String s, int defaultValue) {
		if (s==null) return defaultValue;
		try {
			defaultValue = Integer.parseInt(s);
            } 
        catch (NumberFormatException e) {}
		return defaultValue;			
        }    
    
    static void addNewAstroData()
        {
        astroConverterUpdating = false;
        acc = new AstroConverter(true, true, "MP Coordinate Converter");
        acc.setEnableTimeEntry(false);
        acc.getTAIminusUTC();
        acc.setEnableObjectEntry(true);
        acc.setEnableObservatoryEntry(true);
        acc.showPanel(true);
        
        addAstroDataFrame = new JFrame ("Add astronomical data to table");
        addAstroDataFrame.setIconImage(plotIcon.getImage());
        JPanel addAstroDataPanel = new JPanel (new SpringLayout());
        JScrollPane addAstroDataScrollPane = new JScrollPane(addAstroDataPanel);
        addAstroDataFrame.setDefaultCloseOperation(javax.swing.WindowConstants.DISPOSE_ON_CLOSE);
        addAstroDataFrame.addWindowListener(new WindowAdapter(){
            @Override
            public void windowClosing(WindowEvent e){
                    addAstroDataFrameWasShowing = false;
                    getAstroPanelValues();
                    saveAstroPanelPrefs();
                    if (acc != null) acc.saveAndClose();
                    closeAddAstroDataFrame();
                    }});        
        
        useGJDButton = new JRadioButton("JD (UTC)");
        useGJDButton.setToolTipText("Use geocentric JD column as time base");
        useHJDButton = new JRadioButton("HJD (UTC)");
        useHJDButton.setToolTipText("Use heliocentric JD column as time base");
        useBJDButton = new JRadioButton("BJD (TDB)");
        useBJDButton.setToolTipText("Use barycentric JD column as time base");
        
        if (useGJD)
            {
            useGJDButton.setSelected(true);
            useHJDButton.setSelected(false);
            useBJDButton.setSelected(false);
            }
        else if (useHJD)
            {
            useGJDButton.setSelected(false);
            useHJDButton.setSelected(true);
            useBJDButton.setSelected(false);
            }
        else //useBJD
            {
            useGJDButton.setSelected(false);
            useHJDButton.setSelected(false);
            useBJDButton.setSelected(true);
            }
        useGJD = useGJDButton.isSelected();
        useHJD = useHJDButton.isSelected();
        useBJD = useBJDButton.isSelected();
        
        JDRadioGroup = new ButtonGroup();
        JDRadioGroup.add(useGJDButton);
        JDRadioGroup.add(useHJDButton);
        JDRadioGroup.add(useBJDButton);

        useGJDButton.addActionListener(new ActionListener(){
            public void actionPerformed(ActionEvent ae) {
                useGJD = true;
                useHJD = false;
                useBJD = false;
                jdcolumnbox.setSelectedIndex(jdcolumnbox.getSelectedIndex());
                }});
        useHJDButton.addActionListener(new ActionListener(){
            public void actionPerformed(ActionEvent ae) {
                useGJD = false;
                useHJD = true;
                useBJD = false;
                jdcolumnbox.setSelectedIndex(jdcolumnbox.getSelectedIndex());
                }});    
        useBJDButton.addActionListener(new ActionListener(){
            public void actionPerformed(ActionEvent ae) {
                useGJD = false;
                useHJD = false;
                useBJD = true;
                jdcolumnbox.setSelectedIndex(jdcolumnbox.getSelectedIndex());
                }});         
       
        JPanel jdselectionpanel = new JPanel(new SpringLayout());
        jdselectionpanel.setBorder(BorderFactory.createTitledBorder("Input Time Format"));
        jdselectionpanel.add(useGJDButton);
        jdselectionpanel.add(useHJDButton);
        jdselectionpanel.add(useBJDButton);
        SpringUtil.makeCompactGrid (jdselectionpanel, 1, jdselectionpanel.getComponentCount(), 2,0,2,0);
        addAstroDataPanel.add (jdselectionpanel);         
         
        JPanel jdcolumnpanel = new JPanel(new SpringLayout());
        jdcolumnpanel.setBorder(BorderFactory.createTitledBorder("Date/Time Column From Active Table"));
        String[] tabCols = columns;
        if (columns.length > 1) tabCols = Arrays.copyOfRange(tabCols, 1, tabCols.length - 1);
        MutableComboBoxModel jddefaultmodel = new DefaultComboBoxModel(tabCols);
        jdcolumnbox = new JComboBox(jddefaultmodel);
        jdcolumnbox.setSelectedItem(JDColumn);
        jdcolumnbox.setPrototypeDisplayValue("123456789012345");
        jdcolumnbox.addActionListener(new ActionListener(){
            public void actionPerformed(ActionEvent ae) 
                {
                if (!astroConverterUpdating)
                    {
                    JDColumn = (String) jdcolumnbox.getSelectedItem();
                    int tableLength = table.getCounter();
                    int jdRow = tpanel.getSelectionStart();
                    if (jdRow < 0 || jdRow >= tableLength) jdRow = 0;
                    int jdCol = table.getColumnIndex(JDColumn);
                    if (jdCol == MeasurementTable.COLUMN_NOT_FOUND)
                        {
                        IJ.beep();
                        IJ.showMessage("Error: could not find JD table column '"+JDColumn+"'");
                        return;
                        }     
                    if (tableLength > 0)
                        {
                        double jd = table.getValueAsDouble(jdCol, jdRow);
                        if (JDColumn.contains("-2400000")) jd += 2400000;
                        if (!Double.isNaN(jd))
                            {
                            int returnCode = 0;
                            if (useGJD) 
                                {
                                acc.setTime(jd);
                                returnCode = acc.processManualCoordinates();
                                }
                            else if (useHJD) 
                                {
                                returnCode = acc.setHJDTime(jd);
                                }
                            else 
                                {
                                returnCode = acc.setBJDTime(jd);
                                }
                            if (returnCode == 3)
                                {
                                IJ.beep();
                                IJ.showMessage("ERROR: Ohio State access failed when attempting to retrieve BJD(TDB).");
                                return;
                                }
                            } 
                        }
                    }
                }});

        jdcolumnpanel.add (jdcolumnbox);
        jdcolumnpanel.setPreferredSize(new Dimension(125, 25));
        SpringUtil.makeCompactGrid (jdcolumnpanel, 1, 1, 0,0,0,0);
        addAstroDataPanel.add (jdcolumnpanel); 
        
       
        JLabel dataTitleLabel = new JLabel("Select data to add");
        addAstroDataPanel.add (dataTitleLabel); 
        
        JLabel nameTitleLabel = new JLabel("Enter table column name to add");
        addAstroDataPanel.add (nameTitleLabel);        
                
        airmassCB = new JCheckBox("Airmass", addAirmass);
        airmassCB.setToolTipText("Add Airmass column to active table");
        addAstroDataPanel.add (airmassCB); 
        
        airmassField = new JTextField (airmassName);
        airmassField.setBorder (BorderFactory.createEmptyBorder(0,4,0,4));
        airmassField.setPreferredSize (new Dimension(250,20));
        airmassField.setHorizontalAlignment (JTextField.LEFT);
        airmassField.setToolTipText("Airmass column name");        
        addAstroDataPanel.add (airmassField); 
        
        altitudeCB = new JCheckBox("Altitude", addAltitude);
        altitudeCB.setToolTipText("Add Altitude column to active table");
        addAstroDataPanel.add (altitudeCB); 
        
        altitudeField = new JTextField (altitudeName);
        altitudeField.setBorder (BorderFactory.createEmptyBorder(0,4,0,4));
        altitudeField.setPreferredSize (new Dimension(250,20));
        altitudeField.setHorizontalAlignment (JTextField.LEFT);
        altitudeField.setToolTipText("Altitude/elevation column name");        
        addAstroDataPanel.add (altitudeField); 
                
        azimuthCB = new JCheckBox("Azimuth", addAzimuth);
        azimuthCB.setToolTipText("Add Azumuth column to active table");
        addAstroDataPanel.add (azimuthCB); 
        
        azimuthField = new JTextField (azimuthName);
        azimuthField.setBorder (BorderFactory.createEmptyBorder(0,4,0,4));
        azimuthField.setPreferredSize (new Dimension(250,20));
        azimuthField.setHorizontalAlignment (JTextField.LEFT);
        azimuthField.setToolTipText("Azimuth column name");        
        addAstroDataPanel.add (azimuthField); 
        
        hourAngleCB = new JCheckBox("Hour Angle", addHourAngle);
        hourAngleCB.setToolTipText("Add Hour Angle column to active table");
        addAstroDataPanel.add (hourAngleCB); 
        
        hourAngleField = new JTextField (hourAngleName);
        hourAngleField.setBorder (BorderFactory.createEmptyBorder(0,4,0,4));
        hourAngleField.setPreferredSize (new Dimension(250,20));
        hourAngleField.setHorizontalAlignment (JTextField.LEFT);
        hourAngleField.setToolTipText("Hour Angle column name");        
        addAstroDataPanel.add (hourAngleField); 
        
        zenithDistanceCB = new JCheckBox("Zenith Distance", addZenithDistance);
        zenithDistanceCB.setToolTipText("Add Zenith Distance column to active table");
        addAstroDataPanel.add (zenithDistanceCB); 
        
        zenithDistanceField = new JTextField (zenithDistanceName);
        zenithDistanceField.setBorder (BorderFactory.createEmptyBorder(0,4,0,4));
        zenithDistanceField.setPreferredSize (new Dimension(250,20));
        zenithDistanceField.setHorizontalAlignment (JTextField.LEFT);
        zenithDistanceField.setToolTipText("Zenith Distance column name");        
        addAstroDataPanel.add (zenithDistanceField);  
        
        gjdCB = new JCheckBox("JD_UTC", addGJD);
        gjdCB.setToolTipText("Add Geocentric Julian Date (UTC) column to active table"); 
        addAstroDataPanel.add (gjdCB); 
        
        gjdField = new JTextField (gjdName);
        gjdField.setBorder (BorderFactory.createEmptyBorder(0,4,0,4));
        gjdField.setPreferredSize (new Dimension(250,20));
        gjdField.setHorizontalAlignment (JTextField.LEFT);
        gjdField.setToolTipText("Geocentric Julian Date (UTC) column name");        
        addAstroDataPanel.add (gjdField);        
        
        hjdCB = new JCheckBox("HJD_UTC", addHJD);
        hjdCB.setToolTipText("Add Heliocentric Julian Date (UTC) column to active table"); 
        addAstroDataPanel.add (hjdCB); 
        
        hjdField = new JTextField (hjdName);
        hjdField.setBorder (BorderFactory.createEmptyBorder(0,4,0,4));
        hjdField.setPreferredSize (new Dimension(250,20));
        hjdField.setHorizontalAlignment (JTextField.LEFT);
        hjdField.setToolTipText("Heliocentric Julian Date (UTC) column name");        
        addAstroDataPanel.add (hjdField); 
        
        hjdCorrCB = new JCheckBox("HJD Correction", addHJDCorr);
        hjdCorrCB.setToolTipText("Add Geocentric (UTC) to Heliocentric (UTC) Julian Date Correction column to active table"); 
        addAstroDataPanel.add (hjdCorrCB); 
        
        hjdCorrField = new JTextField (hjdCorrName);
        hjdCorrField.setBorder (BorderFactory.createEmptyBorder(0,4,0,4));
        hjdCorrField.setPreferredSize (new Dimension(250,20));
        hjdCorrField.setHorizontalAlignment (JTextField.LEFT);
        hjdCorrField.setToolTipText("JD to HJD correction column name");        
        addAstroDataPanel.add (hjdCorrField);         
        
        bjdCB = new JCheckBox("BJD_TDB", addBJD);
        bjdCB.setToolTipText("Add Barycentric Julian Date (TDB) column to active table"); 
        addAstroDataPanel.add (bjdCB); 
        
        bjdField = new JTextField (bjdName);
        bjdField.setBorder (BorderFactory.createEmptyBorder(0,4,0,4));
        bjdField.setPreferredSize (new Dimension(250,20));
        bjdField.setHorizontalAlignment (JTextField.LEFT);
        bjdField.setToolTipText("Barycentric Julian Date (TDB) column name");        
        addAstroDataPanel.add (bjdField); 
        
        bjdCorrCB = new JCheckBox("BJD Correction", addBJDCorr);
        bjdCorrCB.setToolTipText("Add Geocentric (UTC) to Barycentric (TDB) Julian Date Correction column to active table"); 
        addAstroDataPanel.add (bjdCorrCB); 
        
        bjdCorrField = new JTextField (bjdCorrName);
        bjdCorrField.setBorder (BorderFactory.createEmptyBorder(0,4,0,4));
        bjdCorrField.setPreferredSize (new Dimension(250,20));
        bjdCorrField.setHorizontalAlignment (JTextField.LEFT);
        bjdCorrField.setToolTipText("JD to BJD correction column name");        
        addAstroDataPanel.add (bjdCorrField);
       
        raNowCB = new JCheckBox("RA Now", addRaNow);
        raNowCB.setToolTipText("Add Right Ascension of target at epoch of observation column to active table"); 
        addAstroDataPanel.add (raNowCB); 
        
        raNowField = new JTextField (raNowName);
        raNowField.setBorder (BorderFactory.createEmptyBorder(0,4,0,4));
        raNowField.setPreferredSize (new Dimension(250,20));
        raNowField.setHorizontalAlignment (JTextField.LEFT);
        raNowField.setToolTipText("Right Ascension of target at epoch of observation column name");        
        addAstroDataPanel.add (raNowField);
        
        decNowCB = new JCheckBox("Declination Now", addDecNow);
        decNowCB.setToolTipText("Add Declination of target at epoch of observation column to active table"); 
        addAstroDataPanel.add (decNowCB); 
        
        decNowField = new JTextField (decNowName);
        decNowField.setBorder (BorderFactory.createEmptyBorder(0,4,0,4));
        decNowField.setPreferredSize (new Dimension(250,20));
        decNowField.setHorizontalAlignment (JTextField.LEFT);
        decNowField.setToolTipText("Declination of target at epoch of observation column name");        
        addAstroDataPanel.add (decNowField); 
        
        ra2000CB = new JCheckBox("RA J2000", addRA2000);
        ra2000CB.setToolTipText("Add Right Ascension of target at epoch J2000 column to active table"); 
        addAstroDataPanel.add (ra2000CB); 
        
        ra2000Field = new JTextField (ra2000Name);
        ra2000Field.setBorder (BorderFactory.createEmptyBorder(0,4,0,4));
        ra2000Field.setPreferredSize (new Dimension(250,20));
        ra2000Field.setHorizontalAlignment (JTextField.LEFT);
        ra2000Field.setToolTipText("Right Ascension of target at epoch J2000 column name");        
        addAstroDataPanel.add (ra2000Field); 
        
        dec2000CB = new JCheckBox("Declination J2000", addDec2000);
        dec2000CB.setToolTipText("Add Declination of target at epoch J2000 column to active table"); 
        addAstroDataPanel.add (dec2000CB); 
        
        dec2000Field = new JTextField (dec2000Name);
        dec2000Field.setBorder (BorderFactory.createEmptyBorder(0,4,0,4));
        dec2000Field.setPreferredSize (new Dimension(250,20));
        dec2000Field.setHorizontalAlignment (JTextField.LEFT);
        dec2000Field.setToolTipText("Declination of target at epoch J2000 column name");        
        addAstroDataPanel.add (dec2000Field);
        
        JLabel lineFeedLabel1 = new JLabel("");
        addAstroDataPanel.add (lineFeedLabel1); 
        JLabel lineFeedLabel2 = new JLabel("");
        addAstroDataPanel.add (lineFeedLabel2);         

        JLabel instructionLabel1L = new JLabel("Setup target and observatory parameters");
        addAstroDataPanel.add (instructionLabel1L); 
        JLabel instructionLabel1R = new JLabel("");
        addAstroDataPanel.add (instructionLabel1R);
        
        JLabel instructionLabel2L = new JLabel("in 'MP Coordinate Converter' window,");
        addAstroDataPanel.add (instructionLabel2L); 
        JLabel instructionLabel2R = new JLabel("");
        addAstroDataPanel.add (instructionLabel2R);        

        JLabel buttonLabel = new JLabel("then press the 'Update Table' button.");
        addAstroDataPanel.add (buttonLabel);         
        
        JPanel OKCancelPanel = new JPanel (new SpringLayout());
        JLabel buttonIndentLabel = new JLabel("");
        OKCancelPanel.add (buttonIndentLabel);        
        
        OKbutton = new JButton("Update Table");
        OKbutton.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e)
                {
                astroConverterUpdating = true;
                JDColumn = (String) jdcolumnbox.getSelectedItem();
                int jdCol = table.getColumnIndex(JDColumn);
                if (jdCol == MeasurementTable.COLUMN_NOT_FOUND)
                    {
                    astroConverterUpdating = false;
                    IJ.beep();
                    IJ.showMessage("Error: could not find table column '"+JDColumn+"'");
                    return;
                    }

                getAstroPanelValues();
                
                int airmassCol = MeasurementTable.COLUMN_NOT_FOUND;
                int altitudeCol = MeasurementTable.COLUMN_NOT_FOUND;
                int azimuthCol = MeasurementTable.COLUMN_NOT_FOUND;
                int hourAngleCol = MeasurementTable.COLUMN_NOT_FOUND;
                int zenithDistanceCol = MeasurementTable.COLUMN_NOT_FOUND;
                int gjdCol = MeasurementTable.COLUMN_NOT_FOUND;
                int hjdCol = MeasurementTable.COLUMN_NOT_FOUND;
                int hjdCorrCol = MeasurementTable.COLUMN_NOT_FOUND;                
                int bjdCol = MeasurementTable.COLUMN_NOT_FOUND;
                int bjdCorrCol = MeasurementTable.COLUMN_NOT_FOUND;
                int raNowCol = MeasurementTable.COLUMN_NOT_FOUND;
                int decNowCol = MeasurementTable.COLUMN_NOT_FOUND;
                int ra2000Col = MeasurementTable.COLUMN_NOT_FOUND;
                int dec2000Col = MeasurementTable.COLUMN_NOT_FOUND;
                
                if (addAirmass) airmassCol = table.getColumnIndex(airmassName);
                if (addAltitude) altitudeCol = table.getColumnIndex(altitudeName);
                if (addAzimuth) azimuthCol = table.getColumnIndex(azimuthName);
                if (addHourAngle) hourAngleCol = table.getColumnIndex(hourAngleName);
                if (addZenithDistance) zenithDistanceCol = table.getColumnIndex(zenithDistanceName);
                if (addGJD) gjdCol = table.getColumnIndex(gjdName);
                if (addHJD) hjdCol = table.getColumnIndex(hjdName);
                if (addHJDCorr) hjdCorrCol = table.getColumnIndex(hjdCorrName);
                if (addBJD) bjdCol = table.getColumnIndex(bjdName);
                if (addBJDCorr) bjdCorrCol = table.getColumnIndex(bjdCorrName);
                if (addRaNow) raNowCol = table.getColumnIndex(raNowName);
                if (addDecNow) decNowCol = table.getColumnIndex(decNowName);
                if (addRA2000) ra2000Col = table.getColumnIndex(ra2000Name);
                if (addDec2000) dec2000Col = table.getColumnIndex(dec2000Name);

                if ((addAirmass && airmassCol != MeasurementTable.COLUMN_NOT_FOUND) ||
                    (addAltitude && altitudeCol != MeasurementTable.COLUMN_NOT_FOUND) ||
                    (addAzimuth && azimuthCol != MeasurementTable.COLUMN_NOT_FOUND) ||
                    (addHourAngle && hourAngleCol != MeasurementTable.COLUMN_NOT_FOUND) ||
                    (addZenithDistance && zenithDistanceCol != MeasurementTable.COLUMN_NOT_FOUND) ||
                    (addGJD && gjdCol != MeasurementTable.COLUMN_NOT_FOUND) ||
                    (addHJD && hjdCol != MeasurementTable.COLUMN_NOT_FOUND) ||
                    (addHJDCorr && hjdCorrCol != MeasurementTable.COLUMN_NOT_FOUND) ||                        
                    (addBJD && bjdCol != MeasurementTable.COLUMN_NOT_FOUND) ||
                    (addBJDCorr && bjdCorrCol != MeasurementTable.COLUMN_NOT_FOUND) ||
                    (addRaNow && raNowCol != MeasurementTable.COLUMN_NOT_FOUND) ||
                    (addDecNow && decNowCol != MeasurementTable.COLUMN_NOT_FOUND) ||
                    (addRA2000 && ra2000Col != MeasurementTable.COLUMN_NOT_FOUND) ||
                    (addDec2000 && dec2000Col != MeasurementTable.COLUMN_NOT_FOUND) )
                    {
                    GenericDialog gd = new GenericDialog ("Over-write existing data?", addAstroDataFrame.getX()+100, addAstroDataFrame.getY()+100);
                    gd.addMessage ((airmassCol != MeasurementTable.COLUMN_NOT_FOUND?"Airmass column: "+airmassName+"\n":"")+
                                   (altitudeCol != MeasurementTable.COLUMN_NOT_FOUND?"Altitude column: "+altitudeName+"\n":"")+
                                   (azimuthCol != MeasurementTable.COLUMN_NOT_FOUND?"Azimuth column: "+azimuthName+"\n":"")+
                                   (hourAngleCol != MeasurementTable.COLUMN_NOT_FOUND?"Hour Angle column: "+hourAngleName+"\n":"")+
                                   (zenithDistanceCol != MeasurementTable.COLUMN_NOT_FOUND?"Zenith Distance column: "+zenithDistanceName+"\n":"")+
                                   (gjdCol != MeasurementTable.COLUMN_NOT_FOUND?"JD column: "+gjdName+"\n":"")+
                                   (hjdCol != MeasurementTable.COLUMN_NOT_FOUND?"HJD column: "+hjdName+"\n":"")+
                                   (hjdCorrCol != MeasurementTable.COLUMN_NOT_FOUND?"HJD Correction column: "+hjdCorrName+"\n":"")+                            
                                   (bjdCol != MeasurementTable.COLUMN_NOT_FOUND?"BJD column: "+bjdName+"\n":"")+
                                   (bjdCorrCol != MeasurementTable.COLUMN_NOT_FOUND?"BJD Correction column: "+bjdCorrName+"\n":"")+
                                   (raNowCol != MeasurementTable.COLUMN_NOT_FOUND?"RA OBS column: "+raNowName+"\n":"")+
                                   (decNowCol != MeasurementTable.COLUMN_NOT_FOUND?"DEC OBS column: "+decNowName+"\n":"")+
                                   (ra2000Col != MeasurementTable.COLUMN_NOT_FOUND?"RA J2000 column: "+ra2000Name+"\n":"")+
                                   (dec2000Col != MeasurementTable.COLUMN_NOT_FOUND?"Dec J2000 column: "+dec2000Name+"\n":"")+
                                    "\n"+
                                    "table column name(s) are already in use.\nPress OK to over-write existing data.");
                    gd.showDialog();
                    if (gd.wasCanceled()) 
                        {
                        astroConverterUpdating = false;
                        return;
                        }
                    }
                defaultOKForeground = OKbutton.getForeground();
                Thread t = new Thread()
                    {
                    public void run()
                        {if (OKbutton != null)
                            {
                            OKbutton.setForeground(Color.RED);
                            OKbutton.setText("working...");
                            OKbutton.paint(OKbutton.getGraphics());
                            }
                        }
                    };

                t.start();
                Thread.yield();                

                Prefs.set("plot2.JDColumn", JDColumn);
                
                saveAstroPanelPrefs();

                int tableLength = table.getCounter();
                double jd;
                for (int i = 0; i<tableLength; i++)
                    {
                    jd = table.getValueAsDouble(jdCol, i);
                    if (JDColumn.contains("-2400000")) jd += 2400000;
                    if (!Double.isNaN(jd))
                        {
                        int returnCode = 0;
                        if (useGJD) 
                            {
                            acc.setTime(jd);
                            returnCode = acc.processManualCoordinates();
                            }
                        else if (useHJD) 
                            {
                            returnCode = acc.setHJDTime(jd);
                            }
                        else 
                            {
                            acc.setBJDTime(jd);
                            returnCode = acc.setBJDTime(jd);
                            }
                        if (returnCode == 3)
                            {
                            astroConverterUpdating = false;
                            IJ.beep();
                            IJ.showMessage("ERROR: Ohio State access failed when attempting to retrieve BJD(TDB).");
                            return;
                            }
                        if (addAirmass) table.setValue(airmassName, i, acc.getAirmass());
                        if (addAltitude) table.setValue(altitudeName, i, acc.getAltitude());
                        if (addAzimuth) table.setValue(azimuthName, i, acc.getAzimuth());
                        if (addHourAngle) table.setValue(hourAngleName, i, acc.getHourAngle());
                        if (addZenithDistance) table.setValue(zenithDistanceName, i, acc.getZenithDistance());
                        if (addGJD) table.setValue(gjdName, i, acc.getJD());
                        if (addHJD) table.setValue(hjdName, i, acc.getHJD());
                        if (addHJDCorr) table.setValue(hjdCorrName, i, acc.getHJDCorrection());  
                        if (addBJD) table.setValue(bjdName, i, acc.getBJD());
                        if (addBJDCorr) table.setValue(bjdCorrName, i, acc.getBJDCorrection());
                        if (addRaNow) table.setValue(raNowName, i, acc.getRAEOI());
                        if (addDecNow) table.setValue(decNowName, i, acc.getDecEOI());
                        if (addRA2000) table.setValue(ra2000Name, i, acc.getRAJ2000());
                        if (addDec2000) table.setValue(dec2000Name, i, acc.getDecJ2000());
                        }
                    else
                        {
                        if (addAirmass) table.setValue(airmassName, i, Double.NaN);
                        if (addAltitude) table.setValue(altitudeName, i, Double.NaN);
                        if (addAzimuth) table.setValue(azimuthName, i, Double.NaN);
                        if (addHourAngle) table.setValue(hourAngleName, i, Double.NaN);
                        if (addZenithDistance) table.setValue(zenithDistanceName, i, Double.NaN);   
                        if (addGJD) table.setValue(gjdName, i, Double.NaN);
                        if (addHJD) table.setValue(hjdName, i, Double.NaN);
                        if (addHJDCorr) table.setValue(hjdCorrName, i, Double.NaN);  
                        if (addBJD) table.setValue(bjdName, i, Double.NaN);
                        if (addBJDCorr) table.setValue(bjdCorrName, i, Double.NaN);
                        if (addRaNow) table.setValue(raNowName, i, Double.NaN);
                        if (addDecNow) table.setValue(decNowName, i, Double.NaN);
                        if (addRA2000) table.setValue(ra2000Name, i, Double.NaN);
                        if (addDec2000) table.setValue(dec2000Name, i, Double.NaN);                            
                        }
                    }
                
                    
                table.show();
                updatePlot();
                Thread t2 = new Thread()
                    {
                    public void run()
                        {
                        if (OKbutton != null)
                            {
                            OKbutton.setForeground(defaultOKForeground);
                            OKbutton.setText("Update Table");
                            OKbutton.paint(OKbutton.getGraphics());
                            }
                        }
                    };

                t2.start();
                Thread.yield();                  
                astroConverterUpdating = false;
                Prefs.set("plot2.JDColumn", JDColumn);                
                }});
       
        OKCancelPanel.add(OKbutton);
        
        JButton cancelbutton = new JButton("   Close    ");
        cancelbutton.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e)
                {
                addAstroDataFrameWasShowing = false;
                getAstroPanelValues();
                saveAstroPanelPrefs();
                if (acc != null) acc.saveAndClose();
                closeAddAstroDataFrame();
                }});
        OKCancelPanel.add(cancelbutton);        
           
        SpringUtil.makeCompactGrid (OKCancelPanel, 1, OKCancelPanel.getComponentCount(), 6,6,6,6);

        addAstroDataPanel.add (OKCancelPanel);                           

        SpringUtil.makeCompactGrid (addAstroDataPanel, addAstroDataPanel.getComponentCount()/2, 2, 6,6,6,6);

        addAstroDataFrame.add (addAstroDataScrollPane);
        addAstroDataFrame.pack();
        addAstroDataFrame.setResizable (true);

        if (rememberWindowLocations)
            {
            Dimension mainScreenSize = Toolkit.getDefaultToolkit().getScreenSize();
            if (!Prefs.isLocationOnScreen(new Point(addAstroDataFrameLocationX,addAstroDataFrameLocationY)))
                {
                addAstroDataFrameLocationX = mainFrame.getX()+100;
                addAstroDataFrameLocationY =  mainFrame.getY()+100;
                }
            addAstroDataFrame.setLocation(addAstroDataFrameLocationX, addAstroDataFrameLocationY);
            }
        addAstroDataFrame.setVisible (true);
        addAstroDataFrameWasShowing = true;
         
        String JDColName = (String) jdcolumnbox.getSelectedItem();
        int col = table.getColumnIndex(JDColName);
        int tableLength = table.getCounter();
        if (col == MeasurementTable.COLUMN_NOT_FOUND)
            {
//            IJ.beep();
//            IJ.showMessage("<html>Could not find table column '"+JDColumn+"'<br>"+
//                                 "Select a valid Date/Time column.</html>");
            }     
        else if (tableLength > 0)
            {
            double jd = table.getValueAsDouble(col, 0);
            if (JDColumn.contains("-2400000")) jd += 2400000;
            if (!Double.isNaN(jd))
                {
                int returnCode = 0;
                if (useGJD) 
                    {
                    acc.setTime(jd);
                    returnCode = acc.processManualCoordinates();
                    }
                else if (useHJD) 
                    {
                    returnCode = acc.setHJDTime(jd);
                    }
                else 
                    {
                    acc.setBJDTime(jd);
                    returnCode = acc.setBJDTime(jd);
                    }
                if (returnCode == 3)
                    {
                    IJ.beep();
                    IJ.showMessage("ERROR: Ohio State access failed when attempting to retrieve BJD(TDB).");
                    }
                } 
            }
     
        } 
    static void getAstroPanelValues()
        {
        addAirmass = airmassCB.isSelected();
        addAltitude = altitudeCB.isSelected();
        addAzimuth = azimuthCB.isSelected();
        addHourAngle = hourAngleCB.isSelected();
        addZenithDistance = zenithDistanceCB.isSelected();
        addGJD = gjdCB.isSelected();
        addHJD = hjdCB.isSelected();
        addHJDCorr = hjdCorrCB.isSelected();                
        addBJD = bjdCB.isSelected();
        addBJDCorr = bjdCorrCB.isSelected();
        addRaNow = raNowCB.isSelected();
        addDecNow = decNowCB.isSelected();
        addRA2000 = ra2000CB.isSelected();
        addDec2000 = decNowCB.isSelected();
        airmassName = airmassField.getText().trim();
        airmassName = airmassName.replace(' ', '_');
        altitudeName = altitudeField.getText().trim();
        altitudeName = altitudeName.replace(' ', '_');
        azimuthName = azimuthField.getText().trim();
        azimuthName = azimuthName.replace(' ', '_');
        hourAngleName = hourAngleField.getText().trim();
        hourAngleName = hourAngleName.replace(' ', '_');
        zenithDistanceName = zenithDistanceField.getText().trim();
        zenithDistanceName = zenithDistanceName.replace(' ', '_');
        gjdName = gjdField.getText().trim();
        gjdName = gjdName.replace(' ', '_');
        hjdName = hjdField.getText().trim();
        hjdName = hjdName.replace(' ', '_');
        hjdCorrName = hjdCorrField.getText().trim();
        hjdCorrName = hjdCorrName.replace(' ', '_');
        bjdName = bjdField.getText().trim();
        bjdName = bjdName.replace(' ', '_');
        bjdCorrName = bjdCorrField.getText().trim();
        bjdCorrName = bjdCorrName.replace(' ', '_');
        raNowName = raNowField.getText().trim();
        raNowName = raNowName.replace(' ', '_');
        decNowName = decNowField.getText().trim();
        decNowName = decNowName.replace(' ', '_');
        ra2000Name = ra2000Field.getText().trim();
        ra2000Name = ra2000Name.replace(' ', '_');
        dec2000Name = dec2000Field.getText().trim();
        dec2000Name = dec2000Name.replace(' ', '_');
        }
    
    public static void updateMPAstroConverter()
        {
        if (acc != null && addAstroDataFrame != null)
            {
            jdcolumnbox.setSelectedItem(jdcolumnbox.getSelectedItem());
            }
        }
    
    static int selectCurve()
        {
        GenericDialog gd;
        String[] curves = new String[maxCurves];
        int num = 0;
        for (int c = 0; c < maxCurves; c++)
            {
            if (plotY[c])
                {
                curves[num] = ""+(c+1);
                num++;
                }
            }
        if (num < 1)
            {
            gd = new GenericDialog ("No curves selected", mainFrame.getX()+100, mainFrame.getY()+100);
            gd.addMessage ("At least one curve must be enabled for plotting.");
            gd.hideCancelButton();
            gd.showDialog();
            return -1;               
            }
        curves = Arrays.copyOf(curves, num);
        gd = new GenericDialog ("Select plot curve", mainFrame.getX()+100, mainFrame.getY()+100);
        gd.addChoice("Add data to table from curve:", curves, curves[0]);
        gd.showDialog();
        if (gd.wasCanceled()) return -1; 
        String selection = gd.getNextChoice();
        return parseInteger(selection, 0) - 1;
        }
    
    static void addNewColumn(int c, boolean main)
        {
        GenericDialog gd;
        if (tableName.equalsIgnoreCase("No Table Selected"))
            {
            gd = new GenericDialog ("No Table Selected", (main?mainFrame.getX():subFrame.getX())+100, (main?mainFrame.getY():subFrame.getY())+100);
            gd.addMessage ("A table must be open and selected to add new columns.");
            gd.hideCancelButton();
            gd.showDialog();
            return;            
            }
        String xHeading = xlabel2[c];
        if (showXAxisAsPhase)
                {
                xHeading = "Phase"; 
                }
        else if (showXAxisAsDaysSinceTc)
                {
                xHeading = "DaysSinceTc"; 
                }         
        else if (showXAxisAsHoursSinceTc)
                {
                xHeading = "HoursSinceTc"; 
                }         
        
        String yHeading = "";
        String yErrHeading = "";
        if (operatorIndex[c] == 5)
            {
            yHeading = "Centroid_Dist_"+xyc1label[c]+"-"+xyc2label[c];
            if (normIndex[c] != 0)
                {
                yHeading += "(norm)";
                }
            else
                {
                yHeading += usePixelScale ? "(arcsecs)" : "(pixels)";
                }
            }
        else
            {
            yHeading=ylabel[c];
            yErrHeading=(errcolumn[c] < 0 || table.getColumnHeading(errcolumn[c])==null?"":table.getColumnHeading(errcolumn[c]));
            if (operatorIndex[c] != 0 && operatorIndex[c] != 6)
                {
                yHeading += opSymbol[operatorIndex[c]]+oplabel[c];
                yErrHeading += opSymbol[operatorIndex[c]]+(operrcolumn[c] < 0 || table.getColumnHeading(operrcolumn[c])==null?"":table.getColumnHeading(operrcolumn[c]));
                }
            if (detrendFitIndex[c] != 0)
                {
                boolean atLeastOne = false;
                for (int v = 0; v < maxDetrendVars; v++)
                    {
                    if (detrendIndex[c][v] != 0)
                        {
                        atLeastOne = true;
                        }
                    }
                if (atLeastOne)
                    {
                    yHeading += "_d";
                    yErrHeading += "_d";
                    }
                if (normIndex[c] != 0)
                    {
                    yHeading += "n";
                    yErrHeading += "n";                    
                    }
                }            
            else if (normIndex[c] != 0)
                {
                yHeading += "_n";
                yErrHeading += "_n";
                }
            }
        if (mmag[c])
            if (mmag[firstCurve] && totalScaleFactor[firstCurve] == 0.001)
                {
                yHeading += "(mmag)";
                yErrHeading += "(mmag)";
                }
            else if (totalScaleFactor[c] == 0.001)
                {
                yHeading += "(mmag)";
                yErrHeading += "(mmag)";
                }
            else
                {
                yHeading += "(mag)";
                yErrHeading += "(mag)";
                }
//        if (showLScaleInfo || showLShiftInfo)
//            yHeading += scaleShiftText(showLScaleInfo, showLShiftInfo, mmag[c], showLmmagInfo, totalScaleFactor[c], totalShiftFactor[c]);
        if ((binSize[c] != 1))
            {
            xHeading += "(binSize="+binSize[c]+")"; 
            yHeading += "(binSize="+binSize[c]+")";
            yErrHeading += "(binSize="+binSize[c]+")";
            }
        else if (showXAxisNormal)
            {
            xHeading += "_B";
            }

        gd = new GenericDialog ("Add new columns to table", (main?mainFrame.getX():subFrame.getX())+100, (main?mainFrame.getY():subFrame.getY())+100);
        gd.addCheckbox("Add new column from X-data", saveNewXColumn);
        gd.addStringField("New column name (from X-data): ", xHeading, 40);
        gd.addCheckbox("Add new column from Y-data", saveNewYColumn);
        gd.addStringField("New column name (from Y-data): ", yHeading, 40);
        if (showErr[c] || showOpErrors[c])
            {
            gd.addCheckbox("Add new column from Y-error data", saveNewYErrColumn);
            gd.addStringField("New column name (from Y-error data): ", yErrHeading, 40);            
            }

        gd.addMessage ("***New data column(s) will be added to the open table.***\n***Save table to save new column(s) to disk.***");
        gd.showDialog();
        if (gd.wasCanceled()) return; 
        
        saveNewXColumn = gd.getNextBoolean();
        xHeading = gd.getNextString();
        saveNewYColumn = gd.getNextBoolean();
        yHeading = gd.getNextString();     
        if (showErr[c] || showOpErrors[c])
            {
            saveNewYErrColumn = gd.getNextBoolean();
            yErrHeading = gd.getNextString();                 
            }        
        
        table.setPrecision(6);
        int xColumn = ResultsTable.COLUMN_IN_USE;
        int yColumn = ResultsTable.COLUMN_IN_USE; 
        int yErrColumn = ResultsTable.COLUMN_IN_USE; 
        if (saveNewXColumn) xColumn = table.getFreeColumn(xHeading);
        if (saveNewYColumn) yColumn = table.getFreeColumn(yHeading);
        
        boolean xUsed = saveNewXColumn && (xColumn == ResultsTable.COLUMN_IN_USE);
        boolean yUsed = saveNewYColumn && (yColumn == ResultsTable.COLUMN_IN_USE);
        
        boolean yErrUsed = false;
        if (showErr[c] || showOpErrors[c])
            {        
            if (saveNewYErrColumn) yErrColumn = table.getFreeColumn(yErrHeading);
            yErrUsed = saveNewYErrColumn && (yErrColumn == ResultsTable.COLUMN_IN_USE);
            }
        
        if (xUsed || yUsed || yErrUsed)
            {
            gd = new GenericDialog ("Over-write existing data?", (main?mainFrame.getX():subFrame.getX())+100, (main?mainFrame.getY():subFrame.getY())+100);
            gd.addMessage ((xUsed?"X-column: "+xHeading+"\n":"")+(yUsed?"Y-column: "+yHeading+"\n":"")+(yErrUsed?"Y-error column: "+yErrHeading+"\n":"")+
                           "name(s) are already in use.\nPress OK to over-write existing data.");
            gd.showDialog();
            if (gd.wasCanceled()) return;
            }
        int skipNum = excludedHeadSamples<nn[c]?excludedHeadSamples:nn[c]-1;
        if (binSize[c] != 1 || skipNum < 0) skipNum = 0;
        if (saveNewXColumn)
            {
            xColumn = table.getColumnIndex(xHeading);
            if (binSize[c] == 1 && skipNum > 0)
                {
                for (int i=0; i<skipNum; i++)
                    {
                    table.setValue(xColumn, i, Double.NaN);
                    }         
                }            
            if (showXAxisNormal && xlabel2[c].trim().startsWith("J.D.-2400000"))
                {
                for (int i=0; i<nn[c]; i++)
                    {
                    table.setValue(xColumn, i+skipNum, x[c][i]+xOffset-2400000.0);
                    }                    
                }
            else if (showXAxisNormal && xlabel2[c].contains("J.D.") || xlabel2[c].contains("JD"))
                {
                for (int i=0; i<nn[c]; i++)
                    {
                    table.setValue(xColumn, i+skipNum, x[c][i]+xOffset);
                    }
                }
            else
                {
                for (int i=0; i<nn[c]; i++)
                    {
                    table.setValue(xColumn, i+skipNum, x[c][i]);
                    }                
                }
            if (nn[c] < table.getCounter()-skipNum)
                {
                for (int i=nn[c]; i<table.getCounter()-skipNum; i++)
                    {
                    table.setValue(xColumn, i+skipNum, Double.NaN);
                    }         
                }
            }
        if (saveNewYColumn)
            {
            yColumn = table.getColumnIndex(yHeading);
            if (binSize[c] == 1 && skipNum > 0)
                {
                for (int i=0; i<skipNum; i++)
                    {
                    table.setValue(yColumn, i, Double.NaN);
                    }         
                }             
            if (ylabel[c].trim().startsWith("J.D.-2400000"))
                {
                for (int i=0; i<nn[c]; i++)
                    {
                    table.setValue(yColumn, i+skipNum, y[c][i]+yOffset-2400000.0);
                    }                    
                }            
            else if (ylabel[c].contains("J.D.") || ylabel[c].contains("JD"))
                {            
                for (int i=0; i<nn[c]; i++)
                    {
                    table.setValue(yColumn, i+skipNum, y[c][i]+yOffset);
                    }
                }
            else
                {
                for (int i=0; i<nn[c]; i++)
                    {
                    table.setValue(yColumn, i+skipNum, y[c][i]);
                    }                
                }
            if (nn[c] < table.getCounter()-skipNum)
                {
                for (int i=nn[c]; i<table.getCounter()-skipNum; i++)
                    {
                    table.setValue(yColumn, i+skipNum, Double.NaN);
                    }         
                }
            }  
        if ((showErr[c] || showOpErrors[c]) && saveNewYErrColumn)
            {
            yErrColumn = table.getColumnIndex(yErrHeading);
            if (binSize[c] == 1 && skipNum > 0)
                {
                for (int i=0; i<skipNum; i++)
                    {
                    table.setValue(yErrColumn, i, Double.NaN);
                    }         
                }             
            for (int i=0; i<nn[c]; i++)
                {
                table.setValue(yErrColumn, i+skipNum, yerr[c][i]);
                }                
            if (nn[c] < table.getCounter()-skipNum)
                {
                for (int i=nn[c]; i<table.getCounter()-skipNum; i++)
                    {
                    table.setValue(yErrColumn, i+skipNum, Double.NaN);
                    }         
                }
            }            
        table.show();
        }

	static void addElement(Vector<Roi> list, Shape shape, Color color, int lineWidth) {
		roi = new ShapeRoi(shape);
		roi.setInstanceColor(color);
		roi.setLineWidth(lineWidth);
		list.addElement(roi);
                }
    
//    static MouseListener panelMouseListener = new MouseListener() {
//        public void mouseClicked(MouseEvent e) 
//            {
//            panelShiftDown = e.isShiftDown();
//            panelAltDown = e.isAltDown();
//            panelControlDown = e.isControlDown();
//            }
//
//        public void mousePressed(MouseEvent e) 
//            {
//            panelShiftDown = e.isShiftDown();
//            panelAltDown = e.isAltDown();
//            panelControlDown = e.isControlDown();
//            }
//
//        public void mouseReleased(MouseEvent e) 
//            {
//            panelShiftDown = e.isShiftDown();
//            panelAltDown = e.isAltDown();
//            panelControlDown = e.isControlDown();            
//            }
//
//        public void mouseEntered(MouseEvent e) 
//            {
//            panelShiftDown = e.isShiftDown();
//            panelAltDown = e.isAltDown();
//            panelControlDown = e.isControlDown();
//            }
//
//        public void mouseExited(MouseEvent e) 
//            {
//            panelShiftDown = e.isShiftDown();
//            panelAltDown = e.isAltDown();
//            panelControlDown = e.isControlDown();          
//            }
//    };    
    
    static MouseMotionListener panelMouseMotionListener = new MouseMotionListener() {
        public void mouseDragged(MouseEvent e) 
            {
            panelShiftDown = e.isShiftDown();
            panelAltDown = e.isAltDown();
            panelControlDown = e.isControlDown(); 
            }
 
        public void mouseMoved(MouseEvent e)
            {
            panelShiftDown = e.isShiftDown();
            panelAltDown = e.isAltDown();
            panelControlDown = e.isControlDown();            
            }
    };
    
    

    static MouseListener plotMouseListener = new MouseListener() {
        public void mouseClicked(MouseEvent e) {

                // mouse clicked code is in mouseReleased() to allow drag/click thresholding
                }

        public void mousePressed(MouseEvent e) {
                startDragX = plotImageCanvas.offScreenX(e.getX());
                startDragY = plotImageCanvas.offScreenY(e.getY());
                startDragScreenX = e.getX();
                startDragScreenY = e.getY();
                startDragSubImageX =plotImageCanvas.getSrcRect().x;
                startDragSubImageY =plotImageCanvas.getSrcRect().y;
                button2Drag = false;
                if (e.isControlDown() && !e.isAltDown())  //handle control click to move trend lines
                    {
                    if ((e.getModifiers() & MouseEvent.BUTTON1_MASK) != 0)
                        handleControlLeftClickDrag(e, startDragX, startDragY);
                    else if ((e.getModifiers() & MouseEvent.BUTTON3_MASK) != 0)
                        handleControlRightClickDrag(e, startDragX, startDragY);
                    }
                }

        public void mouseReleased(MouseEvent e) {
                int screenX = e.getX();
                int screenY = e.getY();
                int imageX = plotImageCanvas.offScreenX(screenX);
                int imageY = plotImageCanvas.offScreenY(screenY);

                double xval = plotMinX + (double)(startDragX - Plot.LEFT_MARGIN) * (plotMaxX - plotMinX)/(double)(plotSizeX - (Plot.LEFT_MARGIN+Plot.RIGHT_MARGIN));
                double yval = plotMaxY - (double)(startDragY - Plot.TOP_MARGIN) * (plotMaxY - plotMinY)/(double)(plotSizeY - (Plot.TOP_MARGIN+Plot.BOTTOM_MARGIN));
                double dxval = (double)(imageX - startDragX) * (plotMaxX - plotMinX)/(double)(plotSizeX - (Plot.LEFT_MARGIN+Plot.RIGHT_MARGIN));
                double dyval = (double)(imageY - startDragY) * (-1.0) * (plotMaxY - plotMinY)/(double)(plotSizeY - (Plot.TOP_MARGIN+Plot.BOTTOM_MARGIN));


                if (((e.getModifiers() & MouseEvent.BUTTON2_MASK) != 0) && button2Drag && !e.isShiftDown() && !e.isControlDown() && !e.isAltDown())
                        {                                                          // measure distance and report in Results Table
                        button2Drag = false;
                        IJ.log("x="+fourPlaces.format(xval)+", y="+fourPlaces.format(yval)+
                                 ", dx="+fourPlaces.format(dxval)+", dy="+fourPlaces.format(dyval));
                        }
                if (Math.sqrt((screenX-startDragScreenX)*(screenX-startDragScreenX) +      //check mouse click/drag threshold
                        (screenY-startDragScreenY)*(screenY-startDragScreenY)) < 4.0)
                        {
                        if(e.getButton() == MouseEvent.BUTTON1)       //left mouse click release 
                            {                                                       
                            if (e.isShiftDown() && !e.isAltDown() && !e.isControlDown())  //shift+left-click (remove selected table row)
                                {
                                selectedRowStart=tpanel.getSelectionStart();
                                selectedRowEnd=tpanel.getSelectionEnd();
                                for (int i=selectedRowEnd; i>=selectedRowStart;i--)
                                    {
                                    deletedRowList.add(tpanel.getLine(i));
                                    }
                                tpanel.clearSelection();
                                table=MeasurementTable.getTable(tableName);
//                                if (Data_Processor.running) Data_Processor.setTable(table);
                                table.show();
                                tpanel=MeasurementTable.getTextPanel(MeasurementTable.longerName(tableName));
                                int lineCount = tpanel.getLineCount();
                                if (selectedRowEnd >= lineCount-excludedTailSamples)
                                    {
                                    selectedRowStart=binSize[firstCurve]*((lineCount-excludedHeadSamples-excludedTailSamples)/binSize[firstCurve])-binSize[firstCurve];
                                    selectedRowEnd=selectedRowStart+binSize[firstCurve]-1;
                                    }
                                tpanel.setSelection(selectedRowStart, selectedRowEnd);
                                boldedDatum=(selectedRowStart-excludedHeadSamples)/binSize[firstCurve];
                                plotcoordlabel.setText("DATA: x="+fourPlaces.format(x[firstCurve][boldedDatum])+", y="+fourPlaces.format(y[firstCurve][boldedDatum]));
                                IJ.showStatus("data values: x="+fourPlaces.format(x[firstCurve][boldedDatum])+", y="+fourPlaces.format(y[firstCurve][boldedDatum]));
                                updatePlot();
                                }
                            else if(!e.isControlDown() && !e.isShiftDown() && !e.isAltDown()) // left mouse click release (zoom in)
                                {
                                zoomControl(e.getX(), e.getY(), -5);
                                }
                            }
                        else if (e.getButton() == MouseEvent.BUTTON2)                     //middle mouse click release
                                {
                                IJ.log("x="+fourPlaces.format(xval)+", y="+fourPlaces.format(yval));
                                }

                        else if (e.getButton() == MouseEvent.BUTTON3 && e.isShiftDown() && !e.isControlDown() && !e.isAltDown() && !deletedRowList.isEmpty()) //shift + right mouse click release
                                {                                                                                     //undo delete selected table row
                                int lastRow = tpanel.getLineCount()-1;
                                 if (selectedRowEnd<=lastRow-binSize[firstCurve])
                                    {  //restore all rows other than the last one OR next to last one when when binze > 1
                                    String[] lines = new String[lastRow-selectedRowStart+1];
                                    for (int i=0; i<lines.length; i++)
                                        {
                                        lines[i] = tpanel.getLine(selectedRowStart+i);
                                        }
                                    tpanel.setSelection(selectedRowStart, lastRow);
                                    tpanel.clearSelection();
                                    for (int i=0; i<binSize[firstCurve];i++)
                                        {
                                        if (!deletedRowList.isEmpty()) tpanel.appendWithoutUpdate(deletedRowList.remove(deletedRowList.size()-1));
                                        }
                                    for (int i=0; i<lines.length; i++)
                                        {
                                        tpanel.appendWithoutUpdate(lines[i]);
                                        }
                                    }
                                else if (selectedRowEnd<lastRow)
                                    { //restore next to last row when binsize > 1
                                    selectedRowStart-=binSize[firstCurve];
                                    selectedRowEnd-=binSize[firstCurve];
                                    String[] lines = new String[lastRow-selectedRowEnd];
                                    for (int i=0; i<lines.length; i++)
                                        {
                                        lines[i] = tpanel.getLine(selectedRowEnd+1+i);
                                        }
                                    tpanel.setSelection(selectedRowEnd+1, lastRow);
                                    tpanel.clearSelection();
                                    for (int i=0; i<binSize[firstCurve];i++)
                                        {
                                        if (!deletedRowList.isEmpty()) tpanel.appendWithoutUpdate(deletedRowList.remove(deletedRowList.size()-1));
                                        }
                                    for (int i=0; i<lines.length; i++)
                                        {
                                        tpanel.appendWithoutUpdate(lines[i]);
                                        }
                                    selectedRowStart+=binSize[firstCurve];
                                    selectedRowEnd+=binSize[firstCurve];
                                    
                                    }
                                else
                                    { //restore last row in table
                                    for (int i=0; i<binSize[firstCurve];i++)
                                        {
                                        if (!deletedRowList.isEmpty()) tpanel.appendWithoutUpdate(deletedRowList.remove(deletedRowList.size()-1));
                                        };
                                    selectedRowStart+=binSize[firstCurve];
                                    }
                                table=MeasurementTable.getTable(tableName);
                                table.show();
                                tpanel=MeasurementTable.getTextPanel(MeasurementTable.longerName(tableName));
                                selectedRowEnd=selectedRowStart+binSize[firstCurve]-1;
                                tpanel.setSelection(selectedRowStart, selectedRowEnd);
                                boldedDatum=(selectedRowStart-excludedHeadSamples)/binSize[firstCurve];
                                plotcoordlabel.setText("DATA: x="+fourPlaces.format(x[firstCurve][boldedDatum])+", y="+fourPlaces.format(y[firstCurve][boldedDatum]));
                                IJ.showStatus("data values: x="+fourPlaces.format(x[firstCurve][boldedDatum])+", y="+fourPlaces.format(y[firstCurve][boldedDatum]));
                                updatePlot();
                                }
                        else if (e.getButton() == MouseEvent.BUTTON3)                     //right mouse click release
                                {
                                totalPanOffsetX=0.0;
                                totalPanOffsetY=0.0;
                                newPanOffsetX=0.0;
                                newPanOffsetY=0.0;
                                zoomControl(screenX, screenY, 10000);
                                }
                        }
                else                                                                      //complete drag operations
                    {
                    if (e.getButton() == MouseEvent.BUTTON1 && !e.isShiftDown() && !e.isControlDown() && !e.isAltDown())                             //left mouse drag release
                        {
                        totalPanOffsetX += newPanOffsetX;
                        totalPanOffsetY += newPanOffsetY;
                        newPanOffsetX = 0.0;
                        newPanOffsetY = 0.0;
                        leftDragReleased = true;
                        }
                    }
                list.clear();
                path.reset();
                plotImageCanvas.repaint();
                }

        public void mouseEntered(MouseEvent e) {}

        public void mouseExited(MouseEvent e) {}
    };


    static MouseMotionListener plotMouseMotionListener = new MouseMotionListener() {
        public void mouseDragged(MouseEvent e) {
                int screenX = e.getX();
                int screenY = e.getY();
                int imageX = plotImageCanvas.offScreenX(e.getX());
                int imageY = plotImageCanvas.offScreenY(e.getY());
                if (Math.sqrt((screenX-startDragScreenX)*(screenX-startDragScreenX) +  //check mouse click/drag threshold
                        (screenY-startDragScreenY)*(screenY-startDragScreenY)) >= 4.0)
                        {
                        if ((e.getModifiers() & MouseEvent.BUTTON3_MASK) != 0)    // dragging with right mouse button
                            {
                            if (!e.isControlDown() && !e.isShiftDown() && !e.isAltDown())
                                {                                                 // measure distance, show on plot and status bar and save to results window
                                list.clear();
                                path.reset();
                                path.moveTo(startDragX, startDragY);
                                path.lineTo(imageX, imageY);

                                addElement(list, path, Color.red, 2);
                                plotImageCanvas.setDisplayList(list);
                                double x = plotMinX + (double)(imageX - (Plot.LEFT_MARGIN)) * (plotMaxX - plotMinX)/(double)(plotSizeX - (Plot.LEFT_MARGIN+Plot.RIGHT_MARGIN));
                                double y = plotMaxY - (double)(imageY - (Plot.TOP_MARGIN)) * (plotMaxY - plotMinY)/(double)(plotSizeY - (Plot.TOP_MARGIN+Plot.BOTTOM_MARGIN));
                                double dx = (double)(imageX - startDragX) * (plotMaxX - plotMinX)/(double)(plotSizeX - (Plot.LEFT_MARGIN+Plot.RIGHT_MARGIN));
                                double dy = (double)(imageY - startDragY) * (-1.0) * (plotMaxY - plotMinY)/(double)(plotSizeY - (Plot.TOP_MARGIN+Plot.BOTTOM_MARGIN));
                                plotbottompanel = (Panel) plotWindow.getComponent(1);
                                plotbottompanel.setSize(600,30);
                                plotcoordlabel = (Label) plotbottompanel.getComponent(4);
                                plotcoordlabel.setSize(400,20);
                                plotcoordlabel.setText("x="+fourPlaces.format(x)+", y="+fourPlaces.format(y)+
                                                       ", dx="+fourPlaces.format(dx)+", dy="+fourPlaces.format(dy));
                                IJ.showStatus("plot coordinates: x="+fourPlaces.format(x)+", y="+fourPlaces.format(y)+
                                              ", dx="+fourPlaces.format(dx)+", dy="+fourPlaces.format(dy));
                                }
                           else if (e.isControlDown() && !e.isAltDown()) // control right-drag (update vertical marker 3 or 4 position)
                                {
                                handleControlRightClickDrag(e, imageX, imageY);
                                }                             
                            }
                        else if ((e.getModifiers() & MouseEvent.BUTTON2_MASK) != 0)        // dragging with middle mouse button
                                {                                                          // measure distance, show on plot and status bar
                                button2Drag = true;                                        // and save to results window when mouse released
                                list.clear();
                                path.reset();
                                path.moveTo(startDragX, startDragY);
                                path.lineTo(imageX, imageY);

                                addElement(list, path, Color.red, 2);
                                plotImageCanvas.setDisplayList(list);
                                double x = plotMinX + (double)(imageX - (Plot.LEFT_MARGIN)) * (plotMaxX - plotMinX)/(double)(plotSizeX - (Plot.LEFT_MARGIN+Plot.RIGHT_MARGIN));
                                double y = plotMaxY - (double)(imageY - (Plot.TOP_MARGIN)) * (plotMaxY - plotMinY)/(double)(plotSizeY - (Plot.TOP_MARGIN+Plot.BOTTOM_MARGIN));
                                double dx = (double)(imageX - startDragX) * (plotMaxX - plotMinX)/(double)(plotSizeX - (Plot.LEFT_MARGIN+Plot.RIGHT_MARGIN));
                                double dy = (double)(imageY - startDragY) * (-1.0) * (plotMaxY - plotMinY)/(double)(plotSizeY - (Plot.TOP_MARGIN+Plot.BOTTOM_MARGIN));
                                plotcoordlabel.setText("x="+fourPlaces.format(x)+", y="+fourPlaces.format(y)+
                                                       ", dx="+fourPlaces.format(dx)+", dy="+fourPlaces.format(dy));
                                IJ.showStatus("plot coordinates: x="+fourPlaces.format(x)+", y="+fourPlaces.format(y)+
                                              ", dx="+fourPlaces.format(dx)+", dy="+fourPlaces.format(dy));
                                }
                        else if ((e.getModifiers() & MouseEvent.BUTTON1_MASK) != 0)            // dragging with left mouse button (pan image)
                                {
                                if (!e.isControlDown() && !e.isAltDown())
                                    {
//                                    if (true)//(zoom != 0.0)
//                                        {
                                    newPanOffsetX=-(plotMaxX - plotMinX)*(screenX-startDragScreenX)/(plotSizeX-(Plot.LEFT_MARGIN+Plot.RIGHT_MARGIN));
                                    newPanOffsetY=(plotMaxY - plotMinY)*(screenY-startDragScreenY)/(plotSizeY-(Plot.TOP_MARGIN+Plot.BOTTOM_MARGIN));
                                    updatePlot();
//                                        }
//                                    else
//                                        {
//                                        totalPanOffsetX=0.0;
//                                        totalPanOffsetY=0.0;
//                                        newPanOffsetX=0.0;
//                                        newPanOffsetY=0.0;
//                                        leftDragReleased=false;
//                                        }
                                    if (e.isShiftDown())
                                        {
                                        plotcoordlabel.setText("DATA: x="+fourPlaces.format(x[firstCurve][boldedDatum])+", y="+fourPlaces.format(y[firstCurve][boldedDatum]));
                                        IJ.showStatus("data values: x="+fourPlaces.format(x[firstCurve][boldedDatum])+", y="+fourPlaces.format(y[firstCurve][boldedDatum]));
                                        }
                                    else
                                        {
                                        double x = plotMinX + (double)(imageX - (Plot.LEFT_MARGIN)) * (plotMaxX - plotMinX)/(double)(plotSizeX - (Plot.LEFT_MARGIN+Plot.RIGHT_MARGIN));
                                        double y = plotMaxY - (double)(imageY - (Plot.TOP_MARGIN)) * (plotMaxY - plotMinY)/(double)(plotSizeY - (Plot.TOP_MARGIN+Plot.BOTTOM_MARGIN));  //was 58
                                        plotcoordlabel.setText("x="+fourPlaces.format(x)+", y="+fourPlaces.format(y));
                                        IJ.showStatus("plot coordinates: x="+fourPlaces.format(x)+", y="+fourPlaces.format(y));
                                        }
                                    }
                                else if (e.isControlDown() && !e.isAltDown()) // control left-drag (update left detrend position)
                                    {
                                    handleControlLeftClickDrag(e, imageX, imageY);   
                                    }                                
                                }
                        }
                }

 
        public void mouseMoved(MouseEvent e) {
                
                int imageX = plotImageCanvas.offScreenX(e.getX());
                int imageY = plotImageCanvas.offScreenY(e.getY());
                double xval = plotMinX + (double)(imageX - (Plot.LEFT_MARGIN)) * (plotMaxX - plotMinX)/(double)(plotSizeX - (Plot.LEFT_MARGIN+Plot.RIGHT_MARGIN));
                double yval = plotMaxY - (double)(imageY - (Plot.TOP_MARGIN)) * (plotMaxY - plotMinY)/(double)(plotSizeY - (Plot.TOP_MARGIN+Plot.BOTTOM_MARGIN));


                if (e.isShiftDown() && !e.isControlDown() && !e.isAltDown())     //select table line nearest x-val of first selected curve
                    {
                    nearestLine=0;
                    nearestX = Math.abs(xval-x[firstCurve][0]);
                    if (nn[firstCurve]>0)
                        {
                        for (int i=1; i<nn[firstCurve];i++)
                            {
                            if (Math.abs(xval-x[firstCurve][i]) < nearestX)
                                {
                                nearestX = Math.abs(xval-x[firstCurve][i]);
                                nearestLine = i;
                                }
                            }
                        }
                    tpanel.setSelection(excludedHeadSamples+nearestLine*binSize[firstCurve], excludedHeadSamples+(nearestLine+1)*binSize[firstCurve]-1);
                    boldedDatum=nearestLine;
                    plotcoordlabel.setText("DATA: x="+fourPlaces.format(x[firstCurve][nearestLine])+", y="+fourPlaces.format(y[firstCurve][nearestLine]));
                    IJ.showStatus("data values: x="+fourPlaces.format(x[firstCurve][nearestLine])+", y="+fourPlaces.format(y[firstCurve][nearestLine]));
                    if (useBoldedDatum) updatePlot();
                    updateMPAstroConverter();
                    if (useUpdateStack) updateStack();
                    }
                else
                    {
                    plotcoordlabel.setText("x="+fourPlaces.format(xval)+", y="+fourPlaces.format(yval));
                    IJ.showStatus("plot coordinates: x="+fourPlaces.format(xval)+", y="+fourPlaces.format(yval));
                    if (boldedDatum!=-1)
                        {
                        boldedDatum = -1;
                        if (useBoldedDatum)updatePlot();
                        }
                    }
                }
    };
        
        static public void handleControlLeftClickDrag(MouseEvent e, int imageX, int imageY)
            {
            double x = plotMinX + (double)(imageX - (Plot.LEFT_MARGIN)) * (plotMaxX - plotMinX)/(double)(plotSizeX - (Plot.LEFT_MARGIN+Plot.RIGHT_MARGIN));
            double y = plotMaxY - (double)(imageY - (Plot.TOP_MARGIN)) * (plotMaxY - plotMinY)/(double)(plotSizeY - (Plot.TOP_MARGIN+Plot.BOTTOM_MARGIN));
            double delta = 0.025*(plotMaxX - plotMinX);
            boolean alreadyMoved = false;
            if (showDMarkers && x > dMarker2Value - delta && x < dMarker2Value + delta)
                {
                dmarker2spinner.setValue(x);
                alreadyMoved = true;
                }
            if (showDMarkers && x > dMarker3Value - delta && x < dMarker3Value + delta)
                {
                dmarker3spinner.setValue(x);
                alreadyMoved = true;
                }
            if (showDMarkers && x > dMarker1Value - delta && x < dMarker1Value + delta)
                {
                dmarker1spinner.setValue(x);
                alreadyMoved = true;
                }               
            if (showDMarkers && x > dMarker4Value - delta && x < dMarker4Value + delta)
                {
                dmarker4spinner.setValue(x);
                alreadyMoved = true;
                }
            if (!alreadyMoved)
                {
                showDMarkersCB.setSelected(true);
                showDMarkers = true;
                if (e.isShiftDown())   //update vertical marker 1 position
                    {
                    useDMarker1CB.setSelected(true);
                    useDMarker1 = true;
                    dmarker1spinner.setValue(x);
                    }
                else                   //update vertical marker 2 position
                    {
                    dmarker2spinner.setValue(x);
                    }
                }
            plotcoordlabel.setText("x="+fourPlaces.format(x)+", y="+fourPlaces.format(y));
            IJ.showStatus("plot coordinates: x="+fourPlaces.format(x)+", y="+fourPlaces.format(y));             
            }
        
        static public void handleControlRightClickDrag(MouseEvent e, int imageX, int imageY)
            {        
            double x = plotMinX + (double)(imageX - (Plot.LEFT_MARGIN)) * (plotMaxX - plotMinX)/(double)(plotSizeX - (Plot.LEFT_MARGIN+Plot.RIGHT_MARGIN));
            double y = plotMaxY - (double)(imageY - (Plot.TOP_MARGIN)) * (plotMaxY - plotMinY)/(double)(plotSizeY - (Plot.TOP_MARGIN+Plot.BOTTOM_MARGIN));
            showDMarkersCB.setSelected(true);
            showDMarkers = true;
            if (e.isShiftDown()) //update vertical marker 4 position
                {
                useDMarker4CB.setSelected(true);
                useDMarker4 = true;                                    
                dmarker4spinner.setValue(x);

                }
            else                 //update vertical marker 3 position
                {
                dmarker3spinner.setValue(x);
                }
            plotcoordlabel.setText("x="+fourPlaces.format(x)+", y="+fourPlaces.format(y));
            IJ.showStatus("plot coordinates: x="+fourPlaces.format(x)+", y="+fourPlaces.format(y));   
            }
        
    public static void updateStack()
        {
        ImageWindow iw = null;
        ImageStack imstack = null;
        iplus = IJ.getImage();
        sliceCol = table.getColumnIndex("slice");
        
        int row1Slice = 0;
        if (sliceCol != MeasurementTable.COLUMN_NOT_FOUND) 
            row1Slice = (int)table.getValueAsDouble(sliceCol, 0);
        if (iplus != null)
            {
            iw = iplus.getWindow();
            imstack = iplus.getStack();
            if (iw != null && imstack != null && imstack.getSize() > 1 &&
               ((row1Slice > 0 && row1Slice <= imstack.getSize()) || imstack.getSize() == table.getCounter()))
                {
//                WindowManager.toFront(iw);
                updateSlice();
                return;
                }
            }
        if (WindowManager.getWindowCount()> 0) 
            {
            int[] ID = WindowManager.getIDList();
            for (int win = 0; win < ID.length; win++)
                {
                iplus = WindowManager.getImage(ID[win]);
                if (iplus != null) 
                    {
                    iw = iplus.getWindow();
                    imstack = iplus.getStack();
                    }
                else
                    iw = null;
                if (iw != null && imstack != null && imstack.getSize() > 1 &&
                   ((row1Slice > 0 && row1Slice <= imstack.getSize()) || imstack.getSize() == table.getCounter()))
                    {
                    WindowManager.setCurrentWindow(iw);
                    WindowManager.toFront(iw);
                    updateSlice();
                    return;
                    }
                }
            }
        }
    
    static void updateSlice()
        {
        Thread t = new Thread()
            {
            public void run()
                {
                Thread.yield();
                ImageStack imstack = iplus.getStack();
                selectedRow = tpanel.getSelectionStart();
                if (selectedRow < 0) selectedRow = 0;
                else if (selectedRow >= imstack.getSize()) selectedRow = imstack.getSize() - 1;
                int newSlice = 0;
                if (sliceCol != MeasurementTable.COLUMN_NOT_FOUND) newSlice = (int)table.getValueAsDouble(sliceCol, selectedRow);
                if (newSlice == 0) newSlice = selectedRow + 1;
                if (newSlice > imstack.getSize()) newSlice = imstack.getSize();
                iplus.setSlice(newSlice);
//                iplus.updateAndDraw();
                }
            };
        t.start();
        }
    
    
    static MouseWheelListener plotMouseWheelListener = new MouseWheelListener() {
        public void mouseWheelMoved( MouseWheelEvent e )
                {
                int screenX = e.getX();
                int screenY = e.getY();
                int magChangeSteps = e.getWheelRotation();
                zoomControl(screenX, screenY, magChangeSteps);
                }
    };

        static public void zoomControl(int screenX, int screenY, int magChangeSteps)
                {
                if (screenX > Plot.LEFT_MARGIN && screenX <= plotSizeX - Plot.RIGHT_MARGIN &&
                    screenY >= Plot.TOP_MARGIN && screenY <= plotSizeY - Plot.BOTTOM_MARGIN)
                    {
                    if (zoom == 0.0)
                        {
                        mouseX = screenX - Plot.LEFT_MARGIN;
                        mouseY = screenY - Plot.TOP_MARGIN;
//                        totalPanOffsetX=0.0;
//                        totalPanOffsetY=0.0;
//                        newPanOffsetX=0.0;
//                        newPanOffsetY=0.0;
                        }
                    zoom -= (1-zoom)*magChangeSteps/25.0;
                    zoom = zoom > 0.99 ? 0.99 : zoom;
                    zoom = zoom < 0.0 ? 0.0 : zoom;
                    updatePlot();
                    }
                }

        static public void zoomIn(int x, int y) {
                double magnification = plotImageCanvas.getMagnification();
                if (magnification>=32) return;
                double newMag = plotImageCanvas.getHigherZoomLevel(magnification);

                        int w = (int)Math.round(plotSizeX/newMag);
                if (w*newMag<plotSizeX) w++;
                int h = (int)Math.round(plotSizeY/newMag);
                if (h*newMag<plotSizeY) h++;
                x = plotImageCanvas.offScreenX(x);
                y = plotImageCanvas.offScreenY(y);
                Rectangle r = new Rectangle(x-w/2, y-h/2, w, h);
                if (r.x<0) r.x = 0;
                if (r.y<0) r.y = 0;
                if (r.x+w>plotSizeX) r.x = plotSizeX-w;
                if (r.y+h>plotSizeY) r.y = plotSizeY-h;
                plotImageCanvas.setSourceRect(r);
                plotImageCanvas.setMagnification(newMag);
                plotImageCanvas.repaint();
                }

static void initializeVariables()
        {
        title = "title";
        titlePosX = .5;
        titlePosY = 0.06;
        subtitle ="subtitle";
        subtitlePosX = 0.5;
        subtitlePosY = 0.1;
        legendPosX = 0.8;
        legendPosY = 0.2;
        button2Drag = false;
        priorityColumns = "j.d.,jd,hjd,bjd,ratio,rel_flux,source-sky,tot_c_cnts,alt,air,ccd,saturated,source_radius,sky_radius,x(fits),y(fits),peak,sky/pixel,x(ij),y(ij)";
        priorityDetrendColumns = "airmass,j.d.,jd,ccd-temp,x(fits)_t1,x(fits)_c1,y(fits)_t1,y(fits)_c1,sky/pixel_t1,width_t1,peak_C2,tot_c_cnts,tot_c_err,x(ij),y(ij)";
        airmassName = "AIRMASS";
        altitudeName = "ALTITUDE";
        azimuthName = "AZIMUTH";
        bjdName = "BJD_TDB_MOBS";
        bjdCorrName = "BJD_CORR";
        decNowName = "DEC_EOD";
        raNowName = "RA_EOD";
        dec2000Name = "DEC_J2000";
        ra2000Name = "RA_J2000";
        gjdName = "JD_UTC_MOBS";
        hjdName = "HJD_UTC_MOBS";
        hjdCorrName = "HJD_CORR";
        hourAngleName = "HOUR_ANGLE";
        zenithDistanceName = "ZENITH_DIST";
        addAirmass = true;
        addAltitude = true;
        addAzimuth = false;
        addBJD = true;
        addBJDCorr = false; 
        addDecNow = false;
        addRaNow = false;
        addDec2000 = false;
        addRA2000 = false;
        addHJD = true;
        addGJD = false;
        addHJDCorr = false;
        addHourAngle = false;
        addZenithDistance = false;        
        prioritizeColumns = true;
        updatePlotEnabled = true;
        disableUpdatePlotBox = false;
        astroConverterUpdating = false;
        xlabeldefault = "J.D.-2400000";
        plotSizeX = 800;
        plotSizeY = 800;
        xExponent = 0;
        yExponent = 0;
        mmagrefs = 10;
        xLegend = "";
        yLegend = "";
        invertYAxis = false;
        invertYAxisSign = 1;
        negateMag = true;
        saveNewXColumn = true;
        saveNewYColumn = true;
        saveNewYErrColumn = true;
        selectAnotherTableCanceled = false;
        vMarker1TopText = "Predicted";
        vMarker1BotText = "Ingress";
        vMarker2TopText = "Predicted";
        vMarker2BotText = "Egress";
        saveImage = true;
        savePlot = true;
        saveConfig = true;
        saveTable = true;
        saveApertures = true;
        saveLog = true;
        saveAllPNG = true;          
        imageSuffix = "_field";
        aperSuffix = "_measurements";
        logSuffix = "_calibration"; 
        plotSuffix = "_lightcurve";
        configSuffix = "_measurements";
        dataSuffix = "_measurements";
        appendDestinationSuffix = "_A";
        appendSourceSuffix = "_B";
        templateDir = "";
        JDColumn = "JD_MOBS";
        combinedTableName = "Measurements";
        legendLeft = false;
        legendRight = false;
        useTitle = false;
        useSubtitle = false;
        useXCustomName = false;
        useYCustomName = false;
        useXColumnName = true;
        useYColumnName = true;
        xTics = true;
        yTics = true;
        xGrid = false;
        yGrid = true;
        xNumbers = true;
        yNumbers = true;
        plotAutoMode = false;
        nextPanel = false;
        canceled = false;
        waitSecs = false;
        openDataSetWindow = true;
        openRefStarWindow = true;
        rememberWindowLocations = true;
        divideNotSubtract = true;
        autoScaleX = true;
        useFirstX = false;
        autoScaleY = true;
        showToolTips = true;
        showXAxisNormal = true;
        showXAxisAsPhase = false;
        showXAxisAsHoursSinceTc = false;
        showXAxisAsDaysSinceTc = false;
        T0 = 0;
        period = 1;
        showXScaleInfo = true;
        showYScaleInfo = true;
        showYNormInfo = true;
        showYShiftInfo = true;
        showLScaleInfo = true;
        showLRelScaleInfo = false;
        showLShiftInfo = true;
        showLRelShiftInfo = false;
        showYBinInfo = true;
        showLBinInfo = true;
        showYmmagInfo = true;
        showLmmagInfo = true;
        showLdetrendInfo = true;
        showLnormInfo = true;
        showYSymbolInfo = false;
        showLSymbolInfo = true;
        useTwoLineLegend = false;
        useWideDataPanel = true;
        subframeWasShowing = true;
        newPlotWindow = false;
        useDefaultSettings = false;
        tableHasText = false;
        setSubpanelVisible = false;
        refStarPanelWasShowing = false;
        addAstroDataFrameWasShowing = false;
        usePixelScale = true;
        deletedRowList = new ArrayList<String>();
        useBoldedDatum = true;
        useUpdateStack = false;
        multiUpdate=false;
        selectedRow = -1;
        panelShiftDown = false;
        panelControlDown = false;
        panelAltDown = false;     
        keepFileNamesOnAppend = true;
        updatePlotRunning = false;
        boldedDatum=-1;
        pixelScale = 1.0;

        plotOptions = 0;
        xMin = 0.0;
        xMax = 0.0;
        xWidth = 0.3;
        yMin = 0.0;
        yMax = 0.0;
        xautoscalemin = 0.0;
        xautoscalemax = 0.0;
        yautoscalemin = 0.0;
        yautoscalemax = 0.0;
        yRange = 0.0;
        zoom = 0.0;

        mainFrameLocationX = 40;
        mainFrameLocationY = 40;
        subFrameLocationX = 40;
        subFrameLocationY = 40;
        refStarFrameLocationX = 40;
        refStarFrameLocationY = 40;     
        addAstroDataFrameLocationX = -100000;  //force offscreen
        addAstroDataFrameLocationY = -100000;  //force offscreen      
        plotFrameLocationX = 40;
        plotFrameLocationY = 40;

        excludedHeadSamples = 0;
        excludedTailSamples = 0;
        holdExcludedHeadSamples = 0;
        holdExcludedTailSamples = 0;
        excludedHeadSamplesStep = 1;
        excludedTailSamplesStep = 1;
        refStarHorzWidth = 30;
        lastRefStar = 1;
        cycleEnabledStarsLess1PressedConsecutive = false;

        xWidthStep = 0.01;
        xMaxStep = 0.01;
        xMinStep = 1.0;
        yMaxStep = 1.0;
        yMinStep = 1.0;
        yMid = 0.0;
        firstXmin = 0.0;
        plotSizeXStep = 10;
        plotSizeYStep = 10;
        vMarkerStep = 0.01;
        dMarkerStep = 0.01;

        mainmenubar = new JMenuBar();
        filemenu = new JMenu("File");
        preferencesmenu = new JMenu("Preferences");

        tableName = "Results";
        list = new Vector<Roi>();
        path = new GeneralPath();

        fourPlaces = new DecimalFormat("0.0000");
        
        delayedUpdateTask = null;
        delayedUpdateTimer = null;        

        markers = new String[]
                {
                "box",
                "circle",
                "cross",
                "dot",
                "line",
                "triangle",
                "X"
                };
        colors = new String[]
                {
                "black",
                "dark gray",
                "gray",
                "light gray",
                "green",
                "dark green",
                "light blue",
                "blue",
                "magenta",
                "pink",
                "red",
                "orange",
                "yellow",
                "brown",
                "purple",
                "teal"
                };
        
        normiconlist = new ImageIcon[]
                {
                createImageIcon("astroj/images/norm_off.png", "Do not use normalization"),                                // 0
                createImageIcon("astroj/images/norm_left.png", "Normalize based on data left of vertical marker 1"),      // 1
                createImageIcon("astroj/images/norm_right.png", "Normalize based on data right of vertical marker 2"),    // 2
                createImageIcon("astroj/images/norm_outside.png", "Normalize based on data outside of vertical markers"), // 3
                createImageIcon("astroj/images/norm_inside.png", "Normalize based on data inside of vertical markers"),   // 4
                createImageIcon("astroj/images/norm_left2.png", "Normalize based on data left of vertical marker 2"),     // 5
                createImageIcon("astroj/images/norm_right2.png", "Normalize based on data right of vertical marker 1"),   // 6
                createImageIcon("astroj/images/norm_all.png", "Normalize based on all data"),                             // 7
                };
        
        detrendiconlist = new ImageIcon[]
                {
                createImageIcon("astroj/images/norm_off.png", "Do not use detrending"),                                 // 0
                createImageIcon("astroj/images/norm_user.png", "User defined detrending constant"),                     // 1
                createImageIcon("astroj/images/norm_left.png", "Detrend based on data left of vertical marker 1"),      // 2
                createImageIcon("astroj/images/norm_right.png", "Detrend based on data right of vertical marker 2"),    // 3
                createImageIcon("astroj/images/norm_outside.png", "Detrend based on data outside of vertical markers"), // 4
                createImageIcon("astroj/images/norm_inside.png", "Detrend based on data inside of vertical markers"),   // 5
                createImageIcon("astroj/images/norm_left2.png", "Detrend based on data left of vertical marker 2"),     // 6
                createImageIcon("astroj/images/norm_right2.png", "Detrend based on data right of vertical marker 1"),   // 7
                createImageIcon("astroj/images/norm_all.png", "Detrend based on all data"),                             // 8
                };        

        operators = new String[]
                {
                "none",
                "divide by",
                "multiply by",
                "subtract",
                "add",
                "centroid distance",
                "custom error"
                };

        opSymbol = new String[]
                {
                "",
                " / ",
                " * ",
                " - ",
                " + ",
                " -> ",
                " (with error) "
                };

        spinnerscalelist = new String[]
                {
                "      0.0000000001",
                "        0.000000001",
                "          0.00000001",
                "            0.0000001",
                "              0.000001",
                "                0.00001",
                "                  0.0001",
                "                    0.001",
                "                    0.010",
                "                    0.100",
                "                    0.500",
                "                    1.000",
                "                  10.000",
                "                100.000",
                "              1000.000",
                "            10000.000",
                "          100000.000",
                "        1000000.000",
                "      10000000.000",
                "    100000000.000",
                "1000000000.000",
                };

        integerspinnerscalelist = new String[]
                {
                "                    1",
                "                    2",
                "                    3",
                "                    4",
                "                    5",
                "                    6",
                "                    7",
                "                    8",
                "                    9",
                "                  10",
                "                  25",
                "                  50",
                "                100",
                "                250",
                "                500",
                "              1000"
                };
            }

    static public void setupArrays()
                {
                maxCurves = (int)Prefs.get("plot.maxCurves", 8);
                maxDetrendVars = (int)Prefs.get("plot.maxDetrendVars", 3);
                if (maxCurves < 1) maxCurves = 1;
                if (maxDetrendVars < 1) maxDetrendVars = 1;
                nn = new int[maxCurves];
                nnr = new int[maxCurves];
                xlabel2 = new String[maxCurves];
                detrendlabel = new String[maxCurves][maxDetrendVars];
                detrendlabelhold = new String[maxCurves][maxDetrendVars];
                detrendcolumn = new int[maxCurves][maxDetrendVars];
                
                setupDataBuffers();
                
                detrendVarDisplayed = new int[maxCurves];
                xlabel = new String[maxCurves];
                ylabel = new String[maxCurves];
                xcolumn = new int[maxCurves];
                ycolumn = new int[maxCurves];
                errcolumn = new int[maxCurves];
                operrcolumn = new int[maxCurves];
                xc1column = new int[maxCurves];
                xc2column = new int[maxCurves];
                yc1column = new int[maxCurves];
                yc2column = new int[maxCurves];
                oplabel = new String[maxCurves];
                opcolumn = new int[maxCurves];
//                errlabel = new String[maxCurves];
//                operrlabel = new String[maxCurves];
                showErrors = new boolean[maxCurves];
                showErr = new boolean[maxCurves];
                showOpErrors = new boolean[maxCurves];
                lines = new boolean[maxCurves];
                marker = new int[maxCurves];
                color = new Color[maxCurves];
                markerIndex = new int[maxCurves];
                colorIndex = new int[maxCurves];
                operatorIndex = new int[maxCurves];
                cblabels = new String[maxCurves];
                moreOptions = new boolean [maxCurves];
                autoScaleFactor = new double[maxCurves];
                autoScaleStep = new double[maxCurves];
                autoShiftFactor = new double[maxCurves];
                autoShiftStep = new double[maxCurves];
                customScaleFactor = new double[maxCurves];
                customScaleStep = new double[maxCurves];
                customShiftFactor = new double[maxCurves];
                customShiftStep = new double[maxCurves];
                detrendFactor = new double[maxCurves][maxDetrendVars];
                detrendFactorStep = new double[maxCurves][maxDetrendVars];                
                baseline = new double[maxCurves];
                subtotalScaleFactor = new double[maxCurves];
                subtotalShiftFactor = new double[maxCurves];
                totalScaleFactor = new double[maxCurves];
                totalShiftFactor = new double[maxCurves];
                yWidth = new double[maxCurves];
                xMinimum = new double[maxCurves];
                xMaximum = new double[maxCurves];
                yMinimum = new double[maxCurves];
                yMn = new double[maxCurves];
                yMaximum = new double[maxCurves];
                yMx = new double[maxCurves];
                yMidpoint = new double[maxCurves];
                mmag = new boolean[maxCurves];
                fromMag = new boolean[maxCurves];
                normIndex = new int[maxCurves];
                detrendFitIndex = new int[maxCurves];
                detrendIndex = new int[maxCurves][maxDetrendVars];
                plotY = new boolean[maxCurves];
                useColumnName = new boolean[maxCurves];
                useLegend = new boolean[maxCurves];
                force = new boolean[maxCurves];
//                detrendFit = new boolean[maxCurves];
                binSize = new int[maxCurves];
                legend = new String[maxCurves];
                xyc1label = new String[maxCurves];
                xyc2label = new String[maxCurves];
                ASInclude = new boolean[maxCurves];
                autoscaleincludebox = new JCheckBox[maxCurves];

                usecurvebox = new JCheckBox[maxCurves];
                errorcolumnbox = new JCheckBox[maxCurves];
//                operrorcolumnbox = new JCheckBox[maxCurves];
                usebinbox = new JCheckBox[maxCurves];
                usemmagbox = new JCheckBox[maxCurves];
                fromMagBox = new JCheckBox[maxCurves];
                normtypecombobox = new JComboBox[maxCurves];
                detrendtypecombobox = new JComboBox[maxCurves];
                uselinesbox = new JCheckBox[maxCurves];
                forcebox = new JCheckBox[maxCurves];
                xdatacolumn = new JComboBox[maxCurves];
                ydatacolumn = new JComboBox[maxCurves];
                operatorselection = new JComboBox[maxCurves];
                operatorcolumn = new JComboBox[maxCurves];
                detrendbox = new JComboBox[maxCurves];
                markersymbolselection = new JComboBox[maxCurves];
                markercolorselection = new JComboBox[maxCurves];
                binsizespinnermodel = new SpinnerModel[maxCurves];
                binsizespinner = new JSpinner[maxCurves];
                morelegendradiopanelgroup = new JPanel[maxCurves];
                autoscalepanelgroup = new JPanel[maxCurves];
                customscalepanelgroup = new JPanel[maxCurves];
                detrendpanelgroup = new JPanel[maxCurves];
                normpanelgroup = new JPanel[maxCurves];
                savecolumnpanelgroup = new JPanel[maxCurves];
                legendnoneButton = new JRadioButton[maxCurves];
                legendcustomNameButton = new JRadioButton[maxCurves];
                legendcolumnNameButton = new JRadioButton[maxCurves];
                detrendVarButton = new JRadioButton[maxCurves][maxDetrendVars];
                autoscaleButton = new JRadioButton[maxCurves];
                customscaleButton = new JRadioButton[maxCurves];
                morelegendRadioGroup = new ButtonGroup[maxCurves];
                detrendVarRadioGroup = new ButtonGroup[maxCurves];
                autoscaleRadioGroup = new ButtonGroup[maxCurves];
                morelegendField = new JTextField[maxCurves];
                autoshiftmodel = new SpinnerModel[maxCurves];
                autoshiftstepmodel = new SpinnerModel[maxCurves];
                autoscalemodel  = new SpinnerModel[maxCurves];
                autoscalestepmodel  = new SpinnerModel[maxCurves];
                customshiftmodel = new SpinnerModel[maxCurves];
                customshiftstepmodel = new SpinnerModel[maxCurves];
                customscalemodel = new SpinnerModel[maxCurves];
                customscalestepmodel = new SpinnerModel[maxCurves];
                detrendfactormodel = new SpinnerModel[maxCurves];
                detrendfactorstepmodel = new SpinnerModel[maxCurves];                
                autoshiftspinner = new JSpinner[maxCurves];
                autoshiftstepspinner = new JSpinner[maxCurves];
                autoscalespinner  = new JSpinner[maxCurves];
                autoscalestepspinner  = new JSpinner[maxCurves];
                customshiftspinner = new JSpinner[maxCurves];
                customshiftstepspinner = new JSpinner[maxCurves];
                customscalespinner = new JSpinner[maxCurves];
                customscalestepspinner = new JSpinner[maxCurves];
                detrendfactorspinner = new JSpinner[maxCurves];
                detrendfactorstepspinner = new JSpinner[maxCurves];                
                autoscalesteppopup = new JPopupMenu[maxCurves];
                autoshiftsteppopup = new JPopupMenu[maxCurves];
                customscalesteppopup = new JPopupMenu[maxCurves];
                customshiftsteppopup = new JPopupMenu[maxCurves];
                detrendfactorsteppopup = new JPopupMenu[maxCurves];
                autoscalesteppanel = new JPanel[maxCurves];
                customscalesteppanel = new JPanel[maxCurves];
                autoshiftsteppanel = new JPanel[maxCurves];
                customshiftsteppanel = new JPanel[maxCurves];
                detrendfactorsteppanel = new JPanel[maxCurves];
                autoscalesteplabel = new JLabel[maxCurves];
                autoshiftsteplabel = new JLabel[maxCurves];
                customscalesteplabel = new JLabel[maxCurves];
                customshiftsteplabel = new JLabel[maxCurves];
                detrendfactorsteplabel = new JLabel[maxCurves];
                curvelabel = new JLabel[maxCurves];
                othercurvelabel = new JLabel[maxCurves];
                grabautopanel = new JPanel[maxCurves];
                grabautobutton = new JButton[maxCurves];
                savecolumnbutton = new JButton[maxCurves];

                customscaleeditor = new JSpinner.NumberEditor[maxCurves];
                customscaleformat = new DecimalFormat[maxCurves];
                customshifteditor = new JSpinner.NumberEditor[maxCurves];
                customshiftformat = new DecimalFormat[maxCurves];
                detrendfactoreditor = new JSpinner.NumberEditor[maxCurves];
                detrendfactorformat = new DecimalFormat[maxCurves];  
              

                //INITIALIZE CERTAIN ARRAYS
                for (int i=0; i<maxCurves; i++)
                        {
                        xlabel[i] = "default";
                        ylabel[i] = "rel_flux_"+(i==0?"T":"C")+(i+1);
//                        errlabel[i] = "rel_flux_err_"+(i==0?"T":"C")+(i+1);
                        oplabel[i] = "source-sky_"+(i==0?"T":"C")+(i+1);
//                        operrlabel[i] = "source_error_"+(i==0?"T":"C")+(i+1);
                        colorIndex[i] = (7+i)%colors.length;  //start with blue
                        markerIndex[i] = 3;   //default to dot
//                        detrendFit[i] = true;
                        force[i] = false;
                        showErrors[i] = false;
                        showErr[i] = false;
                        if (i==0)
                                {
                                plotY[i] = true;
                                ASInclude[i]= true;
                                }
                        else
                                {
                                plotY[i] = false;
                                ASInclude[i] = false;
                                }
                        autoScaleFactor[i] = 0.2;
                        autoScaleStep[i] = 1;
                        autoShiftFactor[i] = 0-i*10;
                        autoShiftStep[i] = 1;
                        customScaleFactor[i] = 1;
                        customScaleStep[i] = 1;
                        customShiftFactor[i] = 0-i*10;
                        customShiftStep[i] = 1;
                        detrendVarDisplayed[i] = 0;
                        binSize[i] = 1;
                        baseline[i] = 0;
                        legend[i] = "Legend"+(i+1);
                        normIndex[i] = 0;
                        detrendFitIndex[i] = 0;
                        for (int v = 0; v <maxDetrendVars; v++)
                            {
                            detrendlabel[i][v] = ""; 
                            detrendlabelhold[i][v] = ""; 
                            detrendFactor[i][v] = 0;  
                            detrendFactorStep[i][v] = 0.01;    
                            detrendIndex[i][v] = 0;
                            }
                        }
                }

    static void setupDataBuffers()
        {
        x = new double[maxCurves][maxColumnLength];
        y = new double[maxCurves][maxColumnLength];
        yerr = new double[maxCurves][maxColumnLength];
        yop = new double[maxCurves][maxColumnLength];
        yoperr = new double[maxCurves][maxColumnLength];
        detrend = new double[maxCurves][maxDetrendVars][maxColumnLength];

        xc1 = new double[maxCurves][maxColumnLength];
        xc2 = new double[maxCurves][maxColumnLength];
        yc1 = new double[maxCurves][maxColumnLength];
        yc2 = new double[maxCurves][maxColumnLength];
        }

    /**
	 * Returns minimum of double array.
	 */
	static double minOf(double[] arr, int n)
		{
        double mn = Double.POSITIVE_INFINITY;
		for (int i=0; i < n; i++)
            {
            if (!Double.isNaN(arr[i]))
                {
                mn = (arr[i] < mn) ? arr[i] : mn;
                }
            }
		return mn;
		}

	/**
	 * Returns maximum of double array.
	 */
	static double maxOf(double[] arr, int n)
		{
		double mx = Double.NEGATIVE_INFINITY;
		for (int i=0; i < n; i++)
            {
            if (!Double.isNaN(arr[i]))
                {
                mx = (arr[i] > mx)? arr[i] : mx;
                }
            }
		return mx;
		}

    /**
	 * Returns minimum of double array including error.
	 */
	static double minOf(double[] arr, double[] err, int n)
		{
        double mn = Double.POSITIVE_INFINITY;
        double value = 0; 
		for (int i=0; i < n; i++)
            {
            if (!Double.isNaN(arr[i]) && !Double.isNaN(err[i]))
                {
                value = arr[i] - err[i];
                mn = (value < mn) ? value : mn;
                }
            }
		return mn;
		}

	/**
	 * Returns maximum of double array including.
	 */
	static double maxOf(double[] arr, double[] err, int n)
		{
		double mx = Double.NEGATIVE_INFINITY;
        double value = 0; 
		for (int i=0; i < n; i++)
            {
            if (!Double.isNaN(arr[i]) && !Double.isNaN(err[i]))
                {
                value = arr[i] + err[i];
                mx = (value > mx)? value : mx;
                }
            }
		return mx;
		}    
    
    
    static Color colorOf(int ci)
            {
            Color cmkr;
            switch (ci)
                {
                case  0: cmkr = java.awt.Color.black;       break;
                case  1: cmkr = java.awt.Color.darkGray;    break;
                case  2: cmkr = java.awt.Color.gray;        break;
                case  3: cmkr = java.awt.Color.lightGray;   break;
                case  4: cmkr = new Color(0,235,0);         break;   //green
                case  5: cmkr = new Color(0,155,0);         break;   //dark green
                case  6: cmkr = new Color(84,201,245);      break;   //light blue
                case  7: cmkr = java.awt.Color.blue;        break;
                case  8: cmkr = java.awt.Color.magenta;     break;
                case  9: cmkr = java.awt.Color.pink;        break;
                case 10: cmkr = java.awt.Color.red;         break;
                case 11: cmkr = java.awt.Color.orange;      break;
                case 12: cmkr = new Color(235,235,0);       break;   //yellow
                case 13: cmkr = new Color(167,131,96);      break;   //brown
                case 14: cmkr = new Color(124,0,255);       break;   //purple
                case 15: cmkr = new Color(3,148,163);       break;   //teal
                default: cmkr =  java.awt.Color.black;      break;
                }
            return cmkr;
                }

        static int markerOf(int mi)
                {
                int mkr = 3;
                if      (mi == 0) mkr = ij.gui.Plot.BOX;
                else if (mi == 1) mkr = ij.gui.Plot.CIRCLE;
                else if (mi == 2) mkr = ij.gui.Plot.CROSS;
                else if (mi == 3) mkr = ij.gui.Plot.DOT;
                else if (mi == 4) mkr = ij.gui.Plot.LINE;
                else if (mi == 5) mkr = ij.gui.Plot.TRIANGLE;
                else if (mi == 6) mkr = ij.gui.Plot.X;
                else              mkr = ij.gui.Plot.DOT;
                return mkr;
                }


        static String scaleShiftText(boolean showScale, boolean showShift, boolean usemmag, boolean showmmaginfo, double scale, double shift)
                {  
                String shiftText = "";
                String scaleText = "";
                NumberFormat Eformatter = new DecimalFormat("0.###E0");
                NumberFormat dformatter = new DecimalFormat("##0.###");
                if (showScale && showmmaginfo && mmag[firstCurve] && usemmag && totalScaleFactor[firstCurve]==0.001 && scale!=0.001)
                    {
                    scale = 1000*scale;
                    if (scale == 1.0)
                        scaleText +="";
                    else if(((scale > 0.001) && (scale < 1000.0)) || ((scale < -0.001) && (scale > -1000.0)))
                        scaleText += " x("+dformatter.format(1/scale)+")";
                    else
                        scaleText += " x("+Eformatter.format(1/scale)+")";
                    }
                else if(showScale && (scale != 1.0))
                    {
                    if (((scale > 0.001) && (scale < 1000.0)) || ((scale < -0.001) && (scale > -1000.0)))
                        scaleText += " x("+dformatter.format(1/scale)+")";
                    else
                        if (usemmag && showmmaginfo && (scale == 0.001))
                            scaleText +="";
                        else
                            scaleText += " x("+Eformatter.format(1/scale)+")";
                    }
                if (showShift)
                        {
                        if (shift < 0.0)
                                {
                                shift *= -1.0;
                                shiftText += " - ";
                                }
                        else if (shift != 0.0)
                                shiftText += " + ";
                        if ((shift >= 0.01) && (shift < 100.0))
                                shiftText += dformatter.format(shift);
                        else if (shift != 0.0)
                                shiftText += Eformatter.format(shift);
                        }
                return scaleText + shiftText;
                }

        static String convertToText(double increment)
                {
                if      (increment == 0.0000000001) return "      0.0000000001";
                else if (increment == 0.000000001)  return "        0.000000001";
                else if (increment == 0.00000001)   return "          0.00000001";
                else if (increment == 0.0000001)    return "            0.0000001";
                else if (increment == 0.000001)     return "              0.000001";
                else if (increment == 0.00001)      return "                0.00001";
                else if (increment == 0.0001)       return "                  0.0001";
                else if (increment == 0.001)        return "                    0.001";
                else if (increment == 0.01)         return "                    0.010";
                else if (increment == 0.1)          return "                    0.100";
                else if (increment == 0.5)          return "                    0.500";
                else if (increment == 1.0)          return "                    1.000";
                else if (increment == 10.0)         return "                  10.000";
                else if (increment == 100.0)        return "                100.000";
                else if (increment == 1000.0)       return "              1000.000";
                else if (increment == 10000.0)      return "            10000.000";
                else if (increment == 100000.0)     return "          100000.000";
                else if (increment == 1000000.0)    return "        1000000.000";
                else if (increment == 10000000.0)   return "      10000000.000";
                else if (increment == 100000000.0)  return "    100000000.000";
                else if (increment == 1000000000.0) return "1000000000.000";
                else return "                    1.000";
                }

        static String intConvertToText(int increment)
                {
                if      (increment == 1)          return "                    1";
                else if (increment == 2)          return "                    2";
                else if (increment == 3)          return "                    3";
                else if (increment == 4)          return "                    4";
                else if (increment == 5)          return "                    5";
                else if (increment == 6)          return "                    6";
                else if (increment == 7)          return "                    7";
                else if (increment == 8)          return "                    8";
                else if (increment == 9)          return "                    9";
                else if (increment == 10)         return "                  10";
                else if (increment == 25)         return "                  25";
                else if (increment == 50)         return "                  50";
                else if (increment == 100)        return "                100";
                else if (increment == 250)        return "                250";
                else if (increment == 500)        return "                500";
                else if (increment == 1000)       return "              1000";
//                else if (increment == 10000)      return "            10000";
//                else if (increment == 100000)     return "          100000";
//                else if (increment == 1000000)    return "        1000000";
//                else if (increment == 10000000)   return "      10000000";
//                else if (increment == 100000000)  return "    100000000";
//                else if (increment == 1000000000) return "1000000000";
                else return "                    1";
                }

       static void checkForUT(JSpinner spinner)
                {
                JSpinner.NumberEditor ed = (JSpinner.NumberEditor)spinner.getEditor();
                String text = ed.getTextField().getText();
                if (Double.isNaN(Tools.parseDouble(text)))
                    {
                    Double value = 0.5;
                    String[] pieces = text.replaceAll("[^0-9\\.:]{1,}", " ").trim().split("[:]{1,}");
                    if (pieces.length > 0) value += Tools.parseDouble(pieces[0], 0)/24.0;
                    if (pieces.length > 1) value += Tools.parseDouble(pieces[1], 0)/1440.0;
                    if (pieces.length > 2) value += Tools.parseDouble(pieces[2], 0)/86400.0;
                    spinner.setValue(value%1);
                    }
                }
       
       static void keepMarkersInOrder(int marker)
           {
           if (marker == 1)
               {
               if (dMarker1Value > dMarker2Value)
                   {
                   dmarker2spinner.setValue(dMarker1Value);
                   dMarker2Value = dMarker1Value;                   
                   }
               if (dMarker1Value > dMarker3Value)
                   {
                   dmarker3spinner.setValue(dMarker1Value);
                   dMarker3Value = dMarker1Value;
                   }
               if (dMarker1Value > dMarker4Value)
                   {
                   dmarker4spinner.setValue(dMarker1Value);
                   dMarker4Value = dMarker1Value;
                   }
               }
           if (marker == 2)
               {
               if (dMarker2Value < dMarker1Value)
                   {
                   dmarker1spinner.setValue(dMarker2Value);
                   dMarker1Value = dMarker2Value;                   
                   }
               if (dMarker2Value > dMarker3Value)
                   {
                   dmarker3spinner.setValue(dMarker2Value);
                   dMarker3Value = dMarker2Value;
                   }
               if (dMarker2Value > dMarker4Value)
                   {
                   dmarker4spinner.setValue(dMarker2Value);
                   dMarker4Value = dMarker2Value;
                   }
               } 
           if (marker == 3)
               {
               if (dMarker3Value < dMarker1Value)
                   {
                   dmarker1spinner.setValue(dMarker3Value);
                   dMarker1Value = dMarker3Value;                   
                   }
               if (dMarker3Value < dMarker2Value)
                   {
                   dmarker2spinner.setValue(dMarker3Value);
                   dMarker2Value = dMarker3Value;
                   }
               if (dMarker3Value > dMarker4Value)
                   {
                   dmarker4spinner.setValue(dMarker3Value);
                   dMarker4Value = dMarker3Value;
                   }
               }    
           if (marker == 4)
               {
               if (dMarker4Value < dMarker1Value)
                   {
                   dmarker1spinner.setValue(dMarker4Value);
                   dMarker1Value = dMarker4Value;                   
                   }
               if (dMarker4Value < dMarker2Value)
                   {
                   dmarker2spinner.setValue(dMarker4Value);
                   dMarker2Value = dMarker4Value;
                   }
               if (dMarker4Value < dMarker3Value)
                   {
                   dmarker3spinner.setValue(dMarker4Value);
                   dMarker3Value = dMarker4Value;
                   }
               }           
           }


       static void showMainJPanel()
                {
// DISPLAY MAIN PANEL

                mainFrame = new JFrame ("Multi-plot Main");
                plotIcon = createImageIcon("astroj/images/plot.png", "Plot Icon");
                mainFrame.setIconImage(plotIcon.getImage());
                mainFrame.setDefaultCloseOperation(javax.swing.WindowConstants.DISPOSE_ON_CLOSE);
                mainFrame.addWindowListener(new WindowAdapter(){
                    @Override
                    public void windowClosing(WindowEvent e){
                    saveAndClose();}});
                
                mainpanel = new JPanel (new SpringLayout());
//                mainpanel.addMouseListener(panelMouseListener);
                mainpanel.addMouseMotionListener(panelMouseMotionListener);
                FileDrop fileDrop = new FileDrop(mainpanel, BorderFactory.createEmptyBorder(),new FileDrop.Listener()
                    {   public void filesDropped( java.io.File[] files )
                        {
                        openDragAndDropFiles(files);
                        }
                    });
               
                
                mainscrollpane = new JScrollPane(mainpanel);
                
                mainmenubar = new JMenuBar();
                filemenu = new JMenu("  File      ");
                preferencesmenu = new JMenu("Preferences    ");
                
                mainpanelgroupa = new JPanel (new SpringLayout());
                
                mainpanelgroupa.setBorder(dataSectionBorder);
                JPanel mainpanelgroupb = new JPanel (new SpringLayout());
                mainpanelgroupb.setBorder(BorderFactory.createTitledBorder("Data Options"));
                JPanel mainpanelgroupc = new JPanel (new SpringLayout());   //TITLE - SUBTITLE - LEGEND GROUP
                JPanel mainpanelgroupd = new JPanel (new SpringLayout());   // X & Y AXIS LEGEND GROUP
                JPanel mainpanelgroupe = new JPanel (new SpringLayout());
                mainpanelgroupe.setBorder(BorderFactory.createTitledBorder("Scaling Options"));
                JPanel mainpanelgroupf = new JPanel (new SpringLayout());

    // BUILD MENUBAR

                opentablemenuitem = new JMenuItem("Select open table...");
                opentablemenuitem.setToolTipText("<html>"+"select a table that is already open for plotting"+"</html>");
                opentablemenuitem.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                        selectAnotherTableCanceled = false;
                        MeasurementTable newTable = selectAnotherTable("");
                        if (newTable == null)
                            {
                            if (!selectAnotherTableCanceled)
                                {
                                makeDummyTable();
                                IJ.beep();
                                IJ.showMessage("No tables open");
                                }
                            selectAnotherTableCanceled = false;
                            return;
                            }
                        newTable.show();
                        setTable(newTable, true);
                        plotWindow.setVisible(true);
                        }});
                filemenu.add(opentablemenuitem);

                JMenuItem appendtablerowsmenuitem = new JMenuItem("Append open table as new rows...");
                appendtablerowsmenuitem.setToolTipText("<html>"+"select a table that is already open to append to current table"+"</html>");
                appendtablerowsmenuitem.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                            appendDataAsRows(false, null); }});
                filemenu.add(appendtablerowsmenuitem);                

                JMenuItem appendtablecolumnssmenuitem = new JMenuItem("Append open table as new columns...");
                appendtablecolumnssmenuitem.setToolTipText("<html>"+"select a table that is already open to append to current table"+"</html>");
                appendtablecolumnssmenuitem.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                            appendDataAsColumns(false, null); }});
                filemenu.add(appendtablecolumnssmenuitem); 
                filemenu.addSeparator();

                opendatamenuitem = new JMenuItem("Open table from file...");
                opendatamenuitem.setToolTipText("<html>"+"default input format = tab delimited, or use"+"<br>"+
                    "filename.csv = comma delimited"+"<br>"+
                    "filename.prn or filename.spc = space delimted"+"<br>"+
                    "---------------------------------------------"+"<br>"+
                    "first line should be column headings delimited as stated above"+"<br>"+
                    "lines starting with # are considered comments and ignored, except"+"<br>"+
                    "the last comment before the first data line can be headings"+"</html>");
                opendatamenuitem.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                            openData(); }});
                filemenu.add(opendatamenuitem);
                
                JMenuItem appenddatarowsmenuitem = new JMenuItem("Append table from file as new rows...");
                appenddatarowsmenuitem.setToolTipText("<html>"+"default input format = tab delimited, or use"+"<br>"+
                    "filename.csv = comma delimited"+"<br>"+
                    "filename.prn or filename.spc = space delimted"+"<br>"+
                    "---------------------------------------------"+"<br>"+
                    "first line should be column headings delimited as stated above"+"<br>"+
                    "lines starting with # are considered comments and ignored, except"+"<br>"+
                    "the last comment before the first data line can be headings"+"</html>");
                appenddatarowsmenuitem.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                            appendDataAsRows(true, null); }});
                filemenu.add(appenddatarowsmenuitem);
                
                JMenuItem appenddatacolumnsmenuitem = new JMenuItem("Append table from file as new columns...");
                appenddatacolumnsmenuitem.setToolTipText("<html>"+"default input format = tab delimited, or use"+"<br>"+
                    "filename.csv = comma delimited"+"<br>"+
                    "filename.prn or filename.spc = space delimted"+"<br>"+
                    "---------------------------------------------"+"<br>"+
                    "first line should be column headings delimited as stated above"+"<br>"+
                    "lines starting with # are considered comments and ignored, except"+"<br>"+
                    "the last comment before the first data line can be headings"+"</html>");
                appenddatacolumnsmenuitem.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                            appendDataAsColumns(true, null); }});
                filemenu.add(appenddatacolumnsmenuitem);
                
                openplotconfigmenuitem = new JMenuItem("Open plot configuration from file...");
                openplotconfigmenuitem.setToolTipText("<html>"+"opens a previously saved plot configuration<br>"+
                                                               "from a user selected file</html>");
                openplotconfigmenuitem.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                            openConfig(false); }});
                filemenu.add(openplotconfigmenuitem);

                opendataconfigmenuitem = new JMenuItem("Open table and plot configuration from file...");
                opendataconfigmenuitem.setToolTipText("<html>"+ "opens a previously saved measurement table"+"<br>"+
                                                                "from a user selected file and attempts to"+"<br>"+
                                                                "open a plot configuration file with the"+"<br>"+
                                                                "same name but ending in .plotcfg"+"</html>");
                opendataconfigmenuitem.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                            openDataAndConfig(null); }});
                filemenu.add(opendataconfigmenuitem);
               
                filemenu.addSeparator();

                savedatamenuitem = new JMenuItem("Save data to file...");
                savedatamenuitem.setToolTipText("<html>"+"saves measurement table data to a user selected file"+"</html>");
                savedatamenuitem.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                            saveData(); }});
                filemenu.add(savedatamenuitem);

                JMenuItem savedatasubsetmenuitem = new JMenuItem("Save data subset to file...");
                savedatasubsetmenuitem.setToolTipText("<html>"+"saves a subset of measurement table data to a user selected file"+"</html>");
                savedatasubsetmenuitem.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                            saveDataSubsetDialog(); }});
                filemenu.add(savedatasubsetmenuitem);
                
                JMenuItem saveimagepngmenuitem = new JMenuItem("Save plot image as PNG...");
                saveimagepngmenuitem.setToolTipText("<html>"+"saves plot image as a .png file"+"</html>");
                saveimagepngmenuitem.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                            savePlotImageAsPng(); }});
                filemenu.add(saveimagepngmenuitem);

                JMenuItem saveimagejpgmenuitem = new JMenuItem("Save plot image as JPG...");
                saveimagejpgmenuitem.setToolTipText("<html>"+"saves plot image as a .jpg file"+"</html>");
                saveimagejpgmenuitem.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                            savePlotImageAsJpg(); }});
                filemenu.add(saveimagejpgmenuitem);

                saveplotconfigmenuitem = new JMenuItem("Save plot configuration...");
                saveplotconfigmenuitem.setToolTipText("<html>"+"saves plot configuration to a user selected file</html>");
                saveplotconfigmenuitem.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                            saveConfig(false); }});
                filemenu.add(saveplotconfigmenuitem);

                filemenu.addSeparator();
                
                openplottemplatemenuitem = new JMenuItem("Open plot configuration template...");
                openplottemplatemenuitem.setToolTipText("<html>"+"opens a plot configuration from a user selected template file path</html>");
                openplottemplatemenuitem.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                            openConfig(true); }});
                filemenu.add(openplottemplatemenuitem);                
                
                saveplottemplatemenuitem = new JMenuItem("Save plot configuration template...");
                saveplottemplatemenuitem.setToolTipText("<html>"+"saves plot configuration to user selected template file path</html>");
                saveplottemplatemenuitem.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                            saveConfig(true); }});
                filemenu.add(saveplottemplatemenuitem);                
                
                filemenu.addSeparator();

                JMenuItem saveAllMenuItem = new JMenuItem("Save all...");
                saveAllMenuItem.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                        if (saveAllPNG)
                            saveAll("png", true);
                        else
                            saveAll("jpg", true); }});
                filemenu.add(saveAllMenuItem);

                JMenuItem saveAllWithOptionsMenuItem = new JMenuItem("Save all (with options)...");
                saveAllWithOptionsMenuItem.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                            setAllDialog(); }});
                filemenu.add(saveAllWithOptionsMenuItem);

                filemenu.addSeparator();

                JMenuItem exitmenuitem = new JMenuItem("Exit");
                exitmenuitem.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                            saveAndClose(); }});
                filemenu.add(exitmenuitem);
                
                mainmenubar.add(filemenu);

                opendatasetCB = new JCheckBoxMenuItem("Open Y-data columns window at startup",openDataSetWindow);
                opendatasetCB.addItemListener(new ItemListener(){
                    public void itemStateChanged(ItemEvent e) {
                        if (e.getStateChange() == ItemEvent.DESELECTED)
                                openDataSetWindow = false;
                        else if (e.getStateChange() == ItemEvent.SELECTED)
                                openDataSetWindow = true;
                        Prefs.set("plot2.openDataSetWindow",openDataSetWindow); }});
                preferencesmenu.add(opendatasetCB); 

                usewidedataCB = new JCheckBoxMenuItem("Use --wide-- Y-data columns window",useWideDataPanel);
                usewidedataCB.addItemListener(new ItemListener(){
                    public void itemStateChanged(ItemEvent e) {
                        if (e.getStateChange() == ItemEvent.DESELECTED)
                                useWideDataPanel = false;
                        else if (e.getStateChange() == ItemEvent.SELECTED)
                                useWideDataPanel = true;
                        Prefs.set("plot2.useWideDataPanel",useWideDataPanel);
                        subframeWasShowing = false;
                        if (subFrame.isShowing())
                            {
                            subFrameLocationX = subFrame.getLocation().x;
                            subFrameLocationY = subFrame.getLocation().y;
                            Prefs.set("plot2.subFrameLocationX",subFrameLocationX);
                            Prefs.set("plot2.subFrameLocationY",subFrameLocationY);
                            subFrame.dispose();
                            subframeWasShowing = true;
                            }

                        if (subframeWasShowing)
                                showMoreCurvesJPanel();
                        }});

                preferencesmenu.add(usewidedataCB);  
                
                preferencesmenu.addSeparator();
                    
                openrefstarCB = new JCheckBoxMenuItem("Open reference star selection window at startup",openRefStarWindow);
                openrefstarCB.addItemListener(new ItemListener(){
                    public void itemStateChanged(ItemEvent e) {
                        if (e.getStateChange() == ItemEvent.DESELECTED)
                                openRefStarWindow = false;
                        else if (e.getStateChange() == ItemEvent.SELECTED)
                                openRefStarWindow = true;
                        Prefs.set("plot2.openRefStarWindow",openRefStarWindow); }});
                preferencesmenu.add(openrefstarCB);                 
                
                setrefstarhorzsizemenuitem= new JMenuItem("Set horizontal width of reference star window...");
                setrefstarhorzsizemenuitem.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                            changeRefStarHorizontalWidth();
                            }});
                preferencesmenu.add(setrefstarhorzsizemenuitem);

                preferencesmenu.addSeparator();

                rememberwindowlocationsCB = new JCheckBoxMenuItem("Open windows at previous locations",rememberWindowLocations);
                rememberwindowlocationsCB.addItemListener(new ItemListener(){
                    public void itemStateChanged(ItemEvent e) {
                        if (e.getStateChange() == ItemEvent.DESELECTED)
                                rememberWindowLocations = false;
                        else if (e.getStateChange() == ItemEvent.SELECTED)
                                rememberWindowLocations = true;
                        Prefs.set("plot2.rememberWindowLocations",rememberWindowLocations); }});
                preferencesmenu.add(rememberwindowlocationsCB);
                

                divideNotSubtractCB = new JCheckBoxMenuItem("Detrend by division (deselect for faster subtraction)",divideNotSubtract);
                divideNotSubtractCB.addItemListener(new ItemListener(){
                    public void itemStateChanged(ItemEvent e) {
                        if (e.getStateChange() == ItemEvent.DESELECTED)
                                divideNotSubtract = false;
                        else if (e.getStateChange() == ItemEvent.SELECTED)
                                divideNotSubtract = true;
                        Prefs.set("plot.divideNotSubtract",divideNotSubtract);
                        updatePlot();   }});
                preferencesmenu.add(divideNotSubtractCB);

                showtooltipsCB = new JCheckBoxMenuItem("Show tooltips help",showToolTips);
                showtooltipsCB.addItemListener(new ItemListener(){
                    public void itemStateChanged(ItemEvent e) {
                        if (e.getStateChange() == ItemEvent.DESELECTED)
                            {
                            showToolTips=false;
                            }
                        else if (e.getStateChange() == ItemEvent.SELECTED)
                            {
                            showToolTips=true;
                            }
                        ToolTipManager.sharedInstance().setEnabled(showToolTips);
                        Prefs.set ("astroIJ.showToolTips",showToolTips);                        
                        }});
                preferencesmenu.add(showtooltipsCB);

                preferencesmenu.addSeparator();

                useBoldedDatumCB = new JCheckBoxMenuItem("Bold data point nearest mouse while holding shift",useBoldedDatum);
                useBoldedDatumCB.addItemListener(new ItemListener(){
                    public void itemStateChanged(ItemEvent e) {
                        if (e.getStateChange() == ItemEvent.DESELECTED)
                                useBoldedDatum = false;
                        else if (e.getStateChange() == ItemEvent.SELECTED)
                                useBoldedDatum = true;
                        Prefs.set("plot2.useBoldedDatum",useBoldedDatum);
                        updatePlot();}});
                preferencesmenu.add(useBoldedDatumCB);
                
                useUpdateStackCB = new JCheckBoxMenuItem("Display slice corresponding to point nearest mouse while holding shift",useUpdateStack);
                useUpdateStackCB.addItemListener(new ItemListener(){
                    public void itemStateChanged(ItemEvent e) {
                        if (e.getStateChange() == ItemEvent.DESELECTED)
                                useUpdateStack = false;
                        else if (e.getStateChange() == ItemEvent.SELECTED)
                                useUpdateStack = true;
                        Prefs.set("plot2.useUpdateStack",useUpdateStack);
                        updatePlot();}});
                preferencesmenu.add(useUpdateStackCB);    
                
                preferencesmenu.addSeparator();

//                setverticalmarkertextmenuitem = new JMenuItem("Change vertical marker text...");
//                setverticalmarkertextmenuitem.addActionListener(new ActionListener() {
//                    public void actionPerformed(ActionEvent e) {
//                            setVMarkerText();
//                            updatePlot();}});
//                preferencesmenu.add(setverticalmarkertextmenuitem);
                
                usePriorityColumnsCB = new JCheckBoxMenuItem("Show specified priority data names first in pulldown lists",prioritizeColumns);
                usePriorityColumnsCB.addItemListener(new ItemListener(){
                    public void itemStateChanged(ItemEvent e) {
                        if (e.getStateChange() == ItemEvent.DESELECTED)
                                prioritizeColumns = false;
                        else if (e.getStateChange() == ItemEvent.SELECTED)
                                prioritizeColumns = true;
                        Prefs.set("plot.prioritizeColumns",prioritizeColumns);
                        oldUnfilteredColumns = null;
                        updatePlot();}});
                preferencesmenu.add(usePriorityColumnsCB);

                changePriorityColumnsmenuitem = new JMenuItem("Change priority data names list...");
                changePriorityColumnsmenuitem.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                            changePriorityColumns();
                            }});
                preferencesmenu.add(changePriorityColumnsmenuitem);

                preferencesmenu.addSeparator();
               
                usepixelscaleCB = new JCheckBoxMenuItem("Use pixel scale when calculating distance",usePixelScale);
                usepixelscaleCB.addItemListener(new ItemListener(){
                    public void itemStateChanged(ItemEvent e) {
                        if (e.getStateChange() == ItemEvent.DESELECTED)
                                usePixelScale = false;
                        else if (e.getStateChange() == ItemEvent.SELECTED)
                                usePixelScale = true;
                        Prefs.set("plot.usePixelScale",usePixelScale);
                        updatePlot();}});
                preferencesmenu.add(usepixelscaleCB);

                changepixelscalemenuitem = new JMenuItem("Set pixel scale ("+pixelScale+")...");
                changepixelscalemenuitem.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                            changePixelScale();
                            updatePlot();}});
                preferencesmenu.add(changepixelscalemenuitem);

                preferencesmenu.addSeparator();

                changemaxdatalengthmenuitem = new JMenuItem("Set minimum data column buffer length...");
                changemaxdatalengthmenuitem.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                            changeMaxDataLength();
                            updatePlot();}});
                preferencesmenu.add(changemaxdatalengthmenuitem);

                usedefaultsettingsCB = new JCheckBoxMenuItem("Reset preferences to default settings (restart required)",useDefaultSettings);
                usedefaultsettingsCB.addItemListener(new ItemListener(){
                    public void itemStateChanged(ItemEvent e) {
                        if (e.getStateChange() == ItemEvent.DESELECTED)
                                useDefaultSettings = false;
                        else if (e.getStateChange() == ItemEvent.SELECTED)
                                useDefaultSettings = true;
                        Prefs.set("plot2.useDefaultSettings",useDefaultSettings); }});
                preferencesmenu.add(usedefaultsettingsCB);

                mainmenubar.add(preferencesmenu);
                
                
                
                
                tablemenu = new JMenu("Table    ");
                
                addnewcolsfromplotmenuitem = new JMenuItem("Add new data columns to table from plot...");
                addnewcolsfromplotmenuitem.setToolTipText("<html>"+"Add new data columns to the current active table from<br>"+
                                                                   "an enabled 'Multi-plot Y-data' plot row.</html>");
                addnewcolsfromplotmenuitem.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                        int c = selectCurve();
                        if (c < 0) return;
                        addNewColumn(c, true);
                        updatePlot();
                        }});
                tablemenu.add(addnewcolsfromplotmenuitem);   
                
                addnewcolsfromastroCCmenuitem = new JMenuItem("Add new astronomical data columns to table...");
                addnewcolsfromastroCCmenuitem.setToolTipText("<html>"+"Add new data columns to the current active table from AstroCC<br>"+
                                                                      "Cordinate Converter (airmass, alternate time formats, etc).</html>");
                addnewcolsfromastroCCmenuitem.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                        if (addAstroDataFrame != null)
                            {
                            addAstroDataFrameWasShowing = false;
                            closeAddAstroDataFrame();
                            if (acc != null) acc.saveAndClose();
                            }
                        addNewAstroData();
                        }});
                tablemenu.add(addnewcolsfromastroCCmenuitem);      
                
                tablemenu.addSeparator();
                
                transposetablemenuitem = new JMenuItem("Transpose currently selected table");
                transposetablemenuitem.setToolTipText("<html>"+"swap rows and columns in current active table"+"</html>");
                transposetablemenuitem.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                        transposeTable();
                        }});
                tablemenu.add(transposetablemenuitem);                
                
                mainmenubar.add(tablemenu);
                
                
                
                
                
                xaxismenu = new JMenu("X-axis    ");
                
                xTicsCB = new JCheckBoxMenuItem ("Show X-axis tick marks",xTics);
                xTicsCB.addItemListener(new ItemListener(){
                    public void itemStateChanged(ItemEvent e) {
                        if (e.getStateChange() == ItemEvent.DESELECTED)
                                xTics = false;
                        else if (e.getStateChange() == ItemEvent.SELECTED)
                                xTics = true;
                        Prefs.set("plot.xTics",xTics);
                        updatePlot();}});
                xaxismenu.add(xTicsCB); 
                
                xGridCB = new JCheckBoxMenuItem ("Show X-axis grid lines",xGrid);
                xGridCB.addItemListener(new ItemListener(){
                    public void itemStateChanged(ItemEvent e) {
                        if (e.getStateChange() == ItemEvent.DESELECTED)
                                xGrid = false;
                        else if (e.getStateChange() == ItemEvent.SELECTED)
                                xGrid = true;
                        Prefs.set("plot.xGrid",xGrid);
                        updatePlot();}});
                xaxismenu.add(xGridCB);   
                
                xNumbersCB = new JCheckBoxMenuItem ("Show X-axis numbers",xNumbers);
                xNumbersCB.addItemListener(new ItemListener(){
                    public void itemStateChanged(ItemEvent e) {
                        if (e.getStateChange() == ItemEvent.DESELECTED)
                                xNumbers = false;
                        else if (e.getStateChange() == ItemEvent.SELECTED)
                                xNumbers = true;
                        Prefs.set("plot.xNumbers",xNumbers);
                        updatePlot();}});
                xaxismenu.add(xNumbersCB);                
                
                xaxismenu.addSeparator();

                if (showXAxisNormal)
                    {
                    showXAxisAsPhase = false;
                    showXAxisAsDaysSinceTc = false;
                    showXAxisAsHoursSinceTc = false; 
                    }
                else if (showXAxisAsPhase)
                    {
                    showXAxisAsDaysSinceTc = false;
                    showXAxisAsHoursSinceTc = false;                    
                    }
                else if (showXAxisAsDaysSinceTc)
                    {
                    showXAxisAsHoursSinceTc = false;                    
                    }
                else if (!showXAxisAsHoursSinceTc)
                    {
                    showXAxisNormal = true;                    
                    }
                Prefs.set("plot.showXAxisNormal",showXAxisNormal);
                Prefs.set("plot.showXAxisAsPhase",showXAxisAsPhase);
                Prefs.set("plot.showXAxisAsDaysSinceTc",showXAxisAsDaysSinceTc);
                Prefs.set("plot.showXAxisAsHoursSinceTc",showXAxisAsHoursSinceTc);                
                
                showXAxisNormalCB = new JRadioButtonMenuItem("Show x-axis unphased", showXAxisNormal);
                showXAxisNormalCB.addActionListener(new ActionListener(){
                        public void actionPerformed(ActionEvent ae) {
                        showXAxisNormal = true;
                        showXAxisAsPhase = false;
                        showXAxisAsDaysSinceTc = false;
                        showXAxisAsHoursSinceTc = false;
                        Prefs.set("plot.showXAxisNormal",showXAxisNormal);
                        Prefs.set("plot.showXAxisAsPhase",showXAxisAsPhase);
                        Prefs.set("plot.showXAxisAsDaysSinceTc",showXAxisAsDaysSinceTc);
                        Prefs.set("plot.showXAxisAsHoursSinceTc",showXAxisAsHoursSinceTc);
                        updatePlot();}});
                xaxismenu.add(showXAxisNormalCB);
                
                showXAxisAsPhaseCB = new JRadioButtonMenuItem("Show x-axis as orbital phase", showXAxisAsPhase);
                showXAxisAsPhaseCB.addActionListener(new ActionListener(){
                        public void actionPerformed(ActionEvent ae) {
                        showXAxisNormal = false;
                        showXAxisAsPhase = true;
                        showXAxisAsDaysSinceTc = false;
                        showXAxisAsHoursSinceTc = false;
                        Prefs.set("plot.showXAxisNormal",showXAxisNormal);
                        Prefs.set("plot.showXAxisAsPhase",showXAxisAsPhase);
                        Prefs.set("plot.showXAxisAsDaysSinceTc",showXAxisAsDaysSinceTc);
                        Prefs.set("plot.showXAxisAsHoursSinceTc",showXAxisAsHoursSinceTc);
                        updatePlot();}});
                xaxismenu.add(showXAxisAsPhaseCB);
                
                showXAxisAsDaysSinceTcCB = new JRadioButtonMenuItem("Show x-axis as days since Tc", showXAxisAsDaysSinceTc);
                showXAxisAsDaysSinceTcCB.addActionListener(new ActionListener(){
                        public void actionPerformed(ActionEvent ae) {
                        showXAxisNormal = false;
                        showXAxisAsPhase = false;
                        showXAxisAsDaysSinceTc = true;
                        showXAxisAsHoursSinceTc = false;
                        Prefs.set("plot.showXAxisNormal",showXAxisNormal);
                        Prefs.set("plot.showXAxisAsPhase",showXAxisAsPhase);
                        Prefs.set("plot.showXAxisAsDaysSinceTc",showXAxisAsDaysSinceTc);
                        Prefs.set("plot.showXAxisAsHoursSinceTc",showXAxisAsHoursSinceTc);
                        updatePlot();}});
                xaxismenu.add(showXAxisAsDaysSinceTcCB);                
                
                showXAxisAsHoursSinceTcCB = new JRadioButtonMenuItem("Show x-axis as hours since Tc", showXAxisAsHoursSinceTc);
                showXAxisAsHoursSinceTcCB.addActionListener(new ActionListener(){
                        public void actionPerformed(ActionEvent ae) {
                        showXAxisNormal = false;
                        showXAxisAsPhase = false;
                        showXAxisAsDaysSinceTc = false;
                        showXAxisAsHoursSinceTc = true;
                        Prefs.set("plot.showXAxisNormal",showXAxisNormal);
                        Prefs.set("plot.showXAxisAsPhase",showXAxisAsPhase);
                        Prefs.set("plot.showXAxisAsDaysSinceTc",showXAxisAsDaysSinceTc);
                        Prefs.set("plot.showXAxisAsHoursSinceTc",showXAxisAsHoursSinceTc);
                        updatePlot();}});
                xaxismenu.add(showXAxisAsHoursSinceTcCB);     
                
                xPhaseGroup = new ButtonGroup();
                xPhaseGroup.add(showXAxisNormalCB);
                xPhaseGroup.add(showXAxisAsPhaseCB);
                xPhaseGroup.add(showXAxisAsDaysSinceTcCB);
                xPhaseGroup.add(showXAxisAsHoursSinceTcCB);
                
                xaxismenu.addSeparator();

                setephemerismenuitem = new JMenuItem("Set epoch and period for X-axis phase calculation...");
                setephemerismenuitem.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                            setEphemeris();
                            updatePlot();
                            }});
                xaxismenu.add(setephemerismenuitem);

//                xaxismenu.addSeparator();
                
                mainmenubar.add(xaxismenu);
                
                
                
                
                yaxismenu = new JMenu("Y-axis    ");
                
                yTicsCB = new JCheckBoxMenuItem ("Show Y-axis tick marks",yTics);
                yTicsCB.addItemListener(new ItemListener(){
                    public void itemStateChanged(ItemEvent e) {
                        if (e.getStateChange() == ItemEvent.DESELECTED)
                                yTics = false;
                        else if (e.getStateChange() == ItemEvent.SELECTED)
                                yTics = true;
                        Prefs.set("plot.yTics",yTics);
                        updatePlot();}});
                yaxismenu.add(yTicsCB); 
                
                yGridCB = new JCheckBoxMenuItem ("Show Y-axis grid lines",yGrid);
                yGridCB.addItemListener(new ItemListener(){
                    public void itemStateChanged(ItemEvent e) {
                        if (e.getStateChange() == ItemEvent.DESELECTED)
                                yGrid = false;
                        else if (e.getStateChange() == ItemEvent.SELECTED)
                                yGrid = true;
                        Prefs.set("plot.yGrid",yGrid);
                        updatePlot();}});
                yaxismenu.add(yGridCB);   
                
                yNumbersCB = new JCheckBoxMenuItem ("Show Y-axis numbers",yNumbers);
                yNumbersCB.addItemListener(new ItemListener(){
                    public void itemStateChanged(ItemEvent e) {
                        if (e.getStateChange() == ItemEvent.DESELECTED)
                                yNumbers = false;
                        else if (e.getStateChange() == ItemEvent.SELECTED)
                                yNumbers = true;
                        Prefs.set("plot.yNumbers",yNumbers);
                        updatePlot();}});
                yaxismenu.add(yNumbersCB);                
                
                yaxismenu.addSeparator();                
                
                JCheckBoxMenuItem invertyaxisCB = new JCheckBoxMenuItem ("Invert Y-axis",invertYAxis);
                invertyaxisCB.addItemListener(new ItemListener(){
                    public void itemStateChanged(ItemEvent e) {
                        if (e.getStateChange() == ItemEvent.DESELECTED)
                            {
                            invertYAxis = false;
                            invertYAxisSign = 1;
                            }
                        else if (e.getStateChange() == ItemEvent.SELECTED)
                            {
                            invertYAxis = true;
                            invertYAxisSign = -1;
                            }
                        for(int c=0; c<maxCurves;c++)
                            {
                            customshiftspinner[c].setModel(new SpinnerNumberModel(new Double(force[c]?autoShiftFactor[c]*100:customShiftFactor[c]), null, null,
                                                         new Double(force[c]?invertYAxisSign*autoShiftStep[c]:invertYAxisSign*customShiftStep[c])));
                            customshiftspinner[c].setEditor(new JSpinner.NumberEditor(customshiftspinner[c], "0.#########"));
                            }
                        Prefs.set("plot.invertYAxis",invertYAxis);
                        updatePlot();}});
                yaxismenu.add(invertyaxisCB);

                JCheckBoxMenuItem negateMagCB = new JCheckBoxMenuItem ("Negate relative magnitude calculations",negateMag);
                negateMagCB.addItemListener(new ItemListener(){
                    public void itemStateChanged(ItemEvent e) {
                        if (e.getStateChange() == ItemEvent.DESELECTED)
                                negateMag = false;
                        else if (e.getStateChange() == ItemEvent.SELECTED)
                                negateMag = true;
                        Prefs.set("plot.negateMag",negateMag);
                        updatePlot();}});
                yaxismenu.add(negateMagCB);

//                preferencesmenu.addSeparator();                
                mainmenubar.add(yaxismenu);
                                
                JMenu helpmenu = new JMenu("Help");
                JMenuItem helpmenuitem = new JMenuItem("General help...");
                helpmenuitem.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                            openHelpPanel(); }});
                helpmenu.add(helpmenuitem);
                
                JMenuItem helpdatamenuitem = new JMenuItem("Data naming convention...");
                helpdatamenuitem.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                            openDataHelpPanel(); }});
                helpmenu.add(helpdatamenuitem);                 

                mainmenubar.add(helpmenu);                



    // MAIN PANEL GROUP A START
                xcolumnlabel = new JLabel ("Default X-data");
                xcolumnlabel.setHorizontalAlignment(JLabel.CENTER);
                mainpanelgroupa.add (xcolumnlabel);

//                ycolumnslabel = new JLabel ("Y-data Panel");
//                ycolumnslabel.setHorizontalAlignment(JLabel.CENTER);
//                mainpanelgroupa.add (ycolumnslabel);

                JLabel dummy50label = new JLabel ("Y-datasets");
                dummy50label.setHorizontalAlignment(JLabel.CENTER);
                mainpanelgroupa.add (dummy50label);
                
                JLabel dummy50Alabel = new JLabel ("Detrend Vars");
                dummy50Alabel.setHorizontalAlignment(JLabel.CENTER);
                mainpanelgroupa.add (dummy50Alabel);                

                JLabel mmagsampleslabel = new JLabel ("Rel. Mag. Reference");
                mmagsampleslabel.setHorizontalAlignment(JLabel.CENTER);
                mmagsampleslabel.setToolTipText("The number of starting samples averaged to calculate relative magnitude reference level");
                mainpanelgroupa.add (mmagsampleslabel);
                
                
                JPanel vmarker1titlepanel = new JPanel();
                vmarker1titlepanel.setLayout(new BoxLayout(vmarker1titlepanel, BoxLayout.LINE_AXIS));
                
                vmarker1titlepanel.add(Box.createGlue());
                vmarker1titlepanel.add(Box.createHorizontalStrut(5));
                
                JLabel vmarker1toplabel = new JLabel ("V. Marker 1");
                vmarker1toplabel.setHorizontalAlignment(JLabel.CENTER);
                vmarker1toplabel.setToolTipText("Enter vertical marker 1 x-axis location");
                vmarker1titlepanel.add (vmarker1toplabel);
                
                vmarker1titlepanel.add(Box.createHorizontalStrut(5));
                
                editTextIcon = createImageIcon("astroj/images/edittext.png", "Edit vertical marker text");
                vmarker1edittextButton = new JButton(editTextIcon);
                vmarker1edittextButton.setToolTipText("Edit vertical marker text");
                vmarker1edittextButton.setMargin(new Insets(0,0,0,0));

                vmarker1edittextButton.addActionListener(new ActionListener(){
                            public void actionPerformed(ActionEvent ae) {
                                setVMarkerText();
                                updatePlot();}});                
                vmarker1titlepanel.add (vmarker1edittextButton);
                vmarker1titlepanel.add(Box.createGlue());
                mainpanelgroupa.add (vmarker1titlepanel);     
                
                JPanel vmarker2titlepanel = new JPanel();
                vmarker2titlepanel.setLayout(new BoxLayout(vmarker2titlepanel, BoxLayout.LINE_AXIS));
                
                vmarker2titlepanel.add(Box.createGlue());                
                vmarker2titlepanel.add(Box.createHorizontalStrut(5));
                
                JLabel vmarker2toplabel = new JLabel ("V. Marker 2");
                vmarker2toplabel.setHorizontalAlignment(JLabel.CENTER);
                vmarker2toplabel.setToolTipText("Enter vertical marker 2 x-axis location");
                vmarker2titlepanel.add (vmarker2toplabel);
                
                vmarker2titlepanel.add(Box.createHorizontalStrut(5));
                
                vmarker2edittextButton = new JButton(editTextIcon);
                vmarker2edittextButton.setToolTipText("Edit vertical marker text");
                vmarker2edittextButton.setMargin(new Insets(0,0,0,0));

                vmarker2edittextButton.addActionListener(new ActionListener(){
                            public void actionPerformed(ActionEvent ae) {
                                setVMarkerText();
                                updatePlot();}});                
                vmarker2titlepanel.add (vmarker2edittextButton);
                vmarker2titlepanel.add(Box.createGlue());
                mainpanelgroupa.add (vmarker2titlepanel);                  

                JPanel xdatacolumnpanel = new JPanel(new SpringLayout());
                MutableComboBoxModel xdatadefaultmodel = new DefaultComboBoxModel(columns);
                xdatacolumndefault = new JComboBox(xdatadefaultmodel);
                xdatacolumndefault.setSelectedItem(xlabeldefault);
                xdatacolumndefault.setPrototypeDisplayValue("123456789012345");
                xdatacolumndefault.addActionListener(new ActionListener(){
                    public void actionPerformed(ActionEvent ae) {
                        xlabeldefault = (String) xdatacolumndefault.getSelectedItem();
                        updatePlot();
                    }});
                        
                xdatacolumnpanel.add (xdatacolumndefault);
                xdatacolumnpanel.setPreferredSize(new Dimension(125, 25));
                SpringUtil.makeCompactGrid (xdatacolumnpanel, 1, 1, 0,0,0,0);
                mainpanelgroupa.add (xdatacolumnpanel);

                JPanel maxcurvespanel = new JPanel(new SpringLayout());
                maxcurvespanel.setBorder(BorderFactory.createLineBorder(Color.lightGray,1));
                maxcurvesmodel = new SpinnerNumberModel(new Integer(maxCurves),new Integer(1), null, new Integer(1));
                maxcurvesspinner = new JSpinner(maxcurvesmodel);
                maxcurvesspinner.setPreferredSize(new Dimension(60,25));
                maxcurvesspinner.addChangeListener (new ChangeListener()
                    {
                    public void stateChanged(ChangeEvent ev)
                        {
                        savePreferences();
                        maxCurves = ((Integer)maxcurvesspinner.getValue()).intValue();
                        Prefs.set("plot.maxCurves", maxCurves);
                        subframeWasShowing = false;
                        if (subFrame.isShowing())
                            {
                            subFrameLocationX = subFrame.getLocation().x;
                            subFrameLocationY = subFrame.getLocation().y;
                            Prefs.set("plot2.subFrameLocationX",subFrameLocationX);
                            Prefs.set("plot2.subFrameLocationY",subFrameLocationY);
                            subFrame.dispose();
                            subframeWasShowing = true;
                            }
                        setupArrays();
                        getPreferences();
                        if (subframeWasShowing)
                                showMoreCurvesJPanel();
                        updatePlot();
                        } });
                maxcurvesspinner.addMouseWheelListener( new MouseWheelListener()
                    {
                    public void mouseWheelMoved( MouseWheelEvent e )
                        {
                        int newValue =  new Integer(((Integer)maxcurvesspinner.getValue()).intValue() - e.getWheelRotation());
                        if (newValue > 0) maxcurvesspinner.setValue(newValue);
                        else maxcurvesspinner.setValue(1);
                        }
                    });
                maxcurvespanel.add (maxcurvesspinner);

                JLabel maxcurveslabel2 = new JLabel("sets");
                maxcurveslabel2.setHorizontalAlignment(JLabel.LEFT);
                maxcurvespanel.add (maxcurveslabel2);
                SpringUtil.makeCompactGrid (maxcurvespanel, 1, 2, 0,0,0,0);
                mainpanelgroupa.add(maxcurvespanel);   
                
                JPanel maxdetrendvarspanel = new JPanel(new SpringLayout());
//                maxdetrendvarspanel.setBorder(BorderFactory.createLineBorder(Color.lightGray,1));
                maxdetrendvarsmodel = new SpinnerNumberModel(new Integer(maxDetrendVars),new Integer(1), null, new Integer(1));
                maxdetrendvarsspinner = new JSpinner(maxdetrendvarsmodel);
                maxdetrendvarsspinner.setPreferredSize(new Dimension(60,25));
                maxdetrendvarsspinner.addChangeListener (new ChangeListener()
                    {
                    public void stateChanged(ChangeEvent ev)
                        {
                        savePreferences();
                        maxDetrendVars = ((Integer)maxdetrendvarsspinner.getValue()).intValue();
                        Prefs.set("plot.maxDetrendVars", maxDetrendVars);
                        for (int c = 0; c < maxCurves; c++)
                            {
                            if (detrendVarDisplayed[c] >= maxDetrendVars)
                                {
                                detrendVarDisplayed[c] = maxDetrendVars - 1;
                                Prefs.set("plot.detrendVarDisplayed"+c, detrendVarDisplayed[c]);
                                }
                            }
                        subframeWasShowing = false;
                        if (subFrame.isShowing())
                            {
                            subFrameLocationX = subFrame.getLocation().x;
                            subFrameLocationY = subFrame.getLocation().y;
                            Prefs.set("plot2.subFrameLocationX",subFrameLocationX);
                            Prefs.set("plot2.subFrameLocationY",subFrameLocationY);
                            subFrame.dispose();
                            subframeWasShowing = true;
                            }
                        setupArrays();
                        getPreferences();
                        if (subframeWasShowing)
                                showMoreCurvesJPanel();
                        updatePlot();
                        } });
                maxdetrendvarsspinner.addMouseWheelListener( new MouseWheelListener()
                    {
                    public void mouseWheelMoved( MouseWheelEvent e )
                        {
                        int newValue =  new Integer(((Integer)maxdetrendvarsspinner.getValue()).intValue() - e.getWheelRotation());
                        if (newValue > 0) maxdetrendvarsspinner.setValue(newValue);
                        else maxdetrendvarsspinner.setValue(1);
                        }
                    });
                maxdetrendvarspanel.add (maxdetrendvarsspinner);

                SpringUtil.makeCompactGrid (maxdetrendvarspanel, 1, maxdetrendvarspanel.getComponentCount(), 0,0,0,0);
                mainpanelgroupa.add(maxdetrendvarspanel);
                

                JPanel mmagsubpanel = new JPanel (new SpringLayout());
                mmagsubpanel.setBorder(BorderFactory.createLineBorder(Color.lightGray,1));
                mmagsubpanel.setToolTipText("The number of starting samples averaged to calculate relative magnitude reference level");
                mmagrefsmodel = new SpinnerNumberModel(new Integer(mmagrefs), new Integer (1), null, new Integer(1));
                mmagrefsspinner = new JSpinner(mmagrefsmodel);
                mmagrefsspinner.setPreferredSize(new Dimension(60, 25));
                mmagrefsspinner.setToolTipText("The number of starting samples averaged to calculate relative magnitude reference level");
                mmagrefsspinner.addChangeListener (new ChangeListener()
                    {
                    public void stateChanged(ChangeEvent ev)
                        {
                        mmagrefs = ((Integer)mmagrefsspinner.getValue()).intValue();
                        updatePlot();
                        }
                    });
                mmagrefsspinner.addMouseWheelListener( new MouseWheelListener()
                    {
                    public void mouseWheelMoved( MouseWheelEvent e )
                        {
                        int newValue =  new Integer(((Integer)mmagrefsspinner.getValue()).intValue() - e.getWheelRotation());
                        if (newValue > 0) mmagrefsspinner.setValue(newValue);
                        }
                    });
                mmagsubpanel.add (mmagrefsspinner);

                mmagrefsspinnerlabel = new JLabel ("samples");
                mmagrefsspinnerlabel.setHorizontalAlignment(JLabel.LEFT);
                mmagsubpanel.add (mmagrefsspinnerlabel);
                SpringUtil.makeCompactGrid (mmagsubpanel, 1, 2, 0,0,0,0);
                mainpanelgroupa.add (mmagsubpanel);

                //VERTICAL MARKER 1 TIME STEPSIZE POPUP

                vmarkersteppopup = new JPopupMenu();
                JPanel vmarkersteppanel = new JPanel();
                vmarkerstepmodel = new SpinnerListModel(spinnerscalelist);
                vmarkerstepspinner = new JSpinner(vmarkerstepmodel);
                vmarkerstepspinner.setValue(convertToText(vMarkerStep));
                vmarkerstepspinner.addChangeListener (new ChangeListener()
                    {
                    public void stateChanged(ChangeEvent ev)
                        {
                        vMarkerStep = Double.parseDouble((String)vmarkerstepspinner.getValue());
                        vmarker1spinner.setModel(new SpinnerNumberModel(new Double(vMarker1Value), null, null, new Double(vMarkerStep)));
                        vmarker2spinner.setModel(new SpinnerNumberModel(new Double(vMarker2Value), null, null, new Double(vMarkerStep)));
                        vmarker1spinner.setEditor(new JSpinner.NumberEditor(vmarker1spinner, "0.#########"));
                        vmarker2spinner.setEditor(new JSpinner.NumberEditor(vmarker2spinner, "0.#########"));
                        Prefs.set("plot.vmarkerStep",vMarkerStep);
                        }
                    });

                JLabel vmarkersteplabel = new JLabel ("Stepsize:");
                vmarkersteppanel.add(vmarkersteplabel);
                vmarkersteppanel.add(vmarkerstepspinner);
                vmarkersteppopup.add(vmarkersteppanel);
                vmarkersteppopup.setLightWeightPopupEnabled(false);
                vmarkersteppopup.addPopupMenuListener (new PopupMenuListener()
                    {
                    public void popupMenuCanceled(PopupMenuEvent popupMenuEvent) {
                            vmarkersteppopup.setVisible(false);
                            }

                    public void popupMenuWillBecomeInvisible(PopupMenuEvent popupMenuEvent) {
                            vmarkersteppopup.setVisible(true);
                            }

                    public void popupMenuWillBecomeVisible(PopupMenuEvent popupMenuEvent) {
                            vmarkersteppopup.setVisible(false);
                            }
                    });

                JPanel vmarker1panel = new JPanel(new SpringLayout());
                vmarker1panel.setBorder(BorderFactory.createLineBorder(Color.lightGray,1));
                showVMarker1CB = new JCheckBox ("", showVMarker1);
                showVMarker1CB.setToolTipText("Select to show vertical marker 1");
                showVMarker1CB.addItemListener(new ItemListener(){
                    public void itemStateChanged(ItemEvent e) {
                        if (e.getStateChange() == ItemEvent.DESELECTED)
                                showVMarker1 = false;
                        else if (e.getStateChange() == ItemEvent.SELECTED)
                                showVMarker1 = true;
                        updatePlot();}});
                showVMarker1CB.setHorizontalAlignment(JLabel.CENTER);
                vmarker1panel.add (showVMarker1CB);

                vmarker1spinnermodel = new SpinnerNumberModel(new Double(vMarker1Value),  null, null, new Double(vMarkerStep));

                vmarker1spinner = new JSpinner(vmarker1spinnermodel);
//                JComponent editor = new JSpinner.NumberEditor(vmarker1spinner);
//                vmarker1spinner.setEditor(editor);
                vmarker1spinner.setEditor(new JSpinner.NumberEditor(vmarker1spinner, "0.#########"));
                vmarker1spinner.setPreferredSize(new Dimension(75, 25));
                vmarker1spinner.setComponentPopupMenu(vmarkersteppopup);
                vmarker1spinner.setToolTipText("<html>"+"Enter vertical marker 1 x-axis location"+"<br>"+
                    "or enter UT time in HH:MM or HH:MM:SS format and press 'Enter'"+"<br>"+
                    "---------------------------------------------"+"<br>"+
                    "Right click to set spinner stepsize"+"</html>");
                vmarker1spinner.addChangeListener (new ChangeListener()
                    {
                    public void stateChanged(ChangeEvent ev)
                        {
                        checkForUT(vmarker1spinner);
                        vMarker1Value = ((Double)vmarker1spinner.getValue()).doubleValue();
                        updatePlot();
                        }
                    });
                vmarker1spinner.addMouseWheelListener( new MouseWheelListener()
                    {
                    public void mouseWheelMoved( MouseWheelEvent e )
                        {
                        vmarker1spinner.setValue(((Double)vmarker1spinner.getValue()).doubleValue()
                                - e.getWheelRotation()*vMarkerStep );
                        }
                    });
                vmarker1panel.add (vmarker1spinner);

                SpringUtil.makeCompactGrid (vmarker1panel, 1, vmarker1panel.getComponentCount(), 2,2,0,0);
                mainpanelgroupa.add (vmarker1panel);



                JPanel vmarker2panel = new JPanel(new SpringLayout());
                vmarker2panel.setBorder(BorderFactory.createLineBorder(Color.lightGray,1));
                showVMarker2CB = new JCheckBox ("", showVMarker2);
                showVMarker2CB.setToolTipText("Select to show vertical marker 2");
                showVMarker2CB.addItemListener(new ItemListener(){
                    public void itemStateChanged(ItemEvent e) {
                        if (e.getStateChange() == ItemEvent.DESELECTED)
                                showVMarker2 = false;
                        else if (e.getStateChange() == ItemEvent.SELECTED)
                                showVMarker2 = true;
                        updatePlot();}});
                showVMarker2CB.setHorizontalAlignment(JLabel.CENTER);
                vmarker2panel.add (showVMarker2CB);

                vmarker2spinnermodel = new SpinnerNumberModel(new Double(vMarker2Value),  null, null, new Double(vMarkerStep));
                vmarker2spinner = new JSpinner(vmarker2spinnermodel);
                vmarker2spinner.setEditor(new JSpinner.NumberEditor(vmarker2spinner, "0.#########"));
                vmarker2spinner.setPreferredSize(new Dimension(75, 25));
                vmarker2spinner.setComponentPopupMenu(vmarkersteppopup);
                vmarker2spinner.setToolTipText("<html>"+"Enter vertical Marker 2 x-axis location"+"<br>"+
                    "or enter UT time in HH:MM or HH:MM:SS format and press 'Enter'"+"<br>"+
                    "---------------------------------------------"+"<br>"+
                    "Right click to set spinner stepsize"+"</html>");
                vmarker2spinner.addChangeListener (new ChangeListener()
                    {
                    public void stateChanged(ChangeEvent ev)
                        {
                        checkForUT(vmarker2spinner);
                        vMarker2Value = ((Double)vmarker2spinner.getValue()).doubleValue();
                        updatePlot();
                        }
                    });
                vmarker2spinner.addMouseWheelListener( new MouseWheelListener()
                    {
                    public void mouseWheelMoved( MouseWheelEvent e )
                        {
                        vmarker2spinner.setValue(((Double)vmarker2spinner.getValue()).doubleValue()
                                - e.getWheelRotation()*vMarkerStep );
                        }
                    });
                vmarker2panel.add (vmarker2spinner);

                SpringUtil.makeCompactGrid (vmarker2panel, 1, vmarker2panel.getComponentCount(), 2,2,0,0);
                mainpanelgroupa.add (vmarker2panel);

                SpringUtil.makeCompactGrid (mainpanelgroupa, 2, mainpanelgroupa.getComponentCount()/2, 5,5,5,5);
                mainpanel.add(mainpanelgroupa);

    // MAIN PANEL GROUP C START

                // TITLE GROUP
                        JPanel titlegroup = new JPanel (new SpringLayout());
                        titlegroup.setBorder (BorderFactory.createEmptyBorder(10,10,10,10));
                        titlegroup.setBorder(BorderFactory.createTitledBorder("Title"));

                                // TITLE RADIO BUTTON GROUP
                                JPanel titleradiopanelgroup = new JPanel (new SpringLayout());
                                        JRadioButton noTitleButton = new JRadioButton("None");
                                        if (!useTitle) noTitleButton.setSelected(true);
                                        else noTitleButton.setSelected(false);
                                titleradiopanelgroup.add (noTitleButton);
                                        JRadioButton useTitleButton = new JRadioButton("Custom");
                                        if (useTitle) useTitleButton.setSelected(true);
                                        else useTitleButton.setSelected(false);
                                titleradiopanelgroup.add (useTitleButton);
                                        ButtonGroup titleRadioGroup = new ButtonGroup();
                                        titleRadioGroup.add(noTitleButton);
                                        titleRadioGroup.add(useTitleButton);
                                        noTitleButton.addActionListener(new ActionListener(){
                                            public void actionPerformed(ActionEvent ae) {
                                                useTitle = false;
                                                updatePlot();}});
                                        useTitleButton.addActionListener(new ActionListener(){
                                            public void actionPerformed(ActionEvent ae) {
                                                useTitle =  true;
                                                updatePlot();}});
                                SpringUtil.makeCompactGrid (titleradiopanelgroup, 1, 2, 0,0,0,0);
                        titlegroup.add (titleradiopanelgroup);

                                // TITLE TEXT FIELD
                                titleField = new JTextField (title);
                                titleField.setBorder (BorderFactory.createEmptyBorder(0,4,0,4));
                                titleField.getDocument().addDocumentListener (new DocumentListener()
                                    {
                                    public void insertUpdate (DocumentEvent ev)
                                        {
                                        title = titleField.getText();
                                        updatePlot();
                                        }
                                    public void removeUpdate (DocumentEvent ev)
                                        {
                                        title = titleField.getText();
                                        updatePlot();
                                        }
                                    public void changedUpdate (DocumentEvent ev)
                                        {
                                        title = titleField.getText();
                                        updatePlot();
                                        }
                                    });
                                titleField.setPreferredSize (new Dimension(250,20));
                                titleField.setHorizontalAlignment (JTextField.LEFT);
                        titlegroup.add (titleField);

                                //MAKE TITLE Y-POSITION SLIDER
                                
                                titlePosYSlider = new JSlider(JSlider.HORIZONTAL, 0,1000, (int)(titlePosY));
                                titlePosYSlider.setBorder (BorderFactory.createEmptyBorder(10,0,0,0));
                                titlePosYSlider.addChangeListener (new ChangeListener()
                                    {
                                    public void stateChanged(ChangeEvent ev)
                                        {
                                        JSlider slider=(JSlider)ev.getSource();
                                        titlePosY=slider.getValue();
                                        updatePlot();
                                        }
                                    });
                                titlePosYSlider.addMouseWheelListener( new MouseWheelListener()
                                    {
                                    public void mouseWheelMoved( MouseWheelEvent e )
                                        {
                                        int newValue =  new Integer(((Integer)titlePosYSlider.getValue()).intValue() + e.getWheelRotation());
                                        if ((newValue >= 0) && (newValue <=1000)) titlePosYSlider.setValue(newValue);
                                        }
                                    });
                                titlePosYSlider.setPreferredSize (new Dimension(250,50));
                                titlePosYSlider.setInverted(false);
                                titlePosYSlider.setMajorTickSpacing(500);
                                titlePosYSlider.setMinorTickSpacing(50);
                                titlePosYSlider.setPaintTicks(true);
                                Hashtable<Integer, JLabel> titlePosYLabel = new Hashtable<Integer, JLabel>();
                                titlePosYLabel.put(0, new JLabel("Top"));
                                titlePosYLabel.put(500, new JLabel("Middle") );
                                titlePosYLabel.put(1000, new JLabel("Bottom") );
                                titlePosYSlider.setLabelTable( titlePosYLabel );
                                titlePosYSlider.setPaintTrack(true);
                                titlePosYSlider.setPaintLabels(true);
                        titlegroup.add (titlePosYSlider);

                                //MAKE TITLE X-POSITION SLIDER
                                
                                titlePosXSlider = new JSlider(0,500, (int) (titlePosX*500.0));
                                titlePosXSlider.setBorder (BorderFactory.createEmptyBorder(10,5,0,5));
                                titlePosXSlider.addChangeListener (new ChangeListener()
                                    {
                                    public void stateChanged(ChangeEvent ev)
                                        {
                                        JSlider slider=(JSlider)ev.getSource();
                                        titlePosX=slider.getValue()/500.0;
                                        updatePlot();
                                        }
                                    });
                                titlePosXSlider.addMouseWheelListener( new MouseWheelListener()
                                    {
                                    public void mouseWheelMoved( MouseWheelEvent e )
                                        {
                                        int newValue =  new Integer(((Integer)titlePosXSlider.getValue()).intValue() + e.getWheelRotation());
                                        if ((newValue >= 0) && (newValue <=500)) titlePosXSlider.setValue(newValue);
                                        }
                                    });
                                titlePosXSlider.setPreferredSize (new Dimension(250,50));
                                titlePosXSlider.setMajorTickSpacing(250);
                                titlePosXSlider.setMinorTickSpacing(25);
                                titlePosXSlider.setPaintTicks(true);
                                Hashtable<Integer, JLabel> titlePosXLabel = new Hashtable<Integer, JLabel>();
                                titlePosXLabel.put( 0, new JLabel("Left"));
                                titlePosXLabel.put( 250, new JLabel("Center") );
                                titlePosXLabel.put( 500, new JLabel("Right") );
                                titlePosXSlider.setLabelTable( titlePosXLabel );
                                titlePosXSlider.setPaintTrack(true);
                                titlePosXSlider.setPaintLabels(true);
                        titlegroup.add (titlePosXSlider);
                        SpringUtil.makeCompactGrid (titlegroup, 4, 1, 0,0,0,0);
		mainpanelgroupc.add (titlegroup);

                        //SUBTITLE GROUP
                        JPanel subtitlegroup = new JPanel (new SpringLayout());
                        subtitlegroup.setBorder(BorderFactory.createTitledBorder("Subtitle"));

                                // SUBTITLE RADIO BUTTON GROUP
                                JPanel subtitleradiopanelgroup = new JPanel (new SpringLayout());
                                        JRadioButton noSubtitleButton = new JRadioButton("None");
                                        if (!useSubtitle) noSubtitleButton.setSelected(true);
                                        else noSubtitleButton.setSelected(false);
                                subtitleradiopanelgroup.add (noSubtitleButton);
                                        JRadioButton useSubtitleButton = new JRadioButton("Custom");
                                        if (useSubtitle) useSubtitleButton.setSelected(true);
                                        else useSubtitleButton.setSelected(false);
                                subtitleradiopanelgroup.add (useSubtitleButton);
                                        ButtonGroup subtitleRadioGroup = new ButtonGroup();
                                        subtitleRadioGroup.add(noSubtitleButton);
                                        subtitleRadioGroup.add(useSubtitleButton);
                                        noSubtitleButton.addActionListener(new ActionListener(){
                                            public void actionPerformed(ActionEvent ae) {
                                                useSubtitle = false;
                                                updatePlot();}});
                                        useSubtitleButton.addActionListener(new ActionListener(){
                                            public void actionPerformed(ActionEvent ae) {
                                                useSubtitle =  true;
                                                updatePlot();}});
                                SpringUtil.makeCompactGrid (subtitleradiopanelgroup, 1, 2, 0,0,0,0);
                        subtitlegroup.add (subtitleradiopanelgroup);

                                // SUBTITLE TEXT FIELD
                                subtitleField = new JTextField (subtitle);
                                subtitleField.setBorder (BorderFactory.createEmptyBorder(0,4,0,4));
                                subtitleField.getDocument().addDocumentListener (new DocumentListener()
                                    {
                                    public void insertUpdate (DocumentEvent ev)
                                        {
                                        subtitle = subtitleField.getText();
                                        updatePlot();
                                        }
                                    public void removeUpdate (DocumentEvent ev)
                                        {
                                        subtitle = subtitleField.getText();
                                        updatePlot();
                                        }
                                    public void changedUpdate (DocumentEvent ev)
                                        {
                                        subtitle = subtitleField.getText();
                                        updatePlot();
                                        }
                                    });
                                subtitleField.setPreferredSize (new Dimension(250,20));
                                subtitleField.setHorizontalAlignment (JTextField.LEFT);

                        subtitlegroup.add (subtitleField);

                                //SUBTITLE Y-POSITION SLIDER
                                
                                subtitlePosYSlider = new JSlider(JSlider.HORIZONTAL, 0,1000, (int)(subtitlePosY));
                                subtitlePosYSlider.setBorder (BorderFactory.createEmptyBorder(10,0,0,0));
                                subtitlePosYSlider.addChangeListener (new ChangeListener()
                                    {
                                    public void stateChanged(ChangeEvent ev)
                                        {
                                        JSlider slider=(JSlider)ev.getSource();
                                        subtitlePosY=slider.getValue();
                                        updatePlot();
                                        }
                                    });
                                subtitlePosYSlider.addMouseWheelListener( new MouseWheelListener()
                                    {
                                    public void mouseWheelMoved( MouseWheelEvent e )
                                        {
                                        int newValue =  new Integer(((Integer)subtitlePosYSlider.getValue()).intValue() + e.getWheelRotation());
                                        if ((newValue >= 0) && (newValue <=1000)) subtitlePosYSlider.setValue(newValue);
                                        }
                                    });
                                subtitlePosYSlider.setPreferredSize (new Dimension(250,50));
                                subtitlePosYSlider.setInverted(false);
                                subtitlePosYSlider.setMajorTickSpacing(500);
                                subtitlePosYSlider.setMinorTickSpacing(50);
                                subtitlePosYSlider.setPaintTicks(true);
                                Hashtable<Integer, JLabel> subtitlePosYLabel = new Hashtable<Integer, JLabel>();

                                subtitlePosYLabel.put(0, new JLabel("Top"));
                                subtitlePosYLabel.put(500, new JLabel("Middle") );
                                subtitlePosYLabel.put(1000, new JLabel("Bottom") );
                                subtitlePosYSlider.setLabelTable( titlePosYLabel );
                                subtitlePosYSlider.setPaintTrack(true);
                                subtitlePosYSlider.setPaintLabels(true);
                        subtitlegroup.add (subtitlePosYSlider);

                                //SUBTITLE X-POSITION SLIDER
                                
                                subtitlePosXSlider = new JSlider(0,500, (int) (subtitlePosX*500.0));
                                subtitlePosXSlider.setBorder (BorderFactory.createEmptyBorder(10,5,0,5));
                                subtitlePosXSlider.addChangeListener (new ChangeListener()
                                    {
                                    public void stateChanged(ChangeEvent ev)
                                        {
                                        JSlider slider=(JSlider)ev.getSource();
                                        subtitlePosX=slider.getValue()/500.0;
                                        updatePlot();
                                        }
                                    });
                                subtitlePosXSlider.addMouseWheelListener( new MouseWheelListener()
                                    {
                                    public void mouseWheelMoved( MouseWheelEvent e )
                                        {
                                        int newValue =  new Integer(((Integer)subtitlePosXSlider.getValue()).intValue() + e.getWheelRotation());
                                        if ((newValue >= 0) && (newValue <=500)) subtitlePosXSlider.setValue(newValue);
                                        }
                                    });
                                subtitlePosXSlider.setPreferredSize (new Dimension(250,50));
                                subtitlePosXSlider.setMajorTickSpacing(250);
                                subtitlePosXSlider.setMinorTickSpacing(25);
                                subtitlePosXSlider.setPaintTicks(true);
                                Hashtable<Integer, JLabel> subtitlePosXLabel = new Hashtable<Integer, JLabel>();
                                subtitlePosXLabel.put( 0, new JLabel("Left"));
                                subtitlePosXLabel.put( 250, new JLabel("Center") );
                                subtitlePosXLabel.put( 500, new JLabel("Right") );
                                subtitlePosXSlider.setLabelTable( titlePosXLabel );
                                subtitlePosXSlider.setPaintTrack(true);
                                subtitlePosXSlider.setPaintLabels(true);
                        subtitlegroup.add (subtitlePosXSlider);
                        SpringUtil.makeCompactGrid (subtitlegroup, 4, 1, 0,0,0,0);
                mainpanelgroupc.add (subtitlegroup);


                        // LEGEND GROUP
                        legendgroup = new JPanel (new SpringLayout());
                        legendgroup.setBorder(BorderFactory.createTitledBorder("Legend"));

                        // LEGEND LABEL POPUP

                        legendpopup = new JPopupMenu();
                        JMenuItem legendshowscalingCB = new JCheckBoxMenuItem("Append scaling factor to legend for absolute curves",showLScaleInfo);
                        legendshowscalingCB.addItemListener(new ItemListener(){
                            public void itemStateChanged(ItemEvent e) {
                                if (e.getStateChange() == ItemEvent.DESELECTED)
                                        showLScaleInfo = false;
                                else if (e.getStateChange() == ItemEvent.SELECTED)
                                        showLScaleInfo = true;
                                updatePlot();}});
                        legendpopup.add(legendshowscalingCB);
                        JMenuItem legendshowrelscalingCB = new JCheckBoxMenuItem("Append scaling factor to legend for relative curves",showLRelScaleInfo);
                        legendshowrelscalingCB.addItemListener(new ItemListener(){
                            public void itemStateChanged(ItemEvent e) {
                                if (e.getStateChange() == ItemEvent.DESELECTED)
                                        showLRelScaleInfo = false;
                                else if (e.getStateChange() == ItemEvent.SELECTED)
                                        showLRelScaleInfo = true;
                                updatePlot();}});
                        legendpopup.add(legendshowrelscalingCB);                        
                        JMenuItem legendshowshiftCB = new JCheckBoxMenuItem("Append shift factor to legend for absolute curves",showLShiftInfo);
                        legendshowshiftCB.addItemListener(new ItemListener(){
                            public void itemStateChanged(ItemEvent e) {
                                if (e.getStateChange() == ItemEvent.DESELECTED)
                                        showLShiftInfo = false;
                                else if (e.getStateChange() == ItemEvent.SELECTED)
                                        showLShiftInfo = true;
                                updatePlot();}});
                        legendpopup.add(legendshowshiftCB);
                        JMenuItem legendshowrelshiftCB = new JCheckBoxMenuItem("Append shift factor to legend for relative curves",showLRelShiftInfo);
                        legendshowrelshiftCB.addItemListener(new ItemListener(){
                            public void itemStateChanged(ItemEvent e) {
                                if (e.getStateChange() == ItemEvent.DESELECTED)
                                        showLRelShiftInfo = false;
                                else if (e.getStateChange() == ItemEvent.SELECTED)
                                        showLRelShiftInfo = true;
                                updatePlot();}});
                        legendpopup.add(legendshowrelshiftCB);  
                        JMenuItem legendshowdetrendCB = new JCheckBoxMenuItem("Append detrend usage to legend",showLdetrendInfo);
                        legendshowdetrendCB.addItemListener(new ItemListener(){
                            public void itemStateChanged(ItemEvent e) {
                                if (e.getStateChange() == ItemEvent.DESELECTED)
                                        showLdetrendInfo = false;
                                else if (e.getStateChange() == ItemEvent.SELECTED)
                                        showLdetrendInfo = true;
                                updatePlot();}});
                        legendpopup.add(legendshowdetrendCB);  
                        JMenuItem legendshownormCB = new JCheckBoxMenuItem("Append normalize usage to legend",showLnormInfo);
                        legendshownormCB.addItemListener(new ItemListener(){
                            public void itemStateChanged(ItemEvent e) {
                                if (e.getStateChange() == ItemEvent.DESELECTED)
                                        showLnormInfo = false;
                                else if (e.getStateChange() == ItemEvent.SELECTED)
                                        showLnormInfo = true;
                                updatePlot();}});
                        legendpopup.add(legendshownormCB);                         
                        JMenuItem legendshowmmagCB = new JCheckBoxMenuItem("Append magnitude usage to legend",showLmmagInfo);
                        legendshowmmagCB.addItemListener(new ItemListener(){
                            public void itemStateChanged(ItemEvent e) {
                                if (e.getStateChange() == ItemEvent.DESELECTED)
                                        showLmmagInfo = false;
                                else if (e.getStateChange() == ItemEvent.SELECTED)
                                        showLmmagInfo = true;
                                updatePlot();}});
                        legendpopup.add(legendshowmmagCB);                        
                        JMenuItem legendshowbinCB = new JCheckBoxMenuItem("Append bin size to legend",showLBinInfo);
                        legendshowbinCB.addItemListener(new ItemListener(){
                            public void itemStateChanged(ItemEvent e) {
                                if (e.getStateChange() == ItemEvent.DESELECTED)
                                        showLBinInfo = false;
                                else if (e.getStateChange() == ItemEvent.SELECTED)
                                        showLBinInfo = true;
                                updatePlot();}});
                        legendpopup.add(legendshowbinCB);
                        JMenuItem legendshowsymbolCB = new JCheckBoxMenuItem("Append symbol description to legend",showLSymbolInfo);
                        legendshowsymbolCB.addItemListener(new ItemListener(){
                            public void itemStateChanged(ItemEvent e) {
                                if (e.getStateChange() == ItemEvent.DESELECTED)
                                        showLSymbolInfo = false;
                                else if (e.getStateChange() == ItemEvent.SELECTED)
                                        showLSymbolInfo = true;
                                updatePlot();}});
                        legendpopup.add(legendshowsymbolCB);
                        JMenuItem twolinelegendCB = new JCheckBoxMenuItem("Use two lines per legend",useTwoLineLegend);
                        twolinelegendCB.addItemListener(new ItemListener(){
                            public void itemStateChanged(ItemEvent e) {
                                if (e.getStateChange() == ItemEvent.DESELECTED)
                                        useTwoLineLegend = false;
                                else if (e.getStateChange() == ItemEvent.SELECTED)
                                        useTwoLineLegend = true;
                                updatePlot();}});
                        legendpopup.add(twolinelegendCB);


                        legendpopup.addPopupMenuListener (new PopupMenuListener() {
                                public void popupMenuCanceled(PopupMenuEvent popupMenuEvent) { }
                                public void popupMenuWillBecomeInvisible(PopupMenuEvent popupMenuEvent) { }
                                public void popupMenuWillBecomeVisible(PopupMenuEvent popupMenuEvent) { } });


                                // LEGEND RADIO BUTTON GROUP
                                JPanel legendradiopanelgroup = new JPanel ();
                                legendradiopanelgroup.setLayout(new BoxLayout(legendradiopanelgroup,BoxLayout.X_AXIS));
                                JLabel legendradiolabel = new JLabel("Align:");
                                legendradiolabel.setToolTipText("Right click to set legend preferences");
                                legendradiopanelgroup.add (legendradiolabel);
                                legendradiopanelgroup.add(Box.createHorizontalStrut(2));
                                        JRadioButton leftButton = new JRadioButton("Left");
                                        if (legendLeft)
                                                leftButton.setSelected(true);
                                        else
                                                leftButton.setSelected(false);
                                legendradiopanelgroup.add (leftButton);
                                legendradiopanelgroup.add(Box.createHorizontalStrut(2));
                                        JRadioButton centerButton = new JRadioButton("Center");
                                        if (!legendLeft && !legendRight)
                                                centerButton.setSelected(true);
                                        else
                                                centerButton.setSelected(false);
                                legendradiopanelgroup.add (centerButton);
                                legendradiopanelgroup.add(Box.createHorizontalStrut(2));
                                        JRadioButton rightButton = new JRadioButton("Right");
                                        if (legendRight)
                                                rightButton.setSelected(true);
                                        else
                                                rightButton.setSelected(false);
                                legendradiopanelgroup.add (rightButton);
                                legendradiopanelgroup.add(Box.createHorizontalStrut(2));
                                legendradiopanelgroup.add(Box.createHorizontalGlue());

                                        ButtonGroup legendRadioGroup = new ButtonGroup();
                                        legendRadioGroup.add(leftButton);
                                        legendRadioGroup.add(centerButton);
                                        legendRadioGroup.add(rightButton);
                                        leftButton.addActionListener(new ActionListener(){
                                            public void actionPerformed(ActionEvent ae) {
                                                legendLeft = true;
                                                legendRight = false;
                                                updatePlot();}});
                                        centerButton.addActionListener(new ActionListener(){
                                            public void actionPerformed(ActionEvent ae) {
                                                legendLeft = false;
                                                legendRight = false;
                                                updatePlot();}});
                                        rightButton.addActionListener(new ActionListener(){
                                            public void actionPerformed(ActionEvent ae) {
                                                legendLeft = false;
                                                legendRight = true;
                                                updatePlot();}});
                                configureIcon = createImageIcon("astroj/images/configure.png", "Configure legend options");
                                
                                legendconfigureButton = new JButton(configureIcon);
                                legendconfigureButton.setToolTipText("Configure legend options");
                                legendconfigureButton.setMargin(new Insets(0,0,0,0));
                                
                                legendconfigureButton.addActionListener(new ActionListener(){
                                            public void actionPerformed(ActionEvent ae) {
                                                legendpopup.show(legendconfigureButton, mainpanel.getX() , mainpanel.getY()+25);}});
                                legendradiopanelgroup.add(legendconfigureButton);
                                legendradiopanelgroup.add(Box.createHorizontalStrut(2));
//                                SpringUtil.makeCompactGrid (legendradiopanelgroup, 1, legendradiopanelgroup.getComponentCount(), 1,1,1,1);
                                leftButton.setComponentPopupMenu(legendpopup);
                                leftButton.setToolTipText("Right click to set legend preferences");
                                rightButton.setComponentPopupMenu(legendpopup);
                                rightButton.setToolTipText("Right click to set legend preferences");
                                centerButton.setComponentPopupMenu(legendpopup);
                                centerButton.setToolTipText("Right click to set legend preferences");
                        legendgroup.add (legendradiopanelgroup);

                                // LEGEND SLIDER LABEL
                                JLabel legendsliderlabel = new JLabel("Position");
                                legendsliderlabel.setBorder (BorderFactory.createEmptyBorder(5,0,0,0));
                                legendsliderlabel.setToolTipText("Right click to set legend preferences");
                                legendsliderlabel.setHorizontalAlignment (JLabel.CENTER);
                                legendsliderlabel.setComponentPopupMenu(legendpopup);

                        legendgroup.add (legendsliderlabel);

                                //LEGEND Y-POSITION SLIDER
                                
                                legendPosYSlider = new JSlider(JSlider.HORIZONTAL, 0,500, (int) (legendPosY*500.0));
                                legendPosYSlider.setBorder (BorderFactory.createEmptyBorder(10,0,0,0));
                                legendPosYSlider.addChangeListener (new ChangeListener()
                                    {
                                    public void stateChanged(ChangeEvent ev)
                                        {
                                        JSlider slider=(JSlider)ev.getSource();
                                        legendPosY=slider.getValue()/500.0;
                                        updatePlot();
                                        }
                                    });
                                legendPosYSlider.addMouseWheelListener( new MouseWheelListener()
                                    {
                                    public void mouseWheelMoved( MouseWheelEvent e )
                                        {
                                        int newValue =  new Integer(((Integer)legendPosYSlider.getValue()).intValue() + e.getWheelRotation());
                                        if ((newValue >= 0) && (newValue <=500)) legendPosYSlider.setValue(newValue);
                                        }
                                    });
                                legendPosYSlider.setPreferredSize (new Dimension(250,50));
                                legendPosYSlider.setInverted(false);
                                legendPosYSlider.setMajorTickSpacing(250);
                                legendPosYSlider.setMinorTickSpacing(25);
                                legendPosYSlider.setPaintTicks(true);
                                Hashtable<Integer, JLabel> legendPosYLabel = new Hashtable<Integer, JLabel>();
                                legendPosYLabel.put( 0, new JLabel("Top"));
                                legendPosYLabel.put( 250, new JLabel("Middle") );
                                legendPosYLabel.put( 500, new JLabel("Bottom") );
                                legendPosYSlider.setLabelTable( titlePosYLabel );
                                legendPosYSlider.setPaintTrack(true);
                                legendPosYSlider.setPaintLabels(true);
                        legendgroup.add (legendPosYSlider);

                                //LEGEND X-POSITION SLIDER
                                
                                legendPosXSlider = new JSlider(0,500, (int) (legendPosX*500.0));
                                legendPosXSlider.setBorder (BorderFactory.createEmptyBorder(10,5,0,5));
                                legendPosXSlider.addChangeListener (new ChangeListener()
                                    {
                                    public void stateChanged(ChangeEvent ev)
                                        {
                                        JSlider slider=(JSlider)ev.getSource();
                                        legendPosX=slider.getValue()/500.0;
                                        updatePlot();
                                        }
                                    });
                                legendPosXSlider.addMouseWheelListener( new MouseWheelListener()
                                    {
                                    public void mouseWheelMoved( MouseWheelEvent e )
                                        {
                                        int newValue =  new Integer(((Integer)legendPosXSlider.getValue()).intValue() + e.getWheelRotation());
                                        if ((newValue >= 0) && (newValue <=500)) legendPosXSlider.setValue(newValue);
                                        }
                                    });
                                legendPosXSlider.setPreferredSize (new Dimension(250,50));
                                legendPosXSlider.setMajorTickSpacing(250);
                                legendPosXSlider.setMinorTickSpacing(25);
                                legendPosXSlider.setPaintTicks(true);
                                Hashtable<Integer, JLabel> legendPosXLabel = new Hashtable<Integer, JLabel>();
                                legendPosXLabel.put( 0, new JLabel("Left"));
                                legendPosXLabel.put( 250, new JLabel("Center") );
                                legendPosXLabel.put( 500, new JLabel("Right") );
                                legendPosXSlider.setLabelTable( titlePosXLabel );
                                legendPosXSlider.setPaintTrack(true);
                                legendPosXSlider.setPaintLabels(true);
                        legendgroup.add (legendPosXSlider);
                        SpringUtil.makeCompactGrid (legendgroup, 4, 1, 0,0,0,0);

                mainpanelgroupc.add (legendgroup);

                SpringUtil.makeCompactGrid (mainpanelgroupc, 1, 3, 0,0,0,0);
                mainpanel.add(mainpanelgroupc);

    // X & Y AXIS LEGEND GROUP (GROUP D)

            //X-AXIS LEGEND GROUP
                        JPanel xlegendgroup = new JPanel (new SpringLayout());
                        xlegendgroup.setBorder(BorderFactory.createTitledBorder("X-Axis Label"));

                        // X-AXIS LABEL POPUP

                        xaxispopup = new JPopupMenu();
                        JMenuItem showxscalingCB = new JCheckBoxMenuItem("Append scaling factor to X-axis label",showXScaleInfo);

                        showxscalingCB.addItemListener(new ItemListener(){
                            public void itemStateChanged(ItemEvent e) {
                                if (e.getStateChange() == ItemEvent.DESELECTED)
                                        showXScaleInfo = false;
                                else if (e.getStateChange() == ItemEvent.SELECTED)
                                        showXScaleInfo = true;
                                updatePlot();}});
                        xaxispopup.add(showxscalingCB);
                        xaxispopup.addPopupMenuListener (new PopupMenuListener() {
                                public void popupMenuCanceled(PopupMenuEvent popupMenuEvent) { }
                                public void popupMenuWillBecomeInvisible(PopupMenuEvent popupMenuEvent) { }
                                public void popupMenuWillBecomeVisible(PopupMenuEvent popupMenuEvent) { } });

                        // X-AXIS LEGEND RADIO BUTTON GROUP
                                JPanel xlegendradiopanelgroup = new JPanel ();
                                xlegendradiopanelgroup.setLayout(new BoxLayout(xlegendradiopanelgroup,BoxLayout.X_AXIS));
                                JRadioButton xnoneButton = new JRadioButton("None");
                                if (!useXCustomName && !useXColumnName)
                                        xnoneButton.setSelected(true);
                                else
                                        xnoneButton.setSelected(false);
                                xlegendradiopanelgroup.add(Box.createHorizontalStrut(2));
                                xlegendradiopanelgroup.add(xnoneButton);
                                xlegendradiopanelgroup.add(Box.createHorizontalStrut(2));
                                JRadioButton xcolumnNameButton = new JRadioButton("Column Label");
                                if (useXColumnName)
                                        xcolumnNameButton.setSelected(true);
                                else
                                        xcolumnNameButton.setSelected(false);
                                xlegendradiopanelgroup.add(xcolumnNameButton);
                                xlegendradiopanelgroup.add(Box.createHorizontalStrut(2));
                                JRadioButton xcustomNameButton = new JRadioButton("Custom Label");
                                if (!useXColumnName && useXCustomName)
                                        xcustomNameButton.setSelected(true);
                                else
                                        xcustomNameButton.setSelected(false);
                                xlegendradiopanelgroup.add(xcustomNameButton);
                                xlegendradiopanelgroup.add(Box.createHorizontalStrut(2));
                                xlegendradiopanelgroup.add(Box.createGlue());
                                ButtonGroup xlegendRadioGroup = new ButtonGroup();
                                xlegendRadioGroup.add(xnoneButton);
                                xlegendRadioGroup.add(xcolumnNameButton);
                                xlegendRadioGroup.add(xcustomNameButton);
                                xnoneButton.addActionListener(new ActionListener(){
                                    public void actionPerformed(ActionEvent ae) {
                                        useXColumnName = false;
                                        useXCustomName = false;
                                        updatePlot();}});
                                xcustomNameButton.addActionListener(new ActionListener(){
                                    public void actionPerformed(ActionEvent ae) {
                                        useXColumnName = false;
                                        useXCustomName = true;
                                        updatePlot();}});
                                xcolumnNameButton.addActionListener(new ActionListener(){
                                    public void actionPerformed(ActionEvent ae) {
                                        useXColumnName = true;
                                        useXCustomName = false;
                                        updatePlot();}});
                                xlegendconfigureButton = new JButton(configureIcon);
                                xlegendconfigureButton.setToolTipText("Set X-axis label preferences");
                                xlegendconfigureButton.setMargin(new Insets(0,0,0,0));

                                xlegendconfigureButton.addActionListener(new ActionListener(){
                                            public void actionPerformed(ActionEvent ae) {
                                                xaxispopup.show(xlegendconfigureButton, mainpanel.getX() , mainpanel.getY()+25);}});
                                xlegendradiopanelgroup.add(xlegendconfigureButton);
                                xlegendradiopanelgroup.add(Box.createHorizontalStrut(2));
//                                SpringUtil.makeCompactGrid (xlegendradiopanelgroup, 1, xlegendradiopanelgroup.getComponentCount(), 1,1,1,1);
                                xnoneButton.setComponentPopupMenu(xaxispopup);
                                xnoneButton.setToolTipText("Right click to set X-axis label preferences");
                                xcustomNameButton.setComponentPopupMenu(xaxispopup);
                                xcustomNameButton.setToolTipText("Right click to set X-axis label preferences");
                                xcolumnNameButton.setComponentPopupMenu(xaxispopup);
                                xcolumnNameButton.setToolTipText("Right click to set X-axis label preferences");
                        xlegendgroup.add (xlegendradiopanelgroup);

                                // X LEGEND TEXT FIELD
                                xlegendField = new JTextField (xLegend);
                                xlegendField.setBorder (BorderFactory.createEmptyBorder(0,4,0,4));
                                xlegendField.getDocument().addDocumentListener (new DocumentListener()
                                    {
                                    public void insertUpdate (DocumentEvent ev)
                                        {
                                        xLegend = xlegendField.getText();
                                        updatePlot();
                                        }
                                    public void removeUpdate (DocumentEvent ev)
                                        {
                                        xLegend = xlegendField.getText();
                                        updatePlot();
                                        }
                                    public void changedUpdate (DocumentEvent ev)
                                        {
                                        xLegend = xlegendField.getText();
                                        updatePlot();
                                        }
                                    });
                                xlegendField.setPreferredSize (new Dimension(250,20));
                                xlegendField.setHorizontalAlignment (JTextField.LEFT);
                                xlegendField.setComponentPopupMenu(xaxispopup);
                                xlegendField.setToolTipText("Right click to set X-axis label preferences");

                        xlegendgroup.add (xlegendField);
                        SpringUtil.makeCompactGrid (xlegendgroup, 2, 1, 0,0,0,0);
                mainpanelgroupd.add (xlegendgroup);

            //Y-AXIS LEGEND GROUP
                        JPanel ylegendgroup = new JPanel (new SpringLayout());
                        ylegendgroup.setBorder(BorderFactory.createTitledBorder("Y-Axis Label"));

                        // Y-AXIS LABEL POPUP

                        yaxispopup = new JPopupMenu();
                        JMenuItem showyscalingCB = new JCheckBoxMenuItem("Append scaling factor to Y-axis label",showYScaleInfo);
                        showyscalingCB.addItemListener(new ItemListener(){
                            public void itemStateChanged(ItemEvent e) {
                                if (e.getStateChange() == ItemEvent.DESELECTED)
                                        showYScaleInfo = false;
                                else if (e.getStateChange() == ItemEvent.SELECTED)
                                        showYScaleInfo = true;
                                updatePlot();}});
                        yaxispopup.add(showyscalingCB);
                        JMenuItem showyshiftCB = new JCheckBoxMenuItem("Append shift factor to Y-axis label",showYShiftInfo);
                        showyshiftCB.addItemListener(new ItemListener(){
                            public void itemStateChanged(ItemEvent e) {
                                if (e.getStateChange() == ItemEvent.DESELECTED)
                                        showYShiftInfo = false;
                                else if (e.getStateChange() == ItemEvent.SELECTED)
                                        showYShiftInfo = true;
                                updatePlot();}});
                        yaxispopup.add(showyshiftCB);

                        JMenuItem showynormCB = new JCheckBoxMenuItem("Append normalization usage to Y-axis label",showYNormInfo);
                        showynormCB.addItemListener(new ItemListener(){
                            public void itemStateChanged(ItemEvent e) {
                                if (e.getStateChange() == ItemEvent.DESELECTED)
                                        showYNormInfo = false;
                                else if (e.getStateChange() == ItemEvent.SELECTED)
                                        showYNormInfo = true;
                                updatePlot();}});
                        yaxispopup.add(showynormCB);                         
                        
                        JMenuItem showymmagCB = new JCheckBoxMenuItem("Append magnitude usage to Y-axis label",showYmmagInfo);
                        showymmagCB.addItemListener(new ItemListener(){
                            public void itemStateChanged(ItemEvent e) {
                                if (e.getStateChange() == ItemEvent.DESELECTED)
                                        showYmmagInfo = false;
                                else if (e.getStateChange() == ItemEvent.SELECTED)
                                        showYmmagInfo = true;
                                updatePlot();}});
                        yaxispopup.add(showymmagCB);
                       
                        JMenuItem showybinCB = new JCheckBoxMenuItem("Append bin size to Y-axis label",showYBinInfo);
                        showybinCB.addItemListener(new ItemListener(){
                            public void itemStateChanged(ItemEvent e) {
                                if (e.getStateChange() == ItemEvent.DESELECTED)
                                        showYBinInfo = false;
                                else if (e.getStateChange() == ItemEvent.SELECTED)
                                        showYBinInfo = true;
                                updatePlot();}});
                        yaxispopup.add(showybinCB);
                        
                        JMenuItem showysymbolCB = new JCheckBoxMenuItem("Append symbol description to Y-axis label",showYSymbolInfo);
                        showysymbolCB.addItemListener(new ItemListener(){
                            public void itemStateChanged(ItemEvent e) {
                                if (e.getStateChange() == ItemEvent.DESELECTED)
                                        showYSymbolInfo = false;
                                else if (e.getStateChange() == ItemEvent.SELECTED)
                                        showYSymbolInfo = true;
                                updatePlot();}});
                        yaxispopup.add(showysymbolCB);
                        yaxispopup.addPopupMenuListener (new PopupMenuListener() {
                                public void popupMenuCanceled(PopupMenuEvent popupMenuEvent) { }
                                public void popupMenuWillBecomeInvisible(PopupMenuEvent popupMenuEvent) { }
                                public void popupMenuWillBecomeVisible(PopupMenuEvent popupMenuEvent) { } });

                        // Y-AXIS LEGEND RADIO BUTTON GROUP
                                JPanel ylegendradiopanelgroup = new JPanel ();
                                ylegendradiopanelgroup.setLayout(new BoxLayout(ylegendradiopanelgroup,BoxLayout.X_AXIS));
                                        JRadioButton ynoneButton = new JRadioButton("None");
                                        if (!useYCustomName && !useYColumnName)
                                                ynoneButton.setSelected(true);
                                        else
                                                ynoneButton.setSelected(false);
                                ylegendradiopanelgroup.add(ynoneButton);
                                ylegendradiopanelgroup.add(Box.createHorizontalStrut(2));
                                        JRadioButton ycolumnNameButton = new JRadioButton("Column Label");
                                        if (useYColumnName)
                                                ycolumnNameButton.setSelected(true);
                                        else
                                                ycolumnNameButton.setSelected(false);
                                ylegendradiopanelgroup.add(ycolumnNameButton);
                                ylegendradiopanelgroup.add(Box.createHorizontalStrut(2));
                                        JRadioButton ycustomNameButton = new JRadioButton("Custom Label");
                                        if (!useYColumnName && useYCustomName)
                                                ycustomNameButton.setSelected(true);
                                        else
                                                ycustomNameButton.setSelected(false);
                                ylegendradiopanelgroup.add(ycustomNameButton);
                                ylegendradiopanelgroup.add(Box.createHorizontalStrut(2));
                                ylegendradiopanelgroup.add(Box.createGlue());

                                ylegendconfigureButton = new JButton(configureIcon);
                                ylegendconfigureButton.setToolTipText("Set Y-axis label preferences");
                                ylegendconfigureButton.setMargin(new Insets(0,0,0,0));

                                ylegendconfigureButton.addActionListener(new ActionListener(){
                                            public void actionPerformed(ActionEvent ae) {
                                                yaxispopup.show(ylegendconfigureButton, mainpanel.getX() , mainpanel.getY()+25);}});
                                ylegendradiopanelgroup.add(ylegendconfigureButton);
                                ylegendradiopanelgroup.add(Box.createHorizontalStrut(2));

                                        ButtonGroup ylegendRadioGroup = new ButtonGroup();
                                        ylegendRadioGroup.add(ynoneButton);
                                        ylegendRadioGroup.add(ycustomNameButton);
                                        ylegendRadioGroup.add(ycolumnNameButton);
                                        ynoneButton.addActionListener(new ActionListener(){
                                            public void actionPerformed(ActionEvent ae) {
                                                useYColumnName = false;
                                                useYCustomName = false;
                                                updatePlot();}});
                                        ycustomNameButton.addActionListener(new ActionListener(){
                                            public void actionPerformed(ActionEvent ae) {
                                                useYColumnName = false;
                                                useYCustomName = true;
                                                updatePlot();}});
                                        ycolumnNameButton.addActionListener(new ActionListener(){
                                            public void actionPerformed(ActionEvent ae) {
                                                useYColumnName = true;
                                                useYCustomName = false;
                                                updatePlot();}});
//                                SpringUtil.makeCompactGrid (ylegendradiopanelgroup, 1, 3, 0,0,0,0);
                                ynoneButton.setComponentPopupMenu(yaxispopup);
                                ynoneButton.setToolTipText("Right click to set Y-axis label preferences");
                                ycustomNameButton.setComponentPopupMenu(yaxispopup);
                                ycustomNameButton.setToolTipText("Right click to set Y-axis label preferences");
                                ycolumnNameButton.setComponentPopupMenu(yaxispopup);
                                ycolumnNameButton.setToolTipText("Right click to set Y-axis label preferences");
                        ylegendgroup.add (ylegendradiopanelgroup);

                                // Y-LEGEND TEXT FIELD
                                ylegendField = new JTextField (yLegend);
                                ylegendField.setBorder (BorderFactory.createEmptyBorder(0,4,0,4));
                                ylegendField.getDocument().addDocumentListener (new DocumentListener()
                                    {
                                    public void insertUpdate (DocumentEvent ev)
                                        {
                                        yLegend = ylegendField.getText();
                                        updatePlot();
                                        }
                                    public void removeUpdate (DocumentEvent ev)
                                        {
                                        yLegend = ylegendField.getText();
                                        updatePlot();
                                        }
                                    public void changedUpdate (DocumentEvent ev)
                                        {
                                        yLegend = ylegendField.getText();
                                        updatePlot();
                                        }
                                    });
                                ylegendField.setPreferredSize (new Dimension(250,20));
                                ylegendField.setHorizontalAlignment (JTextField.LEFT);
                                ylegendField.setComponentPopupMenu(yaxispopup);
                                ylegendField.setToolTipText("Right click to set Y-axis label preferences");

                        ylegendgroup.add (ylegendField);
                        SpringUtil.makeCompactGrid (ylegendgroup, 2, 1, 0,0,0,0);
                mainpanelgroupd.add (ylegendgroup);

            //TRIM DATASET SUBPANEL

                JPanel trimdatapanel = new JPanel (new SpringLayout());
                trimdatapanel.setBorder(BorderFactory.createTitledBorder("Trim Data Samples"));

                //TRIM DATA FROM HEAD STEPSIZE POPUP

                trimdataheadsteppopup = new JPopupMenu();
                JPanel trimdataheadsteppanel = new JPanel();
                trimdataheadstepmodel = new SpinnerListModel(integerspinnerscalelist);
                trimdataheadstepspinner = new JSpinner(trimdataheadstepmodel);
                trimdataheadstepspinner.setValue(intConvertToText(excludedHeadSamplesStep));
                trimdataheadstepspinner.addChangeListener (new ChangeListener()
                    {
                    public void stateChanged(ChangeEvent ev)
                        {
                        excludedHeadSamplesStep = Integer.parseInt(((String)trimdataheadstepspinner.getValue()).trim());
                        trimdataheadspinner.setModel(new SpinnerNumberModel(new Integer(excludedHeadSamples), new Integer(0), null, new Integer(excludedHeadSamplesStep)));
                        Prefs.set("plot.excludedHeadSamplesStep",excludedHeadSamplesStep);
                        }
                    });

                JLabel trimdataheadsteplabel = new JLabel ("Stepsize:");
                trimdataheadsteppanel.add(trimdataheadsteplabel);
                trimdataheadsteppanel.add(trimdataheadstepspinner);
                trimdataheadsteppopup.add(trimdataheadsteppanel);
                trimdataheadsteppopup.setLightWeightPopupEnabled(false);
                trimdataheadsteppopup.addPopupMenuListener (new PopupMenuListener()
                    {
                    public void popupMenuCanceled(PopupMenuEvent popupMenuEvent) {
                            trimdataheadsteppopup.setVisible(false);
                            }

                    public void popupMenuWillBecomeInvisible(PopupMenuEvent popupMenuEvent) {
                            trimdataheadsteppopup.setVisible(true);
                            }

                    public void popupMenuWillBecomeVisible(PopupMenuEvent popupMenuEvent) {
                            trimdataheadsteppopup.setVisible(false);
                            }
                    });

                JLabel trimdataheadlabel = new JLabel ("Head");
                trimdataheadlabel.setHorizontalAlignment(JLabel.RIGHT);
                trimdataheadlabel.setPreferredSize(new Dimension(50, 25));
                trimdataheadlabel.setToolTipText("Right click to set spinner stepsize");
                trimdataheadlabel.setComponentPopupMenu(trimdataheadsteppopup);
                trimdatapanel.add (trimdataheadlabel);

                trimdataheadmodel = new SpinnerNumberModel(new Integer(excludedHeadSamples), new Integer(0), null, new Integer(excludedHeadSamplesStep));
                trimdataheadspinner = new JSpinner(trimdataheadmodel);
                trimdataheadspinner.setComponentPopupMenu(trimdataheadsteppopup);
                trimdataheadspinner.setToolTipText("Right click to set spinner stepsize");
                trimdataheadspinner.addChangeListener (new ChangeListener()
                    {
                    public void stateChanged(ChangeEvent ev)
                        {
                        excludedHeadSamples = ((Integer)trimdataheadspinner.getValue()).intValue();
                        Prefs.set("plot.excludedHeadSamples", excludedHeadSamples);
                        updatePlot();
                        }
                    });
                trimdataheadspinner.addMouseWheelListener( new MouseWheelListener()
                    {
                    public void mouseWheelMoved( MouseWheelEvent e )
                        {
                        int newValue =  new Integer(((Integer)trimdataheadspinner.getValue()).intValue() - e.getWheelRotation());
                        if (newValue >= 0) trimdataheadspinner.setValue(newValue);
                        }
                    });
                trimdatapanel.add (trimdataheadspinner);

                //TRIM DATA FROM TAIL STEPSIZE POPUP

                trimdatatailsteppopup = new JPopupMenu();
                JPanel trimdatatailsteppanel = new JPanel();
                trimdatatailstepmodel = new SpinnerListModel(integerspinnerscalelist);
                trimdatatailstepspinner = new JSpinner(trimdatatailstepmodel);
                trimdatatailstepspinner.setValue(intConvertToText(excludedTailSamplesStep));
                trimdatatailstepspinner.addChangeListener (new ChangeListener()
                    {
                    public void stateChanged(ChangeEvent ev)
                        {
                        excludedTailSamplesStep = Integer.parseInt(((String)trimdatatailstepspinner.getValue()).trim());
                        trimdatatailspinner.setModel(new SpinnerNumberModel(new Integer(excludedTailSamples), new Integer(0), null, new Integer(excludedTailSamplesStep)));
                        Prefs.set("plot.excludedTailSamplesStep",excludedTailSamplesStep);
                        }
                    });

                JLabel trimdatatailsteplabel = new JLabel ("Stepsize:");
                trimdatatailsteppanel.add(trimdatatailsteplabel);
                trimdatatailsteppanel.add(trimdatatailstepspinner);
                trimdatatailsteppopup.add(trimdatatailsteppanel);
                trimdatatailsteppopup.setLightWeightPopupEnabled(false);
                trimdatatailsteppopup.addPopupMenuListener (new PopupMenuListener()
                    {
                    public void popupMenuCanceled(PopupMenuEvent popupMenuEvent) {
                            trimdatatailsteppopup.setVisible(false);
                            }

                    public void popupMenuWillBecomeInvisible(PopupMenuEvent popupMenuEvent) {
                            trimdatatailsteppopup.setVisible(true);
                            }

                    public void popupMenuWillBecomeVisible(PopupMenuEvent popupMenuEvent) {
                            trimdatatailsteppopup.setVisible(false);
                            }
                    });

                JLabel trimdatataillabel = new JLabel ("Tail");
                trimdatataillabel.setHorizontalAlignment(JLabel.RIGHT);
                trimdatataillabel.setPreferredSize(new Dimension(50, 25));
                trimdatataillabel.setToolTipText("Right click to set spinner stepsize");
                trimdatataillabel.setComponentPopupMenu(trimdatatailsteppopup);
                trimdatapanel.add (trimdatataillabel);

                trimdatatailmodel = new SpinnerNumberModel(new Integer(excludedTailSamples), new Integer(0), null, new Integer(excludedTailSamplesStep));
                trimdatatailspinner = new JSpinner(trimdatatailmodel);
                trimdatatailspinner.setComponentPopupMenu(trimdatatailsteppopup);
                trimdatatailspinner.setToolTipText("Right click to set spinner stepsize");
                trimdatatailspinner.addChangeListener (new ChangeListener()
                    {
                    public void stateChanged(ChangeEvent ev)
                        {
                        excludedTailSamples = ((Integer)trimdatatailspinner.getValue()).intValue();
                        Prefs.set("plot.excludedTailSamples", excludedTailSamples);
                        updatePlot();
                        }
                    });
                trimdatatailspinner.addMouseWheelListener( new MouseWheelListener()
                    {
                    public void mouseWheelMoved( MouseWheelEvent e )
                        {
                        int newValue =  new Integer(((Integer)trimdatatailspinner.getValue()).intValue() - e.getWheelRotation());
                        if (newValue >= 0) trimdatatailspinner.setValue(newValue);
                        }
                    });
                trimdatapanel.add (trimdatatailspinner);
                SpringUtil.makeCompactGrid (trimdatapanel, 2, 2, 0,0,0,0);
                mainpanelgroupd.add (trimdatapanel);

                SpringUtil.makeCompactGrid (mainpanelgroupd, 1, 3, 0,0,0,0);
                mainpanel.add(mainpanelgroupd);


    // MAIN PANEL SCALING GROUP START (GROUP E)

            //X-SCALING SUBPANEL

                JPanel xscalingpanel = new JPanel (new SpringLayout());
                xscalingpanel.setBorder(BorderFactory.createTitledBorder("X-axis"));
                autoxButton = new JRadioButton("Auto X-range");
                autoxButton.setHorizontalAlignment(JLabel.CENTER);
                firstxButton = new JRadioButton("First X-value as min");
                firstxButton.setHorizontalAlignment(JLabel.CENTER);
                customxButton = new JRadioButton("Custom X-range");
                customxButton.setHorizontalAlignment(JLabel.CENTER);
                if (autoScaleX)
                        {
                        autoxButton.setSelected(true);
                        firstxButton.setSelected(false);
                        customxButton.setSelected(false);
                        }
                else if (useFirstX)
                        {
                        autoxButton.setSelected(false);
                        firstxButton.setSelected(true);
                        customxButton.setSelected(false);
                        }
                else
                        {
                        autoxButton.setSelected(false);
                        firstxButton.setSelected(false);
                        customxButton.setSelected(true);
                        }

                xscalingpanel.add (autoxButton);
                xscalingpanel.add (firstxButton);
                xscalingpanel.add (customxButton);

                ButtonGroup xaxisradiogroup = new ButtonGroup();
                        xaxisradiogroup.add(autoxButton);
                        xaxisradiogroup.add(firstxButton);
                        xaxisradiogroup.add(customxButton);
                autoxButton.addActionListener(new ActionListener(){
                    public void actionPerformed(ActionEvent ae) {
                        autoScaleX = true;
                        useFirstX = false;
                        updatePlot();}});
                firstxButton.addActionListener(new ActionListener(){
                    public void actionPerformed(ActionEvent ae) {
                        autoScaleX = false;
                        useFirstX = true;
                        updatePlot();}});
                customxButton.addActionListener(new ActionListener(){
                    public void actionPerformed(ActionEvent ae) {
                        autoScaleX = false;
                        useFirstX = false;
                        updatePlot();}});

                ImageIcon grabautoIcon = createImageIcon("astroj/images/grabautoscale.png", "Transfer autoscale values to custom scale values");

                JPanel grabautoxpanel = new JPanel (new SpringLayout());
                grabautoxbutton = new JButton(grabautoIcon);
                grabautoxbutton.setMargin(new Insets(0,0,0,0));
                Dimension d = new Dimension(60,15);
                grabautoxbutton.setMaximumSize(d);
                grabautoxpanel.setBorder(BorderFactory.createEmptyBorder(0, 0, 0, 0));
                grabautoxbutton.setToolTipText("Transfer 'Auto X-range' min and max values to 'Custom X-range'");

                grabautoxbutton.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                        xMin = xautoscalemin;
                        xminspinner.setValue(xMin);
                        xMax = xautoscalemax;
                        xmaxspinner.setValue(xMax);
                        xWidth = xautoscalemax - xautoscalemin;
                        xwidthspinner.setValue(xWidth);
                        autoxButton.setSelected(false);
                        firstxButton.setSelected(false);
                        customxButton.setSelected(true);
                        autoScaleX = false;
                        useFirstX = false;
                        updatePlot();
                    }
                });
                grabautoxpanel.add (grabautoxbutton);

                SpringUtil.makeCompactGrid (grabautoxpanel,1,1, 30,0,10,2);
                xscalingpanel.add (grabautoxpanel);

                //XWIDTH STEPSIZE POPUP

                xwidthsteppopup = new JPopupMenu();
                JPanel xwidthsteppanel = new JPanel();
                xwidthstepmodel = new SpinnerListModel(spinnerscalelist);
                xwidthstepspinner = new JSpinner(xwidthstepmodel);
                xwidthstepspinner.setValue(convertToText(xWidthStep));
                xwidthstepspinner.addChangeListener (new ChangeListener()
                    {
                    public void stateChanged(ChangeEvent ev)
                        {
                        xWidthStep = Double.parseDouble((String)xwidthstepspinner.getValue());
                        xwidthspinner.setModel(new SpinnerNumberModel(new Double(xWidth), new Double(0), null, new Double(xWidthStep)));
                        xwidthspinner.setEditor(new JSpinner.NumberEditor(xwidthspinner, "0.#########"));
                        Prefs.set("plot.xWidthStep",xWidthStep);
                        }
                    });

                JLabel xwidthsteplabel = new JLabel ("Stepsize:");
                xwidthsteppanel.add(xwidthsteplabel);
                xwidthsteppanel.add(xwidthstepspinner);
                xwidthsteppopup.add(xwidthsteppanel);
                xwidthsteppopup.setLightWeightPopupEnabled(false);
                xwidthsteppopup.setLocation(500, 500);
                xwidthsteppopup.addPopupMenuListener (new PopupMenuListener()
                    {
                    public void popupMenuCanceled(PopupMenuEvent popupMenuEvent) {
                            xwidthsteppopup.setVisible(false);
                            }

                    public void popupMenuWillBecomeInvisible(PopupMenuEvent popupMenuEvent) {
                            xwidthsteppopup.setVisible(true);
                            }

                    public void popupMenuWillBecomeVisible(PopupMenuEvent popupMenuEvent) {
                            xwidthsteppopup.setVisible(false);
                            }
                    });

                JPanel xwidthpanel = new JPanel (new SpringLayout());
                JLabel xwidthlabel = new JLabel ("X-width");
                xwidthlabel.setHorizontalAlignment(JLabel.RIGHT);
                xwidthlabel.setPreferredSize(new Dimension(60, 25));
                xwidthlabel.setComponentPopupMenu(xwidthsteppopup);
                xwidthlabel.setToolTipText("Right click to set spinner stepsize");
                xwidthpanel.add (xwidthlabel);
                xwidthspinnermodel = new SpinnerNumberModel(new Double(xWidth), new Double(0.000001), null, new Double(xWidthStep));
                xwidthspinner = new JSpinner(xwidthspinnermodel);
                xwidthspinner.setEditor(new JSpinner.NumberEditor(xwidthspinner, "0.#########"));
                xwidthspinner.setPreferredSize(new Dimension(100, 25));
                xwidthspinner.setToolTipText("Right click to set spinner stepsize");
                xwidthspinner.addChangeListener (new ChangeListener()
                    {
                    public void stateChanged(ChangeEvent ev)
                        {
                        xWidth = ((Double)xwidthspinner.getValue()).doubleValue();
                        updatePlot();
                        }
                    });
                xwidthspinner.addMouseWheelListener( new MouseWheelListener()
                    {
                    public void mouseWheelMoved( MouseWheelEvent e )
                        {
                        double newvalue = ((Double)xwidthspinner.getValue()).doubleValue() - e.getWheelRotation()*xWidthStep ;
                        if (newvalue > 0)
                            xwidthspinner.setValue(newvalue);
                        else
                            xwidthspinner.setValue(0.000001);
                        }
                    });
                xwidthpanel.add (xwidthspinner);
                xwidthspinner.setComponentPopupMenu(xwidthsteppopup);
                SpringUtil.makeCompactGrid (xwidthpanel, 1, 2, 2,2,0,0);
                xscalingpanel.add (xwidthpanel);

                //XMAX STEPSIZE POPUP

                xmaxsteppopup = new JPopupMenu();
                JPanel xmaxsteppanel = new JPanel();
                xmaxstepmodel = new SpinnerListModel(spinnerscalelist);
                xmaxstepspinner = new JSpinner(xmaxstepmodel);
                xmaxstepspinner.setValue(convertToText(xMaxStep));
                xmaxstepspinner.addChangeListener (new ChangeListener()
                    {
                    public void stateChanged(ChangeEvent ev)
                        {
                        xMaxStep = Double.parseDouble((String)xmaxstepspinner.getValue());
                        xmaxspinner.setModel(new SpinnerNumberModel(new Double(xMax), null, null, new Double(xMaxStep)));
                        xmaxspinner.setEditor(new JSpinner.NumberEditor(xmaxspinner, "0.#########"));
                        Prefs.set("plot.xMaxStep",xMaxStep);
                        }
                    });

                JLabel xmaxsteplabel = new JLabel ("Stepsize:");
                xmaxsteppanel.add(xmaxsteplabel);
                xmaxsteppanel.add(xmaxstepspinner);
                xmaxsteppopup.add(xmaxsteppanel);
                xmaxsteppopup.setLightWeightPopupEnabled(false);
                xmaxsteppopup.addPopupMenuListener (new PopupMenuListener()
                    {
                    public void popupMenuCanceled(PopupMenuEvent popupMenuEvent) {
                            xmaxsteppopup.setVisible(false);
                            }

                    public void popupMenuWillBecomeInvisible(PopupMenuEvent popupMenuEvent) {
                            xmaxsteppopup.setVisible(true);
                            }

                    public void popupMenuWillBecomeVisible(PopupMenuEvent popupMenuEvent) {
                            xmaxsteppopup.setVisible(false);
                            }
                    });

                JPanel customxmaxpanel = new JPanel(new SpringLayout());
                JLabel xmaxlabel = new JLabel ("X-max");
                xmaxlabel.setHorizontalAlignment(JLabel.RIGHT);
                xmaxlabel.setPreferredSize(new Dimension(50, 25));
                xmaxlabel.setToolTipText("Right click to set spinner stepsize");
                xmaxlabel.setComponentPopupMenu(xmaxsteppopup);
                customxmaxpanel.add (xmaxlabel);

                xmaxspinnermodel = new SpinnerNumberModel(new Double(xMax),null, null, new Double(xMaxStep));
                xmaxspinner = new JSpinner(xmaxspinnermodel);
                xmaxspinner.setEditor(new JSpinner.NumberEditor(xmaxspinner, "0.#########"));
                xmaxspinner.setPreferredSize(new Dimension(100, 25));
                xmaxspinner.setComponentPopupMenu(xmaxsteppopup);
                xmaxspinner.setToolTipText("Right click to set spinner stepsize");
                xmaxspinner.addChangeListener (new ChangeListener()
                    {
                    public void stateChanged(ChangeEvent ev)
                        {
                        xMax = ((Double)xmaxspinner.getValue()).doubleValue();
                        updatePlot();
                        }
                    });
                xmaxspinner.addMouseWheelListener( new MouseWheelListener()
                    {
                    public void mouseWheelMoved( MouseWheelEvent e )
                        {
                        xmaxspinner.setValue(((Double)xmaxspinner.getValue()).doubleValue()
                                - e.getWheelRotation()*xMaxStep );
                        }
                    });
                customxmaxpanel.add (xmaxspinner);
                SpringUtil.makeCompactGrid (customxmaxpanel, 1, 2, 2,2,0,0);
                xscalingpanel.add (customxmaxpanel);

                JPanel xmultiplierpanel = new JPanel(new SpringLayout());
                xmultiplierpanel.setBorder(BorderFactory.createLineBorder(Color.lightGray));
                JLabel xmultiplierlabel = new JLabel ("X x 1E");
                xmultiplierlabel.setHorizontalAlignment(JLabel.RIGHT);
                xmultiplierpanel.setMaximumSize(new Dimension(250, 30));
                xmultiplierpanel.add (xmultiplierlabel);
                xmultiplierlabel.setToolTipText("Sets the X-axis multiplication factor");
                xmultipliermodel = new SpinnerNumberModel(new Integer(xExponent),null, null, new Integer(1));
                xmultiplierspinner = new JSpinner(xmultipliermodel);
                xmultiplierspinner.setMaximumSize(new Dimension (35,25));
                xmultiplierspinner.setToolTipText("Sets the X-axis multiplication factor");
                xmultiplierspinner.addChangeListener (new ChangeListener()
                    {
                    public void stateChanged(ChangeEvent ev)
                        {
                        xExponent = ((Integer)xmultiplierspinner.getValue()).intValue();
                        updatePlot();
                        }
                    });
                xmultiplierspinner.addMouseWheelListener( new MouseWheelListener()
                    {
                    public void mouseWheelMoved( MouseWheelEvent e )
                        {
                        xmultiplierspinner.setValue( new Integer(((Integer)xmultiplierspinner.getValue()).intValue()
                                - e.getWheelRotation()) );
                        }
                    });
                xmultiplierpanel.add (xmultiplierspinner);
                SpringUtil.makeCompactGrid (xmultiplierpanel, 1,2, 2,2,2,0);
                xscalingpanel.add (xmultiplierpanel);

                //XMIN STEPSIZE POPUP

                xminsteppopup = new JPopupMenu();
                JPanel xminsteppanel = new JPanel();
                xminstepmodel = new SpinnerListModel(spinnerscalelist);
                xminstepspinner = new JSpinner(xminstepmodel);
                xminstepspinner.setValue(convertToText(xMinStep));
                xminstepspinner.addChangeListener (new ChangeListener()
                    {
                    public void stateChanged(ChangeEvent ev)
                        {
                        xMinStep = Double.parseDouble((String)xminstepspinner.getValue());
                        xminspinner.setModel(new SpinnerNumberModel(new Double(xMin), null, null, new Double(xMinStep)));
                        xminspinner.setEditor(new JSpinner.NumberEditor(xminspinner, "0.#########"));
                        Prefs.set("plot.xMinStep",xMinStep);
                        }
                    });

                JLabel xminsteplabel = new JLabel ("Stepsize:");
                xminsteppanel.add(xminsteplabel);
                xminsteppanel.add(xminstepspinner);
                xminsteppopup.add(xminsteppanel);
                xminsteppopup.setLightWeightPopupEnabled(false);
                xminsteppopup.addPopupMenuListener (new PopupMenuListener()
                    {
                    public void popupMenuCanceled(PopupMenuEvent popupMenuEvent) {
                            xminsteppopup.setVisible(false);
                            }

                    public void popupMenuWillBecomeInvisible(PopupMenuEvent popupMenuEvent) {
                            xminsteppopup.setVisible(true);
                            }

                    public void popupMenuWillBecomeVisible(PopupMenuEvent popupMenuEvent) {
                            xminsteppopup.setVisible(false);
                            }
                    });

                ImageIcon grabfirstxIcon = createImageIcon("astroj/images/grabautoscale.png", "Transfer autoscale values to custom scale values");
                JPanel grabfirstxpanel = new JPanel (new SpringLayout());
                JButton grabfirstxbutton = new JButton(grabfirstxIcon);
                grabfirstxbutton.setMargin(new Insets(0,0,0,0));
                grabfirstxbutton.setMaximumSize(new Dimension(60,15));
                grabfirstxpanel.setBorder(BorderFactory.createEmptyBorder(0, 0, 0, 0));
                grabfirstxbutton.setToolTipText("Transfer 'First-X' min and max values to 'Custom X-range'");

                grabfirstxbutton.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                        xMin = firstXmin;
                        xminspinner.setValue(firstXmin);
                        xMax = firstXmin + xWidth;
                        xmaxspinner.setValue(xMax);
                        autoxButton.setSelected(false);
                        firstxButton.setSelected(false);
                        customxButton.setSelected(true);
                        autoScaleX = false;
                        useFirstX = false;
                        updatePlot();
                    }
                });
                grabfirstxpanel.add (grabfirstxbutton);

                SpringUtil.makeCompactGrid (grabfirstxpanel,1,1, 72,2,22,0);                
                xscalingpanel.add (grabfirstxpanel);


                JPanel customxminpanel = new JPanel(new SpringLayout());
                JLabel xminlabel = new JLabel ("X-min");
                xminlabel.setHorizontalAlignment(JLabel.RIGHT);
                xminlabel.setPreferredSize(new Dimension(50, 25));
                xminlabel.setToolTipText("Right click to set spinner stepsize");
                xminlabel.setComponentPopupMenu(xminsteppopup);
                customxminpanel.add (xminlabel);

                xminspinnermodel = new SpinnerNumberModel(new Double(xMin),null, null, new Double(xMinStep));
                xminspinner = new JSpinner(xminspinnermodel);
                xminspinner.setEditor(new JSpinner.NumberEditor(xminspinner, "0.#########"));
                xminspinner.setPreferredSize(new Dimension(100, 25));
                xminspinner.setComponentPopupMenu(xminsteppopup);
                xminspinner.setToolTipText("Right click to set spinner stepsize");
                xminspinner.addChangeListener (new ChangeListener()
                    {
                    public void stateChanged(ChangeEvent ev)
                        {
                        xMin = ((Double)xminspinner.getValue()).doubleValue();
                        updatePlot();
                        }
                    });
                xminspinner.addMouseWheelListener( new MouseWheelListener()
                    {
                    public void mouseWheelMoved( MouseWheelEvent e )
                        {
                        xminspinner.setValue(((Double)xminspinner.getValue()).doubleValue()
                                - e.getWheelRotation()*xMinStep );
                        }
                    });
                customxminpanel.add (xminspinner);
                SpringUtil.makeCompactGrid (customxminpanel, 1, 2, 2,2,0,0);
                xscalingpanel.add (customxminpanel);

                SpringUtil.makeCompactGrid (xscalingpanel, 3, 3, 0,0,0,0);

                mainpanelgroupe.add (xscalingpanel);

            //Y-SCALING SUBPANEL

                JPanel yscalingpanel = new JPanel (new SpringLayout());
                yscalingpanel.setBorder(BorderFactory.createTitledBorder("Y-axis"));
                autoyButton = new JRadioButton("Auto Y-range");
                autoyButton.setHorizontalAlignment(JLabel.CENTER);
                customyButton = new JRadioButton("Custom Y-range");
                customyButton.setHorizontalAlignment(JLabel.CENTER);
                if (autoScaleY)
                        {
                        autoyButton.setSelected(true);
                        customyButton.setSelected(false);
                        }
                else
                        {
                        autoyButton.setSelected(false);
                        customyButton.setSelected(true);
                        }

                yscalingpanel.add (autoyButton);
                yscalingpanel.add (customyButton);

                ButtonGroup yaxisradiogroup = new ButtonGroup();
                        yaxisradiogroup.add(autoyButton);
                        yaxisradiogroup.add(customyButton);
                autoyButton.addActionListener(new ActionListener(){
                    public void actionPerformed(ActionEvent ae) {
                        autoScaleY = true;
                        updatePlot();}});
                customyButton.addActionListener(new ActionListener(){
                    public void actionPerformed(ActionEvent ae) {
                        autoScaleY = false;
                        updatePlot();}});

                JPanel grabautoypanel = new JPanel (new SpringLayout());
                grabautoybutton = new JButton(grabautoIcon);
                grabautoybutton.setMargin(new Insets(0,0,0,0));
                grabautoybutton.setMaximumSize(d);
                grabautoypanel.setBorder(BorderFactory.createEmptyBorder(0, 0, 0, 0));
                grabautoybutton.setToolTipText("Transfer auto scale Y-range values to custom scale Y-range values");
                grabautoybutton.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                        yMin = yautoscalemin;
                        yminspinner.setValue(yMin);
                        yMax = yautoscalemax;
                        ymaxspinner.setValue(yMax);
                        autoyButton.setSelected(false);
                        customyButton.setSelected(true);
                        autoScaleY = false;
                        updatePlot();
                    }
                    });
                grabautoypanel.add (grabautoybutton);
                SpringUtil.makeCompactGrid (grabautoypanel,1,1, 30,0,10,2);
                yscalingpanel.add (grabautoypanel);

                //YMAX STEPSIZE POPUP

                ymaxsteppopup = new JPopupMenu();
                JPanel ymaxsteppanel = new JPanel();
                ymaxstepmodel = new SpinnerListModel(spinnerscalelist);
                ymaxstepspinner = new JSpinner(ymaxstepmodel);
                ymaxstepspinner.setValue(convertToText(yMaxStep));
                ymaxstepspinner.addChangeListener (new ChangeListener()
                    {
                    public void stateChanged(ChangeEvent ev)
                        {
                        yMaxStep = Double.parseDouble((String)ymaxstepspinner.getValue());
                        ymaxspinner.setModel(new SpinnerNumberModel(new Double(yMax), null, null, new Double(yMaxStep)));
                        ymaxspinner.setEditor(new JSpinner.NumberEditor(ymaxspinner, "0.#########"));
                        Prefs.set("plot.yMaxStep",yMaxStep);
                        }
                    });

                JLabel ymaxsteplabel = new JLabel ("Stepsize:");
                ymaxsteppanel.add(ymaxsteplabel);
                ymaxsteppanel.add(ymaxstepspinner);
                ymaxsteppopup.add(ymaxsteppanel);
                ymaxsteppopup.setLightWeightPopupEnabled(false);
                ymaxsteppopup.addPopupMenuListener (new PopupMenuListener()
                    {
                    public void popupMenuCanceled(PopupMenuEvent popupMenuEvent) {
                            ymaxsteppopup.setVisible(false);
                            }

                    public void popupMenuWillBecomeInvisible(PopupMenuEvent popupMenuEvent) {
                            ymaxsteppopup.setVisible(true);
                            }

                    public void popupMenuWillBecomeVisible(PopupMenuEvent popupMenuEvent) {
                            ymaxsteppopup.setVisible(false);
                            }
                    });

                JPanel customymaxpanel = new JPanel(new SpringLayout());
                JLabel ymaxlabel = new JLabel ("Y-max");
                ymaxlabel.setHorizontalAlignment(JLabel.RIGHT);
                ymaxlabel.setPreferredSize(new Dimension(50, 25));
                ymaxlabel.setToolTipText("Right click to set spinner stepsize");
                ymaxlabel.setComponentPopupMenu(ymaxsteppopup);
                customymaxpanel.add (ymaxlabel);

                ymaxspinnermodel = new SpinnerNumberModel(new Double(yMax), null, null, new Double(yMaxStep));
                ymaxspinner = new JSpinner(ymaxspinnermodel);
                ymaxspinner.setEditor(new JSpinner.NumberEditor(ymaxspinner, "0.#########"));
                ymaxspinner.setPreferredSize(new Dimension(100, 25));
                ymaxspinner.setComponentPopupMenu(ymaxsteppopup);
                ymaxspinner.setToolTipText("Right click to set spinner stepsize");
                ymaxspinner.addChangeListener (new ChangeListener()
                    {
                    public void stateChanged(ChangeEvent ev)
                        {
                        yMax = ((Double)ymaxspinner.getValue()).doubleValue();
                        updatePlot();
                        }
                    });
                ymaxspinner.addMouseWheelListener( new MouseWheelListener()
                    {
                    public void mouseWheelMoved( MouseWheelEvent e )
                        {
                        ymaxspinner.setValue(((Double)ymaxspinner.getValue()).doubleValue()
                                - e.getWheelRotation()*yMaxStep );
                        }
                    });
                customymaxpanel.add (ymaxspinner);
                SpringUtil.makeCompactGrid (customymaxpanel, 1, 2, 2,2,0,0);
                yscalingpanel.add (customymaxpanel);

                JPanel ymultiplierpanel = new JPanel(new SpringLayout());
                ymultiplierpanel.setBorder(BorderFactory.createLineBorder(Color.lightGray));
                JLabel ymultiplierlabel = new JLabel ("Y x 1E");
                ymultiplierlabel.setHorizontalAlignment(JLabel.RIGHT);
                ymultiplierpanel.setMaximumSize(new Dimension(250, 30));
                ymultiplierpanel.add (ymultiplierlabel);
                ymultiplierlabel.setToolTipText("Sets the Y-axis multiplication factor");
                ymultipliermodel = new SpinnerNumberModel(new Integer(yExponent),null, null, new Integer(1));
                ymultiplierspinner = new JSpinner(ymultipliermodel);
                ymultiplierspinner.setMaximumSize(new Dimension (35,25));
                ymultiplierspinner.setToolTipText("Sets the Y-axis multiplication factor");
                ymultiplierspinner.addChangeListener (new ChangeListener()
                    {
                    public void stateChanged(ChangeEvent ev)
                        {
                        yExponent = ((Integer)ymultiplierspinner.getValue()).intValue();
                        updatePlot();
                        }
                    });
                ymultiplierspinner.addMouseWheelListener( new MouseWheelListener()
                    {
                    public void mouseWheelMoved( MouseWheelEvent e )
                        {
                        ymultiplierspinner.setValue( new Integer(((Integer)ymultiplierspinner.getValue()).intValue()
                                - e.getWheelRotation()) );
                        }
                    });
                ymultiplierpanel.add (ymultiplierspinner);
                SpringUtil.makeCompactGrid (ymultiplierpanel, 1,2, 2,2,2,0);
                yscalingpanel.add (ymultiplierpanel);

                //YMIN STEPSIZE POPUP

                yminsteppopup = new JPopupMenu();
                JPanel yminsteppanel = new JPanel();
                yminstepmodel = new SpinnerListModel(spinnerscalelist);
                yminstepspinner = new JSpinner(yminstepmodel);
                yminstepspinner.setValue(convertToText(yMinStep));
                yminstepspinner.addChangeListener (new ChangeListener()
                    {
                    public void stateChanged(ChangeEvent ev)
                        {
                        yMinStep = Double.parseDouble((String)yminstepspinner.getValue());
                        yminspinner.setModel(new SpinnerNumberModel(new Double(yMin), null, null, new Double(yMinStep)));
                        yminspinner.setEditor(new JSpinner.NumberEditor(yminspinner, "0.#########"));
                        Prefs.set("plot.yMinStep",yMinStep);
                        }
                    });

                JLabel yminsteplabel = new JLabel ("Stepsize:");
                yminsteppanel.add(yminsteplabel);
                yminsteppanel.add(yminstepspinner);
                yminsteppopup.add(yminsteppanel);
                yminsteppopup.setLightWeightPopupEnabled(false);
                yminsteppopup.addPopupMenuListener (new PopupMenuListener()
                    {
                    public void popupMenuCanceled(PopupMenuEvent popupMenuEvent) {
                            yminsteppopup.setVisible(false);
                            }

                    public void popupMenuWillBecomeInvisible(PopupMenuEvent popupMenuEvent) {
                            yminsteppopup.setVisible(true);
                            }

                    public void popupMenuWillBecomeVisible(PopupMenuEvent popupMenuEvent) {
                            yminsteppopup.setVisible(false);
                            }
                    });


                JPanel customyminpanel = new JPanel(new SpringLayout());
                JLabel yminlabel = new JLabel ("Y-min");
                yminlabel.setHorizontalAlignment(JLabel.RIGHT);
                yminlabel.setComponentPopupMenu(yminsteppopup);
                yminlabel.setPreferredSize(new Dimension(50, 25));
                yminlabel.setToolTipText("Right click to set spinner stepsize");
                customyminpanel.add (yminlabel);

                yminspinnermodel = new SpinnerNumberModel(new Double(yMin),  null, null, new Double(yMinStep));
                yminspinner = new JSpinner(yminspinnermodel);
                yminspinner.setEditor(new JSpinner.NumberEditor(yminspinner, "0.#########"));
                yminspinner.setPreferredSize(new Dimension(100, 25));
                yminspinner.setComponentPopupMenu(yminsteppopup);
                yminspinner.setToolTipText("Right click to set spinner stepsize");
                yminspinner.addChangeListener (new ChangeListener()
                    {
                    public void stateChanged(ChangeEvent ev)
                        {
                        yMin = ((Double)yminspinner.getValue()).doubleValue();
                        updatePlot();
                        }
                    });
                yminspinner.addMouseWheelListener( new MouseWheelListener()
                    {
                    public void mouseWheelMoved( MouseWheelEvent e )
                        {
                        yminspinner.setValue(((Double)yminspinner.getValue()).doubleValue()
                                - e.getWheelRotation()*yMinStep );
                        }
                    });
                customyminpanel.add (yminspinner);
                SpringUtil.makeCompactGrid (customyminpanel, 1, 2, 2,2,0,0);
                yscalingpanel.add (customyminpanel);

                SpringUtil.makeCompactGrid (yscalingpanel, 3, 2, 0,0,0,0);
                mainpanelgroupe.add (yscalingpanel);


            //PLOTSIZE SUBPANEL

                JPanel plotsizepanel = new JPanel (new SpringLayout());
                plotsizepanel.setBorder(BorderFactory.createTitledBorder("Plot Size"));


                JLabel dummylabel10 = new JLabel ("");
                dummylabel10.setHorizontalAlignment(JLabel.CENTER);
                dummylabel10.setPreferredSize(new Dimension(50, 25));
                plotsizepanel.add (dummylabel10);

                JLabel dummylabel11 = new JLabel ("");
                dummylabel11.setHorizontalAlignment(JLabel.CENTER);
                dummylabel11.setPreferredSize(new Dimension(60, 25));
                plotsizepanel.add (dummylabel11);

                //PLOT HEIGHT STEPSIZE POPUP

                plotheightsteppopup = new JPopupMenu();
                JPanel plotheightsteppanel = new JPanel();
                plotheightstepmodel = new SpinnerListModel(integerspinnerscalelist);
                plotheightstepspinner = new JSpinner(plotheightstepmodel);
                plotheightstepspinner.setValue(intConvertToText(plotSizeYStep));
                plotheightstepspinner.addChangeListener (new ChangeListener()
                    {
                    public void stateChanged(ChangeEvent ev)
                        {
                        plotSizeYStep = Integer.parseInt(((String)plotheightstepspinner.getValue()).trim());
                        plotheightspinner.setModel(new SpinnerNumberModel(new Integer(plotSizeY), new Integer(1), null, new Integer(plotSizeYStep)));
                        Prefs.set("plot.plotSizeYStep",plotSizeYStep);
                        }
                    });

                JLabel plotheightsteplabel = new JLabel ("Stepsize:");
                plotheightsteppanel.add(plotheightsteplabel);
                plotheightsteppanel.add(plotheightstepspinner);
                plotheightsteppopup.add(plotheightsteppanel);
                plotheightsteppopup.setLightWeightPopupEnabled(false);
                plotheightsteppopup.addPopupMenuListener (new PopupMenuListener()
                    {
                    public void popupMenuCanceled(PopupMenuEvent popupMenuEvent) {
                            plotheightsteppopup.setVisible(false);
                            }

                    public void popupMenuWillBecomeInvisible(PopupMenuEvent popupMenuEvent) {
                            plotheightsteppopup.setVisible(true);
                            }

                    public void popupMenuWillBecomeVisible(PopupMenuEvent popupMenuEvent) {
                            plotheightsteppopup.setVisible(false);
                            }
                    });

                JLabel plotheightlabel = new JLabel ("Height");
                plotheightlabel.setHorizontalAlignment(JLabel.RIGHT);
                plotheightlabel.setPreferredSize(new Dimension(50, 25));
                plotheightlabel.setToolTipText("Right click to set spinner stepsize");
                plotheightlabel.setComponentPopupMenu(plotheightsteppopup);
                plotsizepanel.add (plotheightlabel);

                plotheightspinnermodel = new SpinnerNumberModel(new Integer(plotSizeY), new Integer(1), null, new Integer(plotSizeYStep));
                plotheightspinner = new JSpinner(plotheightspinnermodel);
                plotheightspinner.setComponentPopupMenu(plotheightsteppopup);
                plotheightspinner.setPreferredSize(new Dimension(50, 25));
                plotheightspinner.setToolTipText("Right click to set spinner stepsize");
                plotheightspinner.addChangeListener (new ChangeListener()
                    {
                    public void stateChanged(ChangeEvent ev)
                        {
                        plotSizeY = ((Integer)plotheightspinner.getValue()).intValue();
                        updatePlot();
                        }
                    });
                plotheightspinner.addMouseWheelListener( new MouseWheelListener()
                    {
                    public void mouseWheelMoved( MouseWheelEvent e )
                        {
                        int newValue =  new Integer(((Integer)plotheightspinner.getValue()).intValue() - e.getWheelRotation()*plotSizeYStep);
                        if (newValue > 0) plotheightspinner.setValue(newValue);
                        }
                    });
                plotsizepanel.add (plotheightspinner);

                //PLOT WIDTH STEPSIZE POPUP

                plotwidthsteppopup = new JPopupMenu();
                JPanel plotwidthsteppanel = new JPanel();
                plotwidthstepmodel = new SpinnerListModel(integerspinnerscalelist);
                plotwidthstepspinner = new JSpinner(plotwidthstepmodel);
                plotwidthstepspinner.setValue(intConvertToText(plotSizeXStep));
                plotwidthstepspinner.addChangeListener (new ChangeListener()
                    {
                    public void stateChanged(ChangeEvent ev)
                        {
                        plotSizeXStep = Integer.parseInt(((String)plotwidthstepspinner.getValue()).trim());
                        plotwidthspinner.setModel(new SpinnerNumberModel(new Integer(plotSizeX), new Integer(1), null, new Integer(plotSizeXStep)));
                        Prefs.set("plot.plotSizeXStep",plotSizeXStep);
                        }
                    });
                JLabel plotwidthsteplabel = new JLabel ("Stepsize:");
                plotwidthsteppanel.add(plotwidthsteplabel);
                plotwidthsteppanel.add(plotwidthstepspinner);
                plotwidthsteppopup.add(plotwidthsteppanel);
                plotwidthsteppopup.setLightWeightPopupEnabled(false);
                plotwidthsteppopup.addPopupMenuListener (new PopupMenuListener()
                    {
                    public void popupMenuCanceled(PopupMenuEvent popupMenuEvent) {
                            plotwidthsteppopup.setVisible(false);
                            }

                    public void popupMenuWillBecomeInvisible(PopupMenuEvent popupMenuEvent) {
                            plotwidthsteppopup.setVisible(true);
                            }

                    public void popupMenuWillBecomeVisible(PopupMenuEvent popupMenuEvent) {
                            plotwidthsteppopup.setVisible(false);
                            }
                    });





                JLabel plotwidthlabel = new JLabel ("Width");
                plotwidthlabel.setHorizontalAlignment(JLabel.RIGHT);
                plotwidthlabel.setPreferredSize(new Dimension(50, 25));
                plotwidthlabel.setToolTipText("Right click to set spinner stepsize");
                plotwidthlabel.setComponentPopupMenu(plotwidthsteppopup);
                plotsizepanel.add (plotwidthlabel);

                plotwidthspinnermodel = new SpinnerNumberModel(new Integer(plotSizeX), new Integer(1), null, new Integer(plotSizeXStep));
                plotwidthspinner = new JSpinner(plotwidthspinnermodel);
                plotwidthspinner.setComponentPopupMenu(plotwidthsteppopup);
                plotwidthspinner.setToolTipText("Right click to set spinner stepsize");
                plotwidthspinner.addChangeListener (new ChangeListener()
                    {
                    public void stateChanged(ChangeEvent ev)
                        {
                        plotSizeX = ((Integer)plotwidthspinner.getValue()).intValue();
                        updatePlot();
                        }
                    });

                plotwidthspinner.addMouseWheelListener( new MouseWheelListener()
                    {
                    public void mouseWheelMoved( MouseWheelEvent e )
                        {
                        int newValue =  new Integer(((Integer)plotwidthspinner.getValue()).intValue() - e.getWheelRotation()*plotSizeXStep);
                        if (newValue > 0) plotwidthspinner.setValue(newValue);
                        }
                    });
                plotsizepanel.add (plotwidthspinner);
                SpringUtil.makeCompactGrid (plotsizepanel, 3, 2, 0,0,0,0);
                mainpanelgroupe.add (plotsizepanel);


        //FINALIZE GROUP E PANEL (SCALING PANEL)

                SpringUtil.makeCompactGrid (mainpanelgroupe, 1, 3, 0,0,0,0);
                mainpanel.add(mainpanelgroupe);
                
        // MAIN PANEL GROUP F START

                JPanel updateplotbuttonpanel = new JPanel (new SpringLayout());
                updateplotbutton = new JButton(" Redraw Plot ");
                updateplotbutton.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                        totalPanOffsetX=0.0;
                        totalPanOffsetY=0.0;
                        newPanOffsetX=0.0;
                        newPanOffsetY=0.0;
                        leftDragReleased=false;
                        zoom = 0.0;
                        updatePlot();
                        plotWindow.toFront();
                    }
                });

                updateplotbuttonpanel.add (updateplotbutton);
                SpringUtil.makeCompactGrid (updateplotbuttonpanel,1,1, 20,15,6,0);
                mainpanelgroupf.add (updateplotbuttonpanel);
                
                
                
                

                JPanel detrendrangepanel = new JPanel (new SpringLayout());
                TitledBorder detrendrangetitle = BorderFactory.createTitledBorder("Detrend and Normalize Region Selection");
                detrendrangetitle.setTitleJustification(TitledBorder.CENTER);
                detrendrangepanel.setBorder(detrendrangetitle);
                
                //DETREND/NORMALIZE STEPSIZE POPUP

                dmarkersteppopup = new JPopupMenu();
                JPanel dmarkersteppanel = new JPanel();
                dmarkerstepmodel = new SpinnerListModel(spinnerscalelist);
                dmarkerstepspinner = new JSpinner(dmarkerstepmodel);
                dmarkerstepspinner.setValue(convertToText(dMarkerStep));
                dmarkerstepspinner.addChangeListener (new ChangeListener()
                    {
                    public void stateChanged(ChangeEvent ev)
                        {
                        dMarkerStep = Double.parseDouble((String)dmarkerstepspinner.getValue());
                        dmarker1spinner.setModel(new SpinnerNumberModel(new Double(dMarker1Value), null, null, new Double(dMarkerStep)));
                        dmarker2spinner.setModel(new SpinnerNumberModel(new Double(dMarker2Value), null, null, new Double(dMarkerStep)));
                        dmarker3spinner.setModel(new SpinnerNumberModel(new Double(dMarker3Value), null, null, new Double(dMarkerStep)));
                        dmarker4spinner.setModel(new SpinnerNumberModel(new Double(dMarker4Value), null, null, new Double(dMarkerStep)));
                        dmarker1spinner.setEditor(new JSpinner.NumberEditor(dmarker1spinner, "0.#########"));
                        dmarker2spinner.setEditor(new JSpinner.NumberEditor(dmarker2spinner, "0.#########"));
                        dmarker3spinner.setEditor(new JSpinner.NumberEditor(dmarker3spinner, "0.#########"));
                        dmarker4spinner.setEditor(new JSpinner.NumberEditor(dmarker4spinner, "0.#########"));
                        Prefs.set("plot.dmarkerStep",dMarkerStep);
                        }
                    });

                JLabel dmarkersteplabel = new JLabel ("Stepsize:");
                dmarkersteppanel.add(dmarkersteplabel);
                dmarkersteppanel.add(dmarkerstepspinner);
                dmarkersteppopup.add(dmarkersteppanel);
                dmarkersteppopup.setLightWeightPopupEnabled(false);
                dmarkersteppopup.addPopupMenuListener (new PopupMenuListener()
                    {
                    public void popupMenuCanceled(PopupMenuEvent popupMenuEvent) {
                            dmarkersteppopup.setVisible(false);
                            }

                    public void popupMenuWillBecomeInvisible(PopupMenuEvent popupMenuEvent) {
                            dmarkersteppopup.setVisible(true);
                            }

                    public void popupMenuWillBecomeVisible(PopupMenuEvent popupMenuEvent) {
                            dmarkersteppopup.setVisible(false);
                            }
                    });
                
                JPanel showdmarkerspanel = new JPanel(new SpringLayout());
                TitledBorder showdmarkerstitle = BorderFactory.createTitledBorder("Show");
                showdmarkerstitle.setTitleJustification(TitledBorder.CENTER);
                showdmarkerspanel.setBorder(showdmarkerstitle);  
                
                JLabel showdmarkerslabel1 = new JLabel("  ");
                showdmarkerspanel.add(showdmarkerslabel1);
                
                showDMarkersCB = new JCheckBox ("", showDMarkers);
                showDMarkersCB.setToolTipText("Show region markers on plot");
                showDMarkersCB.addItemListener(new ItemListener(){
                    public void itemStateChanged(ItemEvent e) {
                        if (e.getStateChange() == ItemEvent.DESELECTED)
                                showDMarkers = false;
                        else if (e.getStateChange() == ItemEvent.SELECTED)
                                showDMarkers = true;
                        updatePlot();}});
                showDMarkersCB.setHorizontalAlignment(JLabel.CENTER);
                showdmarkerspanel.add(showDMarkersCB);
                
                JLabel showdmarkerslabel2 = new JLabel("  ");
                showdmarkerspanel.add(showdmarkerslabel2);                
                
                SpringUtil.makeCompactGrid (showdmarkerspanel, 1, showdmarkerspanel.getComponentCount(), 2,2,0,0);
                detrendrangepanel.add(showdmarkerspanel);                

                JPanel dmarker1panel = new JPanel(new SpringLayout());
                TitledBorder dmarker1title = BorderFactory.createTitledBorder("Left Trim");
                dmarker1title.setTitleJustification(TitledBorder.CENTER);
                dmarker1panel.setBorder(dmarker1title);    
                
                useDMarker1CB = new JCheckBox ("", useDMarker1);
                useDMarker1CB.setToolTipText("Enable left trim");
                useDMarker1CB.addItemListener(new ItemListener(){
                    public void itemStateChanged(ItemEvent e) {
                        if (e.getStateChange() == ItemEvent.DESELECTED)
                                useDMarker1 = false;
                        else if (e.getStateChange() == ItemEvent.SELECTED)
                                useDMarker1 = true;
                        dmarker1spinner.setEnabled(useDMarker1);
                        updatePlot();}});
                useDMarker1CB.setHorizontalAlignment(JLabel.CENTER);
                dmarker1panel.add (useDMarker1CB);

                dmarker1spinnermodel = new SpinnerNumberModel(new Double(dMarker1Value),  null, null, new Double(dMarkerStep));

                dmarker1spinner = new JSpinner(dmarker1spinnermodel);
                dmarker1spinner.setEditor(new JSpinner.NumberEditor(dmarker1spinner, "0.#########"));
//                JComponent editor = new JSpinner.NumberEditor(vmarker1spinner);
//                vmarker1spinner.setEditor(editor);
                dmarker1spinner.setPreferredSize(new Dimension(75, 25));
                dmarker1spinner.setEnabled(useDMarker1);
                dmarker1spinner.setComponentPopupMenu(dmarkersteppopup);
                dmarker1spinner.setToolTipText("<html>"+"Enter detrend/normalize left trim location"+"<br>"+
                    "or enter UT time in HH:MM or HH:MM:SS format and press 'Enter'"+"<br>"+
                    "---------------------------------------------"+"<br>"+
                    "SHORTCUT: &lt;Ctrl&gt;&lt;Shift&gt; Left Click or Drag    "+"<br>"+
                    "---------------------------------------------"+"<br>"+
                    "Right click to set spinner stepsize"+"</html>");
                dmarker1spinner.addChangeListener (new ChangeListener()
                    {
                    public void stateChanged(ChangeEvent ev)
                        {
                        checkForUT(dmarker1spinner);
                        dMarker1Value = ((Double)dmarker1spinner.getValue()).doubleValue();
                        keepMarkersInOrder(1);
                        updatePlot();
                        }
                    });
                dmarker1spinner.addMouseWheelListener( new MouseWheelListener()
                    {
                    public void mouseWheelMoved( MouseWheelEvent e )
                        {
                        if (useDMarker1)
                            {
                            dmarker1spinner.setValue(((Double)dmarker1spinner.getValue()).doubleValue()
                                    - e.getWheelRotation()*dMarkerStep );
                            }
                        }
                    });
                dmarker1panel.add (dmarker1spinner);
                
                SpringUtil.makeCompactGrid (dmarker1panel, 1, dmarker1panel.getComponentCount(), 2,2,0,0);
                detrendrangepanel.add (dmarker1panel);                
                
                JPanel dmarker2panel = new JPanel(new SpringLayout());
//                dmarker2panel.setBorder(BorderFactory.createLineBorder(Color.lightGray,1));
                TitledBorder dmarker2title = BorderFactory.createTitledBorder("Left");
                dmarker2title.setTitleJustification(TitledBorder.CENTER);
//                dmarker2title.setTitlePosition(TitledBorder.BOTTOM);
                dmarker2panel.setBorder(dmarker2title);                

                dmarker2spinnermodel = new SpinnerNumberModel(new Double(dMarker2Value),  null, null, new Double(dMarkerStep));

                dmarker2spinner = new JSpinner(dmarker2spinnermodel);
                dmarker2spinner.setEditor(new JSpinner.NumberEditor(dmarker2spinner, "0.#########"));
//                JComponent editor = new JSpinner.NumberEditor(vmarker1spinner);
//                vmarker1spinner.setEditor(editor);
                dmarker2spinner.setPreferredSize(new Dimension(75, 25));
                dmarker2spinner.setComponentPopupMenu(dmarkersteppopup);
                dmarker2spinner.setToolTipText("<html>"+"Enter detrend/normalize left marker location"+"<br>"+
                    "or enter UT time in HH:MM or HH:MM:SS format and press 'Enter'"+"<br>"+
                    "---------------------------------------------"+"<br>"+
                    "SHORTCUT: &lt;Ctrl&gt; Left Click or Drag    "+"<br>"+
                    "---------------------------------------------"+"<br>"+
                    "Right click to set spinner stepsize"+"</html>");
                dmarker2spinner.addChangeListener (new ChangeListener()
                    {
                    public void stateChanged(ChangeEvent ev)
                        {
                        checkForUT(dmarker2spinner);
                        dMarker2Value = ((Double)dmarker2spinner.getValue()).doubleValue();
                        keepMarkersInOrder(2);
                        updatePlot();
                        }
                    });
                dmarker2spinner.addMouseWheelListener( new MouseWheelListener()
                    {
                    public void mouseWheelMoved( MouseWheelEvent e )
                        {
                        dmarker2spinner.setValue(((Double)dmarker2spinner.getValue()).doubleValue()
                                - e.getWheelRotation()*dMarkerStep );
                        }
                    });
                dmarker2panel.add (dmarker2spinner);
                
                SpringUtil.makeCompactGrid (dmarker2panel, 1, dmarker2panel.getComponentCount(), 2,2,0,0);
                detrendrangepanel.add (dmarker2panel);
                
                
                
                

                JPanel dmarkercopypanel = new JPanel(new SpringLayout());
                TitledBorder dmarkercopytitle = BorderFactory.createTitledBorder("Copy");
                dmarkercopytitle.setTitleJustification(TitledBorder.CENTER);
                dmarkercopypanel.setBorder(dmarkercopytitle);    
                
                copyVMarkersIcon = createImageIcon("astroj/images/copymarkers.png", "Copy V. Markers Icon");
                JButton copyButton = new JButton(copyVMarkersIcon);
                copyButton.setMargin(new Insets(0,0,0,0));
                copyButton.setToolTipText("Copy values from V. Markers 1 and 2 to Left and Right markers");
                copyButton.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                        dmarker2spinner.setValue(vMarker1Value);
                        dmarker3spinner.setValue(vMarker2Value);
                        updatePlot();
                        }});                
                dmarkercopypanel.add(copyButton);
                SpringUtil.makeCompactGrid (dmarkercopypanel, 1, dmarkercopypanel.getComponentCount(), 2,0,2,0);
                detrendrangepanel.add (dmarkercopypanel);                
                
                JPanel dmarker3panel = new JPanel(new SpringLayout());
//                dmarker3panel.setBorder(BorderFactory.createLineBorder(Color.lightGray,1));
                TitledBorder dmarker3title = BorderFactory.createTitledBorder("Right");
                dmarker3title.setTitleJustification(TitledBorder.CENTER);
//                dmarker3title.setTitlePosition(TitledBorder.BOTTOM);
                dmarker3panel.setBorder(dmarker3title);                

                dmarker3spinnermodel = new SpinnerNumberModel(new Double(dMarker3Value),  null, null, new Double(dMarkerStep));

                dmarker3spinner = new JSpinner(dmarker3spinnermodel);
                dmarker3spinner.setEditor(new JSpinner.NumberEditor(dmarker3spinner, "0.#########"));
//                JComponent editor = new JSpinner.NumberEditor(vmarker1spinner);
//                vmarker1spinner.setEditor(editor);
                dmarker3spinner.setPreferredSize(new Dimension(75, 25));
                dmarker3spinner.setComponentPopupMenu(dmarkersteppopup);
                dmarker3spinner.setToolTipText("<html>"+"Enter detrend/normalize right marker location"+"<br>"+
                    "or enter UT time in HH:MM or HH:MM:SS format and press 'Enter'"+"<br>"+
                    "---------------------------------------------"+"<br>"+
                    "SHORTCUT: &lt;Ctrl&gt; Right Click or Drag    "+"<br>"+
                    "---------------------------------------------"+"<br>"+
                    "Right click to set spinner stepsize"+"</html>");
                dmarker3spinner.addChangeListener (new ChangeListener()
                    {
                    public void stateChanged(ChangeEvent ev)
                        {
                        checkForUT(dmarker3spinner);
                        dMarker3Value = ((Double)dmarker3spinner.getValue()).doubleValue();
                        keepMarkersInOrder(3);
                        updatePlot();
                        }
                    });
                dmarker3spinner.addMouseWheelListener( new MouseWheelListener()
                    {
                    public void mouseWheelMoved( MouseWheelEvent e )
                        {
                        dmarker3spinner.setValue(((Double)dmarker3spinner.getValue()).doubleValue()
                                - e.getWheelRotation()*dMarkerStep );
                        }
                    });
                dmarker3panel.add (dmarker3spinner);
                
                SpringUtil.makeCompactGrid (dmarker3panel, 1, dmarker3panel.getComponentCount(), 2,2,0,0);
                detrendrangepanel.add (dmarker3panel);                
                
                JPanel dmarker4panel = new JPanel(new SpringLayout());
//                dmarker4panel.setBorder(BorderFactory.createLineBorder(Color.lightGray,1));
                TitledBorder dmarker4title = BorderFactory.createTitledBorder("Right Trim");
                dmarker4title.setTitleJustification(TitledBorder.CENTER);
//                dmarker4title.setTitlePosition(TitledBorder.BOTTOM);
                dmarker4panel.setBorder(dmarker4title);                  

                dmarker4spinnermodel = new SpinnerNumberModel(new Double(dMarker4Value),  null, null, new Double(dMarkerStep));

                dmarker4spinner = new JSpinner(dmarker4spinnermodel);
                dmarker4spinner.setEditor(new JSpinner.NumberEditor(dmarker4spinner, "0.#########"));
                dmarker4spinner.setPreferredSize(new Dimension(75, 25));
                dmarker4spinner.setEnabled(useDMarker4);
                dmarker4spinner.setComponentPopupMenu(dmarkersteppopup);
                dmarker4spinner.setToolTipText("<html>"+"Enter detrend/normalize right trim location"+"<br>"+
                    "or enter UT time in HH:MM or HH:MM:SS format and press 'Enter'"+"<br>"+
                    "---------------------------------------------"+"<br>"+
                    "SHORTCUT: &lt;Ctrl&gt;&lt;Shift&gt; Right Click or Drag    "+"<br>"+
                    "---------------------------------------------"+"<br>"+
                    "Right click to set spinner stepsize"+"</html>");
                dmarker4spinner.addChangeListener (new ChangeListener()
                    {
                    public void stateChanged(ChangeEvent ev)
                        {
                        checkForUT(dmarker4spinner);
                        dMarker4Value = ((Double)dmarker4spinner.getValue()).doubleValue();
                        keepMarkersInOrder(4);
                        updatePlot();
                        }
                    });
                dmarker4spinner.addMouseWheelListener( new MouseWheelListener()
                    {
                    public void mouseWheelMoved( MouseWheelEvent e )
                        {
                        if (useDMarker4)
                            {
                            dmarker4spinner.setValue(((Double)dmarker4spinner.getValue()).doubleValue()
                                    - e.getWheelRotation()*dMarkerStep );
                            }
                        }
                    });
                dmarker4panel.add (dmarker4spinner);
                
                useDMarker4CB = new JCheckBox ("", useDMarker4);
                useDMarker4CB.setToolTipText("Enable right trim");
                useDMarker4CB.addItemListener(new ItemListener(){
                    public void itemStateChanged(ItemEvent e) {
                        if (e.getStateChange() == ItemEvent.DESELECTED)
                                useDMarker4 = false;
                        else if (e.getStateChange() == ItemEvent.SELECTED)
                                useDMarker4 = true;
                        dmarker4spinner.setEnabled(useDMarker4);
                        updatePlot();}});
                useDMarker4CB.setHorizontalAlignment(JLabel.CENTER);
                dmarker4panel.add (useDMarker4CB);                
                
                SpringUtil.makeCompactGrid (dmarker4panel, 1, dmarker4panel.getComponentCount(), 2,2,0,0);
                detrendrangepanel.add (dmarker4panel);                
                

                SpringUtil.makeCompactGrid (detrendrangepanel, 1,detrendrangepanel.getComponentCount(), 2,2,6,2);
                mainpanelgroupf.add (detrendrangepanel);

//                JPanel closebuttonpanel = new JPanel (new SpringLayout());
//                closebutton = new JButton("        Exit        ");
//                closebutton.setForeground(Color.RED);
//                closebutton.addActionListener(new ActionListener() {
//                    public void actionPerformed(ActionEvent e) {
//                            saveAndClose(); }});
//                closebuttonpanel.add (closebutton);
//                SpringUtil.makeCompactGrid (closebuttonpanel, 1, 1, 6,15,0,0);
//                mainpanelgroupf.add (closebuttonpanel);
                
                
                JPanel morepanelspanel = new JPanel(new SpringLayout());
                TitledBorder morepanelstitle = BorderFactory.createTitledBorder("Open Panels");
                morepanelstitle.setTitleJustification(TitledBorder.CENTER);
                morepanelspanel.setBorder(morepanelstitle); 
//                morepanelspanel.setPreferredSize(new Dimension(125, 25));
                                
                moreybutton = new JButton("Y-data Panel");
                moreybutton.setPreferredSize(new Dimension(150, 20));
                moreybutton.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                        if (subFrame.isShowing())
                            subFrame.setVisible(true);
                        else
                            {
                            setSubpanelVisible = true;
                            showMoreCurvesJPanel();
                            }
                    }
                });
                morepanelspanel.add (moreybutton);
          
                refStarButton = new JButton("Ref. Star Panel");
                refStarButton.setPreferredSize(new Dimension(150, 20));
                refStarButton.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                        if (refStarFrame != null && refStarFrame.isShowing())
                            {
                            refStarFrame.setVisible(true);
                            }
                        else
                            {
                            if (table != null)
                                {
                                showRefStarJPanel();
                                }
                            else
                                {
                                IJ.showMessage("No table is selected in Multi-Plot");
                                }
                            }
                    }
                });

                morepanelspanel.add (refStarButton);
                
                SpringUtil.makeCompactGrid (morepanelspanel, morepanelspanel.getComponentCount(), 1, 2,0,2,2);
                mainpanelgroupf.add (morepanelspanel);

                SpringUtil.makeCompactGrid (mainpanelgroupf, 1, mainpanelgroupf.getComponentCount(), 2,0,6,0);
                mainpanel.add(mainpanelgroupf);

                SpringUtil.makeCompactGrid (mainpanel, 5, 1, 6,6,6,6);

                mainFrame.setJMenuBar(mainmenubar);
                mainFrame.add (mainscrollpane);
                mainFrame.pack();
                mainFrame.setResizable (true);

                if (rememberWindowLocations)
                        {
                        Dimension mainScreenSize = Toolkit.getDefaultToolkit().getScreenSize();
                        if (!Prefs.isLocationOnScreen(new Point(mainFrameLocationX,mainFrameLocationY)))
                            {
                            mainFrameLocationX = mainScreenSize.width/2 - mainFrame.getWidth()/2;
                            mainFrameLocationY = mainScreenSize.height/2 - mainFrame.getHeight()/2;
                            }
                        mainFrame.setLocation(mainFrameLocationX, mainFrameLocationY);
                        }
                if (!plotAutoMode)
                        mainFrame.setVisible (true);
                if (!plotAutoMode && table != null && (openRefStarWindow || refStarPanelWasShowing))
                    {
//                    checkAndLockTable();
                    showRefStarJPanel();
//                    if (table != null) table.setLock(false);
                    refStarPanelWasShowing = true;
                    }
                showMoreCurvesJPanel();
                if (!plotAutoMode && table != null && (addAstroDataFrameWasShowing))
                    {
                    addNewAstroData();
                    addAstroDataFrameWasShowing = true;
                    }                
                }





        static void showMoreCurvesJPanel()
                {
                subFrame = new JFrame ("Multi-plot Y-data");
                subFrame.setIconImage(plotIcon.getImage());
                mainsubpanel = new JPanel (new SpringLayout());
//                mainsubpanel.addMouseListener(panelMouseListener);
                mainsubpanel.addMouseMotionListener(panelMouseMotionListener);                
                FileDrop fileDrop = new FileDrop(mainsubpanel, BorderFactory.createEmptyBorder(),new FileDrop.Listener()
                    {   public void filesDropped( java.io.File[] files )
                        {
                        openDragAndDropFiles(files);
                        }
                    });
                
                subscrollpane = new JScrollPane(mainsubpanel);
                subFrame.setDefaultCloseOperation(javax.swing.WindowConstants.DISPOSE_ON_CLOSE);
                subFrame.addWindowListener(new WindowAdapter(){
                    @Override
                    public void windowClosing(WindowEvent e){
                            subFrameLocationX = subFrame.getLocation().x;
                            subFrameLocationY = subFrame.getLocation().y;
                            setSubpanelVisible = false;
                            Prefs.set("plot2.subFrameLocationX",subFrameLocationX);
                            Prefs.set("plot2.subFrameLocationY",subFrameLocationY);
                            }});

                JPanel mainsubpanelgroupa = new JPanel (new SpringLayout());
                mainsubpanelgroupa.setBorder(BorderFactory.createTitledBorder(""));

                JPanel mainsubpanelgroupb = new JPanel (new SpringLayout());
                mainsubpanelgroupb.setBorder(BorderFactory.createTitledBorder(""));

       // SUBPANEL START

             
                
                if (useWideDataPanel)
                        {
//                        constructTopGroupTopLabels(mainsubpanelgroupa);
//                        constructOtherGroupTopLabels(mainsubpanelgroupa);
                        constructTopGroupBottomLabels(mainsubpanelgroupa);
                        constructOtherGroupBottomLabels(mainsubpanelgroupa);

                        }
                else
                        {
//                        constructTopGroupTopLabels(mainsubpanelgroupa);
                        constructTopGroupBottomLabels(mainsubpanelgroupa);
//                        constructOtherGroupTopLabels(mainsubpanelgroupb);
                        constructOtherGroupBottomLabels(mainsubpanelgroupb);
                        }
                for (cur=0; cur<maxCurves; cur++)
                        {
                        constructTopGroup(mainsubpanelgroupa, cur);
                        if (useWideDataPanel)
                                constructOtherGroup(mainsubpanelgroupa, cur);
                        else
                                constructOtherGroup(mainsubpanelgroupb, cur);
                        }

                if (useWideDataPanel)
                        {
                        SpringUtil.makeCompactGrid (mainsubpanelgroupa, maxCurves+1, mainsubpanelgroupa.getComponentCount()/(maxCurves+1), 5,5,5,5);
                        mainsubpanel.add(mainsubpanelgroupa);
                        SpringUtil.makeCompactGrid (mainsubpanel, 1, 1, 5,5,5,5);
                        }
                else
                        {
                        SpringUtil.makeCompactGrid (mainsubpanelgroupa, maxCurves+1, mainsubpanelgroupa.getComponentCount()/(maxCurves+1), 5,5,5,5);
                        mainsubpanel.add(mainsubpanelgroupa);
                        SpringUtil.makeCompactGrid (mainsubpanelgroupb, maxCurves+1, mainsubpanelgroupb.getComponentCount()/(maxCurves+1), 5,5,5,5);
                        mainsubpanel.add(mainsubpanelgroupb);
                        SpringUtil.makeCompactGrid (mainsubpanel, mainsubpanel.getComponentCount(), 1, 5,5,5,5);
                        }

                
                if (refStarFrame != null && isRefStar != null && isRefStar.length > 0)
                    {
                    
                    for (int r = 0; r < isRefStar.length; r++)
                        {
                        updateTable(isRefStar[r], r);
                        }
                    updatePlotEnabled = false;
                    waitForPlotUpdateToFinish();
                    updateGUI();
                    if (delayedUpdateTimer != null) delayedUpdateTimer.cancel();
                    if (delayedUpdateTask != null) delayedUpdateTask.cancel();                    
                    updatePlotEnabled = true;
                    }
               
                subFrame.add (subscrollpane);
                subFrame.pack();
                subFrame.setResizable (true);

                if (rememberWindowLocations)
                        {
                        Dimension mainScreenSize = Toolkit.getDefaultToolkit().getScreenSize();
                        if (!Prefs.isLocationOnScreen(new Point(subFrameLocationX,subFrameLocationY)))
                            {
                            subFrameLocationX = mainScreenSize.width/2 - subFrame.getWidth()/2;
                            subFrameLocationY = mainScreenSize.height/2 - subFrame.getHeight()/2;
                            }
                        subFrame.setLocation(subFrameLocationX, subFrameLocationY);
                        }

                if (openDataSetWindow || setSubpanelVisible)
                        {
                        subFrame.setVisible (true);
                        }
//                if (table != null)
//                    {
//                    table.setHeading(table.getColumnIndex("rel_flux_C2"), "rel_flux_T2");
//                    xdatacolumndefault.removeItem("rel_flux_C2");
//                    xdatacolumndefault.addItem("rel_flux_T2");
//                    }
                
                }                


        static void constructTopGroupBottomLabels(JPanel mainsubpanelgroup)
                {
                JLabel dummylabel20 = new JLabel ("<HTML><CENTER>Data<BR><CENTER>Set</HTML>");
                dummylabel20.setHorizontalAlignment(JLabel.CENTER);
                dummylabel20.setMaximumSize(new Dimension(35,25));
                mainsubpanelgroup.add (dummylabel20);
                JLabel savenewlabel = new JLabel ("<HTML><CENTER>New<BR><CENTER>Col</HTML>");
                savenewlabel.setToolTipText("Save curve as new table column");
                savenewlabel.setMaximumSize(new Dimension(45,25));
                savenewlabel.setHorizontalAlignment(JLabel.CENTER);
                mainsubpanelgroup.add (savenewlabel);                  
                JLabel dummylabel21 = new JLabel ("Plot");
                dummylabel21.setToolTipText("Show the dataset on the plot");
                dummylabel21.setHorizontalAlignment(JLabel.CENTER);
                dummylabel21.setMaximumSize(new Dimension(45,25));
                mainsubpanelgroup.add (dummylabel21);
                JLabel autoscaleincludelabel = new JLabel ("<HTML><CENTER>Auto<BR><CENTER>Scale</HTML>");
                autoscaleincludelabel.setToolTipText("Include this dataset in X- and Y-axis autoscaling (master autoscaling must also be enabled in MP Main)");
                autoscaleincludelabel.setHorizontalAlignment(JLabel.CENTER);
                autoscaleincludelabel.setMaximumSize(new Dimension(45,25));
                mainsubpanelgroup.add (autoscaleincludelabel);
                xcolumnlabelsub = new JLabel ("X-data");
                xcolumnlabelsub.setHorizontalAlignment(JLabel.CENTER);
                mainsubpanelgroup.add (xcolumnlabelsub);
//                ycolumnlabelsub = new JLabel ("Y-data");
//                ycolumnlabelsub.setHorizontalAlignment(JLabel.CENTER);
//                mainsubpanelgroup.add (ycolumnlabelsub);
                
                JLabel magToFluxlabel = new JLabel ("<HTML><CENTER>Y-input<BR><CENTER>in Mag</HTML>");
                magToFluxlabel.setToolTipText("<HTML>Enable to convert magnitude-based Y-input data (Y-data, Y-error, Y-operand, and Y-operand-error) to flux for additional calculations .<BR>"+
                                             "This option should generally be deselected to plot data generated by Multi-Aperture.</HTML>");
                magToFluxlabel.setHorizontalAlignment(JLabel.CENTER);
                magToFluxlabel.setMaximumSize(new Dimension(45,25));
                mainsubpanelgroup.add (magToFluxlabel);                
                                
                
                JPanel ycolumnlabelgroup = new JPanel();
                ycolumnlabelgroup.setLayout(new BoxLayout(ycolumnlabelgroup,BoxLayout.X_AXIS));
//                legendslabelgroup.add(Box.createGlue());
                ycolumnlabelgroup.add(Box.createHorizontalGlue());

                ycolumnlabelsub = new JLabel ("Y-data");
                ycolumnlabelsub.setHorizontalAlignment(JLabel.CENTER);
                ycolumnlabelgroup.add (ycolumnlabelsub);
                ycolumnlabelgroup.add(Box.createHorizontalStrut(5));
//                legendslabelgroup.add(Box.createGlue());
                helpIcon = createImageIcon("astroj/images/help.png", "Multi-plot Help");
                JButton ycolumnlabelconfigureButton = new JButton(helpIcon);
                ycolumnlabelconfigureButton.setToolTipText("Data naming convention help");
                ycolumnlabelconfigureButton.setMargin(new Insets(0,0,0,0));
                ycolumnlabelconfigureButton.addActionListener(new ActionListener(){
                            public void actionPerformed(ActionEvent ae) {
                                openDataHelpPanel();}});
                ycolumnlabelgroup.add(ycolumnlabelconfigureButton);
                ycolumnlabelgroup.add(Box.createHorizontalGlue());
//                legendslabelgroup.add(Box.createGlue());
                mainsubpanelgroup.add(ycolumnlabelgroup);                
                
                JLabel useerrorlabel = new JLabel ("<HTML><CENTER>Auto<BR><CENTER>Error</HTML>");
                useerrorlabel.setToolTipText("<HTML>Show automatic Y-error bars when dataset has predefined error available.<BR>"+
                                             "Compatible datasets are: Source-Sky_XX, rel_flux_XX, tot_C_cnts.</HTML>");
                useerrorlabel.setHorizontalAlignment(JLabel.CENTER);
                useerrorlabel.setMaximumSize(new Dimension(45,25));
                mainsubpanelgroup.add (useerrorlabel);
//                JLabel errorlabel = new JLabel ("Y-error");
//                errorlabel.setHorizontalAlignment(JLabel.CENTER);
//                errorlabel.setToolTipText("Error associated with Y-data");
//                mainsubpanelgroup.add (errorlabel);
                operatorlabelsub = new JLabel ("Function");
                operatorlabelsub.setHorizontalAlignment(JLabel.CENTER);
                operatorlabelsub.setToolTipText("Performs the selected operation on Y-data Column and Y-operator Column and propagates error if enabled");
                mainsubpanelgroup.add (operatorlabelsub);
                opcolumnlabelsub = new JLabel ("Y-operand");
                opcolumnlabelsub.setHorizontalAlignment(JLabel.CENTER);
                mainsubpanelgroup.add (opcolumnlabelsub);
//                JLabel useoperrorlabel = new JLabel ("Use");
//                useoperrorlabel.setToolTipText("Use Operator Error as error associated with Y-operator Column");
//                useoperrorlabel.setHorizontalAlignment(JLabel.CENTER);
//                mainsubpanelgroup.add (useoperrorlabel);
//                JLabel operrorlabel = new JLabel ("Y-op Error");
//                operrorlabel.setHorizontalAlignment(JLabel.CENTER);
//                mainsubpanelgroup.add (operrorlabel);
                markerlabelsub = new JLabel ("Color");
                markerlabelsub.setHorizontalAlignment(JLabel.CENTER);
                mainsubpanelgroup.add (markerlabelsub);
                markercolorlabelsub = new JLabel ("Symbol");
                markercolorlabelsub.setHorizontalAlignment(JLabel.CENTER);
                mainsubpanelgroup.add (markercolorlabelsub);
                JLabel lineslabel = new JLabel ("Lines");
                lineslabel.setToolTipText("Connect datapoints with lines");
                lineslabel.setHorizontalAlignment(JLabel.CENTER);
                lineslabel.setMaximumSize(new Dimension(45,25));
                mainsubpanelgroup.add (lineslabel);
                JLabel binslabel = new JLabel ("Bin Size");
                binslabel.setMaximumSize(new Dimension(50, 25));
                binslabel.setPreferredSize(new Dimension(50, 25));
                binslabel.setHorizontalAlignment(JLabel.CENTER);
                binslabel.setToolTipText("Plot binned data using the number of samples/bin indicated");
                mainsubpanelgroup.add (binslabel);                
                forceIcon = createImageIcon("astroj/images/grab.png", "Transfer Scale Rel settings to absolute settings");
                insertColumnIcon = createImageIcon("astroj/images/insertcolumn.png", "Save curve as new table column");
                }



        static void constructOtherGroupBottomLabels(JPanel mainsubpanelgroup)
                {
                JLabel seconddatasetlabel = new JLabel ("<HTML><CENTER>Data<BR><CENTER>Set</HTML>");
                seconddatasetlabel.setHorizontalAlignment(JLabel.CENTER);
                seconddatasetlabel.setMaximumSize(new Dimension(35,25));
                if (!useWideDataPanel)
                        mainsubpanelgroup.add (seconddatasetlabel);
                
                JPanel detrendlabelgroup = new JPanel (new SpringLayout());
                detrendlabelgroup.setMaximumSize(new Dimension(195,20));
                detrendlabelgroup.setBorder(BorderFactory.createLineBorder(Color.LIGHT_GRAY));   
                JLabel detrendlabelA = new JLabel ("<HTML><CENTER>Detrend<BR><CENTER>Mode</HTML>");
                detrendlabelA.setToolTipText("Detrend Y-data using a user-defined or fitted constant value times the 'Trend Data' values");
                detrendlabelA.setHorizontalAlignment(JLabel.CENTER);
                detrendlabelA.setPreferredSize(new Dimension(50,16));
                detrendlabelgroup.add (detrendlabelA);
                if (maxDetrendVars>1)
                    {
                    JLabel detrendlabelB = new JLabel ("<HTML><CENTER>Show<BR><CENTER>Variable</HTML>");
                    detrendlabelB.setToolTipText("<html>Click buttons to show the corresponding trend variable coefficient and dataset name.<br>"+
                                                 "If a trend variable has been selected and a detrend mode is enabled,<br>"+
                                                 "the button background will display green. The background will be gray otherwise.<br>"+
                                                 "---------------------------------------------------------------------------<br>"+
                                                 "Select a 'Trend Dataset' for more than one variable to increase the power of its polynomial fit.<br>"+
                                                 "One occurance of a Trend Dataset -> linear fit, two occurances -> quadratic fit, etc.<br>"+
                                                 "---------------------------------------------------------------------------<br>"+
                                                 "If 'Show tooltips help' is enabled under the 'Preferences' menu,<br>"+
                                                 "a mouse-over of a radio button shows the column set for that varaible.<br>"+
                                                 "If a radio button is gray and a tooltip shows a column name enclosed in square brackets,<br>"+
                                                 "that column will be recalled if the radio button is clicked again.<br>"+
                                                 "---------------------------------------------------------------------------<br>"+
                                                 "SHORTCUTS:<br>"+
                                                 "Click an already selected button again to alternately enable/disable the corresponding trend variable<br>"+
                                                 "&lt;SHIFT&gt;Click - selects this variable column for all rows<br>"+
                                                 "&lt;CTRL&gt;Click - copies the selected trend dataset to all rows<br>"+
                                                 "&lt;ALT&gt;Click - alternately enables/disables the corresponding trend variable</html>");
                    detrendlabelB.setPreferredSize(new Dimension(maxDetrendVars*20,25));
                    detrendlabelB.setHorizontalAlignment(JLabel.CENTER);
                    detrendlabelgroup.add (detrendlabelB);       
                    }
                JLabel detrendlabelC = new JLabel ("<HTML><CENTER>Trend<BR><CENTER>Coefficient</HTML>");
                detrendlabelC.setToolTipText("White background indicates user-selected value, gray background indicates fitted constant value");
                detrendlabelC.setPreferredSize(new Dimension(75,25));
                detrendlabelC.setHorizontalAlignment(JLabel.CENTER);
                detrendlabelgroup.add (detrendlabelC);  
                JLabel detrendlabelD = new JLabel ("<HTML><CENTER>Trend<BR><CENTER>Dataset</HTML>");
                detrendlabelD.setToolTipText("<html>Trend data used to detrend Y-data.<br>"+
                                                   "---------------------------------------------------------------------------<br>"+
                                                   "Select a 'Trend Dataset' for more than one variable to increase the power of its polynomial fit.<br>"+
                                                   "One occurance of a Trend Dataset -> linear fit, two occurances -> quadratic fit, etc.</html>");
                detrendlabelD.setHorizontalAlignment(JLabel.CENTER);
                detrendlabelD.setPreferredSize(new Dimension(132,25));
                detrendlabelgroup.add (detrendlabelD);
                
                SpringUtil.makeCompactGrid (detrendlabelgroup, 1, detrendlabelgroup.getComponentCount(), 0,0,0,0);
                mainsubpanelgroup.add (detrendlabelgroup);
                
                

//                JLabel customscalelabel4 = new JLabel ("<HTML><CENTER><u>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;Detrend Y-data&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;</u><BR>"+
//                                                       "&nbsp;&nbsp;&nbsp;&nbsp;Fit&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;" +
//                                                       "Coefficient&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;"+
//                                                       "&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;Trend Data</HTML>");
//                mainsubpanelgroup.add(customscalelabel4);
                JLabel normlabel = new JLabel ("<HTML><CENTER>Norm-<BR><CENTER>alize</HTML>");
                normlabel.setToolTipText("Plot normalized data");
                normlabel.setHorizontalAlignment(JLabel.CENTER);
                normlabel.setMaximumSize(new Dimension(50,25));
                mainsubpanelgroup.add (normlabel);
                JLabel mmaglabel = new JLabel ("Mag");
                mmaglabel.setToolTipText("<html>Plot data in relative magnitude referenced to the average of the first n data samples,<br>"+
                                         "where n is set by the 'Rel. Mag. Reference' spinner on the 'Multi-plot Main' window.</html>");
                mmaglabel.setHorizontalAlignment(JLabel.CENTER);
                mmaglabel.setMaximumSize(new Dimension(30,25));
                mainsubpanelgroup.add (mmaglabel);                

                JPanel customscalelabelgroup = new JPanel();
                customscalelabelgroup.setBorder(BorderFactory.createLineBorder(Color.LIGHT_GRAY)); 
                customscalelabelgroup.setLayout(new BoxLayout(customscalelabelgroup,BoxLayout.X_AXIS));
                customscalelabelgroup.setMaximumSize(new Dimension(195,20));
//                JLabel dummylabel39 = new JLabel ("");
//                customscalelabelgroup.add (dummylabel39);
                customscalelabelgroup.add(Box.createHorizontalStrut(18));
//                customscalelabelgroup.add(Box.createHorizontalGlue());
                JLabel customscalelabel1 = new JLabel ("<HTML><CENTER>Scale<BR><CENTER>Rel</HTML>");
                customscalelabel1.setToolTipText("Select to arbitrarily scale data to fit on plot (\"Scale\" and \"Shift\" by a percentage of the plot Y-range)");
//                customscalelabel1.setHorizontalAlignment(JLabel.LEFT);
                customscalelabelgroup.add(customscalelabel1);
                customscalelabelgroup.add(Box.createHorizontalGlue());
                
                JLabel customscalelabel2 = new JLabel ("Scale");
                customscalelabel2.setToolTipText("Scale dataset by specified value or if \"Scale Rel\" is selected, by specified percentage of the plot Y-range");
//                customscalelabel2.setHorizontalAlignment(JLabel.CENTER);
                customscalelabelgroup.add(customscalelabel2);
                customscalelabelgroup.add(Box.createHorizontalStrut(25));
                
                JLabel customscalelabel3 = new JLabel ("then Shift");
                customscalelabel3.setToolTipText("Add specified constant value to scaled dataset or if \"Scale Rel\" is selected, add specified percentage of the plot Y-range");
//                customscalelabel3.setHorizontalAlignment(JLabel.CENTER);
                customscalelabelgroup.add(customscalelabel3);
                customscalelabelgroup.add(Box.createHorizontalStrut(20));
                mainsubpanelgroup.add (customscalelabelgroup);

                JPanel legendslabelgroup = new JPanel();
                legendslabelgroup.setBorder(BorderFactory.createLineBorder(Color.LIGHT_GRAY)); 
                legendslabelgroup.setLayout(new BoxLayout(legendslabelgroup,BoxLayout.X_AXIS));
                                
                if (useWideDataPanel)
                        legendslabelgroup.setPreferredSize (new Dimension(225,20));
                else
                        legendslabelgroup.setPreferredSize (new Dimension(325,20));
//                legendslabelgroup.add(Box.createGlue());
                legendslabelgroup.add(Box.createHorizontalStrut(1));

                JLabel morelegendslabel = new JLabel("<HTML><CENTER>Legend<BR><CENTER>Type</HTML>");
                morelegendslabel.setHorizontalAlignment(JLabel.CENTER);
                morelegendslabel.setComponentPopupMenu(legendpopup);
                morelegendslabel.setToolTipText("Right click to set legend preferences");
                morelegendslabel.setMaximumSize(new Dimension(75, 45));
                morelegendslabel.setPreferredSize(new Dimension(75, 45));
                morelegendslabel.setMinimumSize(new Dimension(75, 25));
                legendslabelgroup.add (morelegendslabel);
                legendslabelgroup.add(Box.createHorizontalStrut(5));

                legendslabelgroup.add(Box.createGlue());

                JLabel morelegendsnamelabel = new JLabel ("Custom Legend");
                morelegendsnamelabel.setHorizontalAlignment(JLabel.CENTER);
                if (useWideDataPanel)
                        morelegendsnamelabel.setPreferredSize (new Dimension(125,20));
                else
                        morelegendsnamelabel.setPreferredSize (new Dimension(225,20));                
                morelegendsnamelabel.setComponentPopupMenu(legendpopup);
                morelegendsnamelabel.setToolTipText("Right click to set legend preferences");
                legendslabelgroup.add (morelegendsnamelabel);
                legendslabelgroup.add(Box.createGlue());

                legendslabelconfigureButton = new JButton(configureIcon);
                legendslabelconfigureButton.setToolTipText("Set legend preferences");
                legendslabelconfigureButton.setMargin(new Insets(0,0,0,0));
                legendslabelconfigureButton.setMaximumSize(new Dimension(25, 25));
                legendslabelconfigureButton.addActionListener(new ActionListener(){
                            public void actionPerformed(ActionEvent ae) {
                                legendpopup.show(legendslabelconfigureButton, mainsubpanel.getX() , mainsubpanel.getY()+25);}});
                legendslabelgroup.add(legendslabelconfigureButton);
                
                mainsubpanelgroup.add(legendslabelgroup);
                

                if (useWideDataPanel)
                        mainsubpanelgroup.add (seconddatasetlabel);
                }

        static void constructTopGroup(JPanel mainsubpanelgroup, final int c)
                {
                color[c] = colorOf(colorIndex[c]);
                curvelabel[c] = new JLabel (""+(c+1));
                curvelabel[c].setBorder(BorderFactory.createLineBorder(color[c], 2));
                curvelabel[c].setHorizontalAlignment(JLabel.CENTER);
                mainsubpanelgroup.add (curvelabel[c]);
                
//                savecolumnpanelgroup[c] = new JPanel (new SpringLayout());
//                savecolumnpanelgroup[c].setMaximumSize(new Dimension(25,20));
//                savecolumnpanelgroup[c].setBorder(BorderFactory.createEmptyBorder());                 

                savecolumnbutton[c] = new JButton(insertColumnIcon);
                savecolumnbutton[c].setEnabled(plotY[c]);
                savecolumnbutton[c].setMargin(new Insets(0,0,0,0));
                savecolumnbutton[c].setToolTipText("Save curve "+(c+1)+" as new table column");
                savecolumnbutton[c].addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                        addNewColumn(c, false);
                        updatePlot();
                    }
                });
                
//                savecolumnpanelgroup[c].add(savecolumnbutton[c]);
//                SpringUtil.makeCompactGrid (savecolumnpanelgroup[c], 1, savecolumnpanelgroup[c].getComponentCount(), 0,0,0,0);                        

                mainsubpanelgroup.add (savecolumnbutton[c]);                        

                usecurvebox[c] = new JCheckBox ("", plotY[c]);
                usecurvebox[c].addItemListener(new ItemListener(){
                    public void itemStateChanged(ItemEvent e) {
                        if (e.getStateChange() == ItemEvent.DESELECTED)
                                plotY[c] = false;
                        else if (e.getStateChange() == ItemEvent.SELECTED)
                                plotY[c] = true;
                        savecolumnbutton[c].setEnabled(plotY[c]);
                        updatePlot();}});
                usecurvebox[c].setHorizontalAlignment(JLabel.CENTER);
                mainsubpanelgroup.add (usecurvebox[c]);

                autoscaleincludebox[c] = new JCheckBox ("", ASInclude[c]);
                autoscaleincludebox[c].addItemListener(new ItemListener(){
                    public void itemStateChanged(ItemEvent e) {
                        if (e.getStateChange() == ItemEvent.DESELECTED)
                                ASInclude[c] = false;
                        else if (e.getStateChange() == ItemEvent.SELECTED)
                                ASInclude[c] = true;
                        updatePlot();}});
                autoscaleincludebox[c].setHorizontalAlignment(JLabel.CENTER);
                mainsubpanelgroup.add (autoscaleincludebox[c]);
                
                xdatacolumn[c] = new JComboBox(columnswd);
                
                xdatacolumn[c].setSelectedItem(xlabel[c]);
                if (!IJ.isLinux())
                    xdatacolumn[c].setPrototypeDisplayValue("123456789012345678");
                else
                    xdatacolumn[c].setPrototypeDisplayValue("1234567890123456");
                xdatacolumn[c].addActionListener(new ActionListener(){
                    public void actionPerformed(ActionEvent ae) {
                        xlabel[c] = (String) xdatacolumn[c].getSelectedItem();
                        updatePlot();}});
                mainsubpanelgroup.add (xdatacolumn[c]);

                fromMagBox[c] = new JCheckBox ("", fromMag[c]);
                fromMagBox[c].addItemListener(new ItemListener(){
                    public void itemStateChanged(ItemEvent e) {
                        if (e.getStateChange() == ItemEvent.DESELECTED)
                                fromMag[c] = false;
                        else if (e.getStateChange() == ItemEvent.SELECTED)
                                fromMag[c] = true;
                        updatePlot();}});
                fromMagBox[c].setHorizontalAlignment(JLabel.CENTER);
                mainsubpanelgroup.add (fromMagBox[c]);                   

                ydatacolumn[c] = new JComboBox(columns);
                ydatacolumn[c].setSelectedItem(ylabel[c]);
                if (!IJ.isLinux())
                    ydatacolumn[c].setPrototypeDisplayValue("123456789012345678");
                else
                    ydatacolumn[c].setPrototypeDisplayValue("1234567890123456");
                ydatacolumn[c].addActionListener(new ActionListener(){
                    public void actionPerformed(ActionEvent ae) {
                        ylabel[c] = (String) ydatacolumn[c].getSelectedItem();
                        if (!disableUpdatePlotBox)
                            {
                            usecurvebox[c].setSelected(true);
                            plotY[c] = true;
                            savecolumnbutton[c].setEnabled(plotY[c]);
                            }
                        updatePlot();}});
                mainsubpanelgroup.add (ydatacolumn[c]);

                errorcolumnbox[c] = new JCheckBox ("", showErrors[c]);
                errorcolumnbox[c].setToolTipText("<HTML>Show automatic Y-error bars when data has predefined error available.<BR>"+
                                                    "Compatible data columns are: Source-Sky_XX, rel_flux_XX, tot_C_cnts, tot_C_cnts-XX.<BR>"+
                                                    "To display custom user defined error, select the column containing the desired error<BR>"+
                                                    "under 'Y-operand' and then select 'custom error' under 'Function'.</HTML>");                        
                errorcolumnbox[c].addItemListener(new ItemListener(){
                    public void itemStateChanged(ItemEvent e) {
                        if (e.getStateChange() == ItemEvent.DESELECTED)
                                showErrors[c] = false;
                        else if (e.getStateChange() == ItemEvent.SELECTED)
                            {
                            if (operatorIndex[c] == 6)
                                {
                                operatorIndex[c] = 0;
                                operatorselection[c].setSelectedIndex(operatorIndex[c]);
                                }
                            showErrors[c] = true;
                            }
                        updatePlot(); }});
                errorcolumnbox[c].setHorizontalAlignment(JLabel.CENTER);
                mainsubpanelgroup.add (errorcolumnbox[c]);

                operatorselection[c] = new JComboBox(operators);
                operatorselection[c].setSelectedIndex(operatorIndex[c]);
                operatorselection[c].setPrototypeDisplayValue("123456789012");
                operatorselection[c].addActionListener(new ActionListener(){
                    public void actionPerformed(ActionEvent ae) {
                        operatorIndex[c] = operatorselection[c].getSelectedIndex();
                        if (operatorIndex[c] == 6)
                            {
                            showErrors[c] = false;
                            errorcolumnbox[c].setSelected(false);
                            }
                        updatePlot();}});
                mainsubpanelgroup.add (operatorselection[c]);

                operatorcolumn[c] = new JComboBox(columns);
                operatorcolumn[c].setSelectedItem(oplabel[c]);
                if (!IJ.isLinux())
                    operatorcolumn[c].setPrototypeDisplayValue("123456789012345678");
                else
                    operatorcolumn[c].setPrototypeDisplayValue("1234567890123456");
                operatorcolumn[c].addActionListener(new ActionListener(){
                    public void actionPerformed(ActionEvent ae) {
                        oplabel[c] = (String) operatorcolumn[c].getSelectedItem();
                        updatePlot();}});
                mainsubpanelgroup.add (operatorcolumn[c]);


                markercolorselection[c] = new JComboBox(colors);
//                        markercolorselection[c].setPreferredSize(new Dimension(100, 25));
                markercolorselection[c].setSelectedIndex(colorIndex[c]);
                markercolorselection[c].setFont(new Font("Dialog", Font.BOLD, 12));
                markercolorselection[c].setForeground(color[c]);
                markercolorselection[c].setMaximumRowCount(16);
                markercolorselection[c].setPrototypeDisplayValue("12345678");
                markercolorselection[c].addActionListener(new ActionListener(){
                    public void actionPerformed(ActionEvent ae) {
                        colorIndex[c] = markercolorselection[c].getSelectedIndex();
                        color[c] = colorOf(colorIndex[c]);
                        curvelabel[c].setBorder(BorderFactory.createLineBorder(color[c], 2));
                        othercurvelabel[c].setBorder(BorderFactory.createLineBorder(color[c], 2));
                        markercolorselection[c].setForeground(color[c]);
                        markersymbolselection[c].setForeground(color[c]);
                        updatePlot();}});
                mainsubpanelgroup.add (markercolorselection[c]);
                

                markersymbolselection[c] = new JComboBox(markers);
//                        markersymbolselection[c].setPreferredSize(new Dimension(100, 25));
                markersymbolselection[c].setForeground(color[c]);
                markersymbolselection[c].setFont(new Font("Dialog", Font.BOLD, 12));
                markersymbolselection[c].setSelectedIndex(markerIndex[c]);
                markersymbolselection[c].setPrototypeDisplayValue("12345");
                markersymbolselection[c].addActionListener(new ActionListener(){
                    public void actionPerformed(ActionEvent ae) {
                        markerIndex[c] = markersymbolselection[c].getSelectedIndex();
                        updatePlot();}});
                mainsubpanelgroup.add (markersymbolselection[c]);                

                uselinesbox[c] = new JCheckBox ("", lines[c]);
                uselinesbox[c].addItemListener(new ItemListener(){
                    public void itemStateChanged(ItemEvent e) {
                        if (e.getStateChange() == ItemEvent.DESELECTED)
                                lines[c] = false;
                        else if (e.getStateChange() == ItemEvent.SELECTED)
                                lines[c] = true;
                        updatePlot();}});
                uselinesbox[c].setHorizontalAlignment(JLabel.CENTER);
                mainsubpanelgroup.add (uselinesbox[c]);
                
                binsizespinnermodel[c] = new SpinnerNumberModel(new Integer(binSize[c]),new Integer(1), null, new Integer(1));
                binsizespinner[c] = new JSpinner(binsizespinnermodel[c]);
                binsizespinner[c].setMaximumSize(new Dimension(50,25));
                binsizespinner[c].setPreferredSize(new Dimension(50,25));
                binsizespinner[c].addChangeListener (new ChangeListener()
                    {
                    public void stateChanged(ChangeEvent ev)
                        {
                        binSize[c] = ((Integer)binsizespinner[c].getValue()).intValue();
                        Prefs.set("plot.binSize"+c,binSize[c]);
                        updatePlot();
                        }
                    });
                binsizespinner[c].addMouseWheelListener( new MouseWheelListener()
                    {
                    public void mouseWheelMoved( MouseWheelEvent e )
                        {
                        int newValue =  new Integer(((Integer)binsizespinner[c].getValue()).intValue() - e.getWheelRotation());
                        if (newValue > 0) binsizespinner[c].setValue(newValue);
                        }
                    });
                mainsubpanelgroup.add (binsizespinner[c]);  
                
                }

        static void constructOtherGroup(JPanel mainsubpanelgroup, final int c)
                {
                if (!useWideDataPanel)
                        {
                        othercurvelabel[c] = new JLabel (""+(c+1));
                        othercurvelabel[c].setBorder(BorderFactory.createLineBorder(color[c], 2));
                        othercurvelabel[c].setHorizontalAlignment(JLabel.CENTER);
                        mainsubpanelgroup.add (othercurvelabel[c]);
                        }
        //DETREND PANEL GROUP
                
                detrendpanelgroup[c] = new JPanel (new SpringLayout());
                detrendpanelgroup[c].setMaximumSize(new Dimension(350,20));
                detrendpanelgroup[c].setBorder(BorderFactory.createLineBorder(Color.LIGHT_GRAY));
//                customscalepanelgroup[c].setBorder(BorderFactory.createCompoundBorder(BorderFactory.createLineBorder(new Color(122,138,153)),BorderFactory.createLineBorder(new Color(184,207,229))));


        //DETREND MODE


                detrendtypecombobox[c] = new JComboBox(detrendiconlist);
                detrendtypecombobox[c].setSelectedIndex(detrendFitIndex[c]);
                detrendtypecombobox[c].setMaximumRowCount(9);
                detrendtypecombobox[c].setToolTipText("Select 'user' to manually select coefficient or select region type to auto-fit for the best coefficient(s) for the selected trend data column(s).");
                detrendtypecombobox[c].setPreferredSize(new Dimension(50,16));
                detrendtypecombobox[c].addActionListener(new ActionListener(){
                    public void actionPerformed(ActionEvent ae) {
                        detrendFitIndex[c] = detrendtypecombobox[c].getSelectedIndex();
                        detrendfactorspinner[c].setEnabled(detrendFitIndex[c]==1); 
                        Prefs.set("plot.detrendFitIndex"+c,detrendFitIndex[c]);
                        for (int v = 0; v < maxDetrendVars; v++)
                            {
                            if (detrendFitIndex[c] != 0 && detrendIndex[c][v] != 0)
                                detrendVarButton[c][v].setBackground(new Color(0,193,0));
                            else
                                detrendVarButton[c][v].setBackground(defaultBackground);
                            }
                        updatePlot();}});
                detrendpanelgroup[c].add (detrendtypecombobox[c]);    
                
        //DETREND VARIABLE SELECTORS
                
                detrendVarRadioGroup[c] = new ButtonGroup();
                for (int v = 0; v < maxDetrendVars; v++)
                    {
                    constructDetrendVarRadioGroup(c, v);
                    }
                if (c == 1) defaultBackground = detrendVarButton[c][0].getBackground();
              

        //DETREND SPINNER

                detrendfactormodel[c] = new SpinnerNumberModel(new Double(detrendFactor[c][detrendVarDisplayed[c]]), null, null, new Double(detrendFactorStep[c][detrendVarDisplayed[c]]));
                detrendfactorspinner[c] = new JSpinner(detrendfactormodel[c]);
                detrendfactorspinner[c].setToolTipText("<HTML>Y-dataset values are detrended by applying the product(s)<BR>"+
                                                     "of the coefficient(s) and the corresponding 'Trend Dataset(s)'.<BR>"+
                                                     "Right click to set spinner stepsize.</HTML>");
                detrendfactorspinner[c].setPreferredSize(new Dimension(75,25));
                detrendfactorspinner[c].setEnabled(detrendFitIndex[c]==1);
                detrendfactorspinner[c].setEditor(new JSpinner.NumberEditor(detrendfactorspinner[c], "0.#########")); 
                detrendfactorspinner[c].addChangeListener (new ChangeListener(){
                    public void stateChanged(ChangeEvent ev)
                        {
                        if (detrendFitIndex[c]==1)
                            {
                            detrendFactor[c][detrendVarDisplayed[c]]= ((Double)detrendfactorspinner[c].getValue()).doubleValue();
                            Prefs.set("plot.detrendFactor["+c+"]["+detrendVarDisplayed[c]+"]",detrendFactor[c][detrendVarDisplayed[c]]);
                            updatePlot();
                            }
                        }});
                detrendfactorspinner[c].addMouseWheelListener( new MouseWheelListener()
                    {
                    public void mouseWheelMoved( MouseWheelEvent e ) {
                        if (detrendFitIndex[c]==1)
                            {                        
                            detrendfactorspinner[c].setValue(((Double)detrendfactorspinner[c].getValue()).doubleValue()-e.getWheelRotation()*detrendFactorStep[c][detrendVarDisplayed[c]]);
                            }
                        }});
                detrendpanelgroup[c].add (detrendfactorspinner[c]);

        //DETREND FACTOR SPINNER STEPSIZE POPUP

                detrendfactorsteppopup[c] = new JPopupMenu();
                detrendfactorsteppanel[c] = new JPanel();
                detrendfactorstepmodel[c] = new SpinnerListModel(spinnerscalelist);
                detrendfactorstepspinner[c] = new JSpinner(detrendfactorstepmodel[c]);
                detrendfactorstepspinner[c].setValue(convertToText(detrendFactorStep[c][detrendVarDisplayed[c]]));
                detrendfactorstepspinner[c].addChangeListener (new ChangeListener()
                    {
                    public void stateChanged(ChangeEvent ev)
                        {
                        detrendFactorStep[c][detrendVarDisplayed[c]] = Double.parseDouble((String)detrendfactorstepspinner[c].getValue());
                        Prefs.set("plot.detrendFactorStep["+c+"]["+detrendVarDisplayed[c]+"]",detrendFactorStep[c][detrendVarDisplayed[c]]);
                        detrendfactorspinner[c].setModel(new SpinnerNumberModel(new Double(detrendFactor[c][detrendVarDisplayed[c]]), null, null, new Double(detrendFactorStep[c][detrendVarDisplayed[c]])));
                        detrendfactorspinner[c].setEditor(new JSpinner.NumberEditor(detrendfactorspinner[c], "0.#########"));
                        }
                    });
                detrendfactorsteplabel[c] = new JLabel ("Stepsize:");
                detrendfactorsteppanel[c].add(detrendfactorsteplabel[c]);
                detrendfactorsteppanel[c].add(detrendfactorstepspinner[c]);
                detrendfactorsteppopup[c].add(detrendfactorsteppanel[c]);
                detrendfactorsteppopup[c].addPopupMenuListener (new PopupMenuListener() {
                    public void popupMenuCanceled(PopupMenuEvent popupMenuEvent) { }
                    public void popupMenuWillBecomeInvisible(PopupMenuEvent popupMenuEvent) { }
                    public void popupMenuWillBecomeVisible(PopupMenuEvent popupMenuEvent) { } });
                detrendfactorspinner[c].setComponentPopupMenu(detrendfactorsteppopup[c]);
                
        //DETREND COLUMN
                
                detrendbox[c] = new JComboBox(columnsDetrend);
                detrendbox[c].setMaximumRowCount(16);
                for (int v = 0; v < maxDetrendVars; v++)
                    {
                    if (detrendlabel[c][v].trim().equals("") || (table != null && table.getColumnIndex(detrendlabel[c][v]) == MeasurementTable.COLUMN_NOT_FOUND))
                        {
                        detrendbox[c].setSelectedIndex(0);
                        detrendIndex[c][v] = 0;
                        detrendlabel[c][v] = "";
                        }
                    else
                        {
                        detrendbox[c].setSelectedItem(detrendlabel[c][v]); //if in table heading names, should be in detrendbox column names
                        detrendIndex[c][v] = detrendbox[c].getSelectedIndex();
                        }
                    if (detrendFitIndex[c] != 0 && detrendIndex[c][v] != 0)
                        detrendVarButton[c][v].setBackground(new Color(0,193,0));
                    else
                        detrendVarButton[c][v].setBackground(defaultBackground);
                    }
                if (detrendlabel[c][detrendVarDisplayed[c]].trim().equals("") || (table != null && table.getColumnIndex(detrendlabel[c][detrendVarDisplayed[c]]) == MeasurementTable.COLUMN_NOT_FOUND))
                    {
                    detrendbox[c].setSelectedIndex(0);
                    detrendIndex[c][detrendVarDisplayed[c]] = 0;
                    detrendlabel[c][detrendVarDisplayed[c]] = "";                    
                    }
                else
                    {                
                    detrendbox[c].setSelectedItem(detrendlabel[c][detrendVarDisplayed[c]]);
                    detrendIndex[c][detrendVarDisplayed[c]] = detrendbox[c].getSelectedIndex();
                    }
                
                if(!IJ.isLinux())
                    detrendbox[c].setPrototypeDisplayValue("123456789012345678");
                else
                    detrendbox[c].setPrototypeDisplayValue("1234567890123456");
                detrendbox[c].addActionListener(new ActionListener(){
                    public void actionPerformed(ActionEvent ae) {
                        detrendlabel[c][detrendVarDisplayed[c]] = (String) detrendbox[c].getSelectedItem();
                        detrendIndex[c][detrendVarDisplayed[c]] = detrendbox[c].getSelectedIndex();
                        detrendVarButton[c][detrendVarDisplayed[c]].setToolTipText(detrendlabel[c][detrendVarDisplayed[c]].trim().equals("")?(detrendlabelhold[c][detrendVarDisplayed[c]].trim().equals("")?"unused":"["+detrendlabelhold[c][detrendVarDisplayed[c]]+"]"):detrendlabel[c][detrendVarDisplayed[c]]);
                        if (detrendFitIndex[c] != 0 && detrendIndex[c][detrendVarDisplayed[c]] != 0)
                            detrendVarButton[c][detrendVarDisplayed[c]].setBackground(new Color(0,193,0));
                        else
                            detrendVarButton[c][detrendVarDisplayed[c]].setBackground(defaultBackground);
                        updatePlot();}});
                detrendpanelgroup[c].add (detrendbox[c]);                

                SpringUtil.makeCompactGrid (detrendpanelgroup[c], 1, detrendpanelgroup[c].getComponentCount(), 0,0,0,0);
                mainsubpanelgroup.add (detrendpanelgroup[c]);


                normpanelgroup[c] = new JPanel (new SpringLayout());
                normpanelgroup[c].setMaximumSize(new Dimension(50,25));
                normpanelgroup[c].setBorder(BorderFactory.createEmptyBorder());              
                
                normtypecombobox[c] = new JComboBox(normiconlist);
                normtypecombobox[c].setSelectedIndex(normIndex[c]);
                normtypecombobox[c].setPreferredSize(new Dimension(50,16));
                normtypecombobox[c].addActionListener(new ActionListener(){
                    public void actionPerformed(ActionEvent ae) {
                        normIndex[c] = normtypecombobox[c].getSelectedIndex();
                        updatePlot();}});
                normpanelgroup[c].add (normtypecombobox[c]);  
                SpringUtil.makeCompactGrid (normpanelgroup[c], 1, normpanelgroup[c].getComponentCount(), 0,0,0,0);
                mainsubpanelgroup.add (normpanelgroup[c]);                

                usemmagbox[c] = new JCheckBox ("", mmag[c]);
                usemmagbox[c].addItemListener(new ItemListener(){
                    public void itemStateChanged(ItemEvent e) {
                        if (e.getStateChange() == ItemEvent.DESELECTED)
                                mmag[c] = false;
                        else if (e.getStateChange() == ItemEvent.SELECTED)
                                mmag[c] = true;
                        updatePlot();}});
                usemmagbox[c].setHorizontalAlignment(JLabel.CENTER);
                mainsubpanelgroup.add (usemmagbox[c]);                

                Insets buttonMargin = new Insets(0,0,0,0);

                customscalepanelgroup[c] = new JPanel (new SpringLayout());
                customscalepanelgroup[c].setMaximumSize(new Dimension(195,20));
                customscalepanelgroup[c].setBorder(BorderFactory.createLineBorder(Color.LIGHT_GRAY));
//                customscalepanelgroup[c].setBorder(BorderFactory.createCompoundBorder(BorderFactory.createLineBorder(new Color(122,138,153)),BorderFactory.createLineBorder(new Color(184,207,229))));

        // GRAB "FORCE" SCALE AND SHIFT VALUES
//                grabautopanel[c] = new JPanel (new SpringLayout());

                grabautobutton[c] = new JButton(forceIcon);
                grabautobutton[c].setMargin(buttonMargin);
                grabautobutton[c].setToolTipText("Transfer \"Scale Rel\" scale and shift values to absolute values");
                grabautobutton[c].addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                        customScaleFactor[c] = autoScaleFactor[c]*yRange/yWidth[c];
//                        customscalespinner[c].setValue(customScaleFactor[c]);
                        customShiftFactor[c] = yMid + yRange*autoShiftFactor[c] - customScaleFactor[c]*yMidpoint[c];
//                        customshiftspinner[c].setValue(customShiftFactor[c]);
                        force[c] = false;
                        forcebox[c].setSelected(false);
                        customscalespinner[c].setModel(new SpinnerNumberModel(new Double(customScaleFactor[c]), null, null, new Double(customScaleStep[c])));
                        customscalespinner[c].setEditor(new JSpinner.NumberEditor(customscalespinner[c], "0.#########"));
                        customscalestepspinner[c].setValue(convertToText(customScaleStep[c]));
                        customshiftspinner[c].setModel(new SpinnerNumberModel(new Double(customShiftFactor[c]), null, null, new Double(customShiftStep[c])));
                        customshiftspinner[c].setEditor(new JSpinner.NumberEditor(customshiftspinner[c], "0.#########"));
                        customshiftstepspinner[c].setValue(convertToText(customShiftStep[c]));
                        Prefs.set("plot.force"+c,force[c]);
                        updatePlot();
                    }
                });
                customscalepanelgroup[c].add (grabautobutton[c]);
//                SpringUtil.makeCompactGrid (grabautopanel[c],1,1, 0,0,0,0);
//                mainsubpanelgroup.add (grabautopanel[c]);
//                customscalepanelgroup[c].add (grabautopanel[c]);

        //FORCE CHECKBOX

                forcebox[c] = new JCheckBox ("", force[c]);
                forcebox[c].setToolTipText("Select to arbitrarily scale data to fit on plot (\"Scale\" and \"Shift\" by a percentage of the plot Y-range)");
                forcebox[c].addItemListener(new ItemListener(){
                    public void itemStateChanged(ItemEvent e) {
                        if (e.getStateChange() == ItemEvent.DESELECTED)
                            force[c] = false;
                        else if (e.getStateChange() == ItemEvent.SELECTED)
                            force[c] = true;
                        customscalespinner[c].setModel(new SpinnerNumberModel(new Double(force[c]?autoScaleFactor[c]*100.0:customScaleFactor[c]), null, null, new Double(force[c]?autoScaleStep[c]:customScaleStep[c])));
                        customscalespinner[c].setEditor(new JSpinner.NumberEditor(customscalespinner[c], "0.#########"));
                        customscalestepspinner[c].setValue(convertToText(force[c]?autoScaleStep[c]:customScaleStep[c]));
                        customshiftspinner[c].setModel(new SpinnerNumberModel(new Double(force[c]?autoShiftFactor[c]*100:customShiftFactor[c]), null, null, new Double(force[c]?autoShiftStep[c]:customShiftStep[c])));
                        customshiftspinner[c].setEditor(new JSpinner.NumberEditor(customshiftspinner[c], "0.#########"));
                        customshiftstepspinner[c].setValue(convertToText(force[c]?autoShiftStep[c]:customShiftStep[c]));
                        Prefs.set("plot.force"+c,force[c]);
                        updatePlot();}});
//                forcebox[c].setHorizontalAlignment(JLabel.CENTER);
                customscalepanelgroup[c].add (forcebox[c]);

        //SCALE SPINNER

                customscalemodel[c] = new SpinnerNumberModel(new Double(force[c]?autoScaleFactor[c]*100:customScaleFactor[c]), null, null, new Double(force[c]?autoScaleStep[c]:customScaleStep[c]));
                customscalespinner[c] = new JSpinner(customscalemodel[c]);
                customscalespinner[c].setEditor(new JSpinner.NumberEditor(customscalespinner[c], "0.#########"));
                customscalespinner[c].setToolTipText("<HTML>If 'Scale Rel' is deselected, the Y-dataset values are multiplied by this factor<BR>"+
                                                     "(and by the factor of 10 specified in the 'Y-axis' box of the 'Multi-plot Main' panel).<BR>"+
                                                     "If 'Scale Rel' is selected, the dataset values are arbitrarily scaled to fit on a 'Scale' percentage of the Y-axis range.<BR>"+
                                                     "Right click to set spinner stepsize.</HTML>");
                customscalespinner[c].setPreferredSize(new Dimension(75,25));

//                customscaleeditor[c] = (JSpinner.NumberEditor)customscalespinner[c].getEditor();
//                customscaleformat[c] = customscaleeditor[c].getFormat();
//                customscaleformat[c].setMaximumFractionDigits(9);

                customscalespinner[c].addChangeListener (new ChangeListener()
                    {
                    public void stateChanged(ChangeEvent ev)
                        {
                        if (force[c])
                            autoScaleFactor[c]= ((Double)customscalespinner[c].getValue()).doubleValue()/100.0 ;
                        else
                            customScaleFactor[c]= ((Double)customscalespinner[c].getValue()).doubleValue() ;
                        updatePlot();
                        }
                    });
                customscalespinner[c].addMouseWheelListener( new MouseWheelListener()
                    {
                    public void mouseWheelMoved( MouseWheelEvent e )
                        {
                        if (force[c])
                            customscalespinner[c].setValue(((Double)customscalespinner[c].getValue()).doubleValue()-e.getWheelRotation()*autoScaleStep[c]);
                        else
                            customscalespinner[c].setValue(((Double)customscalespinner[c].getValue()).doubleValue()-e.getWheelRotation()*customScaleStep[c]);
                        }
                    });
                customscalepanelgroup[c].add (customscalespinner[c]);

        //SCALE SPINNER STEPSIZE POPUP

                customscalesteppopup[c] = new JPopupMenu();
                customscalesteppanel[c] = new JPanel();
                customscalestepmodel[c] = new SpinnerListModel(spinnerscalelist);
                customscalestepspinner[c] = new JSpinner(customscalestepmodel[c]);
                customscalestepspinner[c].setValue(convertToText(force[c]?autoScaleStep[c]:customScaleStep[c]));
                customscalestepspinner[c].addChangeListener (new ChangeListener()
                    {
                    public void stateChanged(ChangeEvent ev)
                        {
                        if (force[c])
                            {
                            autoScaleStep[c] = Double.parseDouble((String)customscalestepspinner[c].getValue());
                            Prefs.set("plot.autoScaleStep"+c,autoScaleStep[c]);
                            }
                        else
                            {
                            customScaleStep[c] = Double.parseDouble((String)customscalestepspinner[c].getValue());
                            Prefs.set("plot.customScaleStep"+c,customScaleStep[c]);
                            }
                        customscalespinner[c].setModel(new SpinnerNumberModel(new Double(force[c]?autoScaleFactor[c]*100:customScaleFactor[c]), null, null, new Double(force[c]?autoScaleStep[c]:customScaleStep[c])));
                        customscalespinner[c].setEditor(new JSpinner.NumberEditor(customscalespinner[c], "0.#########"));
//                        customscaleeditor[c] = (JSpinner.NumberEditor)customscalespinner[c].getEditor();
//                        customscaleformat[c] = customscaleeditor[c].getFormat();
//                        customscaleformat[c].setMaximumFractionDigits(9);
                        }
                    });
                customscalesteplabel[c] = new JLabel ("Stepsize:");
                customscalesteppanel[c].add(customscalesteplabel[c]);
                customscalesteppanel[c].add(customscalestepspinner[c]);
                customscalesteppopup[c].add(customscalesteppanel[c]);
                customscalesteppopup[c].addPopupMenuListener (new PopupMenuListener() {
                    public void popupMenuCanceled(PopupMenuEvent popupMenuEvent) { }
                    public void popupMenuWillBecomeInvisible(PopupMenuEvent popupMenuEvent) { }
                    public void popupMenuWillBecomeVisible(PopupMenuEvent popupMenuEvent) { } });
                customscalespinner[c].setComponentPopupMenu(customscalesteppopup[c]);

        //SHIFT SPINNER

                customshiftmodel[c] = new SpinnerNumberModel(new Double(force[c]?autoShiftFactor[c]*100:customShiftFactor[c]), null, null,
                                                             new Double(force[c]?invertYAxisSign*autoShiftStep[c]:invertYAxisSign*customShiftStep[c]));
                customshiftspinner[c] = new JSpinner(customshiftmodel[c]);
                customshiftspinner[c].setEditor(new JSpinner.NumberEditor(customshiftspinner[c], "0.#########"));
//                customshifteditor[c] = (JSpinner.NumberEditor)customshiftspinner[c].getEditor();
//                customshiftformat[c] = customshifteditor[c].getFormat();
//                customshiftformat[c].setMaximumFractionDigits(6);


                customshiftspinner[c].setPreferredSize(new Dimension(75,25));
                customshiftspinner[c].setToolTipText("<HTML>If 'Scale Rel' is deselected, the 'Shift' value is added to the scaled Y-dataset values before plotting.<BR>"+
                                                     "If 'Scale Rel' is selected, the dataset values are arbitrarily shifted by 'Shift' percentage of the Y-axis plot range.<BR>"+
                                                     "Right click to set spinner stepsize.</HTML>");
                customshiftspinner[c].addChangeListener (new ChangeListener()
                    {
                    public void stateChanged(ChangeEvent ev)
                        {
                        if (force[c])
                            autoShiftFactor[c]=((Double)customshiftspinner[c].getValue()).doubleValue()/100.0;
                        else
                            customShiftFactor[c]=((Double)customshiftspinner[c].getValue()).doubleValue();
                        updatePlot();
                        }
                    });
                customshiftspinner[c].addMouseWheelListener( new MouseWheelListener()
                    {
                    public void mouseWheelMoved( MouseWheelEvent e )
                        {
                        if (force[c])
                            customshiftspinner[c].setValue(((Double)customshiftspinner[c].getValue()).doubleValue() -
                                      invertYAxisSign*e.getWheelRotation()*autoShiftStep[c]);
                        else
                            customshiftspinner[c].setValue(((Double)customshiftspinner[c].getValue()).doubleValue() -
                                      invertYAxisSign*e.getWheelRotation()*customShiftStep[c]);
                        }
                    });
                customscalepanelgroup[c].add (customshiftspinner[c]);

          //SHIFT SPINNER STEPSIZE POPUP

                customshiftsteppopup[c] = new JPopupMenu();
                customshiftsteppanel[c] = new JPanel();
                customshiftstepmodel[c] = new SpinnerListModel(spinnerscalelist);
                customshiftstepspinner[c] = new JSpinner(customshiftstepmodel[c]);
                customshiftstepspinner[c].setValue(convertToText(force[c]?autoShiftStep[c]:customShiftStep[c]));
                customshiftstepspinner[c].addChangeListener (new ChangeListener()
                    {
                    public void stateChanged(ChangeEvent ev)
                        {
                        if (force[c])
                            {
                            autoShiftStep[c] = Double.parseDouble((String)customshiftstepspinner[c].getValue());
                            Prefs.set("plot.autoShiftStep"+c,autoShiftStep[c]);
                            }
                        else
                            {
                            customShiftStep[c] = Double.parseDouble((String)customshiftstepspinner[c].getValue());
                            Prefs.set("plot.customShiftStep"+c,customShiftStep[c]);
                            }
                        customshiftspinner[c].setModel(new SpinnerNumberModel(new Double(force[c]?autoShiftFactor[c]*100:customShiftFactor[c]), null, null,
                                                             new Double(force[c]?invertYAxisSign*autoShiftStep[c]:invertYAxisSign*customShiftStep[c])));
                        customshiftspinner[c].setEditor(new JSpinner.NumberEditor(customshiftspinner[c], "0.#########"));
//                        customshifteditor[c] = (JSpinner.NumberEditor)customshiftspinner[c].getEditor();
//                        customshiftformat[c] = customshifteditor[c].getFormat();
//                        customshiftformat[c].setMaximumFractionDigits(6);
                        }
                    });
                customshiftsteplabel[c] = new JLabel ("Stepsize:");
                customshiftsteppanel[c].add(customshiftsteplabel[c]);
                customshiftsteppanel[c].add(customshiftstepspinner[c]);
                customshiftsteppopup[c].add(customshiftsteppanel[c]);
                customshiftsteppopup[c].addPopupMenuListener (new PopupMenuListener() {
                    public void popupMenuCanceled(PopupMenuEvent popupMenuEvent) { }
                    public void popupMenuWillBecomeInvisible(PopupMenuEvent popupMenuEvent) { }
                    public void popupMenuWillBecomeVisible(PopupMenuEvent popupMenuEvent) { } });
                customshiftspinner[c].setComponentPopupMenu(customshiftsteppopup[c]);
//                SpringUtil.makeCompactGrid (autoscalepanelgroup[c], 1, 3, 0,0,0,0);
                SpringUtil.makeCompactGrid (customscalepanelgroup[c], 1, customscalepanelgroup[c].getComponentCount(), 0,0,0,0);
//                mainsubpanelgroup.add (autoscalepanelgroup[c]);
                mainsubpanelgroup.add (customscalepanelgroup[c]);

//                
               
                
                
        //LEGEND RADIO GROUP AND TEXT FIELD
                
                morelegendradiopanelgroup[c] = new JPanel ();
                morelegendradiopanelgroup[c].setLayout(new BoxLayout(morelegendradiopanelgroup[c],BoxLayout.X_AXIS));
                                
                if (useWideDataPanel)
                        morelegendradiopanelgroup[c].setPreferredSize (new Dimension(225,20));
                else
                        morelegendradiopanelgroup[c].setPreferredSize (new Dimension(325,20));
//                legendslabelgroup.add(Box.createGlue());
                morelegendradiopanelgroup[c].add(Box.createHorizontalStrut(1));
                morelegendradiopanelgroup[c].setBorder(BorderFactory.createLineBorder(Color.lightGray));
                ImageIcon noneLegendIcon = createImageIcon("astroj/images/nonelegend.png", "Disable legend for this data set");
                ImageIcon columnLegendIcon = createImageIcon("astroj/images/columnlegend.png", "Use the column name as the legend for this data set");
                ImageIcon customLegendIcon = createImageIcon("astroj/images/customlegend.png", "Use a custom legend for this data set");
                legendnoneButton[c] = new JRadioButton(noneLegendIcon);
                legendnoneButton[c].setMaximumSize(new Dimension(25,20));
                if (!useLegend[c] && !useColumnName[c])
                    {
                    legendnoneButton[c].setSelected(true);
                    legendnoneButton[c].setBackground(Color.LIGHT_GRAY);
                    }
                else
                    {
                    legendnoneButton[c].setSelected(false);
                    legendnoneButton[c].setBackground(Color.WHITE);
                    }
                morelegendradiopanelgroup[c].add (legendnoneButton[c]);
                morelegendradiopanelgroup[c].add(Box.createHorizontalStrut(1));
                legendcolumnNameButton[c] = new JRadioButton(columnLegendIcon);
                legendcolumnNameButton[c].setMaximumSize(new Dimension(25,20));
                if (useColumnName[c])
                    {
                    legendcolumnNameButton[c].setSelected(true);
                    legendcolumnNameButton[c].setBackground(Color.LIGHT_GRAY);
                    }
                else
                    {
                    legendcolumnNameButton[c].setSelected(false);
                    legendcolumnNameButton[c].setBackground(Color.WHITE);
                    }
                morelegendradiopanelgroup[c].add (legendcolumnNameButton[c]);
                morelegendradiopanelgroup[c].add(Box.createHorizontalStrut(1));
                legendcustomNameButton[c] = new JRadioButton(customLegendIcon);
                legendcustomNameButton[c].setMaximumSize(new Dimension(25,20));
                if (!useColumnName[c] && useLegend[c])
                    {
                    legendcustomNameButton[c].setSelected(true);
                    legendcustomNameButton[c].setBackground(Color.LIGHT_GRAY);
                    }
                else
                    {
                    legendcustomNameButton[c].setSelected(false);
                    legendcustomNameButton[c].setBackground(Color.WHITE);
                    }
                morelegendradiopanelgroup[c].add (legendcustomNameButton[c]);
                morelegendradiopanelgroup[c].add(Box.createHorizontalStrut(3));

                morelegendRadioGroup[c] = new ButtonGroup();
                morelegendRadioGroup[c].add(legendnoneButton[c]);
                morelegendRadioGroup[c].add(legendcolumnNameButton[c]);
                morelegendRadioGroup[c].add(legendcustomNameButton[c]);
                legendnoneButton[c].addActionListener(new ActionListener(){
                    public void actionPerformed(ActionEvent ae) {
                        useColumnName[c] = false;
                        useLegend[c] = false;
                        legendnoneButton[c].setBackground(Color.LIGHT_GRAY);
                        legendcolumnNameButton[c].setBackground(Color.WHITE);
                        legendcustomNameButton[c].setBackground(Color.WHITE);
                        updatePlot();}});
                legendcustomNameButton[c].addActionListener(new ActionListener(){
                    public void actionPerformed(ActionEvent ae) {
                        useColumnName[c] = false;
                        useLegend[c] = true;
                        legendnoneButton[c].setBackground(Color.WHITE);
                        legendcolumnNameButton[c].setBackground(Color.WHITE);
                        legendcustomNameButton[c].setBackground(Color.LIGHT_GRAY);
                        updatePlot();}});
                legendcolumnNameButton[c].addActionListener(new ActionListener(){
                    public void actionPerformed(ActionEvent ae) {
                        useColumnName[c] = true;
                        useLegend[c] = false;
                        legendnoneButton[c].setBackground(Color.WHITE);
                        legendcolumnNameButton[c].setBackground(Color.LIGHT_GRAY);
                        legendcustomNameButton[c].setBackground(Color.WHITE);
                        updatePlot();}});
                
                legendnoneButton[c].setComponentPopupMenu(legendpopup);
                legendnoneButton[c].setToolTipText("Disable Legend -- Right click to set legend preferences");
                legendcustomNameButton[c].setComponentPopupMenu(legendpopup);
                legendcustomNameButton[c].setToolTipText("Use Custom Legend -- Right click to set legend preferences");
                legendcolumnNameButton[c].setComponentPopupMenu(legendpopup);
                legendcolumnNameButton[c].setToolTipText("Use Column Name as Legend -- Right click to set legend preferences");


                // LEGEND TEXT FIELD
                morelegendField[c] = new JTextField (legend[c]);
                morelegendField[c].setBorder (BorderFactory.createEmptyBorder(0,4,0,4));
                morelegendField[c].getDocument().addDocumentListener (new DocumentListener()
                    {
                    public void insertUpdate (DocumentEvent ev)
                        {
                        legend[c] = morelegendField[c].getText();
                        updatePlot();
                        }
                    public void removeUpdate (DocumentEvent ev)
                        {
                        legend[c] = morelegendField[c].getText();
                        updatePlot();
                        }
                    public void changedUpdate (DocumentEvent ev)
                        {
                        legend[c] = morelegendField[c].getText();
                        updatePlot();
                        }
                    });
                if (useWideDataPanel)
                        morelegendField[c].setPreferredSize (new Dimension(150,20));
                else
                        morelegendField[c].setPreferredSize (new Dimension(250,20));
                morelegendField[c].setHorizontalAlignment (JTextField.LEFT);
                morelegendField[c].setComponentPopupMenu(legendpopup);
                morelegendField[c].setToolTipText("Right click to set legend preferences");
                morelegendradiopanelgroup[c].add(morelegendField[c]);
                
//                SpringUtil.makeCompactGrid (morelegendradiopanelgroup[c], 1, morelegendradiopanelgroup[c].getComponentCount(), 0,0,0,0);
                mainsubpanelgroup.add (morelegendradiopanelgroup[c]);                

                if (useWideDataPanel)
                        {
                        othercurvelabel[c] = new JLabel (""+(c+1));
                        othercurvelabel[c].setBorder(BorderFactory.createLineBorder(color[c], 2));
                        othercurvelabel[c].setHorizontalAlignment(JLabel.CENTER);
                        mainsubpanelgroup.add (othercurvelabel[c]);
                        }
                }
        
        static void constructDetrendVarRadioGroup(final int c, final int v)
            {
            detrendVarButton[c][v] = new JRadioButton("");
            detrendVarButton[c][v].setMaximumSize(new Dimension(25,20));
            detrendVarRadioGroup[c].add(detrendVarButton[c][v]);
            detrendVarButton[c][v].setSelected(detrendVarDisplayed[c] == v);
            detrendVarButton[c][v].setToolTipText(detrendlabel[c][v].trim().equals("")?(detrendlabelhold[c][v].trim().equals("")?"unused":"["+detrendlabelhold[c][v]+"]"):detrendlabel[c][v]);
            detrendVarButton[c][v].addActionListener(new ActionListener(){
                public void actionPerformed(ActionEvent ae) {
                    int mods = ae.getModifiers();
                    if ((mods & ActionEvent.CTRL_MASK) != 0 && (mods & ActionEvent.SHIFT_MASK) == 0 ) //copy this column name to all rows
                        {
                        for (int cc = 0; cc < maxCurves; cc++)
                            {
                            detrendVarDisplayed[cc] = v;
                            detrendVarButton[cc][v].setSelected(true);
                            detrendlabel[cc][detrendVarDisplayed[cc]]=detrendlabel[c][detrendVarDisplayed[c]];
                            detrendFactorStep[cc][detrendVarDisplayed[cc]]=detrendFactorStep[c][detrendVarDisplayed[c]];
                            detrendfactorstepspinner[cc].setValue(convertToText(detrendFactorStep[cc][detrendVarDisplayed[cc]]));
                            if (detrendlabel[cc][detrendVarDisplayed[cc]].trim().equals("") || (table != null && table.getColumnIndex(detrendlabel[cc][detrendVarDisplayed[cc]]) == MeasurementTable.COLUMN_NOT_FOUND))
                                {
                                detrendbox[cc].setSelectedIndex(0);
                                detrendIndex[cc][detrendVarDisplayed[cc]] = 0;
                                detrendlabel[cc][detrendVarDisplayed[cc]] = "";                    
                                }
                            else
                                {                
                                detrendbox[cc].setSelectedItem(detrendlabel[cc][detrendVarDisplayed[cc]]);
                                detrendIndex[cc][detrendVarDisplayed[cc]] = detrendbox[cc].getSelectedIndex();
                                }                            
                            }
                        }
                    else if ((mods & ActionEvent.CTRL_MASK) == 0 && (mods & ActionEvent.SHIFT_MASK) != 0 ) //select this variable for display on all rows
                        {
                        for (int cc = 0; cc < maxCurves; cc++)
                            {
                            detrendVarDisplayed[cc] = v;
                            detrendVarButton[cc][v].setSelected(true);
                            detrendfactorstepspinner[cc].setValue(convertToText(detrendFactorStep[cc][detrendVarDisplayed[cc]]));
                            if (detrendlabel[cc][detrendVarDisplayed[cc]].trim().equals("") || (table != null && table.getColumnIndex(detrendlabel[cc][detrendVarDisplayed[cc]]) == MeasurementTable.COLUMN_NOT_FOUND))
                                {
                                detrendbox[cc].setSelectedIndex(0);
                                detrendIndex[cc][detrendVarDisplayed[cc]] = 0;
                                detrendlabel[cc][detrendVarDisplayed[cc]] = "";                    
                                }
                            else
                                {                
                                detrendbox[cc].setSelectedItem(detrendlabel[cc][detrendVarDisplayed[cc]]);
                                detrendIndex[cc][detrendVarDisplayed[cc]] = detrendbox[cc].getSelectedIndex();
                                }  
                            }
                        }                    
                    else
                        {
                        if (detrendVarDisplayed[c] == v || (mods & ActionEvent.ALT_MASK) != 0)
                            {
                            if (!detrendVarButton[c][v].getBackground().equals(defaultBackground))
                                {
                                detrendlabelhold[c][v] = detrendlabel[c][v];
                                detrendVarDisplayed[c] = v;
                                detrendfactorstepspinner[c].setValue(convertToText(detrendFactorStep[c][detrendVarDisplayed[c]]));
                                detrendbox[c].setSelectedIndex(0);
                                }
                            else
                                {
                                detrendlabel[c][v] = detrendlabelhold[c][v];
                                detrendVarDisplayed[c] = v;
                                detrendfactorstepspinner[c].setValue(convertToText(detrendFactorStep[c][detrendVarDisplayed[c]]));
                                if (detrendlabel[c][v].trim().equals("") || (table != null && table.getColumnIndex(detrendlabel[c][v]) == MeasurementTable.COLUMN_NOT_FOUND))
                                    {
                                    detrendbox[c].setSelectedIndex(0);
                                    detrendIndex[c][v] = 0;
                                    detrendlabel[c][v] = "";                    
                                    }
                                else
                                    {                
                                    detrendbox[c].setSelectedItem(detrendlabel[c][v]);
                                    detrendIndex[c][v] = detrendbox[c].getSelectedIndex();
                                    }  
                                } 
                            }
                        else
                            {
                            detrendVarDisplayed[c] = v;
                            detrendfactorstepspinner[c].setValue(convertToText(detrendFactorStep[c][detrendVarDisplayed[c]]));
                            if (detrendlabel[c][v].trim().equals("") || (table != null && table.getColumnIndex(detrendlabel[c][v]) == MeasurementTable.COLUMN_NOT_FOUND))
                                {
                                detrendbox[c].setSelectedIndex(0);
                                detrendIndex[c][v] = 0;
                                detrendlabel[c][v] = "";                    
                                }
                            else
                                {                
                                detrendbox[c].setSelectedItem(detrendlabel[c][v]);
                                detrendIndex[c][v] = detrendbox[c].getSelectedIndex();
                                }                              
                            }
                        }
                    }});                    
            if (maxDetrendVars > 1)
                {
                detrendpanelgroup[c].add (detrendVarButton[c][v]);
                detrendpanelgroup[c].add(Box.createHorizontalStrut(1));
                }   

            }
        
        static void showRefStarJPanel()
                {
                refStarFrame = new JFrame ("Multi-plot Reference Star Selection");
                refStarFrame.setIconImage(plotIcon.getImage());
                refStarMainPanel = new JPanel (new SpringLayout());
//                mainsubpanel.addMouseListener(panelMouseListener);
//                refStarMainPanel.addMouseMotionListener(panelMouseMotionListener);                
//                FileDrop fileDrop = new FileDrop(refStarMainPanel, BorderFactory.createEmptyBorder(),new FileDrop.Listener()
//                    {   public void filesDropped( java.io.File[] files )
//                        {
//                        openDragAndDropFiles(files);
//                        }
//                    });

                refStarScrollPane = new JScrollPane(refStarMainPanel);
                refStarFrame.setDefaultCloseOperation(javax.swing.WindowConstants.DISPOSE_ON_CLOSE);
                refStarFrame.addWindowListener(new WindowAdapter(){
                    @Override
                    public void windowClosing(WindowEvent e){
                            refStarPanelWasShowing = false;
                            closeRefStarFrame();
                            }});
                FileDrop fileDrop = new FileDrop(refStarMainPanel, BorderFactory.createEmptyBorder(),new FileDrop.Listener()
                    {   public void filesDropped( java.io.File[] files )
                        {
                        openDragAndDropFiles(files);
                        }
                    });
                multiUpdate = false;
                JPanel refStarAPanel = new JPanel (new SpringLayout());
                refStarAPanel.setBorder(BorderFactory.createTitledBorder(""));
                JLabel refstarlabel1 = new JLabel("No table has been selected in 'Multi-plot Main'.");
                if (table != null)
                    {
                    refstarlabel1 = new JLabel("Select reference stars to include in tot_C_cnts and rel_flux calculations");
                    refstarlabel1.setToolTipText("<html>Updates Ts and Cs in the table column headings and the following table values:<br>"+
                                                "rel_flux_Xx, rel_flux_err_Xx, rel_flux_SNR_Xx, tot_C_cnts, tot_C_err, </html>");
                    }
                refStarAPanel.add(refstarlabel1);
                SpringUtil.makeCompactGrid (refStarAPanel, 1, refStarAPanel.getComponentCount(), 6,6,6,6);
                refStarMainPanel.add (refStarAPanel);                

                JPanel refStarBPanel = new JPanel (new SpringLayout());
                refStarBPanel.setBorder(BorderFactory.createTitledBorder("Reference Star Selection"));

                checkAndLockTable();
              

                numAps = 0;
                while(true)
                    {
                    if (table.getColumnIndex("Source-Sky_C"+(numAps+1)) != MeasurementTable.COLUMN_NOT_FOUND || 
                        table.getColumnIndex("Source-Sky_T"+(numAps+1)) != MeasurementTable.COLUMN_NOT_FOUND)
                        {
                        numAps++;
                        }
                    else
                        {
                        break;
                        }
                    }
                if (lastRefStar >= numAps) lastRefStar = numAps - 1;
                if (lastRefStar < 0) lastRefStar = 0;
                if (numAps > 0)
                    {
                    refStarPanel = new JPanel[numAps];
                    refStarLabel = new JLabel[numAps];
                    refStarCB = new JCheckBox[numAps];
                    isRefStar = new boolean[numAps];

                    JPanel allNonePanel = new JPanel(new SpringLayout());
                    JButton noneButton = new JButton("None");
                    noneButton.setToolTipText("Removes all stars from the reference star set"); 
                    noneButton.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                        multiUpdate = true;
                        for (int r = 1; r < numAps; r++)
                            {
                            refStarCB[r].setSelected(false);
                            }
                        multiUpdate = false;
                        cycleEnabledStarsLess1PressedConsecutive = false;
                        updatePlotEnabled = false;
                        waitForPlotUpdateToFinish();
                        checkAndLockTable();                         
                        updateTotals();
                        updateGUI();
                        updatePlotEnabled = true;
                        if (table != null) table.setLock(false);                        
                        table.show();
                        updatePlot();  
                        }});
                    allNonePanel.add(noneButton);                        
                    JButton allButton = new JButton("All");
                    allButton.setToolTipText("Adds all stars to the reference star set"); 
                    allButton.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                        if (numAps < 2) return;
                        multiUpdate = true;
                        for (int r = 1; r < numAps; r++)
                            {
                            refStarCB[r].setSelected(true);
                            }
                        multiUpdate = false;
                        cycleEnabledStarsLess1PressedConsecutive = false;
                        updatePlotEnabled = false;
                        waitForPlotUpdateToFinish();
                        checkAndLockTable();                         
                        updateTotals();
                        updateGUI();
                        updatePlotEnabled = true;
                        if (table != null) table.setLock(false);                        
                        table.show();
                        updatePlot(); 
                        }});
                    allNonePanel.add(allButton);
                    
                    JButton setButton = new JButton("Set");
                    setButton.setToolTipText("Sets the current enabled stars as the 'cycle' set"); 
                    setButton.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                        cycleEnabledStarsLess1PressedConsecutive = false;
                        }});
                    allNonePanel.add(setButton);                    
                    
                    JButton cycleEnabledStarsLess1Button = new JButton("Cycle Enabled Stars Less One");
                    cycleEnabledStarsLess1Button.setToolTipText("Removes one star at a time from the current selected set"); 
                    cycleEnabledStarsLess1Button.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                        
                        if (numAps > 2)
                            {
                            int numEnabled = 0;
                            for (int i = 1; i < numAps; i++)
                                {
                                if (isRefStar[i]) 
                                    {
                                    numEnabled++;
                                    }
                                }
                            if (numEnabled < (cycleEnabledStarsLess1PressedConsecutive?1:2))
                                {
                                return;
                                }

                            multiUpdate = true;
                            boolean foundOne = false;
                            int startingRefStar = lastRefStar;    
                            if (cycleEnabledStarsLess1PressedConsecutive)
                                {
                                lastRefStar = (lastRefStar + 1)%(numAps);
                                if (lastRefStar < 1) lastRefStar = 1; 
                                }
                            if (!isRefStar[lastRefStar])
                                {
                                for (int i = lastRefStar; i < numAps; i++)
                                    {
                                    if (isRefStar[i]) 
                                        {
                                        lastRefStar = i;
                                        foundOne = true;
                                        break;
                                        }
                                    }
                                if (!foundOne)
                                    {
                                    for (int i = 1; i < lastRefStar; i++)
                                        {
                                        if (isRefStar[i]) 
                                            {
                                            lastRefStar = i;
                                            foundOne = true;
                                            break;
                                            }
                                        }                                        
                                    }
                                } 
                            else
                                {
                                foundOne = true;
                                }
                            
                            if (foundOne)
                                {
                                if (cycleEnabledStarsLess1PressedConsecutive) 
                                    {
                                    refStarCB[startingRefStar].setSelected(true);
                                    }
                                refStarCB[lastRefStar].setSelected(false);
                                multiUpdate = false;
                                cycleEnabledStarsLess1PressedConsecutive = true;
                                
                                updatePlotEnabled = false;
                                waitForPlotUpdateToFinish();
                                checkAndLockTable();                                 
                                updateTotals();
                                updateGUI();
                                updatePlotEnabled = true;
                                if (table != null) table.setLock(false);                        
                                table.show();
                                updatePlot();   
                                }
                            }
                        }});
                    allNonePanel.add(cycleEnabledStarsLess1Button);
                    
                    JButton cycleIndividualStarButton = new JButton("Cycle Individual Stars");
                    cycleIndividualStarButton.setToolTipText("Cycles through all stars considering each as an individual reference star");
                    cycleIndividualStarButton.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                        if (numAps < 2) return;
                        multiUpdate = true;
                        for (int r = 1; r < numAps; r++)
                            {
                            if (r == lastRefStar)
                                refStarCB[r].setSelected(true);
                            else
                                refStarCB[r].setSelected(false);
                            }
                        lastRefStar = (lastRefStar + 1)%(numAps);
                        if (lastRefStar < 1) lastRefStar = 1;                        
                        multiUpdate = false;
                        cycleEnabledStarsLess1PressedConsecutive = false;
                        
                        updatePlotEnabled = false;
                        waitForPlotUpdateToFinish();
                        checkAndLockTable();                         
                        updateTotals();
                        updateGUI();
                        updatePlotEnabled = true;
                        if (table != null) table.setLock(false);                                
                        table.show();
                        updatePlot();    
                        }});
                    allNonePanel.add(cycleIndividualStarButton);                    
                    
                    JLabel dummy1Label = new JLabel("");
                    allNonePanel.add(dummy1Label);                        
                    SpringUtil.makeCompactGrid (allNonePanel, 1, allNonePanel.getComponentCount(), 2,2,2,2);
                    refStarBPanel.add(allNonePanel); 
                    
                    JPanel starsPlusSpacerPanel = new JPanel(new SpringLayout());
                    
                    JPanel starsPanel = new JPanel(new SpringLayout());
                    for (int r = 0; r < numAps; r++)
                        {
                        if (table.getColumnIndex("Source-Sky_C"+(r+1)) != MeasurementTable.COLUMN_NOT_FOUND)
                            {
                            refStarLabel[r] = new JLabel("C"+(r+1));
                            isRefStar[r] = true;
                            }
                        else if (table.getColumnIndex("Source-Sky_T"+(r+1)) != MeasurementTable.COLUMN_NOT_FOUND)
                            {
                            refStarLabel[r] = new JLabel("T"+(r+1));
                            isRefStar[r] = false;
                            }                               
                        else
                            {
                            IJ.beep();
                            IJ.showMessage("Error getting apertures from table");
                            closeRefStarFrame();
                            if (table != null) table.setLock(false);
                            return;
                            }
                        
                        refStarPanel[r] = new JPanel(new SpringLayout());
                        refStarLabel[r].setHorizontalAlignment(JLabel.CENTER);
                        refStarPanel[r].add(refStarLabel[r]);
                        refStarCB[r] = new JCheckBox("", isRefStar[r]);
                        setupRefStarCBListener(r);
                        refStarPanel[r].add(refStarCB[r]);                            
                        SpringUtil.makeCompactGrid (refStarPanel[r], refStarPanel[r].getComponentCount(), 1, 0,0,0,0);
                        starsPanel.add(refStarPanel[r]);
                        }
                    
                    if (numAps/refStarHorzWidth > 0)
                        {
                        int fill = refStarHorzWidth - numAps%refStarHorzWidth;
                        if (fill != 0 && fill != refStarHorzWidth)
                            {
                            JPanel[] dummyPanel = new JPanel[fill];
                            for (int i = 0; i < fill; i++)
                                {
                                dummyPanel[i] = new JPanel();
                                dummyPanel[i].setMaximumSize(refStarPanel[0].getSize());
                                starsPanel.add(dummyPanel[i]);
                                }
                            }
                        SpringUtil.makeCompactGrid (starsPanel, starsPanel.getComponentCount()/refStarHorzWidth, refStarHorzWidth, 4,10,4,10);
                        starsPlusSpacerPanel.add(starsPanel);
                        }
                    else
                        {
                        SpringUtil.makeCompactGrid (starsPanel, 1, starsPanel.getComponentCount(), 4,10,4,10);
                        starsPlusSpacerPanel.add(starsPanel);                        
                        }
                    
                    JPanel spacerPanel = new JPanel(new SpringLayout());
                    JLabel spacerDummyLabel = new JLabel("");
                    spacerPanel.add(spacerDummyLabel);
                    SpringUtil.makeCompactGrid (spacerPanel, spacerPanel.getComponentCount(), 1, 0,0,0,0);
                    starsPlusSpacerPanel.add(spacerPanel);
                    
                    SpringUtil.makeCompactGrid (starsPlusSpacerPanel, 1, starsPlusSpacerPanel.getComponentCount(), 0,0,0,0);
                    refStarBPanel.add(starsPlusSpacerPanel);
                    
                    SpringUtil.makeCompactGrid (refStarBPanel, refStarBPanel.getComponentCount(), 1, 6,6,6,6);
                    refStarMainPanel.add (refStarBPanel); 
                    
                    JPanel reminderPanel = new JPanel();
                    reminderPanel.setBorder(BorderFactory.createTitledBorder("Save/Show Current Configuration"));
                    reminderPanel.setLayout(new BoxLayout(reminderPanel, BoxLayout.LINE_AXIS));
                    
                    JButton saveTableButton = new JButton("Save Table");
                    saveTableButton.setForeground(Color.RED);
                    saveTableButton.setToolTipText("Save the measurement table with current reference star configuration"); 
                    saveTableButton.addActionListener(new ActionListener() {
                        public void actionPerformed(ActionEvent e) {
                            saveData();
                            }});
                    reminderPanel.add(saveTableButton); 
                    
                    reminderPanel.add(Box.createHorizontalStrut(10));
                    
                    JButton saveApsButton = new JButton("Save Apertures");
//                    sendApsButton.setForeground(Color.RED);
                    saveApsButton.setToolTipText("Save new aperture configuration to file"); 
                    saveApsButton.addActionListener(new ActionListener() {
                        public void actionPerformed(ActionEvent e) {
                            saveApertures();
                            }});
                    reminderPanel.add(saveApsButton);                    
                    
                    reminderPanel.add(Box.createHorizontalStrut(10));                     
                    
                    JButton sendApsButton = new JButton("Send to Multi-aperture");
//                    sendApsButton.setForeground(Color.RED);
                    sendApsButton.setToolTipText("Send new aperture configuration to Multi-aperture"); 
                    sendApsButton.addActionListener(new ActionListener() {
                        public void actionPerformed(ActionEvent e) {
                            storeApertures();
                            }});
                    reminderPanel.add(sendApsButton);                     
                    
                    reminderPanel.add(Box.createHorizontalStrut(10));                     
                    
                    JButton showApsButton = new JButton("Show Apertures");
//                    sendApsButton.setForeground(Color.RED);
                    showApsButton.setToolTipText("Show apertures in active image"); 
                    showApsButton.addActionListener(new ActionListener() {
                        public void actionPerformed(ActionEvent e) {
                            showApertures();
                            }});
                    reminderPanel.add(showApsButton);                     
//                    JLabel reminderlabel= new JLabel("   ***** Save table to retain reference star configuration *****");
//                    reminderlabel.setForeground(Color.RED);
//                    reminderPanel.add (reminderlabel);  
                    
                    reminderPanel.add(Box.createGlue()); 
                    
                    refStarMainPanel.add (reminderPanel);
                    if (table != null) table.setLock(false);
                    table.show();
                    }
                else
                    {
                    JLabel refstarlabel2= new JLabel("No 'Source-Sky' columns found in table.");
                    refStarBPanel.add(refstarlabel2); 

                    JLabel refstarlabel3= new JLabel("Enable 'aperture net integrated counts' in aperture settings.");
                    refStarBPanel.add(refstarlabel3); 
                    SpringUtil.makeCompactGrid (refStarBPanel, refStarBPanel.getComponentCount(), 1, 6,6,6,6);
                    refStarMainPanel.add (refStarBPanel); 
                    if (table != null) table.setLock(false);
                    }


                   

                SpringUtil.makeCompactGrid (refStarMainPanel, refStarMainPanel.getComponentCount(), 1, 6,6,6,6);

                refStarFrame.add (refStarScrollPane);
                refStarFrame.pack();
                refStarFrame.setResizable (true);

                if (rememberWindowLocations)
                    {
                    Dimension mainScreenSize = Toolkit.getDefaultToolkit().getScreenSize();
                    if (!Prefs.isLocationOnScreen(new Point(refStarFrameLocationX,refStarFrameLocationY)))
                        {
                        refStarFrameLocationX = mainScreenSize.width/2 - refStarFrame.getWidth()/2;
                        refStarFrameLocationY = mainScreenSize.height/2 - refStarFrame.getHeight()/2;
                        }
                    refStarFrame.setLocation(refStarFrameLocationX, refStarFrameLocationY);
                    }
                refStarFrame.setVisible (true);
                refStarPanelWasShowing = true;
              
                }
        
        
        /** Returns an ImageIcon, or null if the path was invalid. */
        static protected ImageIcon createImageIcon(String path, String description) {
            MultiPlot_ m = new MultiPlot_();
            java.net.URL imgURL = m.getClass().getResource(path);
            if (imgURL != null) {
                return new ImageIcon(imgURL, description);
            } else {
                IJ.log("Couldn't find icon file: " + path);
                return null;
            }
        }
        
        static void setupRefStarCBListener(final int r)
            {
            refStarCB[r].addItemListener(new ItemListener(){
            public void itemStateChanged(ItemEvent e) {
                if ((table == null || table.getCounter() < 1) && refStarCB[r].isSelected() != isRefStar[r] )
                    {
                    refStarCB[r].setSelected(isRefStar[r]);
                    return;
                    }
                if (e.getStateChange() == ItemEvent.DESELECTED)
                    {
                    isRefStar[r] = false;
                    refStarLabel[r].setText("T"+(r+1));
                    }
                else if (e.getStateChange() == ItemEvent.SELECTED)
                    {
                    isRefStar[r] = true;
                    refStarLabel[r].setText("C"+(r+1));
                    }
                
                updatePlotEnabled = false;
                waitForPlotUpdateToFinish();
                checkAndLockTable(); 
                
                updateTable(isRefStar[r], r);
                if (!multiUpdate)
                    {
                    lastRefStar = r;
                    cycleEnabledStarsLess1PressedConsecutive = false;
                    updateTotals();
                    updateGUI();
                    table.show();
                    }
                updatePlotEnabled = true;
                if (table != null) table.setLock(false);                
                }});            
            }
        
        static void updateTable(boolean comp, int r)
            {
            if (table != null && table.getCounter() < 1) return;
            int tablen = table.getLastColumn();
            String newSuffix = "_T"+(r+1);
            String oldSuffix = "_C"+(r+1);
            if (comp)
                {
                newSuffix = "_C"+(r+1);
                oldSuffix = "_T"+(r+1);
                }
            int suflen = oldSuffix.length();
            String heading;
            int headlen = 0;
            for (int col = 0; col <= tablen; col++)
                {
                heading = table.getColumnHeading(col);
                headlen = heading.length();
                if (heading.endsWith(oldSuffix))
                    {
                    table.setHeading(col, heading.substring(0,headlen - suflen) + newSuffix);
                    }
                }
            if (xlabeldefault.endsWith(oldSuffix)) xlabeldefault = xlabeldefault.substring(0, xlabeldefault.length() - suflen) + newSuffix;
            for (int c = 0; c < maxCurves; c++)
                {
                if (xlabel[c].endsWith(oldSuffix)) xlabel[c] = xlabel[c].substring(0, xlabel[c].length() - suflen) + newSuffix;
                if (ylabel[c].endsWith(oldSuffix)) ylabel[c] = ylabel[c].substring(0, ylabel[c].length() - suflen) + newSuffix;
                if (oplabel[c].endsWith(oldSuffix)) oplabel[c] = oplabel[c].substring(0, oplabel[c].length() - suflen) + newSuffix;
                for (int v = 0; v < maxDetrendVars; v++)
                    {
                    if (detrendlabel[c][v].endsWith(oldSuffix)) detrendlabel[c][v] = detrendlabel[c][v].substring(0, detrendlabel[c][v].length() - suflen) + newSuffix;
                    if (detrendlabelhold[c][v].endsWith(oldSuffix)) detrendlabelhold[c][v] = detrendlabelhold[c][v].substring(0, detrendlabelhold[c][v].length() - suflen) + newSuffix;
                    }
                }
            }   
        
        
        static void updateTotals()
            {
            if (table == null) return;
            int numRows = table.getCounter();
            if (numRows < 1) return;
            boolean goodErrData = true;
            double[][] source = new double[numAps][numRows];
            double[][] srcvar = new double[numAps][numRows];
            double[][] total = new double[numAps][numRows];
            double[][] totvar = new double[numAps][numRows];
            int totCcntAP = -1;
            for (int ap = 0; ap < numAps; ap++)
                {
                if (!isRefStar[ap])
                    {
                    totCcntAP = ap;
                    break;
                    }
                }
            for (int ap = 0; ap < numAps; ap++)
                {                
                for (int row = 0; row < numRows; row++)
                    {
                    source[ap][row] = 0.0;
                    total[ap][row] = 0.0;
                    srcvar[ap][row] = 0.0;
                    totvar[ap][row] = 0.0;                    
                    }
                }
            double value = 0;
            double errval = 0;
            double ratio = 0;
            double factor = 0;
            double oneOverFactor = 0;
            int col = ResultsTable.COLUMN_NOT_FOUND;
            int errcol = ResultsTable.COLUMN_NOT_FOUND;
            int snrcol = ResultsTable.COLUMN_NOT_FOUND;
            
            for (int ap = 0; ap < numAps; ap++)
                {
                col = table.getColumnIndex ("Source-Sky_"+(isRefStar[ap]?"C":"T")+(ap+1));
                if (col == ResultsTable.COLUMN_NOT_FOUND)
                    {  
                    IJ.beep();
                    IJ.showMessage("Error: could not find data column 'Source-Sky_"+(isRefStar[ap]?"C":"T")+(ap+1)+"'");
                    return;
                    }
                else
                    {
                    errcol = table.getColumnIndex ("Source_Error_"+(isRefStar[ap]?"C":"T")+(ap+1));
                    if (errcol == ResultsTable.COLUMN_NOT_FOUND)
                        {  
                        goodErrData = false;
                        }                    
                    for (int row = 0; row < numRows; row++)
                        {
                        value = table.getValueAsDouble(col, row);
                        source[ap][row] = value;
                        if (goodErrData)
                            {
                            errval = table.getValueAsDouble(errcol, row);
                            srcvar[ap][row] = errval*errval;
                            }
                        if (isRefStar[ap])
                            {
                            for (int i = 0; i < numAps; i++)
                                { 
                                if (i != ap)
                                    {
                                    total[i][row] += value;
                                    totvar[i][row] += srcvar[ap][row];
                                    }
                                }
                            }
                        }
                    }
                }
            if (numAps > 1)
                {
                for (int ap = 0; ap < numAps; ap++)
                    {
                    col = table.getColumnIndex ("rel_flux_"+(isRefStar[ap]?"C":"T")+(ap+1));
                    if (col == ResultsTable.COLUMN_NOT_FOUND)
                        {  
                        col = table.getFreeColumn("rel_flux_"+(isRefStar[ap]?"C":"T")+(ap+1));
                        if (col == ResultsTable.COLUMN_NOT_FOUND)
                            { 
                            IJ.beep();
                            IJ.showMessage("Error: could not create data column 'rel_flux_"+(isRefStar[ap]?"C":"T")+(ap+1)+"'");
                            return;
                            }
                        }
                    if (goodErrData)
                        {
                        errcol = table.getColumnIndex ("rel_flux_err_"+(isRefStar[ap]?"C":"T")+(ap+1));
                        if (col == ResultsTable.COLUMN_NOT_FOUND)
                            {  
                            col = table.getFreeColumn("rel_flux_err_"+(isRefStar[ap]?"C":"T")+(ap+1));
                            if (col == ResultsTable.COLUMN_NOT_FOUND)
                                { 
                                IJ.beep();
                                IJ.showMessage("Error: could not create data column 'rel_flux_err_"+(isRefStar[ap]?"C":"T")+(ap+1)+"'");
                                return;
                                }
                            } 
                        snrcol = table.getColumnIndex ("rel_flux_SNR_"+(isRefStar[ap]?"C":"T")+(ap+1));
                        if (snrcol == ResultsTable.COLUMN_NOT_FOUND)
                            {  
                            snrcol = table.getFreeColumn("rel_flux_SNR_"+(isRefStar[ap]?"C":"T")+(ap+1));
                            if (snrcol == ResultsTable.COLUMN_NOT_FOUND)
                                { 
                                IJ.beep();
                                IJ.showMessage("Error: could not create data column 'rel_flux_SNR_"+(isRefStar[ap]?"C":"T")+(ap+1)+"'");
                                return;
                                }
                            }                     
                        }

                    for (int row = 0; row < numRows; row++)
                        {
                        ratio = total[ap][row] == 0 ? 0: source[ap][row]/total[ap][row];
                        table.setValue(col, row, ratio);
                        if (goodErrData)
                            {
                            if (source[ap][row] == 0 || total[ap][row] == 0)
                                {
                                factor = 0;
                                oneOverFactor = 0;
                                }
                            else
                                {
                                factor = Math.sqrt(srcvar[ap][row]/(source[ap][row]*source[ap][row])+ totvar[ap][row]/(total[ap][row]*total[ap][row]));
                                oneOverFactor = 1.0/factor;
                                }
                            table.setValue(errcol, row, ratio*factor);
                            table.setValue(snrcol, row, oneOverFactor);
                            }
                        }
                    } 

                for (int row = 0; row < numRows; row++)
                    {
                    table.setValue("tot_C_cnts", row, totCcntAP < 0 ? 0.0 : total[totCcntAP][row]);
                    table.setValue("tot_C_err", row, totCcntAP < 0 ? 0.0 : Math.sqrt(totvar[totCcntAP][row]));
                    }
                }
            
            String isRefString = "";   //update prefs to reflect new ref/target star designations
            for (int ap = 0; ap < numAps; ap++)
                {
                if (ap == 0)
                    {
                    isRefString += isRefStar[ap];
                    }
                else
                    {
                    isRefString += ","+isRefStar[ap];
                    }
                }
            Prefs.set (MultiAperture_.PREFS_ISREFSTAR, isRefString);            
            }        
        
        static void updateGUI()
            {
            if (table == null || table.getCounter() < 1) return;
            unfilteredColumns = table.getColumnHeadings().split("\t");
            if (unfilteredColumns == null || unfilteredColumns.length == 0)
                {
                IJ.showMessage ("No data columns in table.");
                makeDummyTable();
                closeRefStarFrame();
                }
            else
                {
                updateColumnLists();
                oldUnfilteredColumns = unfilteredColumns.clone();
                xdatacolumndefault.setModel(new DefaultComboBoxModel(columns));
                xdatacolumndefault.setSelectedItem(xlabeldefault);
                xdatacolumndefault.repaint();
                for (int c = 0; c < maxCurves; c++)
                    {
                    xdatacolumn[c].setModel(new DefaultComboBoxModel(columnswd));
                    xdatacolumn[c].setSelectedItem(xlabel[c]);
                    disableUpdatePlotBox = true;   //disable automatic enable of plot when a new data column has been selected
                    ydatacolumn[c].setModel(new DefaultComboBoxModel(columns));
                    ydatacolumn[c].setSelectedItem(ylabel[c]);
                    disableUpdatePlotBox = false;
                    operatorcolumn[c].setModel(new DefaultComboBoxModel(columns));
                    operatorcolumn[c].setSelectedItem(oplabel[c]);
                    detrendbox[c].setModel(new DefaultComboBoxModel(columnsDetrend));
                    detrendbox[c].setSelectedItem(detrendlabel[c][detrendVarDisplayed[c]]);
                    }
                }
            }
        

        static void saveApertures()
            {
            storeApertures();
            SaveDialog sf = new SaveDialog("Save aperture configuration", MeasurementTable.shorterName(table.shortTitle()), "");
            if (sf.getDirectory() == null || sf.getFileName() == null)
                return;
            String apsPath = sf.getDirectory() + sf.getFileName();
            int location = apsPath.lastIndexOf('.');
            if (location >= 0) apsPath = apsPath.substring(0, location);            
            File outFile = new File(apsPath+".apertures");
            if (outFile == null || outFile.isDirectory())
                {
                IJ.error("bad aperture save filename");
                return;
                }
            if (outFile.isFile()) outFile.delete();
            Properties prefs = new Properties();
            Enumeration e = Prefs.ijPrefs.keys();
            while (e.hasMoreElements()) {
                String key = (String) e.nextElement();
                if (key.startsWith(".aperture.radius") || key.startsWith(".aperture.rback1") ||
                    key.startsWith(".aperture.rback2") || key.startsWith(".aperture.removebackstars") ||
                    key.startsWith(".aperture.backplane") || key.startsWith(".multiaperture.usevarsizeap") ||
                    key.startsWith(".multiaperture.apfwhmfactor") || key.startsWith(".multiaperture.xapertures") ||
                    key.startsWith(".multiaperture.yapertures") || key.startsWith(".multiaperture.isrefstar") || 
                    key.startsWith(".multiaperture.naperturesmax"))
                    prefs.put(key, Prefs.ijPrefs.getProperty(key));
            }
            try
                {
                FileOutputStream fos = new FileOutputStream(apsPath+".apertures");
                BufferedOutputStream bos = new BufferedOutputStream(fos);
                prefs.store(bos, "AstroImageJ Saved Apertures");
                bos.close();
                }
            catch (IOException ioe)
                {
                IJ.beep();
                IJ.showMessage("Error writing apertures to file");
                }
            }

        static void storeApertures()
            {
            int numRows = table.getCounter();
            if (numRows < 1) return;
            
            checkAndLockTable();
            
            double[] X = new double[numAps];
            double[] Y = new double[numAps];
            String xColMiddle = "(IJ)_";
            String yColMiddle = "(IJ)_";
            boolean isXFITS = false;
            boolean isYFITS = false;
            int col;    
            
            for (int ap = 0; ap < numAps; ap++)
                {  
                X[ap] = 0.0;
                Y[ap] = 0.0;
                }          
            
            col = table.getColumnIndex("X(IJ)_"+(isRefStar[0]?"C1":"T1"));
            if (col == ResultsTable.COLUMN_NOT_FOUND)
                { 
                col = table.getColumnIndex("X_"+(isRefStar[0]?"C1":"T1"));
                if (col == ResultsTable.COLUMN_NOT_FOUND)
                    {                  
                    col = table.getColumnIndex("X(FITS)_"+(isRefStar[0]?"C1":"T1"));
                    if (col != ResultsTable.COLUMN_NOT_FOUND)
                        {
                        xColMiddle = "(FITS)_";
                        isXFITS = true;
                        }
                    }
                else
                    {
                    xColMiddle = "_"; 
                    }
                }
            col = table.getColumnIndex("Y(IJ)_"+(isRefStar[0]?"C1":"T1"));
            if (col == ResultsTable.COLUMN_NOT_FOUND)
                { 
                col = table.getColumnIndex("Y_"+(isRefStar[0]?"C1":"T1"));
                if (col == ResultsTable.COLUMN_NOT_FOUND)
                    {                  
                    col = table.getColumnIndex("Y(FITS)_"+(isRefStar[0]?"C1":"T1"));
                    if (col != ResultsTable.COLUMN_NOT_FOUND)
                        {
                        yColMiddle = "(FITS)_";
                        isYFITS = true;                       }
                    }
                else
                    {
                    yColMiddle = "_"; 
                    }
                }
            
            
            
            for (int ap = 0; ap < numAps; ap++)
                {
                col = table.getColumnIndex("X"+xColMiddle+(isRefStar[ap]?"C":"T")+(ap+1));
                if (col != ResultsTable.COLUMN_NOT_FOUND)
                    {
                    X[ap] = table.getValueAsDouble(col, 0); 
                    }
                col = table.getColumnIndex("Y"+yColMiddle+(isRefStar[ap]?"C":"T")+(ap+1));
                if (col != ResultsTable.COLUMN_NOT_FOUND)
                    {
                    Y[ap] = table.getValueAsDouble(col, 0); 
                    }                
                }
            
            String xString = "";
            String yString = "";
            String isRefString = "";
            for (int ap = 0; ap < numAps; ap++)
                {
                if (ap == 0)
                    {
                    xString += (isXFITS?"FITS":"")+(float)X[ap];
                    yString += (isYFITS?"FITS":"")+(float)Y[ap];
                    isRefString += isRefStar[ap];
                    }
                else
                    {
                    xString += ","+(float)X[ap];
                    yString += ","+(float)Y[ap];
                    isRefString += ","+isRefStar[ap];
                    }
                }

            Prefs.set (MultiAperture_.PREFS_PREVIOUS, true);
            Prefs.set (MultiAperture_.PREFS_XAPERTURES, xString);
            Prefs.set (MultiAperture_.PREFS_YAPERTURES, yString);
            Prefs.set (MultiAperture_.PREFS_ISREFSTAR, isRefString);
            
            col = table.getColumnIndex(Aperture_.AP_FWHMMULT);
            if (col != ResultsTable.COLUMN_NOT_FOUND)
                { 
                Prefs.set (MultiAperture_.PREFS_USEVARSIZEAP, true);
                Prefs.set (MultiAperture_.PREFS_APFWHMFACTOR, table.getValueAsDouble(col, 0));
                col = table.getColumnIndex(Aperture_.AP_BRSOURCE);
                if (col != ResultsTable.COLUMN_NOT_FOUND)
                    { 
                    Prefs.set (Aperture_.AP_PREFS_RADIUS, table.getValueAsDouble(col, 0));
                    }                
                } 
            else
                {
                col = table.getColumnIndex(Aperture_.AP_RSOURCE);
                if (col != ResultsTable.COLUMN_NOT_FOUND)
                    { 
                    Prefs.set (MultiAperture_.PREFS_USEVARSIZEAP, false);
                    Prefs.set (Aperture_.AP_PREFS_RADIUS, table.getValueAsDouble(col, 0));
                    }
                }
            
            col = table.getColumnIndex(Aperture_.AP_RBACK1);
            if (col == ResultsTable.COLUMN_NOT_FOUND)
                { 
                col = table.getColumnIndex("Sky_Radius(min)");
                if (col != ResultsTable.COLUMN_NOT_FOUND)
                    {
                    Prefs.set (Aperture_.AP_PREFS_RBACK1, table.getValueAsDouble(col, 0));
                    }
                }
            else
                {
                Prefs.set (Aperture_.AP_PREFS_RBACK1, table.getValueAsDouble(col, 0));
                }
                
            col = table.getColumnIndex(Aperture_.AP_RBACK2);
            if (col == ResultsTable.COLUMN_NOT_FOUND)
                { 
                col = table.getColumnIndex("Sky_Radius(max)");
                if (col != ResultsTable.COLUMN_NOT_FOUND)
                    {
                    Prefs.set (Aperture_.AP_PREFS_RBACK2, table.getValueAsDouble(col, 0));
                    }
                }
            else
                {
                Prefs.set (Aperture_.AP_PREFS_RBACK2, table.getValueAsDouble(col, 0));
                }
            if (table != null) table.setLock(false);
            }
        

        static void showApertures()
            {
//            storeApertures();
//            Frame openFrame = WindowManager.getCurrentWindow();
//            ImagePlus imp = WindowManager.getCurrentImage();
//            Class<?> mainComponentClass = (openFrame == null ? null : openFrame.getClass());
//            if (openFrame != null && imp != null && mainComponentClass != null &&
//                mainComponentClass.getName().contains("AstroStackWindow"))
//                {
//                astroj.AstroStackWindow asw = (astroj.AstroStackWindow) openFrame;
//                asw.openApertures("");
//                }
            
            storeApertures();
            ImagePlus iplus = WindowManager.getCurrentImage();
            ImageWindow iw = null;
            AstroStackWindow asw = null;
            if (iplus != null)
                {
                iw = iplus.getWindow();
                if (iw != null && iw instanceof AstroStackWindow)
                    {
                    asw = (AstroStackWindow)iw;
                    WindowManager.toFront(iw);
                    asw.openApertures("");
                    }
                else if (WindowManager.getWindowCount()> 0) 
                    {
                    int[] ID = WindowManager.getIDList();
                    for (int win = 0; win < ID.length; win++)
                        {
                        iplus = WindowManager.getImage(ID[win]);
                        if (iplus != null) 
                            iw = iplus.getWindow();
                        else
                            iw = null;
                        if (iw != null && iw instanceof AstroStackWindow)
                            {
                            asw = (AstroStackWindow)iw;
                            WindowManager.setCurrentWindow(iw);
                            WindowManager.toFront(iw);
                            asw.openApertures("");
                            break;
                            }
                        }
                    }
                }            
            
            }   
        
        static void waitForPlotUpdateToFinish()
            {
            int cnt = 0;  //timeout after 1 second
            while (cnt < 10 && updatePlotRunning)
                {
                IJ.wait(100);
                cnt++;
//                if (cnt == 10)
//                    {
//                    IJ.log("NOTICE: timed out waiting on plot to finish update");
//                    }
                }
            updatePlotRunning = false;            
            }
        
        static void checkAndLockTable()
            {
            if (table == null) return;
            int cnt = 0;  //timeout after 1 second
            while (table != null && table.isLocked() && cnt < 10)
                {
                IJ.wait(100);
                cnt++;
//                if (cnt == 10)
//                    {
//                    IJ.log("NOTICE: timed out waiting on table to unlock");
//                    }                
                }
            if (table != null) table.setLock(true); 
            }
        
        static void closeRefStarFrame()
            {
//            updatePlotEnabled = true;
            if (refStarFrame.isShowing())
                {
                refStarFrameLocationX = refStarFrame.getLocation().x;
                refStarFrameLocationY = refStarFrame.getLocation().y;
                refStarFrame.setVisible(false);
                Prefs.set("plot2.refStarFrameLocationX",refStarFrameLocationX);
                Prefs.set("plot2.refStarFrameLocationY",refStarFrameLocationY);
                }
//            if (table != null) table.setLock(false);
            
            refStarFrame.dispose();
            refStarFrame = null;
            }
        
        static void closeAddAstroDataFrame()
            {
            astroConverterUpdating = false;
            if (addAstroDataFrame == null) return;
            if (addAstroDataFrame.isShowing())
                {
                addAstroDataFrameLocationX = addAstroDataFrame.getLocation().x;
                addAstroDataFrameLocationY = addAstroDataFrame.getLocation().y;
                addAstroDataFrame.setVisible(false);
                Prefs.set("plot2.addAstroDataFrameLocationX",addAstroDataFrameLocationX);
                Prefs.set("plot2.addAstroDataFrameLocationY",addAstroDataFrameLocationY);
                }
            addAstroDataFrame.dispose();
            addAstroDataFrame = null;
            }

        static void setVMarkerText()
            {
            GenericDialog gd = new GenericDialog ("Set Vertical Marker Text", mainFrame.getX()+mainFrame.getWidth()/2-165,
                                                                              mainFrame.getY()+mainFrame.getHeight()/2-77);
            gd.addStringField("Vertical Marker 1 Text - Line 1:", vMarker1TopText, 40);
            gd.addStringField("Vertical Marker 1 Text - Line 2:", vMarker1BotText, 40);
            gd.addMessage ("");
            gd.addStringField("Vertical Marker 2 Text - Line 1:", vMarker2TopText, 40);
            gd.addStringField("Vertical Marker 2 Text - Line 2:", vMarker2BotText, 40);
            gd.addMessage ("");
            gd.showDialog();
            if (gd.wasCanceled()) return;
            vMarker1TopText = gd.getNextString();
            vMarker1BotText = gd.getNextString();
            vMarker2TopText = gd.getNextString();
            vMarker2BotText = gd.getNextString();
            Prefs.set("plot.vMarker1TopText", vMarker1TopText);
            Prefs.set("plot.vMarker1BotText", vMarker1BotText);
            Prefs.set("plot.vMarker2TopText", vMarker2TopText);
            Prefs.set("plot.vMarker2BotText", vMarker2BotText);
            }

        static void changePixelScale()
            {
            GenericDialog gd = new GenericDialog ("Change Pixel Scale", mainFrame.getX()+mainFrame.getWidth()/2-165,
                                                                        mainFrame.getY()+mainFrame.getHeight()/2-77);

            gd.addNumericField ("Pixel scale: ",pixelScale,4,8, "(seconds of arc per pixel)");
            gd.addMessage ("");

            gd.showDialog();
            if (gd.wasCanceled()) return;

            pixelScale = gd.getNextNumber();
            Prefs.set("plot.pixelScale", pixelScale);
            changepixelscalemenuitem.setText("Change pixel scale ("+pixelScale+")...");
            }

        static void setEphemeris()
            {
            GenericDialog gd = new GenericDialog ("Set Ref. Epoch and Period for x-Axis Phase Display", mainFrame.getX()+mainFrame.getWidth()/2-165,
                                                                        mainFrame.getY()+mainFrame.getHeight()/2-77);

            gd.addNumericField ("Reference Epoch: ", T0, 8, 20, "(days)");
            gd.addNumericField ("Orbital Period: ", period, 8, 20, "(days)");
            gd.addMessage ("");

            gd.showDialog();
            if (gd.wasCanceled()) return;

            T0 = gd.getNextNumber();
            period = gd.getNextNumber();
            Prefs.set("plot.T0", T0);
            Prefs.set("plot.period", period);
            }        
        
        static void changeMaxDataLength()
            {
            GenericDialog gd = new GenericDialog ("Change minimum data column length", mainFrame.getX()+mainFrame.getWidth()/2-165,
                                                                                       mainFrame.getY()+mainFrame.getHeight()/2-77);

            gd.addMessage ("The default length is 1000, but can be made smaller or larger\n"
                + "in order to optimize your system's speed and memory utilization.\n"
                + "The length will automatically double if the set size is reached, however\n"
                + "plot updates will be slightly delayed while memory is being reconfigured.");

            gd.addNumericField("Length: ", maxColumnLength, 0, 10, "(measurement table rows)");
            gd.addMessage ("");

            gd.showDialog();
            if (gd.wasCanceled()) return;

            maxColumnLength = (int)gd.getNextNumber();
            setupDataBuffers();
            Prefs.set("plot.maxColumnLength", maxColumnLength);
            }



        static void changePriorityColumns()
            {
            GenericDialog gd = new GenericDialog ("Change data name priorities in pulldown lists", mainFrame.getX()+mainFrame.getWidth()/2-165,
                                                                                                   mainFrame.getY()+mainFrame.getHeight()/2-77);
            gd.addMessage ("Enter data names to be placed at the top of data selection pulldown lists,\n"
                + "in the order desired, with each name separated by a comma (case insensitive).\n"
                + "All data column names starting with a comma separated value will be included.");
            gd.addStringField("Data Columns:", priorityColumns, 80);
            gd.addStringField("Detrend Columns:", priorityDetrendColumns, 80);
            gd.addMessage ("");

            gd.showDialog();
            if (gd.wasCanceled()) return;

            priorityColumns = gd.getNextString();
            priorityDetrendColumns = gd.getNextString();
            Prefs.set("plot.priorityColumns", priorityColumns);
            Prefs.set("plot.priorityDetrendColumns", priorityDetrendColumns);
            if (prioritizeColumns)
                {
                oldUnfilteredColumns = null;
                updatePlot();
                }            
            }
        
        static void changeRefStarHorizontalWidth()
            {
            GenericDialog gd = new GenericDialog ("Change reference star window width", mainFrame.getX()+mainFrame.getWidth()/2-165,
                                                                                                   mainFrame.getY()+mainFrame.getHeight()/2-77);
            gd.addMessage ("Enter the number of reference stars to display per\n" +
                           "line in the reference star selection window.");
            gd.addNumericField("Stars per line:", refStarHorzWidth, 0);
            gd.addMessage ("");

            gd.showDialog();
            if (gd.wasCanceled()) return;

            refStarHorzWidth = (int)gd.getNextNumber();
            if (refStarHorzWidth < 1) refStarHorzWidth = 1;
            Prefs.set("plot.refStarHorzWidth", refStarHorzWidth);
            closeRefStarFrame();
            if (refStarPanelWasShowing) showRefStarJPanel();
            }        

        static void saveAndClose()
                {
                updatePlotEnabled = true;
                if (table != null) table.setLock(false);
                if (mainFrame != null) 
                    {
                    mainFrameLocationX = mainFrame.getLocation().x;
                    mainFrameLocationY = mainFrame.getLocation().y;
                    Prefs.set("plot2.mainFrameLocationX",mainFrameLocationX);
                    Prefs.set("plot2.mainFrameLocationY",mainFrameLocationY);
                    }
                if (subFrame != null && subFrame.isShowing())
                        {
                        subFrameLocationX = subFrame.getLocation().x;
                        subFrameLocationY = subFrame.getLocation().y;
                        Prefs.set("plot2.subFrameLocationX",subFrameLocationX);
                        Prefs.set("plot2.subFrameLocationY",subFrameLocationY);
                        }
                if (refStarFrame != null)
                        {
                        if (refStarFrame.isShowing())
                            refStarPanelWasShowing = true;
                        else
                            refStarPanelWasShowing = false;
                        closeRefStarFrame();            
                        } 
                if (acc != null)
                    {
                    acc.saveAndClose();
                    }
                if (addAstroDataFrame != null)
                    {
                    if (addAstroDataFrame.isShowing())
                        addAstroDataFrameWasShowing = true;
                    else
                        addAstroDataFrameWasShowing = false;                    
                    closeAddAstroDataFrame();   
                    }
                if (plotImage!=null) 
                    {
                    plotWindow = plotImage.getWindow();
                    if (plotWindow != null)
                            {
                            plotFrameLocationX = plotWindow.getLocation().x;
                            plotFrameLocationY = plotWindow.getLocation().y;
                            Prefs.set("plot2.plotFrameLocationX",plotFrameLocationX);
                            Prefs.set("plot2.plotFrameLocationY",plotFrameLocationY);
                            }
                    }

//                plotImageCanvas.removeMouseMotionListener(plotMouseMotionListener);
//                plotImageCanvas.removeMouseWheelListener(plotMouseWheelListener);
//                plotImageCanvas.removeMouseListener(plotMouseListener);
                
                savePreferences();
                if (delayedUpdateTimer != null) delayedUpdateTimer.cancel();
                if (delayedUpdateTask != null) delayedUpdateTask.cancel();  
                if (table != null) table.setLock(false);
                if (mainFrame != null) mainFrame.dispose();
                if (subFrame != null) subFrame.dispose();
                mainFrame=null;
                subFrame=null;
                }

       static void openDragAndDropFiles(java.io.File[] files)
                {
                dragAndDropFiles = files;
                if (files.length > 0 && files[0].isFile() && files[0].getName().endsWith(".plotcfg"))
                    {
                    OpenDialog.setDefaultDirectory(files[0].getParent());
                    try {
                        InputStream is = new BufferedInputStream(new FileInputStream(files[0].getCanonicalPath()));
                        Prefs.ijPrefs.load(is);
                        is.close();
                        setupArrays();
                        getPreferences();
                        if (table != null) setTable(table, true);
                        if (plotWindow != null) plotWindow.setVisible(true);
                        }
                    catch (Exception e) {
                        IJ.beep();
                        IJ.showMessage("DragAndDrop: Error reading plot configuration file");
                        }
                    }
                else if (files.length > 0 && files[0].isFile() && files[0].getName().endsWith(".apertures"))
                        {
                        OpenDialog.setDefaultDirectory(files[0].getParent());
                        try {
                            InputStream is = new BufferedInputStream(new FileInputStream(files[0].getCanonicalPath()));
                            Prefs.ijPrefs.load(is);
                            is.close();
                            }
                        catch (Exception e) {
                            IJ.beep();
                            IJ.showMessage("DragAndDrop: Error reading aperture file");
                            }            
                        }                        
                else if (files.length > 0 && files[0].isFile() && 
                        (files[0].getName().endsWith(Prefs.get("options.ext", ".xls")) ||
                         files[0].getName().endsWith(".txt") || files[0].getName().endsWith(".csv") ||
                         files[0].getName().endsWith(".prn") || files[0].getName().endsWith(".spc") || 
                         files[0].getName().endsWith(".xls") || files[0].getName().endsWith(".dat") ||
                         files[0].getName().endsWith(".tbl") ))
                    {

                    Thread t = new Thread()
                        {
                        public void run()
                            {
                            IJ.wait(100);
                            int errorCode = 0;
                            try {
                                if (panelShiftDown)
                                    {
                                    appendDataAsRows(true, dragAndDropFiles[0].getCanonicalPath());
                                    } 
                                else if (panelControlDown)
                                    {
                                    appendDataAsColumns(true, dragAndDropFiles[0].getCanonicalPath());
                                    }                         
                                else
                                    {
                                    MeasurementTable newTable = MeasurementTable.getTableFromFile (dragAndDropFiles[0].getCanonicalPath());
                                    errorCode = 1;
                                    if (newTable == null)
                                        {
                                        IJ.beep();
                                        IJ.showMessage("DragAndDrop: Unable to open MeasurementTable "+dragAndDropFiles[0].getCanonicalPath());
                                        }
                                    else
                                        {
                                        errorCode = 2;
                                        newTable.show();
                                        errorCode = 3;
                                        Frame ntf = WindowManager.getFrame(newTable.shortTitle());
                                        errorCode = 4;
                                        if (ntf != null) ntf.setVisible(true);
                                        errorCode = 5;
                                        int lastDot = dragAndDropFiles[0].getCanonicalPath().lastIndexOf('.');
                                        errorCode = 6;
                                        String cfgPath = lastDot > 0 ? dragAndDropFiles[0].getCanonicalPath().substring(0, lastDot) + ".plotcfg":
                                                                       dragAndDropFiles[0].getCanonicalPath() + ".plotcfg";
                                        errorCode = 7;
                                        File cfgFile = new File(cfgPath);
                                        errorCode = 8;
                                        if (cfgFile != null && cfgFile.isFile())
                                            {
                                            errorCode = 9;
                                            InputStream is = new BufferedInputStream(new FileInputStream(cfgPath));
                                            errorCode = 10;
                                            Prefs.ijPrefs.load(is);
                                            errorCode = 11;
                                            is.close();
                                            errorCode = 12;
                                            setupArrays();
                                            errorCode = 13;
                                            getPreferences();
                                            errorCode = 14;
                                            }
                                        errorCode = 15;
                                        setTable(newTable, true);
                                        errorCode = 16;
                                        if (plotWindow != null) plotWindow.setVisible(true);
                                        errorCode = 17;
                                        }
                                    }
                                }
                            catch (Exception e) {
                                IJ.beep();
                                IJ.showMessage("DragAndDrop: Error reading measurements file or table configuration. ErrorCode = "+errorCode);
                                }
                            }
                            
                        };
                    OpenDialog.setDefaultDirectory(files[0].getParent());
                    t.start();
                    Thread.yield();
                    
                    }
                }

        static void openData()
            {
            OpenDialog of = new OpenDialog("Select measurement table to open", "");
            String path = of.getDirectory() + of.getFileName();
            if (of.getDirectory() == null || of.getFileName() == null)
                return;
            try {
                MeasurementTable newTable = MeasurementTable.getTableFromFile (of.getDirectory() + of.getFileName());
                if (newTable == null)
                    {
                    IJ.beep();
                    IJ.showMessage("Unable to open measurement table " + path);
                    }
                else
                    {
                    newTable.show();
                    setTable(newTable, true);
                    plotWindow.setVisible(true);
                    }
                }
            catch (Exception e) {
                IJ.beep();
                IJ.showMessage("Error reading measurements table");
                }
            }
        static void transposeTable()
            {
            MeasurementTable newTable = new MeasurementTable(tableName+"_transposed");
            if (table.getLastColumn()<0 || table.getCounter()<1)
                {
                IJ.error("No data in active table to transpose");
                IJ.beep();
                return;
                }
            for (int row=0; row<table.getCounter(); row++)
                {
                if (table.getLabel(row).equals("") || table.getLabel(row)==null)
                    {
                    if (newTable.getFreeColumn("Col_"+(row+1)) == MeasurementTable.COLUMN_IN_USE) 
                        {
                        IJ.error("Could not make unique column labels from row labels");
                        IJ.beep();
                        return;
                        }                    
                    }
                else if (newTable.getFreeColumn(table.getLabel(row)) == MeasurementTable.COLUMN_IN_USE) 
                    {
                    if (newTable.getFreeColumn(table.getLabel(row)+"_Col_"+(row+1)) == MeasurementTable.COLUMN_IN_USE) 
                        {
                        IJ.error("Could not make unique column labels from row labels");
                        IJ.beep();
                        return;
                        }
                    }
                }
            for (int col=0; col<=table.getLastColumn();col++)
                {
                newTable.incrementCounter();
                newTable.setLabel(table.getColumnHeading(col), col);
                }                
            for (int row=0; row<table.getCounter(); row++)
                {
                for (int col=0; col<=table.getLastColumn();col++)
                    {
                    newTable.setValue(row, col, table.getValueAsDouble(col, row));
                    }                 
                }
            if (newTable != null)
                {
                newTable.show();
                setTable(newTable, true);
                plotWindow.setVisible(true);
                }            
            }
        

        static void appendDataAsRows(boolean fromFile, String inPath)
            {
            String filePath = inPath;
            
            if (fromFile && inPath == null)
                {
                OpenDialog of = new OpenDialog("Select measurement table to append as new rows", "");
                if (of.getDirectory() == null || of.getFileName() == null)
                    return;
                filePath = of.getDirectory() + of.getFileName();
                }

            try {
                MeasurementTable newTable;
                selectAnotherTableCanceled = false;
                if (fromFile)
                    {
                    newTable = MeasurementTable.getTableFromFile (filePath);
                    }
                else
                    {
                    newTable = selectAnotherTable(tableName);
                    }
                if (newTable == null)
                    {
                    if (!selectAnotherTableCanceled)
                        {
                        IJ.beep();
                        if (fromFile)
                            IJ.showMessage("Unable to open measurement table " + filePath);
                        else
                            IJ.showMessage("No additional measurement tables open");
                        }
                    }
                else
                    {
                    newTable.show();
                    
                    combinedTableName = MeasurementTable.shorterName(tableName);
                    if (combinedTableName.endsWith(Prefs.get("options.ext", ".xls")))
                        {
                        int position = combinedTableName.lastIndexOf(Prefs.get("options.ext", ".xls"));
                        if (position >= 0) combinedTableName = combinedTableName.substring(0, position);  
                        }
                    combinedTableName += "_" + MeasurementTable.shorterName(newTable.shortTitle());

                    MeasurementTable combinedTable = MeasurementTable.getTable(combinedTableName, tableName);                    
                    String[] oldColumns = table.getColumnHeadings().split("\t");
                    String[] newColumns = newTable.getColumnHeadings().split("\t");
                    int oldRows = combinedTable.getCounter();
                    int newRows = newTable.getCounter();

                    int combinedRows = oldRows + newRows;
                    int columnIndex = 0;
                    if (newColumns.length > 2 && newRows > 0)
                        {
                        for (int i=0; i<newRows;i++)
                            {
                            combinedTable.incrementCounter();
                            combinedTable.addLabel(newTable.getLabel(i));
                            }
                        
                        for (int i=2;i<oldColumns.length;i++)  //copy all new columns common with old table
                            {
                            columnIndex = newTable.getColumnIndex(oldColumns[i]);
                            for (int j=oldRows; j<combinedRows;j++)
                                {
                                combinedTable.setValue(oldColumns[i], j, columnIndex == ResultsTable.COLUMN_NOT_FOUND?
                                                        Double.NaN : newTable.getValueAsDouble(columnIndex, j-oldRows));
                                }
                            }
                        for (int i=2;i<newColumns.length;i++)  //copy all new columns not common with old table
                            {
                            columnIndex = table.getColumnIndex(newColumns[i]);
                            if (columnIndex == ResultsTable.COLUMN_NOT_FOUND)
                                {
                                for (int j=0; j<combinedRows;j++)
                                    {
                                    combinedTable.setValue(newColumns[i], j, j < oldRows? Double.NaN : newTable.getValue(newColumns[i], j-oldRows));
                                    }                                
                                }             
                            }
                        
                        table.show();
                        combinedTable.show();
                        setTable(combinedTable, true);
                        plotWindow.setVisible(true);
                        }
                    else if (newColumns.length < 3)
                        {
                        IJ.beep();
                        IJ.showMessage("No new columns to append");
                        }
                    else if (newRows < 1)
                        {
                        IJ.beep();
                        IJ.showMessage("No new rows to append");
                        } 
                    else
                        {
                        IJ.beep();
                        IJ.showMessage("Table append error");
                        } 
                    }
                }
            catch (Exception e) {
                IJ.beep();
                IJ.showMessage("Append Rows: Error reading measurement table");
                }
            }      

        static void appendDataAsColumns(boolean fromFile, String inPath)
            {
            String filePath = inPath;
            boolean uniqueHeadings = true;
            
            if (fromFile && inPath == null)
                {
                OpenDialog of = new OpenDialog("Select measurement table to append as new columns", "");
                if (of.getDirectory() == null || of.getFileName() == null)
                    return;
                filePath = of.getDirectory() + of.getFileName();
                }

            try {
                MeasurementTable newTable;
                selectAnotherTableCanceled = false;
                if (fromFile)
                    {
                    newTable = MeasurementTable.getTableFromFile (filePath);
                    }
                else
                    {
                    newTable = selectAnotherTable(tableName);
                    }
                if (newTable == null)
                    {
                    if (!selectAnotherTableCanceled)
                        {
                        IJ.beep();
                        if (fromFile)
                            IJ.showMessage("Unable to open measurement table " + filePath);
                        else
                            IJ.showMessage("No additional measurement tables open");
                        }
                    }
                else
                    {
                    newTable.show();
                    
                    combinedTableName = MeasurementTable.shorterName(tableName);
                    if (combinedTableName.endsWith(Prefs.get("options.ext", ".xls")))
                        {
                        int position = combinedTableName.lastIndexOf(Prefs.get("options.ext", ".xls"));
                        if (position >= 0) combinedTableName = combinedTableName.substring(0, position);  
                        }
                    combinedTableName += "_" + MeasurementTable.shorterName(newTable.shortTitle());

                    if (!appendTableDialog()) return;
                    MeasurementTable combinedTable = MeasurementTable.getTable(combinedTableName, tableName);    
                    String[] oldColumns = table.getColumnHeadings().split("\t");
                    String[] newColumns = newTable.getColumnHeadings().split("\t");
                    int oldRows = combinedTable.getCounter();
                    int newRows = newTable.getCounter();


                    for (int r = 0; r < oldRows; r++)
                        {
                        if (keepFileNamesOnAppend)
                            {                        
                            combinedTable.setLabel(combinedTable.getLabel(r)+(appendDestinationSuffix.equals("") ? " " : "("+appendDestinationSuffix+")") +
                                                (r < newRows ? " "+newTable.getLabel(r)+ "("+appendSourceSuffix+")": ""), r);
                            }
                        else
                            {
                            combinedTable.setLabel("merged", r);
                            }
                        }
                    
                    if (newRows > oldRows)
                        {
                        for (int r = oldRows; r < newRows; r++)
                            {
                            combinedTable.incrementCounter();
                            if (keepFileNamesOnAppend)
                                combinedTable.addLabel(newTable.getLabel(r)+ "("+appendSourceSuffix+")");
                            else
                                combinedTable.addLabel("merged");
                            for (int i=0;i<newColumns.length;i++)
                                {
                                combinedTable.setValue(i, r, Double.NaN);
                                }                            
                            }
                        }
                    if (!appendDestinationSuffix.equals("") && (oldColumns.length > 1))
                        {
                        for (int col=0;col<oldColumns.length - 2;col++)
                            {
                            combinedTable.setHeading(col, combinedTable.getColumnHeading(col)+appendDestinationSuffix);
                            }
                        
                        if (!xlabeldefault.equals("")) xlabeldefault += appendDestinationSuffix;                        
                        for (int curve=0; curve < maxCurves; curve++)
                            {
                            if (!xlabel[curve].equals("") && !xlabel[curve].equals("default")) xlabel[curve] += appendDestinationSuffix;
                            if (!ylabel[curve].equals("")) ylabel[curve] += appendDestinationSuffix;
                            if (!oplabel[curve].equals("")) oplabel[curve] += appendDestinationSuffix;
                            for (int v = 0; v < maxDetrendVars; v++)
                                {
                                if (!detrendlabel[curve][v].equals("")) detrendlabel[curve][v] += appendDestinationSuffix;
                                if (!detrendlabelhold[curve][v].equals("")) detrendlabelhold[curve][v] += appendDestinationSuffix;
                                }
                            }
                        }
                    
                    
                    for (int i=0;i<newColumns.length;i++)
                        {
                        if (combinedTable.getColumnIndex(newColumns[i]+appendSourceSuffix)!=ResultsTable.COLUMN_NOT_FOUND)
                            {
                            uniqueHeadings = false;
                            break;
                            }
                        }
                    int combinedRows = combinedTable.getCounter();
                    if (uniqueHeadings && newColumns.length > 2 && newRows > 0)
                        {
                        for (int i=2;i<newColumns.length;i++)
                            {
                            for (int j=0;j<combinedRows;j++)
                                {
                                if (j < newRows)
                                    combinedTable.setValue(newColumns[i]+appendSourceSuffix, j, newTable.getValue(newColumns[i], j));
                                else
                                    combinedTable.setValue(newColumns[i]+appendSourceSuffix, j, Double.NaN);
                                }
                            }
                        table.show();
                        combinedTable.show();
                        setTable(combinedTable, true);
                        plotWindow.setVisible(true);
                        }
                    else if (!uniqueHeadings)
                        {
                        IJ.beep();
                        IJ.showMessage("Heading names are not unique");
                        } 
                    else if (newColumns.length < 3)
                        {
                        IJ.beep();
                        IJ.showMessage("No new columns to append");
                        }
                    else if (newRows < 1)
                        {
                        IJ.beep();
                        IJ.showMessage("No new rows to append");
                        } 
                    else
                        {
                        IJ.beep();
                        IJ.showMessage("Table append error");
                        } 
                    }
                }
            catch (Exception e) {
                IJ.beep();
                IJ.showMessage("Append Columns: Error reading measurement table");
                }
            }
        
        static boolean appendTableDialog()
            {
            GenericDialog gd = new GenericDialog ("Append settings", mainFrame.getX()+20,
                                                                     mainFrame.getY()+mainFrame.getHeight()/2-77);

            gd.addStringField("Enter name for combined table:", combinedTableName, 80);
            
            gd.addMessage ("The combined dataset column headings must all be unique.\n"+
                           "Enter suffixes to be appended to existing headings to make them unique.\n"+
                           "Either or both of the suffix fields can left blank if the headings are already unique.\n"+
                           "Short suffixes display better in the selection panels.");
            gd.addStringField("Original table column heading suffix:", "_", 80);
            gd.addStringField("Appended table column heading suffix:", "_", 80);
            gd.addCheckbox("Use concatenated source file names as label (label can become very long)***", keepFileNamesOnAppend);
            gd.addMessage("***if deselected, 'merged' will appear in all label columns ");

            gd.showDialog();
            if (gd.wasCanceled()) return false;

            combinedTableName = gd.getNextString().trim();
            appendDestinationSuffix = gd.getNextString().trim();
            appendSourceSuffix = gd.getNextString().trim();
            keepFileNamesOnAppend = gd.getNextBoolean();
            
            return true;
            } 
        
   static MeasurementTable selectAnotherTable(String firstTableName)
        {
        MeasurementTable secondTable;
        String secondTableName = "";
        String[] tables = MeasurementTable.getMeasurementTableNames();
        if (tables == null || tables.length == 0)
            {
            return null;
            }
        else
            {
            //filter out first table and B&C, Log, Errors windows if they exist
            int j = 0;
            for (int i=0; i < tables.length; i++)
                    if (tables[i].equals(firstTableName) || tables[i].equals("B&C") || tables[i].startsWith("Log") || tables[i].equals("Errors"))
                           j++;
            String[] filteredTables = new String[tables.length-j];

            j = 0;
            for (int i=0; i < tables.length; i++)
                if (!tables[i].equals(firstTableName) && !tables[i].equals("B&C") && !tables[i].startsWith("Log") && !tables[i].equals("Errors"))
                    {
                    filteredTables[j] = tables[i];
                    j++;
                    }
            
            if (filteredTables.length == 0)
                {
                return null;
                }
            else                            // IF MORE THAN ONE, ASK WHICH TABLE SHOULD BE USED
                {
                GenericDialog gd = new GenericDialog((firstTableName.equals("") ? "Select table" : "Select table to append"),
                                   mainFrame.getX()+mainFrame.getWidth()/2-165, mainFrame.getY()+mainFrame.getHeight()/2-77);
                gd.addChoice ((firstTableName.equals("") ? "Select table" : "Select table to append"), filteredTables, "Measurements");
                gd.showDialog();
                if (gd.wasCanceled())
                    {
                    selectAnotherTableCanceled = true;
                    return null;
                    }
                selectAnotherTableCanceled = false;
                secondTableName = gd.getNextChoice();
                }
            secondTable = MeasurementTable.getTable(MeasurementTable.longerName(secondTableName));
            return secondTable;
            }
        }           
        
        static void openConfig(boolean template)
            {
            String startingDir = OpenDialog.getDefaultDirectory();
//            String dir = OpenDialog.getLastDirectory();
            if (template && templateDir != null && !templateDir.equals(""))
                {
                OpenDialog.setDefaultDirectory(templateDir);
                }
            OpenDialog of = new OpenDialog("Open plot "+(template?"template":"configuration"), "");
            if (of.getDirectory() == null || of.getFileName() == null)
                {
                if (template) OpenDialog.setDefaultDirectory(startingDir);
                return;
                }
            if (template)
                {
                templateDir = of.getDirectory();
                OpenDialog.setDefaultDirectory(startingDir);
                Prefs.set("plot2.templateDir",templateDir);
                }
            try {
                InputStream is = new BufferedInputStream(new FileInputStream(of.getDirectory()+of.getFileName()));
                Prefs.ijPrefs.load(is);
                is.close();
                setupArrays();
                getPreferences();
                setTable(table, true);
                plotWindow.setVisible(true);
                }
            catch (Exception e) {
                IJ.beep();
                IJ.showMessage("Error reading plot configuration file");
                }
            }

        static public void openDataAndConfig(String filePath)
            {
            String inPath = filePath;
            if (inPath == null)
                {
                OpenDialog of = new OpenDialog("Select measurement table to open", "");
                if (of.getDirectory() == null || of.getFileName() == null)
                    return;
                inPath = of.getDirectory() + of.getFileName();
                }
            try {
                MeasurementTable newTable = MeasurementTable.getTableFromFile (inPath);
                if (newTable == null)
                    {
                    IJ.beep();
                    IJ.showMessage("Unable to open measurement table " + inPath);
                    }
                else
                    {
                    newTable.show();
                    int lastDot = inPath.lastIndexOf('.');
                    String cfgPath = lastDot > 0 ? inPath.substring(0, lastDot) + ".plotcfg": inPath + ".plotcfg";                    
                    File cfgFile = new File(cfgPath);
                    if (cfgFile != null && cfgFile.isFile())
                        {
                        InputStream is = new BufferedInputStream(new FileInputStream(cfgPath));
                        Prefs.ijPrefs.load(is);
                        is.close();
                        setupArrays();
                        getPreferences();
                        }
                    setTable(newTable, true);
                    plotWindow.setVisible(true);
                    }
                }
            catch (Exception e) {
                IJ.beep();
                IJ.showMessage("Open Data and Config: Error reading measurements file");
                }
            }
        
        static public void loadDataOpenConfig(MeasurementTable newTable, String filePath)
            {
            String inPath = filePath;
            String cfgPath = "";
            try {
                int lastDot = inPath.lastIndexOf('.');
                cfgPath = lastDot > 0 ? inPath.substring(0, lastDot) + ".plotcfg": inPath + ".plotcfg";                    
                File cfgFile = new File(cfgPath);
                if (cfgFile != null && cfgFile.isFile())
                    {
                    InputStream is = new BufferedInputStream(new FileInputStream(cfgPath));
                    Prefs.ijPrefs.load(is);
                    is.close();
                    setupArrays();
                    getPreferences();
                    }
                Frame ntf = WindowManager.getFrame(newTable.shortTitle());
                if (ntf != null) ntf.setVisible(true);                
                setTable(newTable, true);
                if (plotWindow != null) plotWindow.setVisible(true);
                }
            catch (Exception e) {
                IJ.beep();
                IJ.showMessage("Error reading plot configuration file " + cfgPath);
                }
            }        


        static void saveData()
            {
            if (table == null || table.getLastColumn()==-1)
                {
                IJ.beep();
                IJ.showMessage("No table to save");
                return;
                }
            SaveDialog sf = new SaveDialog("Save measurement table data",MeasurementTable.shorterName(table.shortTitle()), Prefs.get("options.ext", ".xls"));
            if (sf.getDirectory() == null || sf.getFileName() == null)
                return;
            try {
                table.saveAs(sf.getDirectory() + sf.getFileName());
                }
            catch(IOException ioe)
                {
                IJ.beep();
                IJ.showMessage("Error writing measurement table file");
                }
            }
        
        
        static void saveDataSubsetDialog()
            {
            int maxSubsetColumns = 20;
            boolean saveColumnHeadings = true;
            boolean saveHeadersAsComment = true; 
            boolean saveRowNumbers = true;
            boolean saveRowLabels = true;
            PrintWriter pw = null;            

            String[] subsetColumn = new String[maxSubsetColumns];
            saveColumnHeadings = Prefs.get("plot2.saveColumnHeadings", saveColumnHeadings);
            saveHeadersAsComment = Prefs.get("plot2.saveHeadersAsComment", saveHeadersAsComment);
            saveRowNumbers = Prefs.get("plot2.saveRowNumbers", saveRowNumbers);
            saveRowLabels = Prefs.get("plot2.saveRowLabels", saveRowLabels);
            
            for (int i=0; i<maxSubsetColumns; i++)
                {
                subsetColumn[i] = "";
                }            
            
            GenericDialog gd = new GenericDialog ("Save data subset", mainFrame.getX()+mainFrame.getWidth()/2-165,
                                                                          mainFrame.getY()+mainFrame.getHeight()/2-77);

            for (int i=0; i<maxSubsetColumns; i++)
                {
                subsetColumn[i] = Prefs.get("plot2.subsetColumn"+i, subsetColumn[i]);
                }
            
            gd.addMessage ("Select datasets in the order to be saved.\nLeave extra selections blank if no more data is to be saved.");
            for (int i=0; i<maxSubsetColumns; i++)
                {
                gd.addChoice("", columns, subsetColumn[i]);
                }
            gd.addCheckbox("Save column headings", saveColumnHeadings);
            gd.addCheckbox("Comment headings with '#'", saveHeadersAsComment);
            gd.addCheckbox("Save row numbers", saveRowNumbers);
            gd.addCheckbox("Save row labels", saveRowLabels);            
            
            gd.showDialog();
            if (gd.wasCanceled()) return;
            
            for (int i=0; i<maxSubsetColumns; i++)
                {
                subsetColumn[i] = gd.getNextChoice();
                }
            saveColumnHeadings = gd.getNextBoolean();
            saveHeadersAsComment = gd.getNextBoolean();
            saveRowNumbers = gd.getNextBoolean();
            saveRowLabels = gd.getNextBoolean();

            for (int i=0; i<maxSubsetColumns; i++)
                {
                Prefs.set("plot2.subsetColumn"+i, subsetColumn[i]);
                }            
            Prefs.set("plot2.saveColumnHeadings", saveColumnHeadings);
            Prefs.set("plot2.saveHeadersAsComment", saveHeadersAsComment);
            Prefs.set("plot2.saveRowNumbers", saveRowNumbers);
            Prefs.set("plot2.saveRowLabels", saveRowLabels);
            
            SaveDialog sf = new SaveDialog("Save data subset", MeasurementTable.shorterName(table.shortTitle()), Prefs.get("options.ext", ".xls"));
            if (sf.getDirectory() == null || sf.getFileName() == null)
                {
                IJ.beep();
                IJ.showMessage("Error: No save file path returned. Aborting.");
                return;
                }
            String savePath = sf.getDirectory() + sf.getFileName();
            int numColumns = 0;
            for (int i=0; i<maxSubsetColumns; i++)
                {
                if (!subsetColumn[i].trim().equals(""))
                    {
                    if (table.getColumnIndex(subsetColumn[i])==ResultsTable.COLUMN_NOT_FOUND)
                        {
                        IJ.beep();
                        IJ.showMessage("Error: Table column "+subsetColumn[i]+" not found. Aborting.");
                        return;
                        }
                    else
                        {
                        numColumns++;
                        }
                    }
                }
            if (numColumns == 0 && !saveRowLabels)
                {
                IJ.beep();
                IJ.showMessage("Error: No table columns selected and not saving row labels. Aborting.");
                return;
                }                
            
            int rows = table.getCounter();
            if (rows==0) 
                {
                IJ.beep();
                IJ.showMessage("Error: No rows in table to save. Aborting.");
                return;
                }

            char delimiter = savePath.endsWith(".csv")?',':'\t';
            String formatString = "0.";
            for (int i=0; i<MeasurementTable.DEFAULT_DECIMALS;i++)
                formatString += "0";
            DecimalFormat output = new DecimalFormat(formatString);
            try{
                FileOutputStream fos = new FileOutputStream(savePath);
                BufferedOutputStream bos = new BufferedOutputStream(fos);
                pw = new PrintWriter(bos);
                String line ="";
                String outText = "";
                double value = 0;
                boolean needDelimiter;
                if (saveColumnHeadings) 
                    {
                    line = (saveHeadersAsComment?"#":"")+(saveRowNumbers?delimiter:"")+(saveRowLabels?"Label"+delimiter:"");
                    needDelimiter = false;
                    for (int i=0; i<maxSubsetColumns; i++)
                        {
                        if (!subsetColumn[i].trim().equals(""))
                            {
                            line += (needDelimiter?delimiter:"")+subsetColumn[i];
                            needDelimiter = true;
                            }
                        }                    
                    pw.println(line);
                    }
                for (int row=0; row<rows; row++)
                    {
                    line = (saveRowNumbers?""+(row+1)+delimiter:"")+(saveRowLabels?table.getLabel(row)+delimiter:"");
                    needDelimiter = false;
                    for (int i=0; i<maxSubsetColumns; i++)
                        {
                        if (!subsetColumn[i].trim().equals(""))
                            {
                            value = table.getValue(subsetColumn[i],row);
                            outText = Double.isNaN(value) ? "":output.format(value);
                            line += ""+(needDelimiter?delimiter:"")+outText;
                            needDelimiter = true;
                            }
                        }                     
                    pw.println(line);
                    }
                }
            catch(IOException ioe) 
                {
                IJ.beep();
                IJ.showMessage("Error writing data subset file");
                }
            finally
                {
                pw.close();
                }            
            
            }
        
        static void savePlotImageAsPng()
            {
            ImagePlus image = WindowManager.getImage("Plot of "+tableName);
            if (image == null)
                {
                IJ.beep();
                IJ.showMessage("No plot image to save");
                return;
                }
            SaveDialog sf = new SaveDialog("Save plot image as PNG...",MeasurementTable.shorterName(table.shortTitle()), ".png");
            if (sf.getDirectory() == null || sf.getFileName() == null) return;
            IJ.runPlugIn(image, "ij.plugin.PNG_Writer", sf.getDirectory()+sf.getFileName());
            }

        static void savePlotImageAsJpg()
            {
            ImagePlus image = WindowManager.getImage("Plot of "+tableName);
            if (image == null)
                {
                IJ.beep();
                IJ.showMessage("No plot image to save");
                return;
                }
            SaveDialog sf = new SaveDialog("Save plot image as JPG...",MeasurementTable.shorterName(table.shortTitle()), ".jpg");
            if (sf.getDirectory() == null || sf.getFileName() == null) return;
            IJ.runPlugIn(image, "ij.plugin.JpegWriter", sf.getDirectory()+sf.getFileName());
            }

        static void saveConfig(boolean template)
            {
            String startingDir = OpenDialog.getDefaultDirectory();
            String fileName = table.shortTitle().startsWith("Measurements in ")? table.shortTitle().replace("Measurements in ", ""):table.shortTitle();
            int lastDot = fileName.lastIndexOf('.');
            if (lastDot >= 0) fileName = fileName.substring(0, lastDot);
            SaveDialog sf;
            if (template && templateDir != null && !templateDir.equals(""))
                {
                OpenDialog.setDefaultDirectory(templateDir);
                }
            sf = new SaveDialog("Save plot "+(template?"template":"configuration"), fileName, ".plotcfg");
            if (sf.getDirectory() == null || sf.getFileName() == null || sf.getFileName().equals(""))
                {
                if (template) OpenDialog.setDefaultDirectory(startingDir);
                return;
                }
            String savepath = sf.getDirectory() + sf.getFileName();
            if (template)
                {
                templateDir = sf.getDirectory();
                OpenDialog.setDefaultDirectory(startingDir);
                Prefs.set("plot2.templateDir",templateDir);
                }
            lastDot = savepath.lastIndexOf('.');
            if (lastDot >= 0) savepath = savepath.substring(0, lastDot);
            savepath += ".plotcfg";
            Properties prefs = new Properties();
            Enumeration e = Prefs.ijPrefs.keys();
            while (e.hasMoreElements()) {
                String key = (String) e.nextElement();
                if (key.indexOf(".plot.") == 0)
                    prefs.put(key, Prefs.ijPrefs.getProperty(key));
            }

            try
                {
                FileOutputStream fos = new FileOutputStream(savepath);
                BufferedOutputStream bos = new BufferedOutputStream(fos);
                prefs.store(bos, "AstroImageJ Plot Configuration");
                bos.close();
                }
            catch (IOException ioe)
                {
                IJ.beep();
                IJ.showMessage("Error writing plot configuration file");
                }
            }

        static void saveDataConfig()
            {
            SaveDialog sf = new SaveDialog("Save measurement table data and plot configuration",MeasurementTable.shorterName(table.shortTitle()), Prefs.get("options.ext", ".xls"));
            if (sf.getDirectory() == null || sf.getFileName() == null)
                return;
            String path = sf.getDirectory() + sf.getFileName();
            if (table == null || table.getLastColumn()==-1)
                {
                IJ.beep();
                IJ.showMessage("No table to save");
                return;
                }
            else
                {
                try {
                    table.saveAs(path);
                    }
                catch(IOException ioe)
                    {
                    IJ.beep();
                    IJ.showMessage("Error writing measurement table file");
                    }
                }
            int lastDot = path.lastIndexOf('.');
            if (lastDot >= 0) path = path.substring(0, lastDot);
            path += ".plotcfg";
            File outFile = new File(path);
            if (outFile == null || outFile.isDirectory())
                return;
            if (outFile.isFile()) outFile.delete();
            Properties prefs = new Properties();
            Enumeration e = Prefs.ijPrefs.keys();
            while (e.hasMoreElements()) {
                String key = (String) e.nextElement();
                if (key.indexOf(".plot.") == 0)
                    prefs.put(key, Prefs.ijPrefs.getProperty(key));
            }
            try
                {
                FileOutputStream fos = new FileOutputStream(path);
                BufferedOutputStream bos = new BufferedOutputStream(fos);
                prefs.store(bos, "AstroImageJ Plot Configuration");
                bos.close();
                }
            catch (IOException ioe)
                {
                IJ.beep();
                IJ.showMessage("Error writing plot configuration file");
                }
            }

        static public void saveDataImageConfig(boolean savePlot, boolean saveConfig, boolean saveData, boolean filenamesProvided,
                                               String imageFormat, String plotPath, String configPath, String dataPath)
            {
            SaveDialog sf = null;
            String outPath = "";
            String outBase = "";
            if (!filenamesProvided)
                {
                sf = new SaveDialog("Save "+(saveData?"measurement table data":"")+(saveData&&savePlot?", ":"")+
                    (savePlot?"plot image":"")+(saveData&&savePlot&&saveConfig?",":"")+((saveData||savePlot)&&saveConfig?" and ":"")+
                    (saveConfig?"plot configuration":""),MeasurementTable.shorterName(table.shortTitle()), "");
                if (sf.getDirectory() == null || sf.getFileName() == null)
                    return;
                outPath = sf.getDirectory() + sf.getFileName();
                int location = outPath.lastIndexOf('.');
                if (location >= 0) outPath = outPath.substring(0, location);
                outBase = outPath;
                if (outBase.endsWith(plotSuffix))
                    {
                    location = outBase.lastIndexOf(plotSuffix);
                    if (location >= 0) outBase = outBase.substring(0, location);
                    }
                else if (outBase.endsWith(configSuffix))
                    {
                    location = outBase.lastIndexOf(configSuffix);
                    if (location >= 0) outBase = outBase.substring(0, location);
                    }
                else if (outBase.endsWith(dataSuffix))
                    {
                    location = outBase.lastIndexOf(dataSuffix);
                    if (location >= 0) outBase = outBase.substring(0, location);
                    }
                }

            if (saveData)
                {
                if (table == null || table.getLastColumn()==-1)
                    {
                    IJ.beep();
                    IJ.showMessage("No data table to save");
                    }
                else
                    {
                    try {
                        table.saveAs(filenamesProvided?dataPath:outBase+dataSuffix+Prefs.get("options.ext", ".xls"));
                        }
                    catch(IOException ioe)
                        {
                        IJ.beep();
                        IJ.showMessage("Error writing measurement table file");
                        }
                    }
                }

            if (savePlot)
                {
                String imagepath = filenamesProvided ? plotPath : outBase+plotSuffix + "."+imageFormat;
                ImagePlus image = WindowManager.getImage("Plot of "+tableName);
                if (image == null)
                    {
                    IJ.beep();
                    IJ.showMessage("No plot image to save");
                    }
                else if (imageFormat.toLowerCase().equals("png"))
                    {
                    IJ.runPlugIn(image, "ij.plugin.PNG_Writer", imagepath);
                    }
                else if (imageFormat.toLowerCase().equals("jpg"))
                    {
                    IJ.runPlugIn(image, "ij.plugin.JpegWriter", imagepath);
                    }
                }
            if (saveConfig)
                {
                String cfgpath = filenamesProvided ? configPath : outBase+configSuffix+".plotcfg";
                File outFile = new File(cfgpath);
                if (outFile == null || outFile.isDirectory())
                    {
                    IJ.error("bad configuration save filename");
                    return;
                    }
                if (outFile.isFile()) outFile.delete();
                Properties prefs = new Properties();
                Enumeration e = Prefs.ijPrefs.keys();
                while (e.hasMoreElements()) {
                    String key = (String) e.nextElement();
                    if (key.indexOf(".plot.") == 0)
                        prefs.put(key, Prefs.ijPrefs.getProperty(key));
                }
                try
                    {
                    FileOutputStream fos = new FileOutputStream(cfgpath);
                    BufferedOutputStream bos = new BufferedOutputStream(fos);
                    prefs.store(bos, "AstroImageJ Plot Configuration");
                    bos.close();
                    }
                catch (IOException ioe)
                    {
                    IJ.beep();
                    IJ.showMessage("Error writing plot configuration file");
                    }
                }
            }
        
    static void setAllDialog()
        {
        GenericDialog gd = new GenericDialog ("Save all settings", mainFrame.getX()+mainFrame.getWidth()/2-165, mainFrame.getY()+mainFrame.getHeight()/2-77);
        gd.enableYesNoCancel("Save Files Now", "Save Settings Only");
        saveImage = Prefs.get("Astronomy_Tool.saveImage", saveImage);
        savePlot = Prefs.get("Astronomy_Tool.savePlot", savePlot);
        saveConfig = Prefs.get("Astronomy_Tool.saveConfig", saveConfig);
        saveTable = Prefs.get("Astronomy_Tool.saveTable", saveTable);
        saveApertures = Prefs.get("Astronomy_Tool.saveApertures", saveApertures);
        saveLog = Prefs.get("Astronomy_Tool.saveLog", saveLog);
        imageSuffix = Prefs.get("Astronomy_Tool.imageSuffix", imageSuffix);
        plotSuffix = Prefs.get("Astronomy_Tool.plotSuffix", plotSuffix);
        configSuffix = Prefs.get("Astronomy_Tool.configSuffix", configSuffix);
        dataSuffix = Prefs.get("Astronomy_Tool.dataSuffix", dataSuffix);
        aperSuffix = Prefs.get("Astronomy_Tool.aperSuffix", aperSuffix);
        logSuffix = Prefs.get("Astronomy_Tool.logSuffix", logSuffix);
        saveAllPNG = Prefs.get("Astronomy_Tool.saveAllPNG", saveAllPNG);

        gd.addMessage ("Select items to save when using save all:");
        String[] labels = {"Image","Plot","Plot Config","Data Table","Apertures","Log"};
        boolean[] defaults = {saveImage, savePlot, saveConfig, saveTable, saveApertures, saveLog};
        gd.addCheckboxGroup(1, 6, labels, defaults);
		gd.addStringField("Image suffix:", imageSuffix, 40);
        gd.addStringField("Plot image suffix:", plotSuffix, 40);
        gd.addStringField("Plot config file suffix:", configSuffix, 40);
        gd.addStringField("Data table file suffix:", dataSuffix, 40);
        gd.addStringField("Aperture file suffix:", aperSuffix, 40);
        gd.addStringField("Log file suffix:", logSuffix, 40);
        gd.addCheckbox("Save images in PNG format (uncheck for JPEG format)", saveAllPNG);
        gd.addMessage ("Tip: make plot config and data table suffix the same so that the plot config\n"+
                       "will auto-load when a new data table file is opened by drag and drop.");

		gd.showDialog();
		if (gd.wasCanceled()) return;
        saveImage = gd.getNextBoolean();
        savePlot = gd.getNextBoolean();
        saveConfig = gd.getNextBoolean();
        saveTable = gd.getNextBoolean();
        saveApertures = gd.getNextBoolean();
        saveLog = gd.getNextBoolean(); 
                
		imageSuffix = gd.getNextString();
        plotSuffix = gd.getNextString();
        configSuffix = gd.getNextString();
        dataSuffix = gd.getNextString();
        aperSuffix = gd.getNextString();
        logSuffix = gd.getNextString();
        
        saveAllPNG = gd.getNextBoolean(); 
        
        Prefs.set("Astronomy_Tool.saveImage", saveImage);
        Prefs.set("Astronomy_Tool.savePlot", savePlot);
        Prefs.set("Astronomy_Tool.saveConfig", saveConfig);
        Prefs.set("Astronomy_Tool.saveTable", saveTable);
        Prefs.set("Astronomy_Tool.saveApertures", saveApertures);
        Prefs.set("Astronomy_Tool.saveLog", saveLog);
        Prefs.set("Astronomy_Tool.imageSuffix", imageSuffix);
        Prefs.set("Astronomy_Tool.plotSuffix", plotSuffix);
        Prefs.set("Astronomy_Tool.configSuffix", configSuffix);
        Prefs.set("Astronomy_Tool.dataSuffix", dataSuffix);
        Prefs.set("Astronomy_Tool.aperSuffix", aperSuffix);
        Prefs.set("Astronomy_Tool.logSuffix", logSuffix);
        Prefs.set("Astronomy_Tool.saveAllPNG", saveAllPNG);
        
        if (gd.wasOKed()) 
            {
            if (saveAllPNG)
                saveAll("png", true);
            else
                saveAll("jpg", true);            
            }
        }  
    

static void saveAll(String format, boolean saveAll)
    {
    String outBase = "dataset";
    ImagePlus iplus = WindowManager.getCurrentImage();
    ImageWindow iw = null;
    ImageCanvas ic = null;
    AstroCanvas ac = null;
    Image img = null;
    BufferedImage imageDisplay = null;
    Graphics g = null;
    boolean imageFound = false;
    if (iplus != null)
        {
        iw = iplus.getWindow();
        if (iw != null && iw instanceof AstroStackWindow)
            {
            WindowManager.toFront(iw);
            ic = iplus.getCanvas();
            ac = (AstroCanvas)ic;
            g = ic.getGraphics();
            img = ac.graphicsToImage(g);
            imageDisplay = (BufferedImage)(img);
            imageFound = true;
            }
        else if (WindowManager.getWindowCount()> 0) 
            {
            int[] ID = WindowManager.getIDList();
            for (int win = 0; win < ID.length; win++)
                {
                iplus = WindowManager.getImage(ID[win]);
                if (iplus != null) 
                    iw = iplus.getWindow();
                else
                    iw = null;
                if (iw != null && iw instanceof AstroStackWindow)
                    {
                    WindowManager.setCurrentWindow(iw);
                    WindowManager.toFront(iw);
                    ic = iplus.getCanvas();
                    ac = (AstroCanvas)ic;
                    g = ic.getGraphics();
                    img = ac.graphicsToImage(g);
                    imageDisplay = (BufferedImage)(img);
                    imageFound = true;  
                    break;
                    }
                }
            }
        }

    SaveDialog sf = new SaveDialog(saveAll ? "Save all" : "Save as "+format.toUpperCase(), imageFound ? iplus.getShortTitle() : MeasurementTable.shorterName(table.shortTitle()), "");
    if (sf.getDirectory() == null || sf.getFileName() == null)
        return;
    String outPath = sf.getDirectory() + sf.getFileName();
    int location = outPath.lastIndexOf('.');
    if (location >= 0) outPath = outPath.substring(0, location);
    outBase = outPath;
    if (outBase.endsWith(imageSuffix))
        {
        location = outBase.lastIndexOf(imageSuffix);
        if (location >= 0) outBase = outBase.substring(0, location);
        }
    else if (outBase.endsWith(plotSuffix))
        {
        location = outBase.lastIndexOf(plotSuffix);
        if (location >= 0) outBase = outBase.substring(0, location);
        }
    else if (outBase.endsWith(configSuffix))
        {
        location = outBase.lastIndexOf(configSuffix);
        if (location >= 0) outBase = outBase.substring(0, location);
        }
    else if (outBase.endsWith(dataSuffix))
        {
        location = outBase.lastIndexOf(dataSuffix);
        if (location >= 0) outBase = outBase.substring(0, location);
        }
    else if (outBase.endsWith(aperSuffix))
        {
        location = outBase.lastIndexOf(aperSuffix);
        if (location >= 0) outBase = outBase.substring(0, location);
        }   
    else if (outBase.endsWith(logSuffix))
        {
        location = outBase.lastIndexOf(logSuffix);
        if (location >= 0) outBase = outBase.substring(0, location);
        }    

    if (imageFound && (!saveAll || (saveAll && saveImage)))
        {
        String imagePath = (saveAll ? outBase + imageSuffix : outPath) + "."+format;
        File saveFile = new File(imagePath);
        try {
            if (format.toLowerCase().equals("jpg"))
                {
                Iterator iter = ImageIO.getImageWritersByFormatName("jpeg");
                javax.imageio.ImageWriter writer = (javax.imageio.ImageWriter)iter.next();
                ImageWriteParam iwp = writer.getDefaultWriteParam();
                iwp.setCompressionMode(ImageWriteParam.MODE_EXPLICIT);
                iwp.setCompressionQuality(1.0f);   // an integer between 0 and 1 where 1 specifies minimum compression and maximum quality
                FileImageOutputStream output = new FileImageOutputStream(saveFile);
                writer.setOutput(output);
                IIOImage image = new IIOImage(imageDisplay, null, null);
                writer.write(null, image, iwp);
                output.close();
                writer.dispose();
                }
            else if (format.toLowerCase().equals("png"))
                {
                ImageIO.write(imageDisplay, format, saveFile);
                }
            else
                {
                IJ.error("bad image save format '"+format+"'");
                }

            }
        catch (IOException ex)
            {
            IJ.error("File Write Error", "Error writing image display to file '"+saveFile.toString()+"'");
            }
        }

    if (saveAll && (savePlot || saveConfig || saveTable))
        {
        saveDataImageConfig(savePlot, saveConfig, saveTable, true, format, outBase+plotSuffix+"."+format, outBase+configSuffix+".plotcfg",
                                                   outBase+dataSuffix+Prefs.get("options.ext", ".xls"));
    if (saveAll && saveApertures)
        {
        saveApertures(outBase+aperSuffix+".apertures");
        }
    if (saveAll && saveLog)
        {
        saveLogToFile(outBase+logSuffix+".log");
        }    
    }
    }

static void saveLogToFile(String path)
    {
    String log = IJ.getLog();
    if (log!=null)
        {
        String[] loglines = log.split("\\r?\\n|\\r");
        PrintWriter pw = null;
        try {
            FileOutputStream fos = new FileOutputStream(path);
            BufferedOutputStream bos = new BufferedOutputStream(fos);
            pw = new PrintWriter(bos);
            }
        catch (IOException e) {
            IJ.beep();
            IJ.showMessage("Error writing log file!");
            }
        for (int i=0; i<loglines.length;i++)
            {
            pw.println(loglines[i]);
            }
        pw.close();        
        }
    }

static void saveApertures(String apsPath)
        {
        File outFile = new File(apsPath);
        if (outFile == null || outFile.isDirectory())
            {
            IJ.error("bad aperture save filename");
            return;
            }
        if (outFile.isFile()) outFile.delete();
        Properties prefs = new Properties();
        Enumeration e = Prefs.ijPrefs.keys();
        while (e.hasMoreElements()) {
            String key = (String) e.nextElement();
            if (key.startsWith(".aperture.radius") || key.startsWith(".aperture.rback1") ||
                key.startsWith(".aperture.rback2") || key.startsWith(".aperture.removebackstars") ||
                key.startsWith(".aperture.backplane") || key.startsWith(".multiaperture.usevarsizeap") ||
                key.startsWith(".multiaperture.apfwhmfactor") || key.startsWith(".multiaperture.xapertures") ||
                key.startsWith(".multiaperture.yapertures") || key.startsWith(".multiaperture.isrefstar") || 
                key.startsWith(".multiaperture.naperturesmax"))
                prefs.put(key, Prefs.ijPrefs.getProperty(key));
        }
        try
            {
            FileOutputStream fos = new FileOutputStream(apsPath);
            BufferedOutputStream bos = new BufferedOutputStream(fos);
            prefs.store(bos, "AstroImageJ Saved Apertures");
            bos.close();
            }
        catch (IOException ioe)
            {
            IJ.beep();
            IJ.showMessage("Error writing apertures to file");
            }
        }    
    

//        static void setSaveStateDialog()
//            {
//            GenericDialog gd = new GenericDialog ("File suffix settings", mainFrame.getX()+mainFrame.getWidth()/2-165,
//                                                                          mainFrame.getY()+mainFrame.getHeight()/2-77);
//
//            plotSuffix = Prefs.get("Astronomy_Tool.plotSuffix", plotSuffix);
//            configSuffix = Prefs.get("Astronomy_Tool.configSuffix", configSuffix);
//            dataSuffix = Prefs.get("Astronomy_Tool.dataSuffix", dataSuffix);
//
//            gd.addMessage ("Enter save file suffix settings.");
//            gd.addStringField("Plot image suffix:", plotSuffix, 40);
//            gd.addStringField("Plot config file suffix:", configSuffix, 40);
//            gd.addStringField("Data table file suffix:", dataSuffix, 40);
//            gd.addMessage ("Tip: make plot config and data table suffix the same so that the plot config");
//            gd.addMessage ("will auto-load when a new data table file is opened by drag and drop.");
//
//            gd.showDialog();
//            if (gd.wasCanceled()) return;
//
//            plotSuffix = gd.getNextString();
//            configSuffix = gd.getNextString();
//            dataSuffix = gd.getNextString();
//
//            Prefs.set("Astronomy_Tool.plotSuffix", plotSuffix);
//            Prefs.set("Astronomy_Tool.configSuffix", configSuffix);
//            Prefs.set("Astronomy_Tool.dataSuffix", dataSuffix);
//            }
        
        static void openHelpPanel()
            {
            String filename = "help/multiplot_help.htm";
            new HelpPanel(filename, "Multi-plot");
            }      
        static void openDataHelpPanel()
            {
            String filename = "help/multiplot_data_help.htm";
            new HelpPanel(filename, "Data Naming Convention");
            }         

        static void getPreferences()
                {
                tableName = Prefs.get("plot2.tableName",tableName);
                keepFileNamesOnAppend=Prefs.get("plot2.keepFileNamesOnAppend", keepFileNamesOnAppend);   
                templateDir = Prefs.get("plot2.templateDir",templateDir);
                JDColumn = Prefs.get("plot2.JDColumn",JDColumn);
                
                addAirmass = Prefs.get("plot2.addAirmass", addAirmass);
                addAltitude = Prefs.get("plot2.addAltitude", addAltitude);
                addAzimuth = Prefs.get("plot2.addAzimuth", addAzimuth);
                addBJD = Prefs.get("plot2.addBJD", addBJD);
                addBJDCorr = Prefs.get("plot2.addBJDCorr", addBJDCorr);
                addDecNow = Prefs.get("plot2.addDecNow", addDecNow);
                addRaNow = Prefs.get("plot2.addRaNow", addRaNow);
                addDec2000 = Prefs.get("plot2.addDec2000", addDec2000);
                addRA2000 = Prefs.get("plot2.addRA2000", addRA2000);
                addGJD = Prefs.get("plot2.addGJD", addGJD);
                addHJD = Prefs.get("plot2.addHJD", addHJD);
                addHJDCorr = Prefs.get("plot2.addHJDCorr", addHJDCorr);
                addHourAngle = Prefs.get("plot2.addHourAngle", addHourAngle);
                addZenithDistance = Prefs.get("plot2.addZenithDistance", addZenithDistance);
                
                airmassName = Prefs.get("plot2.airmassName", airmassName);
                altitudeName = Prefs.get("plot2.altitudeName", altitudeName);
                azimuthName = Prefs.get("plot2.azimuthName", azimuthName);
                bjdName = Prefs.get("plot2.bjdName", bjdName);
                bjdCorrName = Prefs.get("plot2.bjdCorrName", bjdCorrName);
                decNowName = Prefs.get("plot2.decNowName", decNowName);
                raNowName = Prefs.get("plot2.raNowName", raNowName);
                dec2000Name = Prefs.get("plot2.dec2000Name", dec2000Name);
                ra2000Name = Prefs.get("plot2.ra2000Name", ra2000Name);
                gjdName = Prefs.get("plot2.gjdName", gjdName);
                hjdName = Prefs.get("plot2.hjdName", hjdName);
                hjdCorrName = Prefs.get("plot2.hjdCorrName", hjdCorrName);
                hourAngleName = Prefs.get("plot2.hourAngleName", hourAngleName);
                zenithDistanceName = Prefs.get("plot2.zenithDistanceName", zenithDistanceName);
                
                useGJD  = Prefs.get("plot2.useGJD", useGJD);
                useHJD  = Prefs.get("plot2.useHJD", useHJD);
                useBJD  = Prefs.get("plot2.useBJD", useBJD);
                
                plotAutoMode=Prefs.get("plot.automode",plotAutoMode);
                xlabeldefault=Prefs.get("plot.xlabeldefault",xlabeldefault);
                title=Prefs.get("plot.title", title);
                useTitle=Prefs.get("plot.useTitle", useTitle);
                subtitle=Prefs.get("plot.subtitle", subtitle);
                useSubtitle=Prefs.get("plot.useSubtitle", useSubtitle);
                titlePosX=Prefs.get("plot.titlePosX",titlePosX);
                titlePosY=Prefs.get("plot.titlePosY",titlePosY);
                subtitlePosX=Prefs.get("plot.subtitlePosX",subtitlePosX);
                subtitlePosY=Prefs.get("plot.subtitlePosY",subtitlePosY);
                legendPosX=Prefs.get("plot.legendPosX",legendPosX);
                legendPosY=Prefs.get("plot.legendPosY",legendPosY);
                xTics=Prefs.get("plot.xTics", xTics);
                yTics=Prefs.get("plot.yTics", yTics);
                xGrid=Prefs.get("plot.xGrid", xGrid);
                yGrid=Prefs.get("plot.yGrid", yGrid);
                xNumbers=Prefs.get("plot.xNumbers", xNumbers);
                yNumbers=Prefs.get("plot.yNumbers", yNumbers);
                autoScaleX=Prefs.get("plot.autoScaleX",autoScaleX);
                useFirstX=Prefs.get("plot.useFirstX",useFirstX);
                autoScaleY=Prefs.get("plot.autoScaleY",autoScaleY);
                xMin=Prefs.get("plot.xMin", xMin);
                xMax=Prefs.get("plot.xMax", xMax);
                xWidth=Prefs.get("plot.xWidth", xWidth);
                yMin=Prefs.get("plot.yMin", yMin);
                yMax=Prefs.get("plot.yMax", yMax);
                plotSizeX=(int)Prefs.get("plot.plotSizeX", plotSizeX);
                plotSizeY=(int)Prefs.get("plot.plotSizeY", plotSizeY);
                mainFrameLocationX=(int)Prefs.get("plot2.mainFrameLocationX", mainFrameLocationX);
                mainFrameLocationY=(int)Prefs.get("plot2.mainFrameLocationY", mainFrameLocationY);
                subFrameLocationX=(int)Prefs.get("plot2.subFrameLocationX", subFrameLocationX);
                subFrameLocationY=(int)Prefs.get("plot2.subFrameLocationY", subFrameLocationY);
                refStarFrameLocationX=(int)Prefs.get("plot2.refStarFrameLocationX", refStarFrameLocationX);
                refStarFrameLocationY=(int)Prefs.get("plot2.refStarFrameLocationY", refStarFrameLocationY);  
                addAstroDataFrameLocationX=(int)Prefs.get("plot2.addAstroDataFrameLocationX", addAstroDataFrameLocationX);
                addAstroDataFrameLocationY=(int)Prefs.get("plot2.addAstroDataFrameLocationY", addAstroDataFrameLocationY);                  
                plotFrameLocationX=(int)Prefs.get("plot2.plotFrameLocationX", plotFrameLocationX);
                plotFrameLocationY=(int)Prefs.get("plot2.plotFrameLocationY", plotFrameLocationY);
                openDataSetWindow=Prefs.get("plot2.openDataSetWindow", openDataSetWindow);
                openRefStarWindow=Prefs.get("plot2.openRefStarWindow", openRefStarWindow);
                rememberWindowLocations=Prefs.get("plot2.rememberWindowLocations", rememberWindowLocations);
                divideNotSubtract=Prefs.get("plot.divideNotSubtract", divideNotSubtract);
                usePixelScale=Prefs.get("plot.usePixelScale", usePixelScale);
                pixelScale=Prefs.get("plot.pixelScale", pixelScale);
                mmagrefs=(int)Prefs.get("plot.mmagrefs", mmagrefs);
                xExponent=(int)Prefs.get("plot.xExponent", xExponent);
                yExponent=(int)Prefs.get("plot.yExponent", yExponent);
                xWidthStep=Prefs.get("plot.xWidthStep", xWidthStep);
                showToolTips=Prefs.get("astroIJ.showToolTips",showToolTips);
                ToolTipManager.sharedInstance().setEnabled(showToolTips);
                showXAxisNormal = Prefs.get("plot.showXAxisNormal", showXAxisNormal);
                showXAxisAsPhase = Prefs.get("plot.showXAxisAsPhase", showXAxisAsPhase);
                showXAxisAsHoursSinceTc = Prefs.get("plot.showXAxisAsHoursSinceTc", showXAxisAsHoursSinceTc);
                showXAxisAsDaysSinceTc = Prefs.get("plot.showXAxisAsDaysSinceTc", showXAxisAsDaysSinceTc);
                T0 = Prefs.get("plot.T0", T0);
                period = Prefs.get("plot.period", period);
                xMaxStep=Prefs.get("plot.xMaxStep", xMaxStep);
                xMinStep=Prefs.get("plot.xMinStep", xMinStep);
                yMaxStep=Prefs.get("plot.yMaxStep", yMaxStep);
                yMinStep=Prefs.get("plot.yMinStep", yMinStep);
                vMarker1Value=Prefs.get("plot.vMarker1Value", vMarker1Value);
                vMarker2Value=Prefs.get("plot.vMarker2Value", vMarker2Value);
                vMarkerStep=Prefs.get("plot.vMarkerStep", vMarkerStep);
                showVMarker1=Prefs.get("plot.showVMarker1", showVMarker1);
                showVMarker2=Prefs.get("plot.showVMarker2", showVMarker2);
                vMarker1TopText=Prefs.get("plot.vMarker1TopText", vMarker1TopText);
                vMarker1BotText=Prefs.get("plot.vMarker1BotText", vMarker1BotText);
                vMarker2TopText=Prefs.get("plot.vMarker2TopText", vMarker2TopText);
                vMarker2BotText=Prefs.get("plot.vMarker2BotText", vMarker2BotText);
                maxDetrendVars = (int)Prefs.get("plot.maxDetrendVars", 3);
                if (maxDetrendVars < 1) maxDetrendVars = 1;
                dMarker1Value=Prefs.get("plot.dMarker1Value", dMarker1Value);
                dMarker2Value=Prefs.get("plot.dMarker2Value", dMarker2Value);
                dMarker3Value=Prefs.get("plot.dMarker3Value", dMarker3Value);
                dMarker4Value=Prefs.get("plot.dMarker4Value", dMarker4Value);   
                dMarkerStep=Prefs.get("plot.dMarkerStep", dMarkerStep);
                useDMarker1=Prefs.get("plot.useDMarker1", useDMarker1);
                useDMarker4=Prefs.get("plot.useDMarker4", useDMarker4);
                showDMarkers=Prefs.get("plot.showDMarkers", showDMarkers);
                invertYAxis=Prefs.get("plot.invertYAxis", invertYAxis);
                invertYAxisSign=invertYAxis?-1:1;
                negateMag=Prefs.get("plot.negateMag", negateMag);
                saveNewXColumn=Prefs.get("plot.saveNewXColumn", saveNewXColumn);
                saveNewYColumn=Prefs.get("plot.saveNewYColumn", saveNewYColumn);
                saveNewYErrColumn=Prefs.get("plot.saveNewYErrColumn", saveNewYErrColumn);
                plotSizeXStep=(int)Prefs.get("plot.plotSizeXStep", plotSizeXStep);
                plotSizeYStep=(int)Prefs.get("plot.plotSizeYStep", plotSizeYStep);
                excludedHeadSamples=(int)Prefs.get("plot.excludedHeadSamples", excludedHeadSamples);
                excludedTailSamples=(int)Prefs.get("plot.excludedTailSamples", excludedTailSamples);
                excludedHeadSamplesStep=(int)Prefs.get("plot.excludedHeadSamplesStep", excludedHeadSamplesStep);
                excludedTailSamplesStep=(int)Prefs.get("plot.excludedTailSamplesStep", excludedTailSamplesStep);
                refStarHorzWidth=(int)Prefs.get("plot.refStarHorzWidth", refStarHorzWidth);
                if (refStarHorzWidth < 1) refStarHorzWidth = 1;
                useXColumnName=Prefs.get("plot.useXColumnName", useXColumnName);
                useYColumnName=Prefs.get("plot.useYColumnName", useYColumnName);
                useXCustomName=Prefs.get("plot.useXCustomName", useXCustomName);
                useYCustomName=Prefs.get("plot.useYCustomName", useYCustomName);
                xLegend=Prefs.get("plot.xLegend", xLegend);
                yLegend=Prefs.get("plot.yLegend", yLegend);
                showXScaleInfo=Prefs.get("plot.showXScaleInfo", showXScaleInfo);
                showYScaleInfo=Prefs.get("plot.showYScaleInfo", showYScaleInfo);
                showYShiftInfo=Prefs.get("plot.showYShiftInfo", showYShiftInfo);
                showLScaleInfo=Prefs.get("plot.showLScaleInfo", showLScaleInfo);
                showLRelScaleInfo=Prefs.get("plot.showLRelScaleInfo", showLRelScaleInfo);
                showLShiftInfo=Prefs.get("plot.showLShiftInfo", showLShiftInfo);
                showLRelShiftInfo=Prefs.get("plot.showLRelShiftInfo", showLRelShiftInfo);
                showYBinInfo=Prefs.get("plot.showYBinInfo", showYBinInfo);
                showLBinInfo=Prefs.get("plot.showLBinInfo", showLBinInfo);
                showYmmagInfo=Prefs.get("plot.showYmmagInfo", showYmmagInfo);
                showLmmagInfo=Prefs.get("plot.showLmmagInfo", showLmmagInfo);
                showLdetrendInfo=Prefs.get("plot.showLdetrendInfo", showLdetrendInfo);
                showLnormInfo=Prefs.get("plot.showLnormInfo", showLnormInfo);
                showYSymbolInfo=Prefs.get("plot.showYSymbolInfo", showYSymbolInfo);
                showLSymbolInfo=Prefs.get("plot.showLSymbolInfo", showLSymbolInfo);
                useTwoLineLegend=Prefs.get("plot.useTwoLineLegend", useTwoLineLegend);
                useWideDataPanel=Prefs.get("plot2.useWideDataPanel", useWideDataPanel);
                useBoldedDatum=Prefs.get("plot2.useBoldedDatum", useBoldedDatum);
                useUpdateStack=Prefs.get("plot2.useUpdateStack", useUpdateStack);
                legendLeft=Prefs.get("plot.legendLeft", legendLeft);
                legendRight=Prefs.get("plot.legendRight", legendRight);
                priorityColumns=Prefs.get("plot.priorityColumns", priorityColumns);
                priorityDetrendColumns=Prefs.get("plot.priorityDetrendColumns", priorityDetrendColumns);
                prioritizeColumns=Prefs.get("plot.prioritizeColumns", prioritizeColumns);
                maxColumnLength=(int)Prefs.get("plot.maxColumnLength", 1000);

                saveImage = Prefs.get("Astronomy_Tool.saveImage", saveImage);
                savePlot = Prefs.get("Astronomy_Tool.savePlot", savePlot);
                saveConfig = Prefs.get("Astronomy_Tool.saveConfig", saveConfig);
                saveTable = Prefs.get("Astronomy_Tool.saveTable", saveTable);
                saveApertures = Prefs.get("Astronomy_Tool.saveApertures", saveApertures);
                saveLog = Prefs.get("Astronomy_Tool.saveLog", saveLog);
                imageSuffix = Prefs.get("Astronomy_Tool.imageSuffix", imageSuffix);
                plotSuffix = Prefs.get("Astronomy_Tool.plotSuffix", plotSuffix);
                configSuffix = Prefs.get("Astronomy_Tool.configSuffix", configSuffix);
                dataSuffix = Prefs.get("Astronomy_Tool.dataSuffix", dataSuffix);
                aperSuffix = Prefs.get("Astronomy_Tool.aperSuffix", aperSuffix);
                logSuffix = Prefs.get("Astronomy_Tool.logSuffix", logSuffix);
                saveAllPNG = Prefs.get("Astronomy_Tool.saveAllPNG", saveAllPNG);

                for (int i=0; i<maxCurves; i++)
                        {
                        moreOptions[i]=Prefs.get("plot.moreOptions"+i,moreOptions[i]);
                        oplabel[i]=Prefs.get("plot.oplabel"+i, oplabel[i]);
                        xlabel[i]=Prefs.get("plot.xlabel"+i,xlabel[i]);
                        ylabel[i]=Prefs.get("plot.ylabel"+i,ylabel[i]);
                        lines[i]=Prefs.get("plot.lines"+i, lines[i]);
                        markerIndex[i]=(int)Prefs.get("plot.markerIndex"+i,markerIndex[i]);
                        colorIndex[i]=(int)Prefs.get("plot.colorIndex"+i, colorIndex[i]);
                        normIndex[i]=(int)Prefs.get("plot.normIndex"+i, normIndex[i]);
                        detrendVarDisplayed[i]=(int)Prefs.get("plot.detrendVarDisplayed"+i, detrendVarDisplayed[i]);
                        if (detrendVarDisplayed[i] >= maxDetrendVars) detrendVarDisplayed[i] = maxDetrendVars - 1;
                        if (detrendVarDisplayed[i] < 0) detrendVarDisplayed[i] = 0;
                        detrendFitIndex[i]=(int)Prefs.get("plot.detrendFitIndex"+i, detrendFitIndex[i]);
                        mmag[i]=Prefs.get("plot.mmag"+i, mmag[i]);
                        fromMag[i]=Prefs.get("plot.fromMag"+i, fromMag[i]);
                        useColumnName[i]=Prefs.get("plot.useColumnName"+i, useColumnName[i]);
                        useLegend[i]=Prefs.get("plot.useLegend"+i, useLegend[i]);
                        legend[i] = Prefs.get("plot.legend"+i, legend[i]);
                        operatorIndex[i]=(int)Prefs.get("plot.operatorIndex"+i, operatorIndex[i]);
                        binSize[i]=(int)Prefs.get("plot.binSize"+i, binSize[i]);
                        plotY[i] = Prefs.get("plot.plotY"+i,plotY[i]);
                        force[i] = Prefs.get("plot.force"+i,force[i]);
                        showErrors[i] = Prefs.get("plot.showErrors"+i,showErrors[i]);
                        autoScaleFactor[i]=Prefs.get("plot.autoScaleFactor"+i,autoScaleFactor[i]);
                        autoScaleStep[i]=Prefs.get("plot.autoScaleStep"+i,autoScaleStep[i]);
                        autoShiftFactor[i]=Prefs.get("plot.autoShiftFactor"+i,autoShiftFactor[i]);
                        autoShiftStep[i]=Prefs.get("plot.autoShiftStep"+i,autoShiftStep[i]);
                        customScaleFactor[i]=Prefs.get("plot.customScaleFactor"+i,customScaleFactor[i]);
                        customScaleStep[i]=Prefs.get("plot.customScaleStep"+i,customScaleStep[i]);
                        customShiftFactor[i]=Prefs.get("plot.customShiftFactor"+i,customShiftFactor[i]);
                        customShiftStep[i]=Prefs.get("plot.customShiftStep"+i,customShiftStep[i]);
                        ASInclude[i]=Prefs.get("plot.ASInclude"+i,ASInclude[i]);
                        for (int v = 0; v < maxDetrendVars; v++)
                            {
                            detrendlabel[i][v]=Prefs.get("plot.detrendlabel["+i+"]["+v+"]",detrendlabel[i][v]);
                            detrendlabelhold[i][v]=Prefs.get("plot.detrendlabelhold["+i+"]["+v+"]",detrendlabelhold[i][v]);
                            detrendFactor[i][v]=Prefs.get("plot.detrendFactor["+i+"]["+v+"]",detrendFactor[i][v]);
                            detrendFactorStep[i][v]=Prefs.get("plot.detrendFactorStep["+i+"]["+v+"]",detrendFactorStep[i][v]);  
                            }
                        }
                }
        
        static void saveAstroPanelPrefs()
                {
                Prefs.set("plot2.addAirmass", addAirmass);
                Prefs.set("plot2.addAltitude", addAltitude);
                Prefs.set("plot2.addAzimuth", addAzimuth);
                Prefs.set("plot2.addBJD", addBJD);
                Prefs.set("plot2.addBJDCorr", addBJDCorr);
                Prefs.set("plot2.addDecNow", addDecNow);
                Prefs.set("plot2.addRaNow", addRaNow);
                Prefs.set("plot2.addDec2000", addDec2000);
                Prefs.set("plot2.addRA2000", addRA2000);
                Prefs.set("plot2.addGJD", addGJD);
                Prefs.set("plot2.addHJD", addHJD);
                Prefs.set("plot2.addHJDCorr", addHJDCorr);
                Prefs.set("plot2.addHourAngle", addHourAngle);
                Prefs.set("plot2.addZenithDistance", addZenithDistance);                
                
                Prefs.set("plot2.airmassName", airmassName);
                Prefs.set("plot2.altitudeName", altitudeName);
                Prefs.set("plot2.azimuthName", azimuthName);
                Prefs.set("plot2.bjdName", bjdName);
                Prefs.set("plot2.bjdCorrName", bjdCorrName);
                Prefs.set("plot2.decNowName", decNowName);
                Prefs.set("plot2.raNowName", raNowName);
                Prefs.set("plot2.dec2000Name", dec2000Name);
                Prefs.set("plot2.ra2000Name", ra2000Name);
                Prefs.set("plot2.gjdName", gjdName);
                Prefs.set("plot2.hjdName", hjdName);
                Prefs.set("plot2.hjdCorrName", hjdCorrName);
                Prefs.set("plot2.hourAngleName", hourAngleName);
                Prefs.set("plot2.zenithDistanceName", zenithDistanceName); 
                
                Prefs.set("plot2.JDColumn",JDColumn);
                Prefs.set("plot2.useGJD", useGJD);
                Prefs.set("plot2.useHJD", useHJD);
                Prefs.set("plot2.useBJD", useBJD);             
                }
        
        
        
        static void savePreferences()
                {
                Prefs.set("plot2.tableName",tableName);
                Prefs.set("plot2.keepFileNamesOnAppend", keepFileNamesOnAppend);
                Prefs.set("plot2.templateDir",templateDir);
                saveAstroPanelPrefs();
                Prefs.set("plot.maxDetrendVars", maxDetrendVars);
                Prefs.set("plot2.useDefaultSettings", useDefaultSettings);
                Prefs.set("plot.xlabeldefault",xlabeldefault);
                Prefs.set("plot.xTics", xTics);
                Prefs.set("plot.yTics", yTics);
                Prefs.set("plot.xGrid", xGrid);
                Prefs.set("plot.yGrid", yGrid);
                Prefs.set("plot.xNumbers", xNumbers);
                Prefs.set("plot.yNumbers", yNumbers);
                Prefs.set("plot.autoScaleX", autoScaleX);
                Prefs.set("plot.autoScaleY", autoScaleY);
                Prefs.set("plot.useFirstX", useFirstX);
                Prefs.set("plot.xWidth", xWidth);
                Prefs.set("plot.xMin", xMin);
                Prefs.set("plot.xMax", xMax);
                Prefs.set("plot.yMin", yMin);
                Prefs.set("plot.yMax", yMax);
                Prefs.set("plot.vMarker1Value", vMarker1Value);
                Prefs.set("plot.vMarker2Value", vMarker2Value);
                Prefs.set("plot.vMarkerStep", vMarkerStep);
                Prefs.set("plot.showVMarker1", showVMarker1);
                Prefs.set("plot.showVMarker2", showVMarker2);
                Prefs.set("plot.vMarker1TopText", vMarker1TopText);
                Prefs.set("plot.vMarker1BotText", vMarker1BotText);
                Prefs.set("plot.vMarker2TopText", vMarker2TopText);
                Prefs.set("plot.vMarker2BotText", vMarker2BotText);
                Prefs.set("plot.dMarker1Value", dMarker1Value);
                Prefs.set("plot.dMarker2Value", dMarker2Value);
                Prefs.set("plot.dMarker3Value", dMarker3Value);
                Prefs.set("plot.dMarker4Value", dMarker4Value);   
                Prefs.set("plot.dMarkerStep", dMarkerStep);
                Prefs.set("plot.useDMarker1", useDMarker1);
                Prefs.set("plot.useDMarker4", useDMarker4);
                Prefs.set("plot.showDMarkers", showDMarkers);                
                Prefs.set("plot.invertYAxis", invertYAxis);
                Prefs.set("plot.negateMag", negateMag);
                Prefs.set("plot.saveNewXColumn", saveNewXColumn);
                Prefs.set("plot.saveNewYColumn", saveNewYColumn);   
                Prefs.set("plot.saveNewYErrColumn", saveNewYErrColumn); 
                Prefs.set("plot.plotSizeX",plotSizeX);
                Prefs.set("plot.plotSizeY",plotSizeY);
                Prefs.set("plot.title",title);
                Prefs.set("plot.useTitle",useTitle);
                Prefs.set("plot.subtitle",subtitle);
                Prefs.set("plot.useSubtitle",useSubtitle);
                Prefs.set("plot.titlePosX",titlePosX);
                Prefs.set("plot.titlePosY",titlePosY);
                Prefs.set("plot.subtitlePosX",subtitlePosX);
                Prefs.set("plot.subtitlePosY",subtitlePosY);
                Prefs.set("plot2.openDataSetWindow",openDataSetWindow);
                Prefs.set("plot2.openRefStarWindow", openRefStarWindow);
                Prefs.set("plot.legendPosX",legendPosX);
                Prefs.set("plot.legendPosY",legendPosY);
                Prefs.set("plot2.rememberWindowLocations",rememberWindowLocations);
                Prefs.set("plot.divideNotSubtract", divideNotSubtract);
                Prefs.set("plot.usePixelScale",usePixelScale);
                Prefs.set("plot.pixelScale",pixelScale);
                Prefs.set("plot.mmagrefs",mmagrefs);
                Prefs.set("plot.xExponent",xExponent);
                Prefs.set("plot.yExponent",yExponent);
                Prefs.set("plot.xWidthStep",xWidthStep);
                Prefs.set("plot.xMaxStep",xMaxStep);
                Prefs.set("plot.xMinStep",xMinStep);
                Prefs.set("plot.yMaxStep",yMaxStep);
                Prefs.set("plot.yMinStep",yMinStep);
                Prefs.set("plot.plotSizeXStep",plotSizeXStep);
                Prefs.set("plot.plotSizeYStep",plotSizeYStep);
                Prefs.set("plot.excludedHeadSamples", excludedHeadSamples);
                Prefs.set("plot.excludedTailSamples", excludedTailSamples);
                Prefs.set("plot.excludedHeadSamplesStep", excludedHeadSamplesStep);
                Prefs.set("plot.excludedTailSamplesStep", excludedTailSamplesStep);
                Prefs.set("plot.refStarHorzWidth", refStarHorzWidth);
                Prefs.set("plot.useXColumnName", useXColumnName);
                Prefs.set("plot.useYColumnName", useYColumnName);
                Prefs.set("plot.useXCustomName", useXCustomName);
                Prefs.set("plot.useYCustomName", useYCustomName);
                Prefs.set("plot.xLegend", xLegend);
                Prefs.set("plot.yLegend", yLegend);
                Prefs.set("plot.showXAxisNormal", showXAxisNormal);
                Prefs.set("plot.showXAxisAsPhase", showXAxisAsPhase);
                Prefs.set("plot.showXAxisAsHoursSinceTc", showXAxisAsHoursSinceTc);
                Prefs.set("plot.showXAxisAsDaysSinceTc", showXAxisAsDaysSinceTc);
                Prefs.set("plot.T0", T0);
                Prefs.set("plot.period", period);                
                Prefs.set("plot.showXScaleInfo", showXScaleInfo);
                Prefs.set("plot.showYScaleInfo", showYScaleInfo);
                Prefs.set("plot.showYShiftInfo", showYShiftInfo);
                Prefs.set("plot.showLScaleInfo", showLScaleInfo);
                Prefs.set("plot.showLRelScaleInfo", showLRelScaleInfo);
                Prefs.set("plot.showLShiftInfo", showLShiftInfo);
                Prefs.set("plot.showLRelShiftInfo", showLRelShiftInfo);
                Prefs.set("plot.showYBinInfo", showYBinInfo);
                Prefs.set("plot.showLBinInfo", showLBinInfo);
                Prefs.set("plot.showYmmagInfo", showYmmagInfo);
                Prefs.set("plot.showLmmagInfo", showLmmagInfo);
                Prefs.set("plot.showLdetrendInfo", showLdetrendInfo);
                Prefs.set("plot.showLnormInfo", showLnormInfo);
                Prefs.set("plot.showYSymbolInfo", showYSymbolInfo);
                Prefs.set("plot.showLSymbolInfo", showLSymbolInfo);
                Prefs.set("plot.showLSymbolInfo", showLSymbolInfo);
                Prefs.set("plot2.useWideDataPanel", useWideDataPanel);
                Prefs.set("plot.useTwoLineLegend", useTwoLineLegend);
                Prefs.set("plot2.useBoldedDatum", useBoldedDatum);
                Prefs.set("plot2.useUpdateStack", useUpdateStack);
                Prefs.set("plot.legendLeft", legendLeft);
                Prefs.set("plot.legendRight", legendRight);
                Prefs.set("plot.priorityColumns", priorityColumns);
                Prefs.set("plot.priorityDetrendColumns", priorityDetrendColumns);
                Prefs.set("plot.prioritizeColumns", prioritizeColumns);
                
                Prefs.set("Astronomy_Tool.saveImage", saveImage);
                Prefs.set("Astronomy_Tool.savePlot", savePlot);
                Prefs.set("Astronomy_Tool.saveConfig", saveConfig);
                Prefs.set("Astronomy_Tool.saveTable", saveTable);
                Prefs.set("Astronomy_Tool.saveApertures", saveApertures);
                Prefs.set("Astronomy_Tool.saveLog", saveLog);
                Prefs.set("Astronomy_Tool.imageSuffix", imageSuffix);
                Prefs.set("Astronomy_Tool.plotSuffix", plotSuffix);
                Prefs.set("Astronomy_Tool.configSuffix", configSuffix);
                Prefs.set("Astronomy_Tool.dataSuffix", dataSuffix);
                Prefs.set("Astronomy_Tool.aperSuffix", aperSuffix);
                Prefs.set("Astronomy_Tool.logSuffix", logSuffix);
                Prefs.set("Astronomy_Tool.saveAllPNG", saveAllPNG);                

                for (int i=0; i<maxCurves; i++)
                        {
                        Prefs.set("plot.xlabel"+i,xlabel[i]);
                        Prefs.set("plot.ylabel"+i,ylabel[i]);
                        Prefs.set("plot.oplabel"+i,oplabel[i]);
                        Prefs.set("plot.lines"+i, lines[i]);
                        Prefs.set("plot.markerIndex"+i,markerIndex[i]);
                        Prefs.set("plot.colorIndex"+i,colorIndex[i]);
                        Prefs.set("plot.normIndex"+i,normIndex[i]);
                        Prefs.set("plot.detrendVarDisplayed"+i, detrendVarDisplayed[i]);
                        Prefs.set("plot.detrendFitIndex"+i,detrendFitIndex[i]);
                        Prefs.set("plot.mmag"+i, mmag[i]);
                        Prefs.set("plot.fromMag"+i, fromMag[i]);
                        Prefs.set("plot.useColumnName"+i,useColumnName[i]);
                        Prefs.set("plot.useLegend"+i,useLegend[i]);
                        Prefs.set("plot.legend"+i,legend[i]);
                        Prefs.set("plot.operatorIndex"+i,operatorIndex[i]);
                        Prefs.set("plot.moreOptions"+i,moreOptions[i]);
                        Prefs.set("plot.binSize"+i,binSize[i]);
                        Prefs.set("plot.plotY"+i,plotY[i]);
                        Prefs.set("plot.force"+i,force[i]);
                        Prefs.set("plot.showErrors"+i,showErrors[i]);
                        Prefs.set("plot.autoScaleFactor"+i,autoScaleFactor[i]);
                        Prefs.set("plot.autoScaleStep"+i,autoScaleStep[i]);
                        Prefs.set("plot.autoShiftFactor"+i,autoShiftFactor[i]);
                        Prefs.set("plot.autoShiftStep"+i,autoShiftStep[i]);
                        Prefs.set("plot.customScaleFactor"+i,customScaleFactor[i]);
                        Prefs.set("plot.customScaleStep"+i,customScaleStep[i]);
                        Prefs.set("plot.customShiftFactor"+i,customShiftFactor[i]);
                        Prefs.set("plot.customShiftStep"+i,customShiftStep[i]);
                        Prefs.set("plot.ASInclude"+i,ASInclude[i]);
                        for (int v = 0; v < maxDetrendVars; v++)
                            {
                            Prefs.set("plot.detrendlabel["+i+"]["+v+"]",detrendlabel[i][v]);
                            Prefs.set("plot.detrendlabelhold["+i+"]["+v+"]",detrendlabelhold[i][v]);
                            Prefs.set("plot.detrendFactor["+i+"]["+v+"]",detrendFactor[i][v]);
                            Prefs.set("plot.detrendFactorStep["+i+"]["+v+"]",detrendFactorStep[i][v]);  
                            }                        
                        }
                }
	}

